/*************************************************************************************\


 #     #   #####   #     #   #####    #####   ######      #  #######  #####   ######
 #     #  #     #  ##    #  #     #  #     #  #     #    #      #    #     #  #     #
 #  #  #  #     #  # #   #  #              #  #     #   #       #          #  #     #
 #  #  #  #######  #  #  #  # #####   #####   #     #  #   #    #     #####   #     #
 #  #  #  #     #  #   # #  #     #        #  #     #  #######  #          #  #     #
 #  #  #  #     #  #    ##  #     #  #     #  #     #      #    #    #     #  #     #      
  ## ##   #     #  #     #   #####    #####   ######       #    #     #####   ######


                                  Preprocessor of
                 3D Wang Tiles with Ellipsoidal Inclusions for T3D
                           Copyright: Daniel Rypl 2018

\*************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <stdarg.h>
#include <limits.h>

/* TODO
	 implement isotropic scale of inclusions
	 implement anisotropic scale of inclusions
	 implement directives as options !!!
	 separate option for absolute epsilon and relative epsilon
	 implement interface */



#define VERSION_MAJOR         1
#define VERSION_MINOR         0

//#define DEBUG

/* basic user customisable directives */

#define MAX_RADIUS_RATE       2.5           /* 2.0 <= MAX_RADIUS_RATE */

/* warning: the larger MAX_TILES the less number of inclusions is allowed to intersect particular tile side */

#define MAX_TILES             100           /* should be power of 10 (real maximal number of tiles is max_tiles - 1) */

#define DEF_REL_EPSILON       0.001         /* 0 < DEF_REL_EPSILON <= MAX_REL_EPSILON */
#define MAX_REL_EPSILON       0.1           /* 0 < MAX_REL_EPSILON << 1.0 */

#define ARBITRARY_TILE_ID                   /* allow arbitrary non-negative numbers for tiles */

//#define SEPARATE_PER_REP                    /* periodic repetition specified separately for each direction ==> increases size of input file */

/* detouching reduces violation of proximity criteria but increseases risk of mutual intersection of inclusions */

//#define DETOUCH                             /* shift not intersected inclusion close to tile side inside tile */  

#ifdef DETOUCH
//#define OVERDETOUCH                         /* shift intersected inclusion close to tile side inside tile */
#endif

#define ALLOW_OVERLAP                       /* defined reference inclusions are allowed to mutually overlap */

#define CHECK_OVERLAP                       /* check for overlap */

#define ALLOW_OUTSIDE                       /* allow for inclusion center of which falls out of rererence cube */

#define FIX_HOLE                            /* fix degeneration of hole */

/* note that isolated nodes may cause problems only on boundary of mesh assembly which is however likely to be subjected to periodic conditions,
	 it is thus probably desirable to keep them in the mesh and thus check for isolated nodes is not necessary */

//#define CHECK_ISOLATED                      /* check for isolated nodes which may cause problems in mesh assembly */

/* there are 3 approaches how to specify keyword tile:
	 1) it is specified only for regions (matrix and inclusions) and then it is propagated in t3d to boundary and fixed entities;
	 note that in this approach there is no need to specify keyword tile for interfaces because they do not have interior nodes;
	 2) it is specified only for matrix region, during t3d input of the regions it is propagated to its subregions 
	 and then it is propagated to boundary and fixed entities;
	 this approach requires that all inclusions are specified as subregions;
	 IMPORTANT: this approach is not applicable when using interfaces !!!
	 tile is propagated from matrix to outer hole subregion, but it is not propagated to inner solid region;
	 CAUTION: this approach is not applicable when subregions are used recursively because region does not store its subregions !!!
	 3) it is specified for all model entitities (except virtual ones);
	 
	 in order to enable simple and safe output of renumbered meshes, it is desirable to ensure that all physical model entities 
	 on t3d input refer to a particular tile;
	 when approach 1 or 2 is adopted, this must be checked after tile propagation (to boundary and fixed entities);
	 in order to avoid writing special check routine, the check can be performed during discretization;
	 the check during input can be used only to trigger the tile propagation to boundary and fixed entities (if tile specification 
	 is missing on any physical entity), but it is not guaranteed that all entities will be visited during propagation;
	 in this approach, missing tile specification for region (in 3D problem) or patch (in 2D problem) is considered an error;
	 it implies that propagation to subregion makes no sense in such a case;
	 when approach 3 is adopted, t3d should check whether keyword tile is present on input for each physical model entity;
	 if this check is performed also for interfaces, then tile keyword must be specified also for interfaces;

	 currently approach 3 is implemented;
	 the major disadvantage of approach 3 is that it unneccessarily increases size of t3d input file (by 2-4%);

	 note that combination of approaches 1 and 2 can be implemented;
	 it means that tile is specified for matrix region and regions of inclusions with interface */

/* keyword TILE must be used for t3d4wt;
	 it makes no sense to use ordinary t3d because it is likely to fail when handling inexact ellipsoids */

#define USE_TILE                            /* keep defined !!! */

/* similarly there are also 3 approaches how to specify keyword side:
	 1) it is specified only for tile side patches and for intersection ellipse patches and then it is cumulatively 
	 propagated in t3d to boundary and fixed entities;
	 in order to allow for processing also 2D tiles, only nonzero side is propagated;
	 2) it is specified only for tile side patches, during t3d input of patch, it is propagated to its subpatches
	 and then it is propagated to boundary and fixed entities;
	 this approach requires that all intersection ellipses are specified as subpatches;
	 IMPORTANT: this approach is not applicable when using interfaces !!!
	 side is propagated from matrix to outer hole subpatch, but it is not propagated to inner solid patch;
	 CAUTION: this approach is not applicable when subpatches are used recursively because patch does not store its subpatches !!!
	 3) it is specified for all relevant model entities

	 approaches 1 and 2 are easier in the sense that only simple side (face) specification is required;
	 approach 3 is more complicated in the sense that simple (face) as well as combined (edge and corner) side specification is required;

	 since there is no easy apriori check, that nonzero side specification is missing at particular model entity,
	 approaches 1 and 2 are considered better and safer;
	 note that coordinate based check may be nontrivial if more tiles are processed together which is however a typical case;

	 currently approach 1 is implemented;

	 note that combination of approaches 1 and 2 can be implemented;
	 it means that side is specified for tile side patches and for intersection ellipse patches of inclusions with interface;

	 since side can be easily retrieved from patch id, keyword side may not be used !!!
	 if USE_SIDE is not defined, it must be ensured that id of patches not lying on tile sides does not end by 1, 2, 3, 4, 5, 6;
	 currently patches are used only on tile sides */
 
#define USE_SIDE                            /* specify tile sides explicitly */

/* note: codes need not be used;
	 they are currently used by t3d for coloring of matching sides only;
	 the original motivation of using codes was to enable to set up mirroring by t3d;
	 but this requires unique global patch numbering which may no be available if tiles are prepared by other preprocessor;
	 therefore the mirroring is handled here */

#define USE_CODE                            /* specify tile codes */

/* simple mirroring can be used to override incompatibility of matching entities due to round-off error
	 or numerical instability (see t3d_split_curve_start|end in curve.c);
	 the problem of numerical instability is temporarily (but not generally) solved in t3d4WT */

//#define USE_SIMPLE                          /* use explicit simple mirroring */

/* max shift (possibly simultaneously in each direction) when resolving (a) touching of ellipsoid with tile sides, (b) touching of intersection ellipse 
	 with tile edges and (c) intersection of ellipsoid with tile edges is defined as MULTIPLE_EPS * epsilon;
	 note that this maximal shift is not considered when performing shift due to side touching but is considered when performing overdetouch;
	 this implies that MULTIPLE_EPS must be at least 1 !!!
	 value smaller than 2 may prevent overdetouch !!!
	 note that it is not likely that ellipsoid shifted when resolving side touching will be shifted in the same direction when resolving touching of intersection 
	 ellipse with tile edges and when resolving intersection of ellipsoid with tile edges;
	 but this cannot be guaranteed !!! */

#define MULTIPLE_EPS          2             /* MULTIPLE_EPS >= 1 (recommended value 2) */

/* warning: meshes obtained for GRID_3D, GRID_2D and DIAGONAL setup are generally different */

/* note: the rules below concerning the size of the octree (and consequently min_epsilon) for GRID_3D and GRID_2D are valid only 
	 if the size of the octree does not account for the elimination of mesh size octree dispersion (in other words only when codes = 0);
	 if octree dispersion is handled the size of octree for GRID_2D may be the same as the size of octree for GRID_3D if
	 k_GRID_3D = k_GRID_2D = k, where 2^k >= num_tiles_1D;
	 k_GRID_3D = (int)(log((int)(cbrt(tiles - 0.5)) + 1) / log(2) + 0.99999999) = (int)(log((int)(sqrt(tiles - 0.5)) + 1) / log(2) + 0.99999999) = k_GRID_2D;
	 for example: tiles = 4: 1 == 1 ===> GRID_3D = GRID_2D
	              tiles = 5: 1 != 2 ===> GRID_3D > GRID_2D
								...
								tiles = 8: 1 != 2 ===> GRID_3D > GRID_2D
								tiles = 9: 2 == 2 ===> GRID_3D = GRID_2D
								...
								tiles = 16: 2 == 2 ==> GRID_3D = GRID_2D
                tiles = 17: 2 != 3 ==> GRID_3D > GRID_2D */

#define GRID_3D               1             /* tiles are setuped into reqular 3D grid ==> smaller octree ==> smaller min_epsion */
#define GRID_2D               2             /* tiles are setuped into reqular 2D grid ==> larger octree ==> larger min_epsilon */
#define DIAGONAL              3             /* tiles are setuped along spacial diagonal of the 1st octant of GCS ==> even larger octree ==> even larger min_epsilon */

#define SETUP                 GRID_3D       /* GRID_3D or GRID_2D or DIAGONAL */

//#define USE_VIRTUAL_ELLIPSOID               /* to overcome singularity along base of biquadratic inexact elliptical surface */

/* WARNING: directive HALF_MAX_TILES is not recommended !

	 if HALF_MAX_TILES is not defined, maximum number of inclusions that may intersect tile side is reduced up to half
	 if all intersecting inclusions have interface;

	 if HALF_MAX_TILES is defined, MAX_TILES may need to be enlarged by one order of magnitude reducing max_side_intersections
	 by factor of 10 !!! */

//#define HALF_MAX_TILES                      /* real number of tiles may not exceed max_tiles / 2 - 1 */

/* advanced user customisable directives */

//#define MIN_MAX_HALVING                   /* halving of intersection ellipse crossing tile edges (also) with respect to min and max diameter (NOT FULLY IMPLEMENTED !!!) */

/* warning: different meshes are obtained for MAX_ALLOWED_SIZE and MAX_PRESENT_SIZE excced size
	 because of different size of root octant and consequently also different min_epsilon */

#define MAX_ALLOWED_SIZE      1             /* maximal allowable size of inclusions is considered ==> lager offset */
#define MAX_PRESENT_SIZE      2             /* maximal present size of inclusion (although possibly not used) is considered ==> smaller offset */
//#define MAX_REAL_EXCESS       3             /* hard to implement ==> not implemented */

#define EXCESS_SIZE           MAX_PRESENT_SIZE      /* MAX_ALLOWED_SIZE or MAX_PRESENT_SIZE */

/* warning: recommended values of offset are based on the assumption that virtual entities do not exceed out of the tile;
	 but this is generally not true and computed offset must be appropriately increased */

#define OFFSET                0.5           /* relative (with respect to tile size) distance between tiles (0.5 <= OFFSET <= 1.0) */

#if SETUP == GRID_2D

/* at least tile offset will ensure the cubic tiles will not overlap in hydrostatic view (vi nor 1 1 1; vi vup 0 0 1);
	 (the best light direction is then render ldir 1 2 8);
	 but this increases octree and thus also min_epsilon;
	 it makes sense to use it only if tiles are cubes */

//#define TILE_OFFSET                         /* if the computed offset is smaller than tile size, use tile size as offset */
#endif

#define MAX_RATE              0.95          /* maximal permissible rate (0.0 << MAX_RATE < 1.0) */

/* the larger INC_RATE the less danger of singularity along base but the large deviation along curves "at" base */

#define INC_RATE              0.05          /* relative increment of rate for base shift (0.0 < INC_RATE << 1.0) */

/* do not modify setting below unless necessary */

/* note: if zero CURVE_RATE or CURVE_RATE >> 1 is used, there is a risk that elliptical arcs might be discretized just
	 by a single segment coinciding with tile edge;
	 therefore density 2.0 specification is used for elliptical arcs starting and ending at the same tile edge;
	 it may be desirable to specify appropriate density also for some arcs starting and ending on different tile edges;
	 in such a case, however, the orientation of the intersection ellipse with respect to tile edges is relevant for approriate setting;
	 I do rely on the fact that surface rate will ensure that these arcs are discretized by sufficient number of segments;
	 if not, appropriate CURVE_RATE (e.g. 1.5) should be used */

#define CURVE_RATE            0.0           /* default rate for t3d curves */

#define RIGHT_ANGLE           0.001         /* cos(89.94 degs) */
#define ZERO_ANGLE            0.9999995     /* cos(0.057 degs) */

#define MAX_ITER              100           /* maximum number of iterations for projection to ellipsoid and ellipse */

//#define PARAMETRIC_ROUTINES                 /* not really needed */

#define ITEMS_PER_LINE        10            /* number of ids on t3d input line */

#define BUFFER_SIZE           1024

#define INDENT                "       "

#define REMARK                '#'
#define NEWLINE               '\n'
#define CR                    '\r'
#define SPACE                 ' '

#define SEP_STRING            "################################################################################"

/* there may be realized some (not negligible) savings on the size of produced t3d input file
	 by replacing format %.13e by %.14g (both output 14 significant digits);
	 precision of g format should not exceed 17 (it seems that double is stored using 17 significant digits);
	 when using larger precision, a random rubish is printed which prevents removal of zeroes from fractional part
	 resulting in undesirable increase of file size;
	 for example:
	 printf("%.17g\n", 1.8); ==> 1.8
	 printf("%.18g\n", 1.8); ==> 1.80000000000000007
	 note: when using format %A.Be, A=B+7 assuming that only two-digit exponent is used */

//#define FORMAT                "%20.13e"
//#define FORMAT                "%.13e"
#define FORMAT                "%.15g"

#define FMT1                  FORMAT
#define FMT2                  FORMAT" "FORMAT
#define FMT3                  FORMAT" "FORMAT" "FORMAT



#define NOTIFICATION          0
#define WARNING               1
#define ALARM                 2
#define GENERAL_ERROR         3
#define FILE_OPEN_ERROR       4
#define FILE_READ_ERROR       5
#define FILE_WRITE_ERROR      6
#define MEMORY_ERROR          7
#define INPUT_ERROR           8

#define EPSILON               1.0e-10
#define EPSILON_T             1.0e-5
#define EPSILON_U             1.0e-5
#define EPSILON_V             1.0e-5

#define SQRT2                 1.41421356237309504880168872420969808          /* sqrt(2.0) */
#define SQRT1_2               0.707106781186547524400844362104849039         /* sqrt(1.0 / 2.0) */
#define PI_2                  1.57079632679489661923132169163975144          /* PI / 2.0 */
#define PI_4                  0.78539816339744830961566084581987572          /* PI / 4.0 */
#define LN_2                  0.693147180559945309417232121458176568         /* log(2.0) */

#ifdef DEBUG
#define DEBUG_PRINT

#ifdef ELIXIR
#define DEBUG_DRAW
#endif

//#define INEXACT_BG_SURFACE                  /* enforce inexact representation even for ellipsoid with center on tile boundary (just for testing) */

//#define IS_TESTING                          /* intersection testing (for is_*.geom) */

#ifdef IS_TESTING

#if 1
#undef ZERO_ANGLE
#define ZERO_ANGLE            0.9           /* for some test examples only (see is_x_*.geom) */
#endif

/* rather than preparing intersection tests for all possible positions of inclusion in the tile,
	 only basic positions (corresponding to MIRROR_X = MIRROR_Y = MIRROR_Z = 0) are prepared
	 which can be (multiply) mirrored according to MIRROR_X|Y|Z setting;
	 IMPORTANT: since the mirroring is implemented as change of sign, it is applicable only if tile center is at origin !!! */ 

#define MIRROR_X              0             /* 0 or 1 */
#define MIRROR_Y              0             /* 0 or 1 */
#define MIRROR_Z              0             /* 0 or 1 */

/* rather than preparing intersection tests for all tile sides (x, y, z),
	 only intersections with side x (corresponding to CYCLE_COORD = 0) are prepared
	 which can be cyclically (x -> y -> z -> x -> ...) exchanged CYCLE_COORD % 3 times;
	 IMPORTANT: cyclic exchange is applicable only for tile centered at origin !!! */

#define CYCLE_COORD           0             /* 0 or 1 or 2 */

/* to check all positions at once use PER_REP_ALL;
	 this requires that inclusion in particular tile is used only once !!! */

#define PER_REP_ALL

#ifndef PER_REP_ALL

/* rather than modifying intersection tests for periodic repetitions
	 only primary instance of inclusion (corresponding to PER_REP_X = PER_REP_Y = PER_REP_Z = 0) is used
	 which can be periodically repeated according to PER_REP_X|Y|Z setting */

#define PER_REP_X             0             /* 0 or 1 */
#define PER_REP_Y             0             /* 0 or 1 */
#define PER_REP_Z             0             /* 0 or 1 */
#endif

#endif /* IS_TESTING */

//#define SPECIAL_TESTING                     /* special handling testing (for special_*.geom) */

#ifdef SPECIAL_TESTING

#undef MAX_RATE
#define MAX_RATE              0.90          /* for test of special handling only !!! (see special_intersection_*.geom) */

#define XX                    0
#define YY                    1
#define ZZ                    2

#define SPLITTING_PLANE       XX   /* XX or YY or ZZ */     /* do not use X, Y, Z which are defined in enum which is not known in preprocessing */
#define OPPOSITE_PLANE        0   /* 0 or 1 */              /* surprisingly it worked with YES and NO (maybe because of Elixir) */

#endif /* SPECIAL_TESTING */
#else /* DEBUG */

#ifdef ELIXIR
#undef ELIXIR                               /* suppress graphics if not in DEBUG mode */
#endif

#endif /* DEBUG */

#ifdef ELIXIR

#include "Esimple.h"
#undef NO
#undef YES

#endif /* ELIXIR */

/* IMPORTANT: it is generally not possible to use several geometrically identical inclusions occupying identical position (including periodic repetition);
	 since the inclusions are processed in order corresponding to their id, it is likely that sides with the same code using different inclusion 
	 would be topologically not identical and thus not mirrorable;

	 o---11111---222-----333333---o    222 - intersection of inclusion 2 with top side
   |                            |          inclusions 2 is processed as the 2nd inclusion (from 3 inclusions intersecting top side)

	 o---11111---444-----333333---o    444 - intersection of inclusion 4 with top side
   |                            |          inclusions 4 is processed as the 3rd inclusion (from 3 inclusions intersecting top side)

#ifdef DEBUG
	 if ALLOW_OVERLAP is not defined and CHECK_OVERLAP is defined, then identical inclusion occupying identical
	 position (including periodic repetition) should be detected in the check
#endif */

/* IMPORTANT: almost touching intersection must be resolved on the level of model;
	 it is not enough to only degenerate intersection into touching point;
	 this could be done only if inclusion exceeds tile side by less than min_epsilon;
	 for example if inclusion exceeds tile side by less than epsilon but its intersection with tile side is (possibly much) larger than epsilon
	 then it is not possible to pretend that the intersection takes place at the touching point only;
	 if the elements to be generated are smaller than the real intersection, the mesh generated over the inclusion would penetrate
	 through tile side, which is not permissible;
	 it is therefore necessary to perform some action, eg. shifting, isotropic scaling, anisotropic scaling, .... 


            ____-----------____             ----                              _o------o_     _o------o_
       __---                   ---__           | < eps    generated         _-          -_ _-          -_
   ---x-----------------------------x---    ----           ======>     -------------------x---------------------
     -                               -                      mesh         _-            touching            -_
                                                                        o               point                o
      |        >> epsilon           |
      |-----------------------------|          
                > msz

   note that there is currently no way how to resolve the case when inclusion exceeds tile by more than epsilon but its intersection with
	 tile side is less than minimum (problem driven) epsilon;

	 IMPORTANT: there is currently no good way how to ensure that boundary nodes on the same entity (matrix) will get closer to each other than min_epsilon
	 (which is likely to result in t3d error) if inclusion is touching or intersecting tile side;
	 for example

      ____                       ____
          o----____     ____----o           ---- < min_epsilon
	 -------o----------x----------o--------   ----

	 the only solution is to use OVERDETOUCH */

/* concept of halving solves for the problem when elliptical arc to be constructed has end points on (or very close to) the diameter of the ellipse;
	 such arc cannot be constructed by fixing a curve between the end points to a quadratic curve constructed using appropriate extreme points
	 of ellipse with respect to 2 perpendicular directions because the weight of the ctrl point (of the fixed curve) is approaching zero, while
	 the coordinates of the ctrl point are approaching infinity, which makes the geometry unstable;
	 such ellipse must be halved, which results in spliting of the elliptical arc;
	 the halving of the ellipse is performed either with respect to ellipse x|y|z-extreme points or min|max-extreme points;
	 note that I am not using cubic representation for half-ellipticxal physical curve (on ellipse intersecting tile edges)
	 because the decision whether the elliptic arc is half-ellipse is based on floating point operations;
	 moreover this case is not very likely */


#define B_0(T) (1.0-2.0*(T)+(T)*(T))
#define B_1(T) (2.0*(1.0-(T))*(T))
#define B_2(T) ((T)*(T))

#define dB_0(T) (-2.0+2.0*(T))
#define dB_1(T) (2.0-4.0*(T))
#define dB_2(T) (2.0*(T))


#define swap(VAL1, VAL2, TMP)              {(TMP) = (VAL1); (VAL1) = (VAL2); (VAL2) = (TMP);}

#define size_vec(VEC)                      ((VEC).x * (VEC).x + (VEC).y * (VEC).y + (VEC).z * (VEC).z)

#define mul_vec(VEC, VALUE)                {(VEC).x *= (VALUE); (VEC).y *= (VALUE); (VEC).z *= (VALUE);}
#define div_vec(VEC, VALUE)                {(VEC).x /= (VALUE); (VEC).y /= (VALUE); (VEC).z /= (VALUE);}

#define copy_vec(DEST, SRC)                {(DEST).x = (SRC).x; (DEST).y = (SRC).y; (DEST).z = (SRC).z;}

#define copy_mul_vec(DEST, SRC, VALUE)     {(DEST).x = (SRC).x * (VALUE); (DEST).y = (SRC).y * (VALUE); (DEST).z = (SRC).z * (VALUE);}
#define copy_div_vec(DEST, SRC, VALUE)     {(DEST).x = (SRC).x / (VALUE); (DEST).y = (SRC).y / (VALUE); (DEST).z = (SRC).z / (VALUE);}

#define add_vec(RES, VEC1, VEC2)           {(RES).x = (VEC1).x + (VEC2).x; \
                                            (RES).y = (VEC1).y + (VEC2).y; \
                                            (RES).z = (VEC1).z + (VEC2).z;}
#define sub_vec(RES, VEC1, VEC2)           {(RES).x = (VEC1).x - (VEC2).x; \
                                            (RES).y = (VEC1).y - (VEC2).y; \
                                            (RES).z = (VEC1).z - (VEC2).z;}

#define add_mul_vec(RES, VEC1, VEC2, VAL1, VAL2)   {(RES).x = (VEC1).x * (VAL1) + (VEC2).x * (VAL2); \
		                                                (RES).y = (VEC1).y * (VAL1) + (VEC2).y * (VAL2); \
                                                    (RES).z = (VEC1).z * (VAL1) + (VEC2).z * (VAL2);}

#define dot_product(VEC1, VEC2)            ((VEC1).x * (VEC2).x + (VEC1).y * (VEC2).y + (VEC1).z * (VEC2).z)
#define cross_product(RES, VEC1, VEC2)     {(RES).x = (VEC1).y * (VEC2).z - (VEC1).z * (VEC2).y; \
                                            (RES).y = (VEC1).z * (VEC2).x - (VEC1).x * (VEC2).z; \
                                            (RES).z = (VEC1).x * (VEC2).y - (VEC1).y * (VEC2).x;}
#define comp_product(RES, VEC1, VEC2)      {(RES).x = (VEC1).x * (VEC2).x; \
                                            (RES).y = (VEC1).y * (VEC2).y; \
                                            (RES).z = (VEC1).z * (VEC2).z;}

#define dist_point(PNT1, PNT2)             (((PNT1).x - (PNT2).x) * ((PNT1).x - (PNT2).x) + \
                                            ((PNT1).y - (PNT2).y) * ((PNT1).y - (PNT2).y) + \
                                            ((PNT1).z - (PNT2).z) * ((PNT1).z - (PNT2).z))
#define aver_point(RES, PNT1, PNT2)        {(RES).x = ((PNT1).x + (PNT2).x) / 2.0; \
																						(RES).y = ((PNT1).y + (PNT2).y) / 2.0; \
                                            (RES).z = ((PNT1).z + (PNT2).z) / 2.0;}


#define array(MAT,ROWS,COLS,ROW,COL)       MAT[(ROW)*(COLS)+(COL)]

#define min_val(VAL1, VAL2)                ((VAL1) < (VAL2) ? (VAL1) : (VAL2))
#define max_val(VAL1, VAL2)                ((VAL1) > (VAL2) ? (VAL1) : (VAL2))



#ifdef PARAMETRIC_ROUTINES
/* caution: shift, scale and expand macros do not modify ctrl points of ellipsoid and ellipse */
#endif

/* caution: shift, scale and expand macros do not modify extreme points and diffs */

#define shift_ellipsoid(ELLIPSOID, SHIFT_X, SHIFT_Y, SHIFT_Z)   {(ELLIPSOID).center.x += SHIFT_X; \
                                                                 (ELLIPSOID).center.y += SHIFT_Y; \
                                                                 (ELLIPSOID).center.z += SHIFT_Z;}

#define scale_ellipsoid(ELLIPSOID, SCALE)  {(ELLIPSOID).max *= SCALE; \
                                            (ELLIPSOID).min *= SCALE; \
                                            (ELLIPSOID).mid *= SCALE;}

#define EXPAND_ELLIPSOID(ELLIPSOID, EPS)   {(ELLIPSOID).max += EPS; \
                                            (ELLIPSOID).min *= EPS; \
                                            (ELLIPSOID).mid *= EPS;}


#define shift_ellipse(ELLIPSE, SHIFT_X, SHIFT_Y, SHIFT_Z)   {(ELLIPSE).center.x += SHIFT_X; \
                                                             (ELLIPSE).center.y += SHIFT_Y; \
                                                             (ELLIPSE).center.z += SHIFT_Z;}

#define scale_ellipse(ELLIPSE, SCALE)      {(ELLIPSE).max *= SCALE;	\
                                            (ELLIPSE).min *= SCALE;}

#define expand_ellipse(ELLIPSE, EPS)       {(ELLIPSE).max += EPS;	\
                                            (ELLIPSE).min -= EPS;}


/* unique global numbering is introduced to allow for efficient mirroring;
	 since the mirroring is applied only between side paches (tile side, intersection of inclusions with tile side)
	 it is enough to use unique global numbering for patches;

	 1st concept:
	 inclusion_id | repetition | tile_id | face     for intersection of inclusion with tile side
            	               | tile_id | face     for rest of tile side

   face: 1 - 6 : 1 decadic digit
   tile_id: 1 - (MAX_TILES - 1) : typically 2 (MAX_TILES == 100) or 3 (MAX_TILES == 1000) decadic digits
	 repetition: 0 - 3 : 1 decadic digit (repetition is reduced from 8 cases in 3D space to 4 cases on particular 2D face)
	 inclusion_id: 1 - 21474 (if MAX_TILES == 1000) for 32 bit signed long ==> insufficient !!!
                 1 - 214747 (if MAX_TILES == 100) for 32 bit signed long ==> may be sufficient if boundary inclusions are numbered first

	 note: I cannot used unsigned long as t3d does not support unsigned long ids
	       I cannot rely on 64 bit platform as long has 32 bits on LLP64 (Windows)
         ==> max signed long = 2147483648

	 2nd concept:
	 local_patch_id | tile_id | face 

	 local_patch_id: 1 - 214748 (if MAX_TILES == 1000 and if NWT <= 364) for 32 bit signed long ==> reasonable
                   1 - 214747 (if MAX_TILES == 1000 and if NWT > 364) for 32 bit signed long ==> reasonable
									 1 - 2147483 (if MAX_TILES == 100 and if NWT <= 64) for 32 bit signed long ==> sufficient
									 1 - 2147482 (if MAX_TILES == 100 and if NWT > 64) for 32 bit signed long ==> sufficient

	 local_patch_id is separate for each side !!!
                   
	 note: this implies that tile side may be intersected by up to 2147483 inclusions (including repeated inclusions) ===> max_side_intersections

#ifdef HALF_MAX_TILES
	 alternatively it may be possible to use

	 local_patch_id | tiles + tile_id | face

	 but this is possible only if tiles < MAX_TILES / 2 (so that tiles + tile_id < MAX_TILES);
#endif

   number of possible code combinations in a single direction: codes^2;
	 number of possible code combinations in a all directions: codes_x^2 * codes_y^2 * codes_z^2;
	 number of possible code combinations in a all directions for codes_x = codes_y = codes_z = codes: codes^6;

#ifdef HALF_MAX_TILES 
	 for codes = 1: 1^6 = 1 ==> MAX_TILES = 10
	 for codes = 2: 2^6 = 64 >= 100 / 2 ==> MAX_TILES = 1000 ==> max_side_intersections = 214748 ==> reasonable
	 for codes = 3: 3^6 = 729 >= 1000 / 2 ==> MAX_TILES = 10000 ==> max_side_intersections = 21474 ==> insufficient !!!
#else 	 
	 for codes = 1: 1^6 = 1 ==> MAX_TILES = 10
	 for codes = 2: 2^6 = 64 ==> MAX_TILES = 100 ==> max_side_intersections = 2147483/2 = 1073741 ==> sufficient
	 for codes = 3: 3^6 = 729 ==> MAX_TILES = 1000 ==> max_side_intersections = 214747/2 = 107373 ==> reasonable
#endif

   IMPORTANT: digits for side identification (face) should be nonzero, because zero side specification is reserved (in t3d) for entities
	 not being on tile side !!! */

#ifdef ARBITRARY_TILE_ID
/* IMPORTANT: macros get_x|y|z_side_patch_id are using fictitious tile id !!! */
#endif

#if 0

#ifdef HALF_MAX_TILES
#define get_x_side_patch_id(IND) ((x_side_patches[IND] + 1) * max_tiles + tile_id) * 10 + ((IND == 0) ? XN : XP + 1);
#define get_y_side_patch_id(IND) ((y_side_patches[IND] + 1) * max_tiles + tile_id) * 10 + ((IND == 0) ? YN : YP + 1);
#define get_z_side_patch_id(IND) ((z_side_patches[IND] + 1) * max_tiles + tile_id) * 10 + ((IND == 0) ? ZN : ZP + 1);
#else
#define get_x_side_patch_id(IND) ((x_side_patches_total[IND] + 1) * max_tiles + tile_id) * 10 + ((IND == 0) ? XN : XP + 1);
#define get_y_side_patch_id(IND) ((y_side_patches_total[IND] + 1) * max_tiles + tile_id) * 10 + ((IND == 0) ? YN : YP + 1);
#define get_z_side_patch_id(IND) ((z_side_patches_total[IND] + 1) * max_tiles + tile_id) * 10 + ((IND == 0) ? ZN : ZP + 1);
#endif

#else

#ifdef HALF_MAX_TILES
#define get_x_side_patch_id(IND) ((x_side_patches[IND] + 1) * max_tiles + tile_id) * 10 + 2 * X + (IND) + 1;
#define get_y_side_patch_id(IND) ((y_side_patches[IND] + 1) * max_tiles + tile_id) * 10 + 2 * Y + (IND) + 1;
#define get_z_side_patch_id(IND) ((z_side_patches[IND] + 1) * max_tiles + tile_id) * 10 + 2 * Z + (IND) + 1;
#else
#define get_x_side_patch_id(IND) ((x_side_patches_total[IND] + 1) * max_tiles + tile_id) * 10 + 2 * X + (IND) + 1;
#define get_y_side_patch_id(IND) ((y_side_patches_total[IND] + 1) * max_tiles + tile_id) * 10 + 2 * Y + (IND) + 1;
#define get_z_side_patch_id(IND) ((z_side_patches_total[IND] + 1) * max_tiles + tile_id) * 10 + 2 * Z + (IND) + 1;
#endif

#endif


typedef enum logic logic;
typedef enum axes axes;
typedef enum face face;
typedef enum position position;

typedef enum vertex_status vertex_status;
typedef enum corner_status corner_status;

typedef struct vector_rec vector_rec;
typedef struct vector_rec point_rec;
typedef struct ellipsoid_rec ellipsoid_rec;
typedef struct ellipse_rec ellipse_rec;

typedef struct tile_rec tile_rec;
typedef struct code_rec code_rec;
typedef struct inclusion_rec inclusion_rec;

typedef struct is_point_rec is_point_rec;
typedef struct is_pnt_rec is_pnt_rec;
typedef struct face_is_rec face_is_rec;
typedef struct edge_is_rec edge_is_rec;

typedef struct trans_matrix trans_matrix;


enum logic{
	NO = 0,
	YES = 1
};


enum axes{
	X = 0,
	Y = 1,
	Z = 2
};


#ifdef MIN_MAX_HALVING

/* IMPORTANT: MIN and MAX are defined as continuation of axes */

#define MIN    3
#define MAX    4
#endif


enum face{
	XN = 0,                             /* x negative */
	XP = 1,                             /* x positive */
	YN = 2,                             /* y negative */
	YP = 3,                             /* y positive */
	ZN = 4,                             /* z negative */
	ZP = 5                              /* z positive */
};


/* ordering of quantities related to y and z for x_ellipse */

#define XY      0
#define XZ      1

/* ordering of quantities related to z and x for y_ellipse */

#define YZ      0
#define YX      1

/* ordering of quantities related to x and y for z_ellipse */

#define ZX      0
#define ZY      1


enum position{
	INDEF = 0,                          /* indefinite */
	POS = 1,                            /* positive */
	NEG = 2,                            /* negative */
	BOTH = 3                            /* positive and negative */
};


enum vertex_status{
	START = -1,                         /* starting */
	TOUCH = 0,                          /* touching */
	END = 1                             /* ending */
};


enum corner_status{
	IN = -1,                            /* inside */       /* CAUTION: INSIDE conflicts with Elixir */
	BND = 0,                            /* boundary */
	OUT = 1                             /* outside */
};


/* shifting strategies are cycled according to their ordering in the enumeration;
	 ==> do not assign numbers to shifting strategies !!!
	 any strategy can be chosen as initial one;
	 strategies to be not used should be placed after MAX_STRATEGY and before DUMMY_STRATEGY */

enum shift_strategy{                  /* IMPORTANT: do not assign numbers !!! */
	TOUCHING,                           /* try enforce ellipse touching edge - cannot be used as the only strategy !!! */
	TOTAL,                              /* minimal total shift */
	OPPOSITE,                           /* opposite shift (usually behaves as TOTAL) */
	MINIMAL,                            /* minimal shift */
	MAXIMAL,                            /* maximal shift */

	MAX_STRATEGY,                       /* number of available strategies */

/* ANY_TO_BE NOT_USED, */

	DUMMY_STRATEGY                      /* must be last */
};


struct vector_rec{                    /* <==> point_rec */
	double x;
	double y;
	double z;
};


struct ellipsoid_rec{
	point_rec center;                   /* ellipsoid center - may not fall outside reference tile */
	vector_rec major, minor, middle;    /* righthanded Cartesian LCS */
	double max, min, mid;               /* size of half-axes (max < (size of tile - epsilon) / MAX_RADIUS_RATE, min > 10.0 * epsilon, min <= mid <= max) */
#ifdef PARAMETRIC_ROUTINES
	point_rec pnt00, pnt01, pnt02;      /* ctrl points in GCS for positive 1/8 (in LCS) of ellipsoid modelled as bi-quadratic surface */
	point_rec pnt10, pnt11, pnt12;
	point_rec pnt20, pnt21, pnt22;
                                                                                
#endif
};


struct ellipse_rec{
	point_rec center;                   /* ellipse center */
	vector_rec major, minor, normal;    /* righthanded Cartesian LCS */
	double max, min;                    /* size of half-axes (min <= max) (both zero if degenerated into point) */
#ifdef PARAMETRIC_ROUTINES
	point_rec pnt0, pnt1, pnt2;         /* ctrl points in GCS for positive 1/4 (n LCS) of ellipse modelled as quadratic curve */
#endif
};


struct code_rec{
	long id;                            /* code number */
	long tile_id;                       /* tile id (fictitious) using code for the first time */
	long tile_id_orig;                  /* tile id (real) using code for the first time */
	face side;                          /* tile side */
	long patches;                       /* number of side patches */
	long vertices;                      /* number of side vertices (fixed) */
	long curves[4];                     /* number of edge curves ordered as
																				 xz-, xz+, xy-, xy+ for x-side
																				 yx-, yx+, yz-, yz+ for y-side
																				 zy-, zy+, zx-, zx+ for z-side
																				 where ab- means edge shared by a-side and by negative b-side */
};


struct tile_rec{
	long id;                            /* tile id */
	point_rec size;                     /* size of reference tile */
	point_rec center;                   /* center of reference tile */
	point_rec corner[2];                /* opposite (min and max) corners of reference tile */
	double imsz, bmsz;                  /* internal and boundary default mesh size of reference tile */
	int mat;                            /* matrix material (0 < mat) */
	int code[6];                        /* side codes of particular tile  */
	long inclusions;                    /* number of inclusions in particular tile (on input) */
	long points;                        /* number of ctrl points in particular tile (on input) */
};


struct inclusion_rec{
	long id;                            /* inclusion id */
	ellipsoid_rec ellipsoid;            /* inclusion geometry */
	double imsz, bmsz;                  /* internal and boundary default mesh size of inclusion */
	int mat;                            /* inclusion material (0 <==> hole) used as inclusion property id */
	int ifmat;                          /* interface material (0 <==> no interface) used as interface property id */
	face_is_rec *face_is[3];            /* intersection with tile faces (NULL <==> no intersection) */
	edge_is_rec *edge_is[3];            /* intersection with tile edges (NULL <==> no intersection) */
	corner_status corner;               /* position of tile corner with respect to inclusion */
	logic active[8];                    /* active periodic repetitions */ 
	logic support;                      /* flag whether ellipsoid geometry was already output */
	logic shift;                        /* flag whether inclusion was shifted ==> intersection check needed even if overlapping in reference tile is not allowed */
#ifdef ALLOW_OUTSIDE
	logic out[3];                       /* flag whether inclusion center is outside of reference tile */
#endif
};


/* IMPORTANT: since vertex id is valid when processing a single inclusion only (then all vertex ids corresponding to all
	 tile edge intersection points of that inclusion are zeroed again), additional array x|y|z_edge_vertex_id_array has been introduced;
	 to save some memory, struct is_pnt_rec was introduced to allow to store x|y|z_edge_vertex_data_array without vertex id */

struct is_pnt_rec{                    /* intersection point data */    
	double coord;                       /* coordinate along edge */
	vertex_status status;               /* status */
};
	

struct is_point_rec{                  /* intersection point */
	int id;                             /* vertex id - CAUTION: valid for processing of a single inclusion only */
	is_pnt_rec data;                    /* intersection point data */  
};


struct face_is_rec{                   /* intersection with tile face */
	ellipse_rec ellipse;                /* intersection ellipse */
	position side[2];                   /* position of ellipse center with respect to tile corner ordered as
                                         y, z for x_ellipse (XY, XZ)
														      			 z, x for y_ellipse (YZ, YX)
														      			 x, y foz z_ellipse (ZX, ZY) */
	logic edge_half[2];                 /* flag that ellipse is (almost) halved by tile edge ordered as
                                         y, z for x_ellipse (XY, XZ)
														      			 z, x for y_ellipse (YZ, YX)
														      			 x, y foz z_ellipse (ZX, ZY) */
	logic cross_half;                   /* flag that ellipse is (almost) halved by connection of adjacent intersection points */
	int cross_halving;                  /* halving (X or Y or Z or MIN or MAX) if cross_half == YES */
	point_rec *extreme_pnt;             /* extreme points of intersection ellipse ordered as
                                         y-, y+, z-, z+ for x_ellipse (2*XY+0, 2*XY+1, 2*XZ+0, 2*XZ+1)
                                         z-, z+, x-, x+ for y_ellipse (2*YZ+0, 2*YZ+1, 2*YX+0, 2*YX+1)
                                         x-, x+, y-, y+ for z_ellipse (2*ZX+0, 2*ZX+1, 2*ZY+0, 2*ZY+1)
														      			 needed really only if intersection ellipse is intersecting or touching tile edge(s) */
	double weight;                      /* weight for arc between the 1st (2nd) and 4th (3rd) extreme point;
														      			 weight for arc between the 3rd (4th) and 1st (2nd) extreme point = sqrt(1.0 - weight * weight) */
};


struct edge_is_rec{                   /* intersection with tile edge */
	int count;                          /* number of intersection points */
	is_point_rec *point;                /* intersection points ordered increasingly with respect to coord */
	position side;                      /* position of intersection points with respect to tile corner */
};


struct trans_matrix{
	vector_rec x;
	vector_rec y;
	vector_rec z;
};


static char *geom_file_name = NULL;
static char *t3d_in_file_name = NULL;

static FILE *geom_file = NULL;
static FILE *t3d_in_file = NULL;

static double epsilon = 0.0;
static double min_epsilon = 0.0;
static int max_subdivisions = 0;

static char line_buffer[BUFFER_SIZE];

static tile_rec tile;

static inclusion_rec *inclusion_array = NULL;
static code_rec *code_array = NULL;

static char hole_spec[8], output_spec[16], mirror_spec[32], side_spec[8];

static double min_sz, max_sz;

/* static global variables for routines
	 get_inclusion_tile_intersection,
	 process_x|y|z_side_geometry, 
	 check_inclusion_outside_tile_x|y|z_side;

	 IMPORTANT: be aware of eventual variable shadowing !!! */

static long max_tiles = 1;                                                      /* maximum number of tiles */
static long max_side_intersections = 0;                                         /* maximum number of inclusions intersecting a single tile side */

static long max_tile_regions = 0;                                               /* maximum number of regions in tile */
static long max_side_patches[3] = {0, 0, 0};                                    /* maximum number of patches on tile side */                     
static long max_side_vertices[3] = {0, 0, 0};                                   /* maximum number of fixed vertices on tile side */
static long max_edge_vertices[3] = {0, 0, 0};                                   /* maximum number of fixed vertices on tile edge */
static long max_edge_curves_1[3] = {0, 0, 0};                                   /* maximum number of segments correspodning to intersection of inclusion with tile edge */
static long max_edge_curves_2[3] = {1, 1, 1};                                   /* maximum number of segments between tile corners and intersections and 
																																									 touchings of inclusion with tile edge */

static long tile_regions;                                                       /* related to tile_region_array */
static long x_side_patches[2], y_side_patches[2], z_side_patches[2];            /* related to x|y|z_side_patch_array */
static long x_side_vertices[2], y_side_vertices[2], z_side_vertices[2];         /* related to x|y|z_side_vertex_array */
static long x_edge_vertices[4], y_edge_vertices[4], z_edge_vertices[4];         /* related to x|y|z_edge_vertex_array */
static long x_edge_curves_1[4], y_edge_curves_1[4], z_edge_curves_1[4];         /* related to x|y|z_edge_curve_1_array */
static long x_edge_curves_2[4], y_edge_curves_2[4], z_edge_curves_2[4];         /* related to x|y|z_edge_curve_2_array */

#ifndef HALF_MAX_TILES
static long x_side_patches_total[2], y_side_patches_total[2], z_side_patches_total[2];    /* total number of side patches needed to generate global unique patch id */
#endif

static long *tile_region_array = NULL;
static long *x_side_patch_array[2] = {NULL, NULL};                              /* ordering x-, x+ */ 
static long *y_side_patch_array[2] = {NULL, NULL};                              /* ordering y-, y+ */
static long *z_side_patch_array[2] = {NULL, NULL};                              /* ordering z-, z+ */
static long *x_side_vertex_array[2] = {NULL, NULL};                             /* ordering x-, x+ */
static long *y_side_vertex_array[2] = {NULL, NULL};                             /* ordering y-, y+ */
static long *z_side_vertex_array[2] = {NULL, NULL};                             /* ordering z-, z+ */
static long *x_edge_vertex_id_array[4] = {NULL, NULL, NULL, NULL};              /* ordering y-z-, y+z-, y-z+, +y+z */
static long *y_edge_vertex_id_array[4] = {NULL, NULL, NULL, NULL};              /* ordering z-x-, z+x-, z-x+, +z+x */
static long *z_edge_vertex_id_array[4] = {NULL, NULL, NULL, NULL};              /* ordering x-y-, x+y-, x-y+, +x+y */
static is_pnt_rec **x_edge_vertex_data_array[4] = {NULL, NULL, NULL, NULL};     /* ordering y-z-, y+z-, y-z+, +y+z */
static is_pnt_rec **y_edge_vertex_data_array[4] = {NULL, NULL, NULL, NULL};     /* ordering z-x-, z+x-, z-x+, +z+x */
static is_pnt_rec **z_edge_vertex_data_array[4] = {NULL, NULL, NULL, NULL};     /* ordering x-y-, x+y-, x-y+, +x+y */
static long *x_edge_curve_1_array[4] = {NULL, NULL, NULL, NULL};                /* ordering y-z-, y+z-, y-z+, +y+z */
static long *y_edge_curve_1_array[4] = {NULL, NULL, NULL, NULL};                /* ordering z-x-, z+x-, z-x+, +z+x */
static long *z_edge_curve_1_array[4] = {NULL, NULL, NULL, NULL};                /* ordering x-y-, x+y-, x-y+, +x+y */
static long *x_edge_curve_2_array[4] = {NULL, NULL, NULL, NULL};                /* ordering y-z-, y+z-, y-z+, +y+z */
static long *y_edge_curve_2_array[4] = {NULL, NULL, NULL, NULL};                /* ordering z-x-, z+x-, z-x+, +z+x */
static long *z_edge_curve_2_array[4] = {NULL, NULL, NULL, NULL};                /* ordering x-y-, x+y-, x-y+, +x+y */


static long tile_id;

static long vertex_id = 0;
static long curve_id = 0;
static long surface_id = 0;
static long region_id = 0;
static long patch_id = 0;
static long shell_id = 0;
static long interface_id = 0;

static long origin_id, corner_id, cor_id;

static long x_patch_id, y_patch_id, z_patch_id;
static long x_curve_id, y_curve_id, z_curve_id;

static int shifting_x[8] = {0, 1, 0, 1, 0, 1, 0, 1};
static int shifting_y[8] = {0, 0, 1, 1, 0, 0, 1, 1};
static int shifting_z[8] = {0, 0, 0, 0, 1, 1, 1, 1};

static int x_ind, y_ind, z_ind;                      /* indices of tile sides {0;1} closer to inclusion center (actual position) */

static double shift_x, shift_y, shift_z;             /* shift of actual inclusion */

static long x_side_curve_array[3], x_side_curves;
static long y_side_curve_array[3], y_side_curves;
static long z_side_curve_array[3], z_side_curves;

static long side_vertex_array[3], side_vertices;

static face_is_rec *x_face_is = NULL, *y_face_is = NULL, *z_face_is = NULL;
static edge_is_rec *x_edge_is = NULL, *y_edge_is = NULL, *z_edge_is = NULL;

static int xx_ind, yy_ind, zz_ind;

static logic no_shift = NO, shift_done = NO;

static int ref_position[2] = {POS, NEG}, sign[2] = {1, -1};

static inclusion_rec *inclusion = NULL;
static ellipsoid_rec *ellipsoid = NULL;
static code_rec *code = NULL;

static logic x_touching, y_touching, z_touching;     /* flags that ellipsoid is touching tile side */
static logic x_intersection, y_intersection, z_intersection;   /* flags that ellipsoid is intersecting tile side */
static logic x_axis, y_axis, z_axis;                 /* flags that intersection with tile edge has beeen processed */

static vector_rec dir_x = {1.0, 0.0, 0.0};           /* GCS base vectors */
static vector_rec dir_y = {0.0, 1.0, 0.0}; 
static vector_rec dir_z = {0.0, 0.0, 1.0};

static vector_rec dir_1, dir_2, dir_3;               /* GCS base vectors in LCS of ellipsoid, then transformed to sphere space */
static vector_rec vec_x, vec_y, vec_z;               /* normals of tile sides in LCS of ellipsoid after being transformed to sphere space */
static vector_rec dir_a, dir_b, dir_c;               /* 3rd of base vectors of tiles sides in LCS of ellipsoid  after being transformed to sphere space */

static double diff_x, diff_y, diff_z;                /* differences of ellipsoid from tile sides */
static double diff_xy, diff_xz;                      /* differences of intersection x-ellipse from tile edges */
static double diff_yz, diff_yx;                      /* differences of intersection y-ellipse from tile edges */
static double diff_zx, diff_zy;                      /* differences of intersection z-ellipse from tile edges */
static double x_diff, y_diff, z_diff;                /* differences of intersection with tile edges from tile corner */
static double x_coord[2], y_coord[2], z_coord[2];    /* intersections with tile edges */

static int ind_x, ind_y, ind_z;                      /* indices of tile sides {0;1} closer to ellipsoid center (reference position) */
static int x_count, y_count, z_count;                /* number of intersection of elliposid with tile edges */

static point_rec ext_point_x[2], ext_point_y[2], ext_point_z[2];  /* extreme points of ellipsoid in GCS */
static point_rec extreme_pnt_xy, extreme_pnt_xz;     /* relevant extreme points of x_ellipse in GCS */
static point_rec extreme_pnt_yz, extreme_pnt_yx;     /* relevant extreme points of y_ellipse in GCS */
static point_rec extreme_pnt_zx, extreme_pnt_zy;     /* relevant extreme_points of z_ellipse in GCS */
static point_rec corner_point[2];                    /* opposite tile corner points in LCS and transformed to sphere space */
static point_rec par_corner[2];                      /* opposite (min and max) corners of particular tile subjected to appropriate shift */

static ellipse_rec x_ellipse, y_ellipse, z_ellipse;  /* intersection ellipses */

static trans_matrix trans_g2l, trans_l2g;            /* transformations between GCS and LCS of ellipsoid */
static vector_rec scale_e2s, scale_s2e;              /* transformations between ellipsoid and sphere in LCS */

static logic x_trans, y_trans, z_trans;              /* flags that transformation between GCS and LCS of x|y|z_ellipse is available */

static trans_matrix x_trans_g2l, x_trans_l2g;        /* transformation between GCS and LCS of x_ellipse */
static trans_matrix y_trans_g2l, y_trans_l2g;        /* transformation between GCS and LCS of y_ellipse */
static trans_matrix z_trans_g2l, z_trans_l2g;        /* transformation between GCS and LCS of z_ellipse */

static double radius;                                /* radius of ellipsoid transformed to sphere */
static vector_rec shift_vec;                         /* shifting vector */
static vector_rec total_shift;                       /* total shifting vector */

static logic corner_hole[8];                         /* flag indicating whether tile corner is inside hole */
static logic inside_hole = NO;                       /* flag indicating that there exist a hole inclusion with tile corner inside */
#ifdef CHECK_ISOLATED
static logic corner_isolated[8];                     /* flag indicating whether the tile will contain isolated vertex at the corner */
static logic boundary_hole = NO;                     /* flag indicating that there exist a hole inclusion with tile corner on boundary
																												while each tile edge is intersected twice */
#endif

#ifdef ELIXIR

#define FONT                  "-adobe-courier-bold-r-normal--*-120-*-*-m-*-iso8859-1"

#define TILE_LAYER            1
#define INCLUSION_LAYER       2

//#define SOLID                 /* ellipsoids drawn by solid surface */

#ifdef SOLID
//#define FILLED                /* ellipsoids drawn by solid filled surface */
#endif

#ifdef DEBUG_DRAW
#define DRAW_INCLUSION_CENTER
#define DRAW_ELLIPSE
#define DRAW_EXTREME_POINT_ELLIPSOID
#define DRAW_EXTREME_POINT_ELLIPSE
#define DRAW_TOUCHING_POINT_ELLIPSOID
#define DRAW_INTERSECTING_POINT
#endif

static EPixel green, red, yellow, blue, cyan, magenta, orange, black, white;
static EPixel bgcolor;
static Font font;
static logic draw = NO;

static void
draw_number(point_rec *pnt, int layer, EPixel color, int number, char *prefix);
static void
draw_point(point_rec *pnt, int layer, EPixel color);
static void
draw_line(point_rec *pnt1, point_rec *pnt2, int layer, EPixel color);
static void
draw_elliptic_arc(point_rec *pnt1, point_rec *pnt2, point_rec *pnt3, double w1, double w2, double w3, int layer, EPixel color);
static void
draw_ellipsoidal_surface(point_rec *pnt, double *w, int layer, EPixel color);

static void
draw_elliptic_arc_segment(ellipse_rec *ellipse, int q, double ta, double tb, point_rec *pointa, point_rec *pointb, 
													trans_matrix *trans, int layer, EPixel color);
static void
draw_elliptic_arc_start_segment(ellipse_rec *ellipse, int q, double t, point_rec *point, trans_matrix *trans, int layer, EPixel color);
static void
draw_elliptic_arc_end_segment(ellipse_rec *ellipse, int q, double t, point_rec *point, trans_matrix *trans, int layer, EPixel color);

static void
draw_tile(tile_rec *tile, int layer, EPixel color);
static void
draw_inclusion(inclusion_rec *inclusion, int layer, EPixel color);
static void
draw_ellipsoid(ellipsoid_rec *ellipsoid, int layer, EPixel color);
static void
draw_ellipse(ellipse_rec *ellipse, int layer, EPixel color);

#ifdef __cplusplus
#define CONST_CHAR_PTR           (char *)
#else
#define CONST_CHAR_PTR 
#endif

static void
augment_command_table(void);

static void
select_distance_points(void);
static void
distance_handler(Widget w, EView *v_p, XEvent *event);

#endif

static logic
get_inclusion_tile_intersection(void);

/* special routines called from get_inclusion_tile_intersection
	 and dependent on static global variables */

static void
resolve_ellipsoid_x_side_touching(void);
static void
resolve_ellipsoid_y_side_touching(void);
static void
resolve_ellipsoid_z_side_touching(void);

static void
setup_ellipse_touching_point(ellipse_rec *ellipse, point_rec *point, vector_rec *dir);

static void
get_ellipsoid_x_side_intersection(void);
static void
get_ellipsoid_y_side_intersection(void);
static void
get_ellipsoid_z_side_intersection(void);

static void
get_ellipsoid_xyz_side_intersection(ellipse_rec *ellipse, point_rec *point, vector_rec *nor, vector_rec *tan_1, vector_rec *tan_2);

static void
resolve_ellipsoid_yz_edge_touching(logic min, logic max);
static void
resolve_ellipsoid_zx_edge_touching(logic min, logic max);
static void
resolve_ellipsoid_xy_edge_touching(logic min, logic max);

static logic
get_ellipsoid_shift(double diff_0, double diff_1, double diff_2, double *shift_pos, double *shift_neg, double eps);

static void
get_ellipsoid_yz_edge_intersection(logic check);
static void
get_ellipsoid_zx_edge_intersection(logic check);
static void
get_ellipsoid_xy_edge_intersection(logic check);

static logic
check_ellipsoid_shift_x_corner(double shift, double eps, int *phase);
static logic
check_ellipsoid_shift_y_corner(double shift, double eps, int *phase);
static logic
check_ellipsoid_shift_z_corner(double shift, double eps, int *phase);

static void
ellipsoid_back_shift_x_corner(double shift, int phase);
static void
ellipsoid_back_shift_y_corner(double shift, int phase);
static void
ellipsoid_back_shift_z_corner(double shift, int phase);

static double
get_ellipse_weight(ellipse_rec *ellipse, point_rec *point1, point_rec *point2, trans_matrix *trans_g2l);

static void
process_x_side_geometry(void);
static void
process_y_side_geometry(void);
static void
process_z_side_geometry(void);

static logic
check_inclusion_outside_tile_x_side(void);
static logic
check_inclusion_outside_tile_y_side(void);
static logic
check_inclusion_outside_tile_z_side(void);

static void
output_patch_boundary(long *edge_curves, long **edge_curve_array, int *edge_curve_sign, long side_patches, long *side_patch_array, long side_vertices, long *side_vertex_array);
static void
output_region_boundary(long *side_patches, long **side_patch_array, int *side_patch_sign, long tile_regions, long *tile_region_array);

/* end of special routines */

#ifdef CHECK_OVERLAP
static void
check_active_inclusion_overlap(ellipsoid_rec *ellipsoid1, int per_x, int per_y, int per_z, long i);
static void
check_inclusion_periodic_overlap(ellipsoid_rec *ellipsoid1, int per_x, int per_y, int per_z, long i);
#endif

static logic
check_ellipsoid_ellipsoid_overlap(ellipsoid_rec *ellipsoid1, ellipsoid_rec *ellipsoid2, double eps);

/* 3D functions */

static logic
check_point_inside_ellipsoid(point_rec *point, ellipsoid_rec *ellipsoid, double eps, double *value);

static double
get_ellipsoid_plane_distance(ellipsoid_rec *ellipsoid, point_rec *point, vector_rec *normal);
static double
get_ellipsoid_line_square_distance(ellipsoid_rec *ellipsoid, point_rec *point, vector_rec *dir);

static logic
get_ellipsoid_plane_intersection(ellipsoid_rec *ellipsoid, point_rec *point, vector_rec *normal, ellipse_rec *ellipse);
static int
get_ellipsoid_line_intersection(ellipsoid_rec *ellipsoid, point_rec *point, vector_rec *dir, point_rec *intersection);
static void
get_ellipsoid_line_center_intersection(ellipsoid_rec *ellipsoid, vector_rec *dir, point_rec *pnt1, point_rec *pnt2);
static logic
get_plane_plane_intersection(point_rec *center1, vector_rec *normal1, point_rec *center2, vector_rec *normal2, 
														 point_rec *point, vector_rec *dir);
static logic
get_plane_line_intersection(point_rec *point, vector_rec *normal, point_rec *pnt, vector_rec *dir, point_rec *intersection);

static double
project_point_to_plane(point_rec *pnt, point_rec *point, vector_rec *normal, point_rec *projection);
static double
project_point_to_line(point_rec *pnt, point_rec *point, vector_rec *dir, point_rec *projection);

static void
project_point_to_ellipsoid(point_rec *point, ellipsoid_rec *ellipsoid, point_rec *pnt);
static void
get_ellipsoid_normal_point(ellipsoid_rec *ellipsoid, point_rec *point, vector_rec *normal);

#ifdef PARAMETRIC_ROUTINES
static int
project_point_to_ellipsoid_uv_octant(point_rec *point, ellipsoid_rec *ellipsoid, point_rec *pnt, double *u_par, double *v_par);

static void
get_ellipsoid_point_uv_octant(ellipsoid_rec *ellipsoid, int octant, double u, double v, point_rec *pnt);
static void
get_ellipsoid_normal_uv_octant(ellipsoid_rec *ellipsoid, int octant, double u, double v, vector_rec *normal);

static void
get_ellipsoid_point_uv(ellipsoid_rec *ellipsoid, double u, double v, point_rec *pnt);
static double
get_ellipsoid_u_tangent_uv(ellipsoid_rec *ellipsoid, double u, double v, vector_rec *u_grad);
static double
get_ellipsoid_v_tangent_uv(ellipsoid_rec *ellipsoid, double u, double v, vector_rec *v_grad);
static void
get_ellipsoid_normal_uv(ellipsoid_rec *ellipsoid, double u, double v, vector_rec *normal);
#endif

/* 2D functions */

static logic
check_point_inside_ellipse(point_rec *point, ellipse_rec *ellipse, double *value);

static int
get_ellipse_line_intersection(ellipse_rec *ellipse, point_rec *point, vector_rec *dir, point_rec *intersection, logic point_in_plane);
static void
get_ellipse_line_center_intersection(ellipse_rec *ellipse, vector_rec *dir, point_rec *pnt1, point_rec *pnt2);

#ifdef PARAMETRIC_ROUTINES
static int
get_ellipse_ellipse_intersection(ellipse_rec *ellipse1, ellipse_rec *ellipse2, point_rec *intersection);

static void
get_ellipse_ellipse_interval_intersection(ellipse_rec *ellipse1, ellipse_rec *ellipse2, int q, double t, double val, double dt,
																					point_rec *point);
#endif

static void
project_point_to_ellipse(point_rec *point, ellipse_rec *ellipse, point_rec *pnt);
static void
get_ellipse_normal_point(ellipse_rec *ellipse, point_rec *point, vector_rec *normal);

static double
get_quadratic_elliptic_arc_weight(ellipse_rec *ellipse, point_rec *point1, point_rec *point2);

#ifdef PARAMETRIC_ROUTINES
static int
project_point_to_ellipse_t_quadrant(point_rec *point, ellipse_rec *ellipse, point_rec *pnt, double *t_par, logic spec);

static void
get_ellipse_point_t_quadrant(ellipse_rec *ellipse, int quadrant, double t, point_rec *pnt);
static double
get_ellipse_tangent_t_quadrant(ellipse_rec *ellipse, int quadrant, double t, vector_rec *grad);
static void
get_ellipse_normal_t_quadrant(ellipse_rec *ellipse, int quadrant, double t, vector_rec *normal);

static void
get_ellipse_point_t(ellipse_rec *ellipse, double t, point_rec *pnt);
static double
get_ellipse_tangent_t(ellipse_rec *ellipse, double t, vector_rec *grad);
static void
get_ellipse_normal_t(ellipse_rec *ellipse, double t, vector_rec *normal);

static logic
check_point_on_ellipse_interval(int q, double t, int q1, double t1, int q2, double t2);
#endif

static void
transform_vector_from_global_to_local(vector_rec *global, vector_rec *local, trans_matrix *trans_g2l);
static void
transform_vector_from_local_to_global(vector_rec *local, vector_rec *global, trans_matrix *trans_l2g);

static void
transform_point_from_global_to_local(point_rec *global, point_rec *local, point_rec *origin, trans_matrix *trans_g2l);
static void
transform_point_from_local_to_global(point_rec *local, point_rec *global, point_rec *origin, trans_matrix *trans_l2g);

static void
invert_rotation_transformation(trans_matrix *trans);
static void
get_inverted_rotation_transformation(trans_matrix *trans_in, trans_matrix *trans_out);

static void
transform_local_from_ellipsoid_to_sphere(point_rec *ellipsoidal, point_rec *spherical, vector_rec *trans_e2s);
static void
transform_local_from_sphere_to_ellipsoid(point_rec *spherical, point_rec *ellipsoidal, vector_rec *trans_s2e);

static void
invert_scale_transformation(vector_rec *trans);
static void
get_inverted_scale_transformation(vector_rec *trans_in, vector_rec *trans_out);

static char *
get_next_record_file(FILE *file, char *file_name, char *err_msg);
static char *
get_next_relevant_record_file(FILE *file, char *file_name, char *err_msg);

static void
write_record_file(FILE *file, char *file_name, char *format, ...);

static char *
get_next_record(char *err_msg);
static char *
get_next_relevant_record(char *err_msg);

static void
write_record(char *format, ...);

static void
error_message(int exit_code, char *format, ...);

static void
sort_id_coord(long n, long *id, is_pnt_rec **pnt);


void
print_geometry(void);
void
print_setting(void);


int
main(int argc, char** argv)
{
	logic relative, help = NO;
	face opposite[6] = {XP, XN, YP, YN, ZP, ZN};
	point_rec origin = {0.0, 0.0, 0.0}, center, pnt, point, point1, point2, point3, point4, point5, point6;
	point_rec x_extreme_point, y_extreme_point, z_extreme_point;
	vector_rec vec_max, vec_min, vec_mid, vec, vec1, vec2, vector, vec_tmp, normal;
	ellipse_rec xyz_ellipse;
	double msz, size, weight, val, per_shift_x, per_shift_y, per_shift_z, offset, max_rad = 0.0;
	double rate, rate_x, rate_y, rate_z, tile_shift_x, tile_shift_y, tile_shift_z, root_size, excess_size;
	double a_diff_xy, a_diff_xz, len_xy, len_xz, len_x;
	double a_diff_yz, a_diff_yx, len_yz, len_yx, len_y;
	double a_diff_zx, a_diff_zy, len_zx, len_zy, len_z;
	long id, tiles, inclusions, codes, num_tiles_1D = 0;
	long ver_id, ver_id1, ver_id2, ver_id3, ver_id4, cur_id1, cur_id2, cur_id3, cur_id4, cur_id5, cur_id6;
	long sur_id1, sur_id2, ptch_id1, ptch_id2, ptch_id3, ptch_id4, ptch_id5, ptch_id6, shll_id1, shll_id2;
	long inclusion_id, xyz_patch_id, xyz_curve_id[2];
	long i, j, k, m, n, kk;
	long *edge_curve_array[8], edge_curves[8], *side_patch_array[6], side_patches[6]; 
	int power, incr = 0, per_x, per_y, per_z, per_xyz, double_size, count, cnt, ind, side;
	int edge_curve_sign[8], side_patch_sign[6];
	char text[1024], suffix[][4] = {"-st", "-nd", "-rd", "-th"}, command[512];
	char facespec[][12] = {"negative x", "positive x", "negative y", "positive y", "negative z", "positive z"};
	char Facespec[][12] = {"Negative x", "Positive x", "Negative y", "Positive y", "Negative z", "Positive z"};
	vertex_status status;
	logic opposite_x, opposite_y, opposite_z, opposite_xyz, xyz_intersection;
	int ind_xx, ind_yy, ind_zz;
	
#ifdef ARBITRARY_TILE_ID
	long *tile_array = NULL;
#else
	logic *tile_array = NULL;
#endif
#if SETUP == GRID_3D || SETUP == GRID_2D
	int grid_size, row, column, layer;
#endif
#ifdef ALLOW_OUTSIDE
	logic outside;
#endif
#ifdef ELIXIR
	EView *view = NULL;
	BOOLEAN success;
#endif
#if 0
	double min_msz;
	int bits;
#endif

	if(MAX_RADIUS_RATE < 2.0)error_message(GENERAL_ERROR, "Invalid MAX_RADIUS_RATE");
	if(MULTIPLE_EPS < 1.0)error_message(GENERAL_ERROR, "Invalid MULTIPLE_EPS");

#ifdef IS_TESTING
	error_message(WARNING, "Directive IS_TESTING is active");

	if(MIRROR_X != 0 && MIRROR_X != 1)error_message(GENERAL_ERROR, "Invalid MIRROR_X");
	if(MIRROR_Y != 0 && MIRROR_Y != 1)error_message(GENERAL_ERROR, "Invalid MIRROR_Y");
	if(MIRROR_Z != 0 && MIRROR_Z != 1)error_message(GENERAL_ERROR, "Invalid MIRROR_Z");

#ifndef PER_REP_ALL
	if(PER_REP_X != 0 && PER_REP_X != 1)error_message(GENERAL_ERROR, "Invalid PER_REP_X");
	if(PER_REP_Y != 0 && PER_REP_Y != 1)error_message(GENERAL_ERROR, "Invalid PER_REP_Y");
	if(PER_REP_Z != 0 && PER_REP_Z != 1)error_message(GENERAL_ERROR, "Invalid PER_REP_Z");
#endif

	if(CYCLE_COORD < 0.0)error_message(GENERAL_ERROR, "Invalid CYCLE_COORD");
	if(CYCLE_COORD - (int)CYCLE_COORD != 0.0)error_message(GENERAL_ERROR, "Invalid CYCLE_COORD");
#endif
#ifdef SPECIAL_TESTING
	error_message(WARNING, "Directive SPECIAL_TESTING is active");

	if(SPLITTING_PLANE != XX && SPLITTING_PLANE != YY && SPLITTING_PLANE != ZZ)error_message(GENERAL_ERROR, "Invalid SPLITTING_PLANE");
	if(OPPOSITE_PLANE != 0 && OPPOSITE_PLANE != 1)error_message(GENERAL_ERROR, "Invalid OPPOSITE_PLANE");
#endif

	if(argc > 1){
		if(strcmp(argv[1], "-h") == 0)help = YES;
	}

	if(help == NO){
		if(argc > 1){
			if(strcmp(argv[1], "-X") == 0){
				incr = 1;
				
#ifdef ELIXIR
				draw = YES;
#endif
				
				argc--;
			}
		}
	}

	fprintf(stdout, "\n");
	fprintf(stdout, "Wang3d4T3d - Preprocessor of 3D Wang Tiles for T3d\n");
	fprintf(stdout, "             (c) Daniel Rypl 2018 - Version %d.%d\n", VERSION_MAJOR, VERSION_MINOR);
	fprintf(stdout, "--------------------------------------------------\n");

	if(help == YES || argc < 3){
		fprintf(stderr, "\n"); 
#ifdef ELIXIR
		fprintf(stderr, "Usage: %s [-h] [-X] geom_file t3d_file [epsilon [subdivisions]]\n\n", argv[0]);
#else
		fprintf(stderr, "Usage: %s [-h] geom_file t3d_file [epsilon [subdivisions]]\n\n", argv[0]);
#endif
		fprintf(stderr, INDENT "geom_file: input geometry\n");
		fprintf(stderr, INDENT "t3d_file: output model for T3d4WT\n");
		fprintf(stderr, INDENT "epsilon: minimal distance of inclusion from tile sides\n");
		fprintf(stderr, INDENT "         minimal distance of inclusion intersection with tile sides from tile edges\n");
		fprintf(stderr, INDENT "         minimal distance of inclusion intersection with tile edges from tile corners\n");
		fprintf(stderr, INDENT "    pos. value is considered as absolute epsilon\n");
		fprintf(stderr, INDENT "         (absolute epsilon <= min(dimension[xyz]) x MAX_REL_EPSILON)\n");
		fprintf(stderr, INDENT "    neg. value is considered as relative epsilon with respect to min(dimension[xyz])\n");
		fprintf(stderr, INDENT "         (relative epsilon <= MAX_REL_EPSILON)\n");
		fprintf(stderr, INDENT "    zero value is replaced by min(dimension[xyz]) x DEF_REL_EPSILON\n");
		fprintf(stderr, INDENT "    missing value is considered as zero value\n");
		fprintf(stderr, INDENT "subdivisions: number of times epsilon may be halved to satisfy proximity criteria \n");
		fprintf(stderr, INDENT "              for each particular inclusion\n");
		fprintf(stderr, INDENT "              missing value is considered as zero value\n");
		fprintf(stderr, INDENT "              use negative value to prevent inclusion shifting\n\n");
		if(help == YES){
			print_geometry();
			print_setting();
		}
		exit(1);
	}

	power = (int)log10(MAX_TILES - 0.5) + 1;
	while(power--)max_tiles *= 10;

#if 0
#if 1
	bits = sizeof(long) * CHAR_BIT;
#else /* LONGBITS is defined only in obsolete values.h */
	bits = LONGBITS;
#endif
	max_side_intersections = (long)(((1 << (bits - 1)) - 1) / (10 * max_tiles)) - 1;
#else
	max_side_intersections = (long)(LONG_MAX / (10 * max_tiles)) - 1;
#endif

	geom_file_name = argv[incr + 1];
	t3d_in_file_name = argv[incr + 2];

	if(argc > 3)epsilon = atof(argv[incr + 3]);
	if(argc > 4)max_subdivisions = atoi(argv[incr + 4]);

	if(max_subdivisions < 0){
		no_shift = YES;
		max_subdivisions = 0;
	}

	if((geom_file = fopen(geom_file_name, "r")) == NULL)
		error_message(FILE_OPEN_ERROR, "File %s opening error", geom_file_name);
	if((t3d_in_file = fopen(t3d_in_file_name, "w")) == NULL)
		error_message(FILE_OPEN_ERROR, "File %s opening error", t3d_in_file_name);
	
#ifdef ELIXIR
	if(draw == YES){
		if(ESIBuildInterface(ESI_GRAPHIC_EDITOR_MASK, 1, argv) != 0){
			fprintf(stderr, "Graphic interface error\n");
			exit(1);
		}
		augment_command_table();
		ESIPopup();
		view = ElixirNewView("Wang3D", "Wang3D", "midnightblue", "white", 400, 400);   /* use as bgcolor */
		EMAttachView(ESIModel(), view);
		EVSetViewOrientation(view, VIEW_ORIENT_ISO);
		EVShowAxes(view, YES);
		ESIHandleCmd("render ambient 0.1");
#ifdef SOLID
		EVSetRenderMode(view, WIRE_RENDERING);
#else
		EVSetRenderMode(view, NORMAL_RENDERING);
#endif
		EVSetLayerOnOff(view, TILE_LAYER, YES);
		EVSetLayerOnOff(view, INCLUSION_LAYER, YES);

#ifdef IS_TESTING
		switch(((int)CYCLE_COORD) % 3){
		case 0:                                /* default setting */
			ESIHandleCmd("view normal 3 2 1");
			ESIHandleCmd("view vup 0 0 1");
			break;
		case 1:
			ESIHandleCmd("view normal 1 3 2");
			ESIHandleCmd("view vup 1 0 0");
			break;
		case 2:
			ESIHandleCmd("view normal 2 1 3");
			ESIHandleCmd("view vup 0 1 0");
			break;
		default:
			break;
		}
#endif
		
		green = ColorGetPixelFromString("green", &success);
		red = ColorGetPixelFromString("red", &success);
		yellow = ColorGetPixelFromString("yellow", &success);
		blue = ColorGetPixelFromString("blue", &success);
		cyan = ColorGetPixelFromString("cyan", &success);
		magenta = ColorGetPixelFromString("magenta", &success);
		orange = ColorGetPixelFromString("orange", &success);
		black = ColorGetPixelFromString("black", &success);
		white = ColorGetPixelFromString("white", &success);
#if 1
		bgcolor = ColorGetPixelFromString("midnightblue", &success);
#else
		bgcolor = ColorGetPixelFromString("grey", &success);
#endif

		font = FontGetFontFromString(FONT, &success);
		if(success == NO)font = FontDefaultFont();
	}
#endif

	get_next_relevant_record("Missing problem identification");
	strcpy(text, line_buffer);

	get_next_relevant_record("Missing overall numbers");
	if(sscanf(line_buffer, "%ld %ld %ld", &tiles, &inclusions, &codes) != 3)
		error_message(INPUT_ERROR, "Invalid record - overall numbers");
	
	if(tiles <= 0 || tiles >= max_tiles)
		error_message(INPUT_ERROR, "Invalid number of Wang tiles: NWT = %ld (0 < NWT < %ld)", tiles, max_tiles);
	if(inclusions <= 0)
		error_message(INPUT_ERROR, "Invalid number of reference inclusions: NRI = %ld (0 < NRI)", inclusions);
	if(codes < 0)
		error_message(INPUT_ERROR, "Invalid number of side codes: NSC = %ld (0 <= NSC)", codes);
	
/* only array of usage of tile ids is allocated to save space */

/* IMPORTANT: I do rely on fact that calloc was used to allocate code_array */

#ifdef ARBITRARY_TILE_ID
	if((tile_array = calloc(tiles, sizeof(long))) == NULL)
		error_message(MEMORY_ERROR, "Memory allocation error");
#else
	if((tile_array = calloc(tiles, sizeof(logic))) == NULL)
		error_message(MEMORY_ERROR, "Memory allocation error");
#endif

/* array of complete data structures allocated */

	if((inclusion_array = calloc(inclusions, sizeof(inclusion_rec))) == NULL)
		error_message(MEMORY_ERROR, "Memory allocation error");

	if(codes != 0){

/* array of complete data structures allocated (small number of codes expected */

		if((code_array = calloc(codes, sizeof(code_rec))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
	}
	
/* reference tile must be read before inclusions to have epsilon available */

	get_next_relevant_record("Missing reference tile definition");
	if(sscanf(line_buffer, "%lf %lf %lf %lf %lf %lf %d %lf %lf", 
						&(tile.center.x), &(tile.center.y), &(tile.center.z), &(tile.size.x), &(tile.size.y), &(tile.size.z), 
						&(tile.mat), &(tile.imsz), &(tile.bmsz)) != 9)
		error_message(INPUT_ERROR, "Invalid record - definition of reference tile");

#ifdef IS_TESTING
	if(MIRROR_X != 0 || MIRROR_Y != 0 || MIRROR_Z != 0 || ((int)CYCLE_COORD) % 3 != 0){
		if(tile.center.x != 0.0 || tile.center.y != 0.0 || tile.center.z != 0.0)error_message(GENERAL_ERROR, "Tile center must be at origin");
	}
#endif

	if(tile.size.x <= 0.0)
		error_message(INPUT_ERROR, "Invalid tile dimension in x-direction: TDX = %e (0 < DX)", tile.size.x);
	if(tile.size.y <= 0.0)
		error_message(INPUT_ERROR, "Invalid tile dimension in y-direction: TDY = %e (0 < DY)", tile.size.y);
	if(tile.size.z <= 0.0)
		error_message(INPUT_ERROR, "Invalid tile dimension in z-direction: TDZ = %e (0 < DZ)", tile.size.z);
	if(tile.mat == 0)
		error_message(INPUT_ERROR, "Invalid matrix material: MMAT = %d (0 < MMAT)", tile.mat);
	if(tile.imsz < 0.0)
		error_message(INPUT_ERROR, "Invalid matrix internal mesh size: MIMSZ = %e (0 < MIMSZ)", tile.imsz);
	if(tile.bmsz < 0.0)
		error_message(INPUT_ERROR, "Invalid matrix boundary mesh size: MBMSZ = %e (0 < MBMSZ)", tile.bmsz);

	min_sz = max_sz = tile.size.x;
	if(tile.size.y < min_sz)min_sz = tile.size.y;
	if(tile.size.z < min_sz)min_sz = tile.size.z;
	if(tile.size.y > max_sz)max_sz = tile.size.y;
	if(tile.size.z > max_sz)max_sz = tile.size.z;

	if(tile.imsz == 0.0)
		tile.imsz = min_sz;
	else{
		if(tile.imsz > min_sz)tile.imsz = min_sz;
	}
	if(tile.bmsz == 0.0)
		tile.bmsz = min_sz;
	else{
		if(tile.bmsz > min_sz)tile.bmsz = min_sz;
	}

#if 0 /* not needed */
	min_msz = (tile.imsz < tile.bmsz) ? tile.imsz : tile.bmsz;
#endif

	if(epsilon == 0.0){
		epsilon = min_sz * DEF_REL_EPSILON;
		error_message(NOTIFICATION, "Epsilon set to %e", epsilon);
	}
	else if(epsilon < 0.0){
		if(-epsilon >= MAX_REL_EPSILON)error_message(GENERAL_ERROR, "Too large relative epsilon");
		epsilon *= -min_sz;
	}
	else{
		if(epsilon >= min_sz * MAX_REL_EPSILON)error_message(GENERAL_ERROR, "Too large absolute epsilon");
	}

	tile.corner[0].x = tile.center.x - tile.size.x / 2.0;
	tile.corner[0].y = tile.center.y - tile.size.y / 2.0;
	tile.corner[0].z = tile.center.z - tile.size.z / 2.0;

	tile.corner[1].x = tile.center.x + tile.size.x / 2.0;
	tile.corner[1].y = tile.center.y + tile.size.y / 2.0;
	tile.corner[1].z = tile.center.z + tile.size.z / 2.0;

#ifdef ELIXIR
	if(draw == YES){
		draw_tile(&tile, TILE_LAYER, green);
		EVFitAllIntoView(view);
			
#ifdef DEBUG_DRAW
		ESIEventLoop(YES, "Tile\n\n");
#endif
			
	}
#endif

	for(i = 0; i < inclusions; i++){
		k = (i <= 3) ? i : 3;

		get_next_relevant_record("Missing inclusion definition");
		if(sscanf(line_buffer, "%ld", &inclusion_id) != 1)
			error_message(INPUT_ERROR, "Invalid record - definition of the %ld%s inclusion", i + 1, suffix[k]);

		relative = NO;
		if(inclusion_id < 0){
			inclusion_id = -inclusion_id;
			relative = YES;
		}

		if(inclusion_id <= 0 || inclusion_id > inclusions)
			error_message(INPUT_ERROR, "Invalid id of the %ld%s inclusion: IID = %ld (0 < IID <= %ld)", i + 1, suffix[k], inclusion_id, inclusions);

		inclusion = &inclusion_array[inclusion_id - 1];
		if(inclusion -> id != 0)
			error_message(INPUT_ERROR, "Multiple definition of inclusion %ld encountered", inclusion_id);

		ellipsoid = &(inclusion -> ellipsoid);
		inclusion -> id = inclusion_id;

		if(sscanf(line_buffer, "%ld %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %d %d", &id,    /* IMPORTANT: do not overwrite inclusion -> id */
							&(ellipsoid -> center.x), &(ellipsoid -> center.y), &(ellipsoid -> center.z), 
							&(ellipsoid -> major.x), &(ellipsoid -> major.y), &(ellipsoid -> major.z), 
							&(ellipsoid -> minor.x), &(ellipsoid -> minor.y), &(ellipsoid -> minor.z), 
							&(ellipsoid -> max), &(ellipsoid -> min), &(ellipsoid -> mid),
							&(inclusion -> imsz), &(inclusion -> bmsz),  
							&(inclusion -> mat), &(inclusion -> ifmat)) != 17){
			error_message(INPUT_ERROR, "Invalid record - definition of inclusion %ld", inclusion_id);
		}
		
#ifdef IS_TESTING
		ellipsoid -> center.x *= sign[MIRROR_X];
		ellipsoid -> center.y *= sign[MIRROR_Y];
		ellipsoid -> center.z *= sign[MIRROR_Z];
		ellipsoid -> major.x *= sign[MIRROR_X];
		ellipsoid -> major.y *= sign[MIRROR_Y];
		ellipsoid -> major.z *= sign[MIRROR_Z];
		ellipsoid -> minor.x *= sign[MIRROR_X];
		ellipsoid -> minor.y *= sign[MIRROR_Y];
		ellipsoid -> minor.z *= sign[MIRROR_Z];
		
		switch(((int)CYCLE_COORD) % 3){
		case 0:
			break;
		case 1:
			swap(ellipsoid -> center.y, ellipsoid -> center.z, val);
			swap(ellipsoid -> center.x, ellipsoid -> center.y, val);
			swap(ellipsoid -> major.y, ellipsoid -> major.z, val);
			swap(ellipsoid -> major.x, ellipsoid -> major.y, val);
			swap(ellipsoid -> minor.y, ellipsoid -> minor.z, val);
			swap(ellipsoid -> minor.x, ellipsoid -> minor.y, val);
			break;
		case 2:
			swap(ellipsoid -> center.x, ellipsoid -> center.y, val);
			swap(ellipsoid -> center.y, ellipsoid -> center.z, val);
			swap(ellipsoid -> major.x, ellipsoid -> major.y, val);
			swap(ellipsoid -> major.y, ellipsoid -> major.z, val);
			swap(ellipsoid -> minor.x, ellipsoid -> minor.y, val);
			swap(ellipsoid -> minor.y, ellipsoid -> minor.z, val);
			break;
		default:
			break;
		}
#endif

#ifdef ALLOW_OUTSIDE

/* ellipsoid center of reference inclusion can be outside tile not more than by epsilon;
	 if outside in particular direction, it is shifted inside in that direction and flag for periodic repetition in that direction is inverted */

		if(ellipsoid -> center.x < tile.corner[0].x - epsilon || ellipsoid -> center.x > tile.corner[1].x + epsilon ||
			 ellipsoid -> center.y < tile.corner[0].y - epsilon || ellipsoid -> center.y > tile.corner[1].y + epsilon ||
			 ellipsoid -> center.z < tile.corner[0].z - epsilon || ellipsoid -> center.z > tile.corner[1].z + epsilon)
			error_message(INPUT_ERROR, "Invalid coordinates of center of inclusion %d", inclusion_id); 
		
		outside = NO;
		if(ellipsoid -> center.x < tile.corner[0].x || ellipsoid -> center.x > tile.corner[1].x){
			outside = inclusion -> out[X] = YES;
			if(ellipsoid -> center.x < tile.center.x)
				ellipsoid -> center.x += tile.size.x;
			else
				ellipsoid -> center.x -= tile.size.x;
		}
		if(ellipsoid -> center.y < tile.corner[0].y || ellipsoid -> center.y > tile.corner[1].y){
			outside = inclusion -> out[Y] = YES;
			if(ellipsoid -> center.y < tile.center.y)
				ellipsoid -> center.y += tile.size.y;
			else
				ellipsoid -> center.y -= tile.size.y;
		}
		if(ellipsoid -> center.z < tile.corner[0].z || ellipsoid -> center.z > tile.corner[1].z){
			outside = inclusion -> out[Z] = YES;
			if(ellipsoid -> center.z < tile.center.z)
				ellipsoid -> center.z += tile.size.z;
			else
				ellipsoid -> center.z -= tile.size.z;
		}

		if(outside == YES)error_message(WARNING, "Reference inclusion %ld subjected to periodic repetition (%d %d %d)", inclusion_id, 
																		(inclusion -> out[X] == YES) ? 1 : 0, (inclusion -> out[Y] == YES) ? 1 : 0, (inclusion -> out[Z] == YES) ? 1 : 0);
#else

/* ellipsoid center of reference inclusion cannot be outside tile;
	 it can be outside only for periodically repeated inclusions which are processed as shifted reference inclusion */

		if(ellipsoid -> center.x < tile.corner[0].x || ellipsoid -> center.x > tile.corner[1].x ||
			 ellipsoid -> center.y < tile.corner[0].y || ellipsoid -> center.y > tile.corner[1].y ||
			 ellipsoid -> center.z < tile.corner[0].z || ellipsoid -> center.z > tile.corner[1].z)
			error_message(INPUT_ERROR, "Invalid coordinates of center of inclusion %d", inclusion_id); 
#endif
		
		if(ellipsoid -> max <= 0.0 || ellipsoid -> max > (min_sz - epsilon) / MAX_RADIUS_RATE)
			error_message(INPUT_ERROR, "Invalid size of major half-axis of inclusion %d: MAX = %e (%e <= MAX <= %e)", 
										inclusion_id, ellipsoid -> max, epsilon * 10.0, (min_sz - epsilon) / MAX_RADIUS_RATE);
		if(ellipsoid -> min <= epsilon * 10.0 || ellipsoid -> min > ellipsoid -> max)
			error_message(INPUT_ERROR, "Invalid size of minor half-axis of inclusion %d: MIN = %e (%e < MIN <= %e)", 
										inclusion_id, ellipsoid -> min, epsilon * 10.0, ellipsoid -> max);
		if(ellipsoid -> mid < ellipsoid -> min || ellipsoid -> mid > ellipsoid -> max)
			error_message(INPUT_ERROR, "Invalid size of middle half-axis of inclusion %d: MID = %e (%e <= MID <= %e)", 
										inclusion_id, ellipsoid -> mid, ellipsoid -> min, ellipsoid -> max);

		if(inclusion -> imsz < 0.0)
			error_message(INPUT_ERROR, "Invalid internal mesh size IIMSZ of inclusion %ld: IIMSZ = %e (0 <= IIMSZ)", inclusion_id, inclusion -> imsz);
		if(inclusion -> bmsz < 0.0)
			error_message(INPUT_ERROR, "Invalid boundary mesh size IBMSZ of inclusion %ld: IBMSZ = %e (0 <= IBMSZ)", inclusion_id, inclusion -> bmsz);
		if(inclusion -> mat < 0)
			error_message(INPUT_ERROR, "Invalid inclusion material: IMAT = %d (0 <= IMAT)", inclusion_id, inclusion -> mat);
		if(inclusion -> ifmat < 0)
			error_message(INPUT_ERROR, "Invalid interface material: IFMAT = %d (0 <= IFMAT)", inclusion_id, inclusion -> ifmat);

		inclusion -> ifmat = 0;   /* interfaces not supported yet */

		if(inclusion -> imsz == 0.0)
			inclusion -> imsz = min_sz;
		else{
			if(inclusion -> imsz > min_sz)inclusion -> imsz = min_sz;
		}
		if(inclusion -> bmsz == 0.0)
			inclusion -> bmsz = min_sz;
		else{
			if(inclusion -> bmsz > min_sz)inclusion -> bmsz = min_sz;
		}

#if EXCESS_SIZE == MAX_PRESENT_SIZE
		if(max_rad < ellipsoid -> max)max_rad = ellipsoid -> max;
#endif

#if 0 /* not needed */
		if(min_msz > inclusion -> bmsz)min_msz = inclusion -> bmsz;
		if(ellipsoid -> mat != 0){
			if(min_msz > inclusion -> imsz)min_msz = inclusion -> imsz;
		}
#endif

		if(relative == NO)sub_vec(ellipsoid -> major, ellipsoid -> major, ellipsoid -> center);
		size = sqrt(size_vec(ellipsoid -> major));
		if(size < EPSILON)
			error_message(INPUT_ERROR, "Zero major axis of inclusion %ld", inclusion_id);
		div_vec(ellipsoid -> major, size);
		if(relative == NO)sub_vec(ellipsoid -> minor, ellipsoid -> minor, ellipsoid -> center);
		size = sqrt(size_vec(ellipsoid -> minor));
		if(size < EPSILON)
			error_message(INPUT_ERROR, "Zero minor axis of inclusion %ld", inclusion_id);   /* not correct */
		div_vec(ellipsoid -> minor, size);
		cross_product(ellipsoid -> middle, ellipsoid -> major, ellipsoid -> minor);
		size = sqrt(size_vec(ellipsoid -> middle));
		if(size < EPSILON)
			error_message(INPUT_ERROR, "Colinear major and minor axes of inclusion %ld", inclusion_id);   /* not correct */
		div_vec(ellipsoid -> middle, size);
		cross_product(ellipsoid -> minor, ellipsoid -> middle, ellipsoid -> major);

#ifdef PARAMETRIC_ROUTINES
		copy_mul_vec(vec_max, ellipsoid -> major, ellipsoid -> max);
		copy_mul_vec(vec_min, ellipsoid -> minor, ellipsoid -> min);
		copy_mul_vec(vec_mid, ellipsoid -> middle, ellipsoid -> mid);

		add_vec(ellipsoid -> pnt00, ellipsoid -> center, vec_max);
		copy_vec(ellipsoid -> pnt01, ellipsoid -> pnt00);
		copy_vec(ellipsoid -> pnt02, ellipsoid -> pnt00);

		add_vec(ellipsoid -> pnt20, ellipsoid -> center, vec_min);
		add_vec(ellipsoid -> pnt22, ellipsoid -> center, vec_mid);
		add_vec(ellipsoid -> pnt21, ellipsoid -> pnt20, vec_mid);
		
		add_vec(ellipsoid -> pnt10, ellipsoid -> pnt20, vec_max);
		add_vec(ellipsoid -> pnt11, ellipsoid -> pnt21, vec_max);
		add_vec(ellipsoid -> pnt12, ellipsoid -> pnt22, vec_max);
#endif

		inclusion -> corner = OUT;
	}

/* compute min_epsilon after min_sz, max_sz and max_rad have been determined and before inclusions are going to be processed
	 (namely before get_inclusion_tile_intersection is called);
	 therefore all inclusions are firstly read in to allow for determination of max_rad and after that they are processed (computing intersection);
	 this implies that I cannot consider real excess of present inclusions out of the tile, because the real excess can be computed in
	 get_inclusion_tile_intersection, which requires min_epsilon, which is dependent to the overall problem size, which is in turn dependent 
	 on the offset, which is finally dependent on the excess */

#if SETUP == GRID_3D
	num_tiles_1D = grid_size = (int)(cbrt(tiles - 0.5)) + 1;
#endif
#if SETUP == GRID_2D
	num_tiles_1D = grid_size = (int)(sqrt(tiles - 0.5)) + 1;
#endif
#if SETUP == DIAGONAL
	num_tiles_1D = tiles;
#endif

#if EXCESS_SIZE == MAX_ALLOWED_SIZE
	max_rad = (min_sz - epsilon) / MAX_RADIUS_RATE;
#endif

#ifdef USE_VIRTUAL_ELLIPSOID

/* virtual surfaces may exceed out of the tile up the whole size of inclusion */

	excess_size = max_rad * 2.0;
#else

/* virtual surfaces may exceed out of the tile only up the half of the whole size of inclusion */

	excess_size = max_rad;
#endif

/* I can allow for min_sz based offset because all imsz and bmsz are guaranteed to be min_sz or smaller unless not modified manually !!! */

/* IMPORTANT: it is dangerous to compute offset from min_msz, because min_msz does not account for curvature based refinement !!! */

/*     o-------------o                o-------------o
       |             |                |             |
     __o             o--_           __o             o--_
    -  |             |   |         -  |             |   |
   |   |             |__-         |   |             |__-        e_sz = excess_size
    -__|             o             -__|             o           o_sz = OFFSET * min_sz
       o             |                o             |
       |             |                |             |
       o-------------o                o-------------o

    e_sz    max_sz    e_sz  o_sz  e_sz     max_sz    e_sz
   |---|-------------|---|--------|---|-------------|---|
   |   |                              |                 |
   |   |------------------------------|                 |
   |                 offset                             |
   |                                                    |
   |----------------------------------------------------|
	                   root_size (obsolete) */

	offset = max_sz + OFFSET * min_sz + 2.0 * excess_size;

#ifdef TILE_OFFSET
	if(tiles != 1){
		if(max_sz - min_sz < max_sz * EPSILON)error_message(WARNING, "TILE_OFFSET applied to tile of not cubic shape");
	}

	if(offset < 2.0 * max_sz)offset = 2.0 * max_sz;
#endif

	if(tiles == 1 || codes == 0){

/* if codes == 0 dispersion is irrelevant as tile sides are not mirrored;
	 otherwise root octant must be shifted so that tile is centered in the root octant !!!
	 this should prevent octree dispersion when opposite sides of tile are mirrored */

		root_size = num_tiles_1D * max_sz + (num_tiles_1D - 1) * (offset - max_sz) + 2.0 * excess_size;
	}
	else{

/* in order to eliminate dispersion of mesh size due to octree mirroring of tile sides it is necessary to ensure that
	 each tile is placed at the same relative position with respect to the smallest octant containing the whole tile !!!
	 this implies that size of root must be 2^k x offset, where 2^k >= num_tiles_1D;
	 although this leads to larger (smaller) root (min_epsilon) than it is really necessary, the elimination of dispersion is of the highest priority !!!
	 root octant must be shifted so that each tile is centered in the smallest octant containing the whole tile !!!
	 this should prevent octree dispersion when opposite sides of tile are mirrored */

		power = 1;
		while(power < num_tiles_1D)power *= 2;
	
		root_size = power * offset;
	}

/* compute the epsilon as t3d does but make it 10 times larger */

	min_epsilon = root_size;
	double_size = 2.0 * (sizeof(double) - 1) - 1;
	while(double_size--)min_epsilon /= 10.0;

/* make epsilon at least min_epsilon * 10 */

	if(epsilon < min_epsilon * 10.0){
		epsilon = min_epsilon * 10.0;
		error_message(NOTIFICATION, "Epsilon increased to min epsilon %e", min_epsilon * 10.0);
		if(max_subdivisions != 0){
			max_subdivisions = 0;
			error_message(NOTIFICATION, "Number of subdivisions reduced to 0");
		}
	}

	inclusion = inclusion_array;
	for(i = 0; i < inclusions; i++, inclusion++){
		if(get_inclusion_tile_intersection() == YES){
			if(inclusion -> mat == 0 || inclusion -> ifmat != 0){
				if(inclusion -> corner == IN)inside_hole = YES;
#ifdef CHECK_ISOLATED
				if(inclusion -> corner == BND){
					if(x_count == 2 && y_count == 2 && z_count == 2)boundary_hole = YES;
				}
#endif
			}

/* resolve side flags;
	 compute intersection ellipse extreme points */

/* note: inclusion -> face_is[X|Y|Z] -> side[*] is used for detection whether output of intersection x|y|z-ellipse should be performed
	 only if the x|y|z_ellipse does not intersect tile edge (x|y|z_count = 0);
	 this implies that center of such ellipse is always inside tile side and it is therefore safe to use just inequalities;
	 if the center of the ellipse is on tile edge, the side should be set to BOTH but since side is not used in such a case, it is not done ! */

			if(x_intersection == YES || x_touching == YES){
#ifdef DRAW_ELLIPSE
				draw_point(&(x_ellipse.center), 0, green);
#endif
				if(x_ellipse.center.y > tile.corner[ind_y].y)
					inclusion -> face_is[X] -> side[XY] = POS;
				else
					inclusion -> face_is[X] -> side[XY] = NEG;
				
				if(x_ellipse.center.z > tile.corner[ind_z].z)
					inclusion -> face_is[X] -> side[XZ] = POS;
				else
					inclusion -> face_is[X] -> side[XZ] = NEG;
			}
			
			if(y_intersection == YES || y_touching == YES){
#ifdef DRAW_ELLIPSE
				draw_point(&(y_ellipse.center), 0, cyan);
#endif
				if(y_ellipse.center.z > tile.corner[ind_z].z)
					inclusion -> face_is[Y] -> side[YZ] = POS;
				else
					inclusion -> face_is[Y] -> side[YZ] = NEG;
				
				if(y_ellipse.center.x > tile.corner[ind_x].x)
					inclusion -> face_is[Y] -> side[YX] = POS;
				else
					inclusion -> face_is[Y] -> side[YX] = NEG;
			}

			if(z_intersection == YES || z_touching == YES){
#ifdef DRAW_ELLIPSE
				draw_point(&(z_ellipse.center), 0, magenta);
#endif
				if(z_ellipse.center.x > tile.corner[ind_x].x)
					inclusion -> face_is[Z] -> side[ZX] = POS;
				else
					inclusion -> face_is[Z] -> side[ZX] = NEG;
				
				if(z_ellipse.center.y > tile.corner[ind_y].y)
					inclusion -> face_is[Z] -> side[ZY] = POS;
				else
					inclusion -> face_is[Z] -> side[ZY] = NEG;
			}

/* WARNING: max_numbers can overflow to negative numbers if too large total number of inclusions is processed
	 or if too large number of inclusions intersects or touches tiles sides */

/* number of potential instances of an inclusion "in" tile equals to 2^(number_of_side_intersections)
	 decreased by 1 if inclusion intersects all tile sides but tile corner is outside inclusion */

			count = 0;
			if(x_intersection == YES)count++;
			if(y_intersection == YES)count++;
			if(z_intersection == YES)count++;

			max_tile_regions += 1 << count;
			
			if(count == 3){
				if(inclusion -> corner == OUT)max_tile_regions -= 1;
			}

			if(count != 0){

/* number of potential intersections of an inclusion with tile side equals to 2^(number_of_other_side_intersections)
	 decreased by 1 if inclusion intersects all tile sides but tile corner is outside inclusion */
		
				cnt = 1 << (count - 1);
				if(count == 3){
					if(inclusion -> corner == OUT)cnt--;
				}
			
				if(x_intersection == YES)max_side_patches[X] += cnt;
				if(y_intersection == YES)max_side_patches[Y] += cnt;
				if(z_intersection == YES)max_side_patches[Z] += cnt;
			}			

/* number of potential touching points on tile side;
	 periodic repetitions needs not be accounted for;
	 periodic repetition can influence the number only if touching happens exactly at tile edge;
	 but then such a point is vertex fixed to curve or tile vertex (and not vertex fixed to patch) */

			if(x_touching == YES)max_side_vertices[X]++;
			if(y_touching == YES)max_side_vertices[Y]++;
			if(z_touching == YES)max_side_vertices[Z]++;
			
/* number of potential points on tile edge
	 periodic repetitions are naturally accounted for;
	 intersection points coinciding with tile corners should not be acounted for but this is NOT done !!! */

			max_edge_vertices[X] += x_count;
			max_edge_vertices[Y] += y_count;
			max_edge_vertices[Z] += z_count;

/* number of potential inclusion segments on tile edge;
	 number of potential segments between inclusions on tile edge;
	 periodic repetition must be accounted for edge intersection side == BOTH */

			for(j = 0; j < 3; j++){
				if(inclusion -> edge_is[j] == NULL)continue;
				switch(inclusion -> edge_is[j] -> side){
				case POS:
				case NEG:
					if(inclusion -> edge_is[j] -> count == 2)max_edge_curves_1[j]++;
					
/* note: max_edge_curves_2 should not be incremented if one of 2 intersections with tile edge is passing through tile corner
	 but this is NOT done */

					max_edge_curves_2[j]++;
					break;
				case BOTH:
					max_edge_curves_1[j] += 2;
					break;
				default:
					break;
				}
			}
		}
		else{

/* WARNING: max_number can overflow to negative numbers if too large total number of inclusions is processed */

			max_tile_regions++;
		}
		
/* IMPORTANT: inclusion can be shifted even if there is no intersection with tile (eg. during overdetouch) */

		size = size_vec(total_shift);
		if(size != 0.0){
			inclusion -> shift = shift_done = YES;
			error_message(WARNING, "Inclusion %ld shifted by [%e %e %e]", inclusion -> id, total_shift.x, total_shift.y, total_shift.z);
			
#ifdef DEBUG
			if(fabs(total_shift.x) > MULTIPLE_EPS * epsilon)error_message(WARNING, "Shift in x-direction violates maximal permissible shift"); 
			if(fabs(total_shift.y) > MULTIPLE_EPS * epsilon)error_message(WARNING, "Shift in y-direction violates maximal permissible shift"); 
			if(fabs(total_shift.z) > MULTIPLE_EPS * epsilon)error_message(WARNING, "Shift in z-direction violates maximal permissible shift"); 
#endif
		}

#if 0
		{
			vector_rec dir;
			point_rec point, p1;
			double dist;

			point.x = 1.0;
			point.y = 0.6;
			point.z = 0.2;

			dir.x = -1.0;
			dir.y = 0.1;
			dir.z = 0.3;

			dist = sqrt(size_vec(dir));
			div_vec(dir, dist);

			mul_vec(dir, 3.0);
			add_vec(p1, point, dir);
			div_vec(dir, 3.0);
			draw_line(&point, &p1, 0, green);
			dist = get_ellipsoid_line_square_distance(&(inclusion -> ellipsoid), &point, &dir);
			fprintf(stderr, "%20.13e\n", sqrt(dist));
		}
#endif

#ifdef ELIXIR
		if(draw == YES){
			draw_inclusion(inclusion, INCLUSION_LAYER, red);
			
#ifdef DEBUG_DRAW
			ESIEventLoop(YES, "Inclusion\n\n");
#endif
			
		}
#endif

	}

/* IMPORTANT: allocate the array larger by 1 item so that there is no segmentation fault failure when max_number is once exceeded */

/* note: max_side_patches[X|Y|Z] can be legally larger than max_side_intersections
	 because not all potential intersections will occur (not all inclusions must be used, not all periodic repetitions must be used) */

/* note: there may occur memory allocation failure despite the fact that the problem can fit in the memory;
	 this due to the fact that max_numbers are likely to be larger than the really needed dimension of the array
	 because not all inclusions must be used and not all periodic repetitions must be used */

/* note: even if max_number + 1 is negative (due to overflow), it does not mean that the problem does not fit in the memory;
	 this due to the fact that max_numbers are likely to be larger than the really needed dimension of the array
	 because not all inclusions must be used and not all periodic repetitions must be used;
	 in such a case it may be enough to allocate memory for LONG_MAX - 1 but this is likely to result in memory allocation failure;
	 this is how much memory would be required (on 32-bit machine) if LONG_MAX is used for all arrays:
	 
	 tile_region_array = 4 x LONG_MAX

	 x_side_patch_array = 4 x LONG_MAX x 2
	 y_side_patch_array = 4 x LONG_MAX x 2
	 z_side_patch_array = 4 x LONG_MAX x 2

	 x_side_vertex_array = 4 x LONG_MAX x 2
	 y_side_vertex_array = 4 x LONG_MAX x 2
	 z_side_vertex_array = 4 x LONG_MAX x 2

	 x_edge_vertex_id_array = 4 x LONG_MAX x 4
	 y_edge_vertex_id_array = 4 x LONG_MAX x 4
	 z_edge_vertex_id_array = 4 x LONG_MAX x 4

	 x_edge_vertex_data_array = (8 + 4) x LONG_MAX x 4
	 y_edge_vertex_data_array = (8 + 4) x LONG_MAX x 4
	 z_edge_vertex_data_array = (8 + 4) x LONG_MAX x 4

	 x_edge_curve_1_array = 4 x LONG_MAX x 4
	 y_edge_curve_1_array = 4 x LONG_MAX x 4
	 z_edge_curve_1_array = 4 x LONG_MAX x 4

	 x_edge_curve_2_array = 4 x LONG_MAX x 4
	 y_edge_curve_2_array = 4 x LONG_MAX x 4
	 z_edge_curve_2_array = 4 x LONG_MAX x 4

	 suma = LONG_MAX x 340 = 2147483647 x 340 = 7.3 x 10^11 = 680 GB */

/* WARNING: max_number will be positive if overwlow repeats even number of times;
	 this will be fatal for processing resulting in seqmentation fault !!! */

	if((long)(max_tile_regions + 1) < 0)error_message(GENERAL_ERROR, "Too large problem");
	for(i = 0; i < 3; i++){
		if((long)(max_side_patches[i] + 1) < 0)error_message(GENERAL_ERROR, "Too large problem");
		if((long)(max_side_vertices[i] + 1) < 0)error_message(GENERAL_ERROR, "Too large problem");
		if((long)(max_edge_vertices[i] + 1) < 0)error_message(GENERAL_ERROR, "Too large problem");
	}

	if((tile_region_array = (long *)calloc(max_tile_regions + 1, sizeof(long))) == NULL)
		error_message(MEMORY_ERROR, "Memory allocation error");

	for(i = 0; i < 2; i++){
		if((x_side_patch_array[i] = (long *)calloc(max_side_patches[X] + 1, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
		if((y_side_patch_array[i] = (long *)calloc(max_side_patches[Y] + 1, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
		if((z_side_patch_array[i] = (long *)calloc(max_side_patches[Z] + 1, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");

		if((x_side_vertex_array[i] = (long *)calloc(max_side_vertices[X] + 1, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
		if((y_side_vertex_array[i] = (long *)calloc(max_side_vertices[Y] + 1, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
		if((z_side_vertex_array[i] = (long *)calloc(max_side_vertices[Z] + 1, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
	}

/* IMPORTANT: allocate  x|y|z_edge_vertex_id_array and x|y|z_edge_vertex_data_array larger by 2 items 
	 so that there is no segmentation fault failure when max_number is twice exceeded;
	 note that when processing single side intersection, up to 2 intersection points on a single axis may be processed */

	for(i = 0; i < 4; i++){
		if((x_edge_vertex_id_array[i] = (long *)calloc(max_edge_vertices[X] + 2, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
		if((y_edge_vertex_id_array[i] = (long *)calloc(max_edge_vertices[Y] + 2, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
		if((z_edge_vertex_id_array[i] = (long *)calloc(max_edge_vertices[Z] + 2, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");

		if((x_edge_vertex_data_array[i] = (is_pnt_rec **)calloc(max_edge_vertices[X] + 2, sizeof(is_pnt_rec *))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
		if((y_edge_vertex_data_array[i] = (is_pnt_rec **)calloc(max_edge_vertices[Y] + 2, sizeof(is_pnt_rec *))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
		if((z_edge_vertex_data_array[i] = (is_pnt_rec **)calloc(max_edge_vertices[Z] + 2, sizeof(is_pnt_rec *))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");

		if((x_edge_curve_1_array[i] = (long *)calloc(max_edge_curves_1[X] + 1, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
		if((y_edge_curve_1_array[i] = (long *)calloc(max_edge_curves_1[Y] + 1, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
		if((z_edge_curve_1_array[i] = (long *)calloc(max_edge_curves_1[Z] + 1, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");

		if((x_edge_curve_2_array[i] = (long *)calloc(max_edge_curves_2[X] + 1, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
		if((y_edge_curve_2_array[i] = (long *)calloc(max_edge_curves_2[Y] + 1, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
		if((z_edge_curve_2_array[i] = (long *)calloc(max_edge_curves_2[Z] + 1, sizeof(long))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
	}

#ifdef CHECK_OVERLAP
#ifndef ALLOW_OVERLAP

/* check for overlap of all inclusions in reference tile */

	for(i = 1; i < inclusions; i++){

/* primary instance */

		inclusion = &inclusion_array[i];
		ellipsoid = &(inclusion -> ellipsoid);
		check_inclusion_periodic_overlap(ellipsoid, 0, 0, 0, i);

/* secondary instances */

		if(inclusion -> face_is[X] == NULL)
			per_shift_x = 0.0;
		else{
			if(ellipsoid -> center.x < tile.center.x)           /* XN side */
				per_shift_x = tile.size.x;
			else                                                /* XP side */
				per_shift_x = -tile.size.x;

			shift_ellipsoid(*ellipsoid, per_shift_x, 0.0, 0.0);
			check_inclusion_periodic_overlap(ellipsoid, 1, 0, 0, i);
			shift_ellipsoid(*ellipsoid, -per_shift_x, 0.0, 0.0);
		}
		if(inclusion -> face_is[Y] == NULL)
			per_shift_y = 0.0;
		else{
			if(ellipsoid -> center.y < tile.center.y)           /* YN side */
				per_shift_y = tile.size.y;
			else                                                /* YP side */
				per_shift_y = -tile.size.y;

			shift_ellipsoid(*ellipsoid, 0.0, per_shift_y, 0.0);
			check_inclusion_periodic_overlap(ellipsoid, 0, 1, 0, i);
			shift_ellipsoid(*ellipsoid, 0.0, -per_shift_y, 0.0);
		}
		if(inclusion -> face_is[Z] == NULL)
			per_shift_z = 0.0;
		else{
			if(ellipsoid -> center.z < tile.center.z)           /* ZN side */
				per_shift_z = tile.size.z;
			else                                                /* ZP side */
				per_shift_z = -tile.size.z;

			shift_ellipsoid(*ellipsoid, 0.0, 0.0, per_shift_z);
			check_inclusion_periodic_overlap(ellipsoid, 0, 0, 1, i);
			shift_ellipsoid(*ellipsoid, 0.0, 0.0, -per_shift_z);
		}
			
		if(per_shift_x != 0.0 && per_shift_y != 0.0){
			shift_ellipsoid(*ellipsoid, per_shift_x, per_shift_y, 0.0);
			check_inclusion_periodic_overlap(ellipsoid, 1, 1, 0, i);
			shift_ellipsoid(*ellipsoid, -per_shift_x, -per_shift_y, 0.0);
		}
		if(per_shift_y != 0.0 && per_shift_z != 0.0){
			shift_ellipsoid(*ellipsoid, 0.0, per_shift_y, per_shift_z);
			check_inclusion_periodic_overlap(ellipsoid, 0, 1, 1, i);
			shift_ellipsoid(*ellipsoid, 0.0, -per_shift_y, -per_shift_z);
		}
		if(per_shift_z != 0.0 && per_shift_x != 0.0){
			shift_ellipsoid(*ellipsoid, per_shift_x, 0.0, per_shift_z);
			check_inclusion_periodic_overlap(ellipsoid, 1, 0, 1, i);
			shift_ellipsoid(*ellipsoid, -per_shift_x, 0.0, -per_shift_z);
		}

		if(per_shift_x != 0.0 && per_shift_y != 0.0 && per_shift_z != 0.0){
			if(inclusion_array[i].corner == IN){
				shift_ellipsoid(*ellipsoid, per_shift_x, per_shift_y, per_shift_z);
				check_inclusion_periodic_overlap(ellipsoid, 1, 1, 1, i);
				shift_ellipsoid(*ellipsoid, -per_shift_x, -per_shift_y, -per_shift_z);
			}
		}
	}
#endif
#endif

	if(inside_hole == NO){
		for(j = 0; j < 8; j++)corner_hole[j] = NO;
	}
#ifdef CHECK_ISOLATED
	if(boundary_hole == NO){
		for(j = 0; j < 8; j++)corner_isolated[j] = NO;
	}
#endif

	if(strlen(text) + 2 > strlen(SEP_STRING)){
		write_record("#\n");
		write_record("# %s", text);
		write_record("#\n\n");
	}
	else{
		write_record(SEP_STRING "\n");
		write_record("# %s", text);
		write_record(SEP_STRING "\n\n");
	}

#ifdef FIX_HOLE
	sprintf(command, "t3d -d %g -t 0 -G 3 -e %e", min_sz, min_epsilon);
#else
	sprintf(command, "t3d -d %g -t 0 -G 1 -e %e", min_sz, min_epsilon);
#endif
	
	if(tiles == 1){     /* keep tile (PUC) in original position */
		if(codes == 0)
			write_record("# %s\n\n", command);
		else{

/* root octant is shifted so that tile is centered in root octant */

			write_record("# %s -R "FMT1" -U "FMT3"\n\n", command, root_size, 
									 tile.center.x - root_size / 2.0, tile.center.y - root_size / 2.0, tile.center.z - root_size / 2.0);
			write_record("# options -R and -U are relevant to prevent octree dispersion due to application of periodic BCs\n\n");
		}
		
		tile_shift_x = tile_shift_y = tile_shift_z = 0.0;
	}
	else{               /* shift x-y-z- corner of the 1st tile to origin of GCS */
		if(codes == 0)
			write_record("# %s\n\n", command);
		else{

/* root octant is shifted so that each tile is centered in the smallest octant containing the whole tile */

			write_record("# %s -R "FMT1" -U "FMT3"\n\n", command, root_size,
									 -(offset - tile.size.x) / 2.0, -(offset - tile.size.y) / 2.0, -(offset - tile.size.z) / 2.0);
			write_record("# options -R and -U are relevant to prevent octree dispersion due to application of generalized periodic BCs\n\n");
		}

		tile_shift_x = tile.size.x * 0.5 - tile.center.x;       
		tile_shift_y = tile.size.y * 0.5 - tile.center.y;
		tile_shift_z = tile.size.z * 0.5 - tile.center.z;

#if SETUP == GRID_3D || SETUP == GRID_2D
		row = column = layer = 0;
#endif
	}

	write_record("dimension 3\n");

/* CAUTION: tiles are processed according to their order (not id) in input file !!! */

	for(i = 0; i < tiles; i++){
		k = (i <= 3) ? i : 3;

		get_next_relevant_record("Missing inclusion definition");
		if(sscanf(line_buffer, "%ld", &(tile.id)) != 1)        /* caution: use tile.id */
			error_message(INPUT_ERROR, "Invalid record - definition of the %ld%s particular tile", i + 1, suffix[k]);
		
#ifdef ARBITRARY_TILE_ID
		if(tile.id < 0)
			error_message(INPUT_ERROR, "Invalid id of the %ld%s particular tile: TID = %ld (0 <= TID)", i + 1, suffix[k], tile.id);

		for(j = 0; j < i; j++){
			if(tile_array[j] == tile.id)
				error_message(INPUT_ERROR, "Multiple description of particular tile %ld encountered", tile.id);
		}
		tile_array[i] = tile.id;
		tile_id = i + 1;               /* IMPORTANT: I am using fictitious tile numbering to ensure tile_id < max_tiles */
#else
		if(tile.id <= 0 || tile.id >= max_tiles)
			error_message(INPUT_ERROR, "Invalid id of the %ld%s particular tile: TID = %ld (0 < TID <= %ld)", i + 1, suffix[k], tile.id, tiles);

		if(tile_array[tile.id - 1] != NO)
			error_message(INPUT_ERROR, "Multiple description of particular tile %ld encountered", tile.id);
		tile_array[tile.id - 1] = YES;
		tile_id = tile.id;
#endif

/* caution: when reporting error, use tile.id (not tile_id) which always contains real tile id no matter
	 whether aribitrary tile numbering is allowed or not */

		if(sscanf(line_buffer, "%ld %d %d %d %d %d %d %ld %ld", &(tile.id),        /* caution: use tile.id */
							&(tile.code[XN]), &(tile.code[XP]), &(tile.code[YN]), &(tile.code[YP]), &(tile.code[ZN]), &(tile.code[ZP]), 
							&(tile.inclusions), &(tile.points)) != 9)
			error_message(INPUT_ERROR, "Invalid record - definition of particular tile %ld", tile.id);        /* caution: use tile.id */
		
/* note: the codes can be generated automatically - NOT IMPLEMENTED;
	 this would require, however, that inclusions (including their periodic repetitions) intersecting or touching at least one side 
	 are remembered by particular tiles;
	 since I do not store particular tiles, there is an alternative, in which inclusions (including their periodic repetitions) intersecting or touching 
	 at least one side remember tiles they are used by;
	 to make the check more efficient each inclusion intersecting or touching at least one side will have 8 (for each possible periodic repetition) 
	 logical arrays of size equal to number of tiles;
	 then when processing a new particular tile, all previously processed particular tiles must be checked
	 whether the same set of inclusions intersecting a particular side is used;
	 since the code must be known in advance (to setup mirroring), it is necessary to process the check before the geometry of the tile is processed */

/* IMPORTANT: I do rely on fact that calloc was used to allocate code_array */

		for(j = 0; j < 6; j++){
			if(tile.code[j] < 0 || tile.code[j] > codes)
				error_message(INPUT_ERROR, "Invalid code for %s side of particular tile %ld: CODE = %d (0 < CODE < %ld)", 
											facespec[j], tile.id, tile.code[j], codes);        /* caution: use tile.id */

/* note: if both opposite tile sides have the same code, then the stored side corresponds to negative side;
	 this is consequence of the fact that codes are processed according to face ordering (in which negative side always preceeds opposite positive side);
	 I do rely on this fact when specifying mirroring between patches corresponding to periodically repeated inclusions */

			if(tile.code[j] == 0)continue;
			code = &code_array[tile.code[j] - 1];
			if(code -> tile_id != 0){
				if(code -> side != j && code -> side != opposite[j])
					error_message(INPUT_ERROR, "Code for %s side of particular tile %ld already used as code for %s side of particular tile %ld",
												facespec[j], tile.id, facespec[code -> side], code -> tile_id);        /* caution: use tile.id */
			}
			else{
				code -> id = tile.code[j];
				code -> tile_id = tile_id;        /* caution: store fictitious tile id if arbitrary tile numbering is allowed */
				code -> tile_id_orig = tile.id;   /* caution: store real tile id */
				code -> side = j;
			}
		}
					
		if(tile.inclusions < 0)
			error_message(INPUT_ERROR, "Invalid number of inclusions of particular tile %ld: NTI = %ld (0 <= NTI)", tile.id, tile.inclusions);        /* caution: use tile.id */
		if(tile.points < 0)
			error_message(INPUT_ERROR, "Invalid number of ctrl points of particular tile %ld: NCP = %ld (0 <= NCP)", tile.id, tile.points);        /* caution: use tile.id */
 
/* IMPORTANT: I cannot process geometry of inclusions here because the ordering of inclusions in particular tile is arbitrary
	 and I need to process inclusions according to their id;
	 therefore inclusions in particular tile are only marked as active and then are processed (see below) according to their id */

		for(j = 0; j < tile.inclusions; j++){
			kk = (j <= 3) ? j : 3;
			get_next_relevant_record("Missing inclusion id");
#ifdef SEPARATE_PER_REP
			if(sscanf(line_buffer, "%ld %d %d %d", &inclusion_id, &per_x, &per_y, &per_z) != 4)
				error_message(INPUT_ERROR, "Invalid record - specification of the %ld%s inclusion of particular tile %ld", j + 1, suffix[kk], tile.id);        /* caution: use tile.id */
#else
			if(sscanf(line_buffer, "%ld %d", &inclusion_id, &per_xyz) != 2)
				error_message(INPUT_ERROR, "Invalid record - specification of the %ld%s inclusion of particular tile %ld", j + 1, suffix[kk], tile.id);        /* caution: use tile.id */
#endif

			if(inclusion_id <= 0 || inclusion_id > inclusions)
				error_message(INPUT_ERROR, "Invalid id of the %ld%s inclusion of particular tile %ld: IID = %ld (0 < IID <= %ld)", 
											j + 1, suffix[kk], tile.id, inclusion_id, inclusions);        /* caution: use tile.id */

			inclusion = &inclusion_array[inclusion_id - 1];
			ellipsoid = &(inclusion -> ellipsoid);

#ifndef PER_REP_ALL
#ifdef IS_TESTING
			per_x = (int)PER_REP_X;
			per_y = (int)PER_REP_Y;
			per_z = (int)PER_REP_Z;
#else
#ifdef SEPARATE_PER_REP
			if(per_x != 0 && per_x != 1)
				error_message(INPUT_ERROR, "Invalid periodic repetition in x-direction of inclusion %ld of particular tile %ld: PRX = %d (PRX = {0;1})", 
											inclusion_id, tile.id, per_x);        /* caution: use tile.id */
			if(per_y != 0 && per_y != 1)
				error_message(INPUT_ERROR, "Invalid periodic repetition in y-direction of inclusion %ld of particular tile %ld: PRY = %d (PRY = {0;1})", 
											inclusion_id, tile.id, per_y);        /* caution: use tile.id */
			if(per_z != 0 && per_z != 1)
				error_message(INPUT_ERROR, "Invalid periodic repetition in z-direction of inclusion %ld of particular tile %ld: PRZ = %d (PRZ = {0;1})", 
											inclusion_id, tile.id, per_z);        /* caution: use tile.id */
#else
			if(per_xyz < 0 || per_xyz > 7)
				error_message(INPUT_ERROR, "Invalid periodic repetition iof inclusion %ld of particular tile %ld: PRXYZ = %d (PRXYZ = {0;1;2;3;4;5;6;7})", 
											inclusion_id, tile.id, per_xyz);      /* caution: use tile.id */
			per_x = (per_xyz & 1) >> 0;
			per_y = (per_xyz & 2) >> 1;
			per_z = (per_xyz & 4) >> 2;
#endif
#endif
#ifdef ALLOW_OUTSIDE
			if(inclusion -> out[X] == YES)per_x = 1 - per_x;
			if(inclusion -> out[Y] == YES)per_y = 1 - per_y;
			if(inclusion -> out[Z] == YES)per_z = 1 - per_z;
#endif
#else
			for(per_x = 0; per_x <= 1; per_x++){
				for(per_y = 0; per_y <= 1; per_y++){
					for(per_z = 0; per_z <= 1; per_z++){
#endif				
						if(per_x != 0 || per_y != 0 || per_z != 0){

/* check that inclusion is intersecting or touching boundary with respect to periodic repetition */

							if(per_x == 1){
								if(inclusion -> face_is[X] == NULL){
									error_message(NOTIFICATION, "Inclusion %ld (%d %d %d) is outside tile ==> ignored", inclusion_id, per_x, per_y, per_z);
									continue;
								}
							}
							if(per_y == 1){
								if(inclusion -> face_is[Y] == NULL){
									error_message(NOTIFICATION, "Inclusion %ld (%d %d %d) is outside tile ==> ignored", inclusion_id, per_x, per_y, per_z);
									continue;
								}
							}
							if(per_z == 1){
								if(inclusion -> face_is[Z] == NULL){
									error_message(NOTIFICATION, "Inclusion %ld (%d %d %d) is outside tile ==> ignored", inclusion_id, per_x, per_y, per_z);
									continue;
								}
							}

/* ind_x|y|z defines portion of inclusion falling inside reference tile;
	 e.g. ind_z == 0 indicates that upper portion of inclusion above ZN side falls inside reference tile
	 while ind_z == 1 indicates that lower portion of inclusion below ZP side falls inside reference tile */
							
							if(ellipsoid -> center.x < tile.center.x)           /* XN side */
								ind_x = 0;
							else                                                /* XP side */
								ind_x = 1;
							if(ellipsoid -> center.y < tile.center.y)           /* YN side */
								ind_y = 0;
							else                                                /* YP side */
								ind_y = 1;
							if(ellipsoid -> center.z < tile.center.z)           /* ZN side */
								ind_z = 0;
							else                                                /* ZP side */
								ind_z = 1;

/* x|y|z_ind defines portion of inclusion (possibly periodically repeated) falling in particular tile;
	 e.g. z_ind == 0 indicates that upper portion of inclusion above ZN side falls inside particular tile
	 while z_ind == 1 indicates that lower portion of inclusion below ZP side falls inside particular tile */
	 
/* ind_z == z_ind <==> per_z = 0  .... primary instance of inclusion
	 ind_z != z_ind <==> per_z = 1  .... secondary instance of inclusion */
			
							if(per_x == 0)
								x_ind = ind_x;
							else{
								//x_ind = ind_x + sign[ind_x];
								x_ind = 1 - ind_x;
							}
							if(per_y == 0)
								y_ind = ind_y;
							else{
								//y_ind = ind_y + sign[ind_y];
								y_ind = 1 - ind_y;
							}
							if(per_z == 0)
								z_ind = ind_z;
							else{
								//z_ind = ind_z + sign[ind_z];
								z_ind = 1 - ind_z;
							}
							
/* check that inclusion is intersecting or touching tile */

							x_count = y_count = z_count = 0;

							if(inclusion -> edge_is[X] != NULL)x_count = inclusion -> edge_is[X] -> count;
							if(inclusion -> edge_is[Y] != NULL)y_count = inclusion -> edge_is[Y] -> count;
							if(inclusion -> edge_is[Z] != NULL)z_count = inclusion -> edge_is[Z] -> count;

/* CAUTION: inclusion touching tile from outside is not classified as outside !!! */

							if(check_inclusion_outside_tile_x_side() == YES){
								if(check_inclusion_outside_tile_y_side() == YES){
									if(check_inclusion_outside_tile_z_side() == YES){
										error_message(NOTIFICATION, "Inclusion %ld (%d %d %d) is outside tile ==> ignored", inclusion_id, per_x, per_y, per_z);
										continue;
									}
								}
							}
						}

						if(inclusion -> active[(per_x << 0) + (per_y << 1) + (per_z << 2)] == YES){
#ifdef IS_TESTING
							error_message(WARNING, "Directive IS_TESTING is active");
#endif
							error_message(INPUT_ERROR, "Multiple use of inclusion %ld with periodic repetition %d %d %d in particular tile %ld", 
														inclusion_id, per_x, per_y, per_z, tile.id);        /* caution: use tile.id */
						}

						inclusion -> active[(per_x << 0) + (per_y << 1) + (per_z << 2)] = YES;

#ifdef PER_REP_ALL
					}
				}
			}
#endif
		}
		
		tile_regions = 0;
		for(j = 0; j < 2; j++){
			x_side_patches[j] = y_side_patches[j] = z_side_patches[j] = 0;
			x_side_vertices[j] = y_side_vertices[j] = z_side_vertices[j] = 0;
#ifndef HALF_MAX_TILES
			x_side_patches_total[j] = y_side_patches_total[j] = z_side_patches_total[j] = 0;
#endif
		}
		for(j = 0; j < 4; j++){
			x_edge_vertices[j] = y_edge_vertices[j] = z_edge_vertices[j] = 0;
			x_edge_curves_1[j] = y_edge_curves_1[j] = z_edge_curves_1[j] = 0;
			x_edge_curves_2[j] = y_edge_curves_2[j] = z_edge_curves_2[j] = 1;
		}

		if(inside_hole == YES){
			for(j = 0; j < 8; j++)corner_hole[j] = NO;

/* check for tile corner inside hole;
	 note that this check is undesirable overhead because it is (indirectly) also performed when processing individual tile inclusions;
	 in the past I needed the information before tile corners were defined;
	 now I can suppress output at all corner vertices and after the whole tile is processed (and tile corners inside holes are checked)
	 I can process corner not inside the hole simply by fixing another vertex to the original corner vertex and enforcing output for the fixed vertex */

			count = 0;
			for(j = 0; j < inclusions; j++){
				inclusion = &inclusion_array[j];
				if(inclusion -> corner == IN){
					if(inclusion -> mat == 0 || inclusion -> ifmat != 0){
						for(k = 0; k < 8; k++){
							if(inclusion -> active[k] == YES){
								if(corner_hole[k] == NO){
									corner_hole[k] = YES;
									count++;
								}
							}
						}
						if(count == 8)break;
					}
				}
			}
		}

#ifdef CHECK_ISOLATED

/* CAUTION: there is a pathological case when tile corner is on boundary of hole inclusion and each of the tile edges is intersected twice;
	 an equivalent situation in 2D looks like this:

          |  /h
	   t3   | /h   t2
          |/h
   -------o--------   hole inclusion is on side marked by h
         /|C
     t4 /h|    t1
       /h |

	 node at corner C needs to be generated even in tile t1 in order to allow for connection of t1 to t2 or t4;
	 an exception is when only a single tile is processed in which case the node is not needed;
	 note that this node in tile t1 will be isolated (not bounding any edge);
	 such node will cause problems in FE computation if t1 is the left-upper tile in an assembly, as there will be no stiffness assigned to the node;
	 isolated may be be therefore considered for removal from the assembled mesh !!! 
	 therefore a warning is printed;
	 note also that this problem is related not only to a single node by also to group of isolated elements;
	 for example:

          |     /h
	   t3   | t2 /h 
         C|   /h
   -------o--o-----   hole inclusion is on side marked by h
          | /h
     t4   |/h  t1
          oh
         /|
        /h|

	 I do not check for such a case;
	 note that it is very likely that periodic BCs are prescribed for the whole assembly;
	 then the isolated node or grou[ of elements is connected to the opposite side(s) of the tile and everything is ok */
							
		if(boundary_hole == YES){
			for(j = 0; j < 8; j++)corner_isolated[j] = NO;

			count = 0;
			for(j = 0; j < inclusions; j++){
				inclusion = &inclusion_array[j];
				if(inclusion -> corner == BND){
					if(inclusion -> mat == 0 || inclusion -> ifmat != 0){
						for(k = 0; k < 8; k++){
							if(inclusion -> active[k] == YES){
								if(corner_isolated[k] == NO){
									if(inclusion -> edge_is[X] -> count != 2)continue;
									if(inclusion -> edge_is[Y] -> count != 2)continue;
									if(inclusion -> edge_is[Z] -> count != 2)continue;

/* check that center of inclusion is inside of processed tile */

									if(inclusion -> ellipsoid.center.x  < tile.corner[0].x || inclusion -> ellipsoid.center.x  > tile.corner[1].x)continue;
									if(inclusion -> ellipsoid.center.y  < tile.corner[0].y || inclusion -> ellipsoid.center.y  > tile.corner[1].y)continue;
									if(inclusion -> ellipsoid.center.z  < tile.corner[0].z || inclusion -> ellipsoid.center.z  > tile.corner[1].z)continue;

									corner_isolated[k] = YES;
									count++;
								}
							}
						}
						if(count == 8)break;
					}
				}
			}

			if(count != 0){
#if 0 /* not recommended */

/* warning: it is not desirable to suppress output of nodes at isolated vertices of a single tile
	 because it is very likely that there are periodic BCs !!! */

				if(tiles == 1){

/* in case when a single tile is processed vertices corresponding to isolated corners are not subjected to output */

					for(j = 0; j < 8; j++){
						if(corner_isolated[j] == YES)corner_hole[j] = YES;
					}
				}
				else{
#endif
					for(j = 0; j < 8; j++){
						if(corner_isolated[j] == YES)error_message(WARNING, "Isolated node at corner %d of tile %ld\n", j, tile.id);        /* caution: use tile.id */
					}
#if 0 /* not recommended */
				}
#endif
			}
		}
#endif

/* IMPORTANT: tile corner vertices must be ordered so that x|y|z-coordinates of latter from any two vertices 
	 on the same edge is larger or equal to coordinates of former from these two vertices;
	 thus vertices are ordered according to their periodic repetition;
	         id:  1  2  3  4  5  6  7  8
	 shifting_x:  0  1  0  1  0  1  0  1
	 shifting_y:  0  0  1  1  0  0  1  1
	 shifting_z:  0  0  0  0  1  1  1  1 */

#ifdef ARBITRARY_TILE_ID
		write_record("\n# tile %ld (%ld)\n\n", tile_id, tile.id);
		write_record("tile %ld ref_id %ld", tile_id, tile.id);
#else
		write_record("\n# tile %ld\n\n", tile_id);
		write_record("tile %ld", tile_id);
#endif

/* caution: use backward shift back to reference position */

		write_record(" shift "FMT3, -tile_shift_x, -tile_shift_y,  -tile_shift_z);

#ifdef USE_CODE
		write_record(" code %d %d %d %d %d %d\n\n", tile.code[0], tile.code[1], tile.code[2], tile.code[3], tile.code[4], tile.code[5]); 
#else
		write_record("\n\n");
#endif 
		
		origin_id = vertex_id + 1;       /* remember id of the 1st tile vertex */

/* calculate coordinates of shifted particular tile */

		for(j = 0; j < 2; j++){
			par_corner[j].x = tile.corner[j].x + tile_shift_x;
			par_corner[j].y = tile.corner[j].y + tile_shift_y;
			par_corner[j].z = tile.corner[j].z + tile_shift_z;
		}

		for(j = 0; j < 8; j++){
			if(corner_hole[j] == NO)
				write_record("vertex %ld xyz "FMT3" tile %ld\n", ++vertex_id, 
										 par_corner[shifting_x[j]].x, 
										 par_corner[shifting_y[j]].y, 
										 par_corner[shifting_z[j]].z, tile_id);
			else
				write_record("vertex %ld xyz "FMT3" output no tile %ld\n", ++vertex_id, 
										 par_corner[shifting_x[j]].x, 
										 par_corner[shifting_y[j]].y, 
										 par_corner[shifting_z[j]].z, tile_id);
		}

/* IMPORTANT: process (active) inclusions according to their id */

		for(j = 0; j < inclusions; j++){
			inclusion = &inclusion_array[j];
			ellipsoid = &(inclusion -> ellipsoid);

/* ind_x|y|z defines portion of inclusion falling inside reference tile;
	 e.g. ind_z == 0 indicates that upper portion of inclusion above ZN side falls inside reference tile
	 while ind_z == 1 indicates that lower portion of inclusion below ZP side falls inside reference tile */

			if(ellipsoid -> center.x < tile.center.x)           /* XN side */
				ind_x = 0;
			else                                                /* XP side */
				ind_x = 1;
			if(ellipsoid -> center.y < tile.center.y)           /* YN side */
				ind_y = 0;
			else                                                /* YP side */
				ind_y = 1;
			if(ellipsoid -> center.z < tile.center.z)           /* ZN side */
				ind_z = 0;
			else                                                /* ZP side */
				ind_z = 1;

/* x|y|z_ind defines portion of inclusion (possibly periodically repeated) falling in particular tile;
	 e.g. z_ind == 0 indicates that upper portion of inclusion above ZN side falls inside particular tile
	 while z_ind == 1 indicates that lower portion of inclusion below ZP side falls inside particular tile */
	 
/* ind_z == z_ind <==> per_z = 0  .... primary instance of inclusion
	 ind_z != z_ind <==> per_z = 1  .... secondary instance of inclusion */
			
			for(k = 0; k < 8; k++){
				if(inclusion -> active[k] == NO)continue;
				inclusion -> active[k] = NO;

				if((per_x = shifting_x[k]) == 0){
					per_shift_x = 0.0;
					x_ind = ind_x;
				}
				else{
					per_shift_x = sign[ind_x] * tile.size.x;
					//x_ind = ind_x + sign[ind_x];
					x_ind = 1 - ind_x;
				}
				
				if((per_y = shifting_y[k]) == 0){
					per_shift_y = 0.0;
					y_ind = ind_y;
				}
				else{
					per_shift_y = sign[ind_y] * tile.size.y;
					//y_ind = ind_y + sign[ind_y];
					y_ind = 1 - ind_y;
				}
				
				if((per_z = shifting_z[k]) == 0){
					per_shift_z = 0.0;
					z_ind = ind_z;
				}
				else{
					per_shift_z = sign[ind_z] * tile.size.z;
					//z_ind = ind_z + sign[ind_z];
					z_ind = 1 - ind_z;
				}
				
				corner_id = origin_id + (z_ind << 2) + (y_ind << 1) + (x_ind << 0);
				cor_id = 0;

				write_record("\n# inclusion %ld (%d %d %d)\n\n", j + 1, per_x, per_y, per_z);

#ifdef CHECK_OVERLAP
#ifndef ALLOW_OVERLAP

/* apply the check also if ALLOW_OVERLAP is not defined (inclusions are not allowed to overlap in the reference tile) if inclusion shifting took place */

				if(shift_done == YES){
#endif

/* check for overlap of used inclusions in particular tile */

					shift_ellipsoid(*ellipsoid, per_shift_x, per_shift_y, per_shift_z);
					check_active_inclusion_overlap(ellipsoid, per_x, per_y, per_z, j);
					shift_ellipsoid(*ellipsoid, -per_shift_x, -per_shift_y, -per_shift_z);
#ifndef ALLOW_OVERLAP
				}
#endif
#endif

				shift_vec.x = shift_x = tile_shift_x + per_shift_x;
				shift_vec.y = shift_y = tile_shift_y + per_shift_y;
				shift_vec.z = shift_z = tile_shift_z + per_shift_z;

				x_count = y_count = z_count = 0;

				if(inclusion -> edge_is[X] != NULL)x_count = inclusion -> edge_is[X] -> count;
				if(inclusion -> edge_is[Y] != NULL)y_count = inclusion -> edge_is[Y] -> count;
				if(inclusion -> edge_is[Z] != NULL)z_count = inclusion -> edge_is[Z] -> count;

#ifdef DEBUG_PRINT
				fprintf(stderr, "\nProcessing inclusion %ld (%d %d %d)\n", j + 1, per_x, per_y, per_z);
				fprintf(stderr, "==================================================\n\n");
				fprintf(stderr, "x|y|z_count: %d %d %d\n", x_count, y_count, z_count);
				fprintf(stderr, "corner: %s\n\n", (inclusion -> corner < 0) ? "in" : (inclusion -> corner > 0) ? "out" : "bnd");
#endif

/* resolve x|y|z-side geometry;
	 note: x|y|z_intersection, x|y|z_touching, x|y|z_face_is and x|y|z_edge_is is set on process_x|y|z_geometry */

				x_patch_id = y_patch_id = z_patch_id = 0;
				x_curve_id = y_curve_id = z_curve_id = 0;
				x_side_curves = y_side_curves = z_side_curves = 0;
				side_vertices = 0;

				xx_ind = (z_ind << 1) + (y_ind << 0);
				yy_ind = (x_ind << 1) + (z_ind << 0);
				zz_ind = (y_ind << 1) + (x_ind << 0);

/* note that tile_id of inclusion patch (intersection of inclusion with tile side) is printed no matter whether hole or not;
	 similarly, tile_id of inclusion curve (intersection of inclusion with tile edge) is printed no matter whether outputed or not */

				if(inclusion -> mat == 0){
					sprintf(hole_spec, "hole");
					sprintf(output_spec, "output no");
				}
				else{
					*hole_spec = '\0';
					*output_spec = '\0';
				}

#ifndef USE_SIDE
				*side_spec = '\0';
#endif

/* IMPORTANT: the order x --> y --> z is relevant as it influences in which order touching vertices are stored in side_vertex_array */

				process_x_side_geometry();
				process_y_side_geometry();
				process_z_side_geometry();

#ifdef IS_TESTING
#ifdef DEBUG_PRINT
				fprintf(stderr, "x-side vertices: %ld %ld\n", x_side_vertices[0], x_side_vertices[1]);
				fprintf(stderr, "y-side vertices: %ld %ld\n", y_side_vertices[0], y_side_vertices[1]);
				fprintf(stderr, "z-side vertices: %ld %ld\n\n", z_side_vertices[0], z_side_vertices[1]);

				fprintf(stderr, "Cumulative values:\n\n");

				fprintf(stderr, "x-edge vertices: %ld %ld %ld %ld\n", x_edge_vertices[0], x_edge_vertices[1], x_edge_vertices[2], x_edge_vertices[3]);
				fprintf(stderr, "y-edge vertices: %ld %ld %ld %ld\n", y_edge_vertices[0], y_edge_vertices[1], y_edge_vertices[2], y_edge_vertices[3]);
				fprintf(stderr, "z-edge vertices: %ld %ld %ld %ld\n\n", z_edge_vertices[0], z_edge_vertices[1], z_edge_vertices[2], z_edge_vertices[3]);

				fprintf(stderr, "x-edge curves_1: %ld %ld %ld %ld\n", x_edge_curves_1[0], x_edge_curves_1[1], x_edge_curves_1[2], x_edge_curves_1[3]);
				fprintf(stderr, "y-edge curves_1: %ld %ld %ld %ld\n", y_edge_curves_1[0], y_edge_curves_1[1], y_edge_curves_1[2], y_edge_curves_1[3]);
				fprintf(stderr, "z-edge curves_1: %ld %ld %ld %ld\n\n", z_edge_curves_1[0], z_edge_curves_1[1], z_edge_curves_1[2], z_edge_curves_1[3]);

				fprintf(stderr, "x-edge curves_2: %ld %ld %ld %ld\n", x_edge_curves_2[0], x_edge_curves_2[1], x_edge_curves_2[2], x_edge_curves_2[3]);
				fprintf(stderr, "y-edge curves_2: %ld %ld %ld %ld\n", y_edge_curves_2[0], y_edge_curves_2[1], y_edge_curves_2[2], y_edge_curves_2[3]);
				fprintf(stderr, "z-edge curves_2: %ld %ld %ld %ld\n\n", z_edge_curves_2[0], z_edge_curves_2[1], z_edge_curves_2[2], z_edge_curves_2[3]);
#endif
#endif

#if 0   // shift including ctrl points - dodelat
#ifdef ELIXIR
				if(draw == YES){
					draw_inclusion(inclusion, INCLUSION_LAYER, red);
					
#ifdef DEBUG_DRAW
					ESIEventLoop(YES, "Inclusion\n\n");
#endif

				}
#endif
#endif

				if(x_intersection == YES || y_intersection == YES || z_intersection == YES || x_touching == YES || y_touching == YES || z_touching == YES){
					side = 0;
					per_xyz = 0;
					
/* check for case when ellipsoid center is on at least one tile side;
	 such ellipsoid can be treated in the same way as if per_x|y|z != 0;
	 CAUTION: floating point based check on equality !!! */

					if(per_x == 0 && per_y == 0 && per_z == 0){
						if(ellipsoid -> center.x == tile.corner[x_ind].x)
							per_xyz = side = X + 1;
						else if(ellipsoid -> center.y == tile.corner[y_ind].y)
							per_xyz = side = Y + 1;
						else if(ellipsoid -> center.z == tile.corner[z_ind].z)
							per_xyz = side = Z + 1;
						
#ifdef INEXACT_BG_SURFACE
						per_xyz = 0;
#endif
					}
					else{
						if(per_x == 1)
							per_xyz = side = X + 1;
						else if(per_y == 1)
							per_xyz = side = Y + 1;
						else
							per_xyz = side = Z + 1;
					}
					
/* duplication of code from get_inclusion_tile_intersection !!! */

					copy_vec(trans_g2l.x, ellipsoid -> major);
					copy_vec(trans_g2l.y, ellipsoid -> minor);
					copy_vec(trans_g2l.z, ellipsoid -> middle);
					
					get_inverted_rotation_transformation(&trans_g2l, &trans_l2g);
					
					transform_vector_from_global_to_local(&dir_x, &dir_1, &trans_g2l);
					transform_vector_from_global_to_local(&dir_y, &dir_2, &trans_g2l);
					transform_vector_from_global_to_local(&dir_z, &dir_3, &trans_g2l);
					
/* compute relevant extreme points in GCS;
	 duplication of code from get_inclusion_tile_intersection !!!
	 IMPORTANT: this must be done before vectors dir_1|2|3 used to compute intersection ellipse are transformed to space of sphere !!! 

	 note: there is an overhead because if periodic repetition is applied I need only one relevant extreme point in the direction of
	 periodic repetition */
					
					vec.x = ellipsoid -> max;
					vec.y = ellipsoid -> min;
					vec.z = ellipsoid -> mid;
					
					comp_product(vector, vec, vec);
					
/* try to avoid overhead by computing only really relevant extreme point(s);
	 if periodic repetition is applied only one extreme point inside the tile is needed;
	 otherwise all extreme points over intersected or touching side(s) outside the tile are needed;
	 caution: extreme points are computed for inclusion in reference position !!! */

					if(per_xyz == X + 1 || (per_xyz == 0 && (x_intersection == YES || x_touching == YES))){
						comp_product(vec_x, dir_1, vec);
						val = 1.0 / sqrt(size_vec(vec_x));
						if(x_ind == 0)val = -val;             /* extreme outside the tile */
						if(per_xyz != 0)val = -val;           /* extreme inside the tile */
						copy_mul_vec(vec_tmp, vector, val);
						comp_product(point, dir_1, vec_tmp);
						
						transform_point_from_local_to_global(&point, &x_extreme_point, &(ellipsoid -> center), &trans_l2g);
					}
					if(per_xyz == Y + 1 || (per_xyz == 0 && (y_intersection == YES || y_touching == YES))){
						comp_product(vec_y, dir_2, vec);
						val = 1.0 / sqrt(size_vec(vec_y));
						if(y_ind == 0)val = -val;             /* extreme outside the tile */
						if(per_xyz != 0)val = -val;           /* extreme inside the tile */
						copy_mul_vec(vec_tmp, vector, val);
						comp_product(point, dir_2, vec_tmp);
						
						transform_point_from_local_to_global(&point, &y_extreme_point, &(ellipsoid -> center), &trans_l2g);
					}
					if(per_xyz == Z + 1 || (per_xyz == 0 && (z_intersection == YES || z_touching == YES))){
						comp_product(vec_z, dir_3, vec);
						val = 1.0 / sqrt(size_vec(vec_z));
						if(z_ind == 0)val = -val;             /* extreme outside the tile */
						if(per_xyz != 0)val = -val;           /* extreme inside the tile */
						copy_mul_vec(vec_tmp, vector, val);
						comp_product(point, dir_3, vec_tmp);
						
						transform_point_from_local_to_global(&point, &z_extreme_point, &(ellipsoid -> center), &trans_l2g);
					}

					if(per_xyz == 0){

/* processing octant containing ellipsoid center (but not on boundary) */

/* note that ellipsoid is not shifted due to periodic repetition
	 and that x|y|z_ind = ind_x|y|z */

/* bg surface for shell is a single biquadratic surface (inexact ellipsoid) with base parallel to side corresponding to minimum rate */

						rate_x = rate_y = rate_z = 1.0;

						if(x_intersection == YES)rate_x = (tile.corner[x_ind].x - ellipsoid -> center.x) / (x_extreme_point.x - ellipsoid -> center.x);
						if(y_intersection == YES)rate_y = (tile.corner[y_ind].y - ellipsoid -> center.y) / (y_extreme_point.y - ellipsoid -> center.y);
						if(z_intersection == YES)rate_z = (tile.corner[z_ind].z - ellipsoid -> center.z) / (z_extreme_point.z - ellipsoid -> center.z);
						
						if(rate_x < MAX_RATE || rate_y < MAX_RATE || rate_z < MAX_RATE){
							if(rate_x < rate_y && rate_x < rate_z)
								side = X + 1;
							else if(rate_y < rate_z)
								side = Y + 1;
							else
								side = Z + 1;
						}
					}

/* precompute complete orthonormal bases of relevant tile side(s) (either above selected side or all sides) transformed to space of sphere;
	 duplication of code from get_inclusion_tile_intersection !!! */

					radius = ellipsoid -> max;     /* sphere radius */
					//radius = ellipsoid -> min;     /* sphere radius */
					//radius = ellipsoid -> mid;     /* sphere radius */
					
					scale_e2s.x = radius / ellipsoid -> max;
					scale_e2s.y = radius / ellipsoid -> min;
					scale_e2s.z = radius / ellipsoid -> mid;
					
					get_inverted_scale_transformation(&scale_e2s, &scale_s2e);
					
					transform_local_from_ellipsoid_to_sphere(&dir_1, &dir_1, &scale_e2s);
					transform_local_from_ellipsoid_to_sphere(&dir_2, &dir_2, &scale_e2s);
					transform_local_from_ellipsoid_to_sphere(&dir_3, &dir_3, &scale_e2s);
					
					size = sqrt(size_vec(dir_1));
					div_vec(dir_1, size);
					
					size = sqrt(size_vec(dir_2));
					div_vec(dir_2, size);
					
					size = sqrt(size_vec(dir_3));
					div_vec(dir_3, size);
					
					if(side == X + 1 || side == 0){
						cross_product(vec_x, dir_2, dir_3);
						size = sqrt(size_vec(vec_x));
						div_vec(vec_x, size);
						
						cross_product(dir_c, vec_x, dir_2);
					}
					if(side == Y + 1 || side == 0){
						cross_product(vec_y, dir_3, dir_1);
						size = sqrt(size_vec(vec_y));
						div_vec(vec_y, size);
						
						cross_product(dir_a, vec_y, dir_3);
					}
					if(side == Z + 1 || side == 0){
						cross_product(vec_z, dir_1, dir_2);
						size = sqrt(size_vec(vec_z));
						div_vec(vec_z, size);
						
						cross_product(dir_b, vec_z, dir_1);
					}
					
					if(per_xyz == 0){

/* processing octant containing ellipsoid center (but not on boundary) */

/* note that ellipsoid is not shifted due to periodic repetition
	 and that x|y|z_ind = ind_x|y|z */

/* if part of the ellipsoid outside of the processed octant is not large enough on any side (rate >= MAX_RATE)
	 side is not selected (side = 0) and special handling is required;
	 this is also the case when only touching occurs */

						if(side != 0){

/* IMPORTANT: do not use tile side as intersection plane;
	 shift the plane slightly out of tile to eliminate singularity and shape deformation along base of biquadratic surface;
	 shift is accomplished by increasing the rate;
	 I do rely on fact that orthonormal base was already precomputed for particular side */

							copy_vec(point, tile.center);
							
							if(side == X + 1){
								ind = x_ind;
								sub_vec(vec, ellipsoid -> center, x_extreme_point);
								
								rate = rate_x + (1.0 - rate_x) * INC_RATE;
								point.x = ellipsoid -> center.x - rate * vec.x;
								
								transform_point_from_global_to_local(&point, &pnt, &(ellipsoid -> center), &trans_g2l);
								transform_local_from_ellipsoid_to_sphere(&pnt, &pnt, &scale_e2s);
								
								get_ellipsoid_xyz_side_intersection(&xyz_ellipse, &pnt, &vec_x, &dir_2, &dir_c);
								copy_vec(xyz_ellipse.normal, dir_x);
								
#ifdef DRAW_EXTREME_POINT_ELLIPSOID
								draw_point(&x_extreme_point, 0, orange);
#endif
							}
							else if(side == Y + 1){
								ind = y_ind;
								sub_vec(vec, ellipsoid -> center, y_extreme_point);
								
								rate = rate_y + (1.0 - rate_y) * INC_RATE;
								point.y = ellipsoid -> center.y - rate * vec.y;
								
								transform_point_from_global_to_local(&point, &pnt, &(ellipsoid -> center), &trans_g2l);
								transform_local_from_ellipsoid_to_sphere(&pnt, &pnt, &scale_e2s);
								
								get_ellipsoid_xyz_side_intersection(&xyz_ellipse, &pnt, &vec_y, &dir_3, &dir_a);
								copy_vec(xyz_ellipse.normal, dir_y);
							
#ifdef DRAW_EXTREME_POINT_ELLIPSOID
								draw_point(&y_extreme_point, 0, orange);
#endif
							}
							else{
								ind = z_ind;
								sub_vec(vec, ellipsoid -> center, z_extreme_point);
								
								rate = rate_z + (1.0 - rate_z) * INC_RATE;
								point.z = ellipsoid -> center.z - rate * vec.z;
								
								transform_point_from_global_to_local(&point, &pnt, &(ellipsoid -> center), &trans_g2l);
								transform_local_from_ellipsoid_to_sphere(&pnt, &pnt, &scale_e2s);
								
								get_ellipsoid_xyz_side_intersection(&xyz_ellipse, &pnt, &vec_z, &dir_1, &dir_b);
								copy_vec(xyz_ellipse.normal, dir_z);
							
#ifdef DRAW_EXTREME_POINT_ELLIPSOID
								draw_point(&z_extreme_point, 0, orange);
#endif
							}

							weight = -0.5 * ((2.0 + SQRT2) * rate + SQRT2);
							rate = 1.0 + (1.0 + SQRT2) * (1.0 + rate) / weight;
							add_mul_vec(point, ellipsoid -> center, vec, 1.0, rate);
							
#ifdef DRAW_ELLIPSE
							draw_point(&point, 0, orange);
							draw_ellipse(&xyz_ellipse, 0, orange);
							if(draw == YES)ESIEventLoop(YES, "Base ellipse of inexact ellipsoid + ctrl point\n\n");
#endif
							copy_mul_vec(vec_max, xyz_ellipse.major, xyz_ellipse.max);
							copy_mul_vec(vec_min, xyz_ellipse.minor, xyz_ellipse.min);
							
							add_vec(center, xyz_ellipse.center, shift_vec);            /* shift ellipse center */
						
							add_vec(point1, center, vec_max);
							sub_vec(point2, center, vec_max);
							add_vec(point3, center, vec_min);
							sub_vec(point4, center, vec_min);
							
							write_record("vertex %ld xyz "FMT3" virtual\n", ver_id1 = ++vertex_id, point1.x, point1.y, point1.z);
							write_record("vertex %ld xyz "FMT3" virtual\n", ver_id2 = ++vertex_id, point2.x, point2.y, point2.z);
							write_record("vertex %ld xyz "FMT3" virtual\n", ver_id3 = ++vertex_id, point3.x, point3.y, point3.z);
							write_record("vertex %ld xyz "FMT3" virtual\n", ver_id4 = ++vertex_id, point4.x, point4.y, point4.z);
							
							add_vec(point5, point1, vec_min);
							add_vec(point6, point2, vec_min);
							
							write_record("curve %ld vertex %ld %ld order 3 virtual\n", cur_id1 = ++curve_id, ver_id1, ver_id3);
							write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, SQRT1_2);
							write_record("curve %ld vertex %ld %ld order 3 virtual\n", cur_id2 = ++curve_id, ver_id3, ver_id2);
							write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, SQRT1_2);
							
							sub_vec(point5, point1, vec_min);
							sub_vec(point6, point2, vec_min);
							
							write_record("curve %ld vertex %ld %ld order 3 virtual\n", cur_id3 = ++curve_id, ver_id2, ver_id4);
							write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, SQRT1_2);
							write_record("curve %ld vertex %ld %ld order 3 virtual\n", cur_id4 = ++curve_id, ver_id4, ver_id1);
							write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, SQRT1_2);
							
							if(ind == 0)
								write_record("surface %ld curve %ld %ld %ld %ld", ++surface_id, cur_id1, cur_id4, cur_id3, cur_id2);
							else
								write_record("surface %ld curve %ld %ld %ld %ld", ++surface_id, cur_id1, cur_id2, cur_id3, cur_id4);
#ifdef USE_VIRTUAL_ELLIPSOID
							write_record(" rate 0 virtual\n");
#else
							write_record(" virtual\n");
#endif

							add_vec(point, point, shift_vec);
							write_record("polygon 1 1 xyz "FMT3" weight "FMT1"\n", point.x, point.y, point.z, weight);

							if(inclusion -> support == NO){
								inclusion -> support = YES;

								write_record("ellipsoid %ld max "FMT1" min "FMT1" mid "FMT1" \\\n", inclusion -> id, ellipsoid -> max, ellipsoid -> min, ellipsoid -> mid);
								write_record("   major "FMT3" \\\n", ellipsoid -> major.x, ellipsoid -> major.y, ellipsoid -> major.z);
								write_record("   minor "FMT3" \\\n", ellipsoid -> minor.x, ellipsoid -> minor.y, ellipsoid -> minor.z);
								write_record("   middle "FMT3"\n", ellipsoid -> middle.x, ellipsoid -> middle.y, ellipsoid -> middle.z);
							}

							write_record("shell %ld bgsurface %ld boundary curve", ++shell_id, surface_id);
							
							for(m = 0; m < x_side_curves; m++)write_record(" %ld", sign[x_ind] * x_side_curve_array[m]);
							for(m = 0; m < y_side_curves; m++)write_record(" %ld", sign[y_ind] * y_side_curve_array[m]);
							for(m = 0; m < z_side_curves; m++)write_record(" %ld", sign[z_ind] * z_side_curve_array[m]);
							
							if(side_vertices != 0){
								write_record(" fixed vertex");
								
								for(m = 0; m < side_vertices; m++)write_record(" %ld", side_vertex_array[m]);
							}

							add_vec(center, ellipsoid -> center, shift_vec);            /* shift ellipsoid center */
							
/* caution: center specification must follow ellipsoid specification immediately !!! */

							write_record(" rate 0 size %g tile %ld ellipsoid %ld center "FMT3"\n", 
													 inclusion -> bmsz, tile_id, inclusion -> id, center.x, center.y, center.z);
							
							write_record("region %ld boundary patch", ++region_id);
							if(x_patch_id != 0)write_record(" %ld", sign[1 - x_ind] * x_patch_id);
							if(y_patch_id != 0)write_record(" %ld", sign[1 - y_ind] * y_patch_id);
							if(z_patch_id != 0)write_record(" %ld", sign[1 - z_ind] * z_patch_id);
							if(inclusion -> mat == 0)
								write_record(" boundary shell %ld hole tile %ld\n", shell_id, tile_id);
							else
								write_record(" boundary shell %ld size %g tile %ld property %d\n", shell_id, inclusion -> imsz, tile_id, inclusion -> mat);
							
							tile_region_array[tile_regions++] = region_id;
						}
						else{

/* IMPORTANT: touching must be also considered to setup properly splitting plane */

							count = 0;
							if(x_intersection == YES || x_touching == YES)count++;
							if(y_intersection == YES || y_touching == YES)count++;
							if(z_intersection == YES || z_touching == YES)count++;

/* determine indices of fixed vertices corresponding to touching */

							ind_xx = 0;
							ind_yy = 1;
							ind_zz = 2;
							if(x_touching == NO){
								ind_yy--;
								ind_zz--;
							}
							if(y_touching == NO)ind_zz--;

							if(count == 1){

/* if ellipsoid intersects just one side, ellipsoid is split by plane parallel with the side and passing through ellipsoid center */

/* I do rely on fact that orthonormal base was already precomputed for particular side */

								copy_vec(point, tile.center);
								
								if(x_intersection == YES || x_touching == YES){
									if((xyz_intersection = x_intersection) == YES){
										xyz_patch_id = x_patch_id;
										for(m = 0; m < 2; m++)xyz_curve_id[m] = x_side_curve_array[m];
									}
									
									ind = x_ind;
									sub_vec(vec, x_extreme_point, ellipsoid -> center);
									
									point.x = ellipsoid -> center.x;
									
									transform_point_from_global_to_local(&point, &pnt, &(ellipsoid -> center), &trans_g2l);
									transform_local_from_ellipsoid_to_sphere(&pnt, &pnt, &scale_e2s);
									
									get_ellipsoid_xyz_side_intersection(&xyz_ellipse, &pnt, &vec_x, &dir_2, &dir_c);
									copy_vec(xyz_ellipse.normal, dir_x);
									
#ifdef DRAW_EXTREME_POINT_ELLIPSOID
									draw_point(&x_extreme_point, 0, orange);
#endif
								}
								else if(y_intersection == YES || y_touching == YES){
									if((xyz_intersection = y_intersection) == YES){
										xyz_patch_id = y_patch_id;
										for(m = 0; m < 2; m++)xyz_curve_id[m] = y_side_curve_array[m];
									}
									
									ind = y_ind;
									sub_vec(vec, y_extreme_point, ellipsoid -> center);
									
									point.y = ellipsoid -> center.y;
									
									transform_point_from_global_to_local(&point, &pnt, &(ellipsoid -> center), &trans_g2l);
									transform_local_from_ellipsoid_to_sphere(&pnt, &pnt, &scale_e2s);
									
									get_ellipsoid_xyz_side_intersection(&xyz_ellipse, &pnt, &vec_y, &dir_3, &dir_a);
									copy_vec(xyz_ellipse.normal, dir_y);
									
#ifdef DRAW_EXTREME_POINT_ELLIPSOID
									draw_point(&y_extreme_point, 0, orange);
#endif
								}
								else{
									if((xyz_intersection = z_intersection) == YES){
										xyz_patch_id = z_patch_id;
										for(m = 0; m < 2; m++)xyz_curve_id[m] = z_side_curve_array[m];
									}
									
									ind = z_ind;
									sub_vec(vec, z_extreme_point, ellipsoid -> center);
									
									point.z = ellipsoid -> center.z;
									
									transform_point_from_global_to_local(&point, &pnt, &(ellipsoid -> center), &trans_g2l);
									transform_local_from_ellipsoid_to_sphere(&pnt, &pnt, &scale_e2s);
									
									get_ellipsoid_xyz_side_intersection(&xyz_ellipse, &pnt, &vec_z, &dir_1, &dir_b);
									copy_vec(xyz_ellipse.normal, dir_z);
									
#ifdef DRAW_EXTREME_POINT_ELLIPSOID
									draw_point(&z_extreme_point, 0, orange);
#endif
								}
								
#ifdef DRAW_ELLIPSE
								if(draw == YES){
									draw_ellipse(&xyz_ellipse, 0, orange);
									ESIEventLoop(YES, "Splitting ellipse\n\n");
								}
#endif

/* node: since x|y|z_extreme_point is relevant extreme point outside the tile, vec_mid is pointing outside the tile;
	 change its orientation for negative side to ensure that defined surfaces have normal pointing out of inclusion */

								if(ind == 0)mul_vec(vec, -1.0);
								
								copy_mul_vec(vec_max, xyz_ellipse.major, xyz_ellipse.max);
								copy_mul_vec(vec_min, xyz_ellipse.minor, xyz_ellipse.min);
								copy_vec(vec_mid, vec);
								
								mul_vec(vec_min, 2.0);
								mul_vec(vec_mid, 4.0);
								
								add_vec(center, ellipsoid -> center, shift_vec);            /* shift ellipsoid center */
								
								add_vec(point1, center, vec_max);
								sub_vec(point2, center, vec_max);
								
								write_record("vertex %ld xyz "FMT3" tile %ld\n", ver_id1 = ++vertex_id, point1.x, point1.y, point1.z, tile_id);
								write_record("vertex %ld xyz "FMT3" tile %ld\n", ver_id2 = ++vertex_id, point2.x, point2.y, point2.z, tile_id);
								
								add_vec(point3, point1, vec_min);
								add_vec(point4, point2, vec_min);
								
								write_record("curve %ld vertex %ld %ld order 4 tile %ld\n", cur_id1 = ++curve_id, ver_id1, ver_id2, tile_id);
								write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point3.x, point3.y, point3.z, 1.0 / 3.0);
								write_record("polygon 2 xyz "FMT3" weight "FMT1"\n", point4.x, point4.y, point4.z, 1.0 / 3.0);
								
								sub_vec(point5, point1, vec_min);
								sub_vec(point6, point2, vec_min);
								
								write_record("curve %ld vertex %ld %ld order 4 tile %ld\n", cur_id2 = ++curve_id, ver_id1, ver_id2, tile_id);
								write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, 1.0 / 3.0);
								write_record("polygon 2 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, 1.0 / 3.0);
								
								write_record("curve %ld vertex %ld %ld order 4 tile %ld\n", cur_id3 = ++curve_id, ver_id1, ver_id1, tile_id);
								write_record("polygon 1 poly 0 weight "FMT1"\n", 1.0 / 3.0);
								write_record("polygon 2 poly 3 weight "FMT1"\n", 1.0 / 3.0);
								
								write_record("curve %ld vertex %ld %ld order 4 tile %ld\n", cur_id4 = ++curve_id, ver_id2, ver_id2, tile_id);
								write_record("polygon 1 poly 0 weight "FMT1"\n", 1.0 / 3.0);
								write_record("polygon 2 poly 3 weight "FMT1"\n", 1.0 / 3.0);
								
								add_vec(point3, point3, vec_mid);
								add_vec(point4, point4, vec_mid);
								add_vec(point5, point5, vec_mid);
								add_vec(point6, point6, vec_mid);
								
								if(ind == 0)
									write_record("surface %ld curve %ld %ld %ld %ld size %g tile %ld\n", sur_id1 = ++surface_id, cur_id1, cur_id3, cur_id2, cur_id4, inclusion -> bmsz, tile_id);
								else
									write_record("surface %ld curve %ld %ld %ld %ld size %g virtual\n", sur_id1 = ++surface_id, cur_id1, cur_id3, cur_id2, cur_id4, inclusion -> bmsz);
								write_record("polygon 1 1 xyz "FMT3" weight "FMT1"\n", point3.x, point3.y, point3.z, 1.0 / 9.0);
								write_record("polygon 2 1 xyz "FMT3" weight "FMT1"\n", point4.x, point4.y, point4.z, 1.0 / 9.0);
								write_record("polygon 1 2 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, 1.0 / 9.0);
								write_record("polygon 2 2 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, 1.0 / 9.0);
								
								mul_vec(vec_mid, 2.0);
								
								sub_vec(point3, point3, vec_mid);
								sub_vec(point4, point4, vec_mid);
								sub_vec(point5, point5, vec_mid);
								sub_vec(point6, point6, vec_mid);
								
								if(ind == 1)
									write_record("surface %ld curve %ld %ld %ld %ld size %g tile %ld\n", sur_id2 = ++surface_id, cur_id2, cur_id3, cur_id1, cur_id4, inclusion -> bmsz, tile_id);
								else
									write_record("surface %ld curve %ld %ld %ld %ld size %g virtual\n", sur_id2 = ++surface_id, cur_id2, cur_id3, cur_id1, cur_id4, inclusion -> bmsz);
								write_record("polygon 1 1 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, 1.0 / 9.0);
								write_record("polygon 2 1 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, 1.0 / 9.0);
								write_record("polygon 1 2 xyz "FMT3" weight "FMT1"\n", point3.x, point3.y, point3.z, 1.0 / 9.0);
								write_record("polygon 2 2 xyz "FMT3" weight "FMT1"\n", point4.x, point4.y, point4.z, 1.0 / 9.0);
								
								if(ind == 0){
									if(xyz_intersection == YES){
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld %ld %ld rate 0 size %g tile %ld\n",
																 ++shell_id, sur_id2, -cur_id1, cur_id2, xyz_curve_id[0], xyz_curve_id[1], inclusion -> bmsz, tile_id);
										
										write_record("region %ld boundary patch %ld boundary surface %ld boundary shell %ld", 
																 ++region_id, sign[1 - ind] * xyz_patch_id, sur_id1, shell_id);
									}
									else{
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld fixed vertex %ld rate 0 size %g tile %ld\n",
																 ++shell_id, sur_id2, -cur_id1, cur_id2, side_vertex_array[0], inclusion -> bmsz, tile_id);
										
										write_record("region %ld boundary surface %ld boundary shell %ld", ++region_id, sur_id1, shell_id);
									}
								}
								else{
									if(xyz_intersection == YES){
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld %ld %ld rate 0 size %g tile %ld\n",
																 ++shell_id, sur_id1, cur_id1, -cur_id2, -xyz_curve_id[0], -xyz_curve_id[1], inclusion -> bmsz, tile_id);
										
										write_record("region %ld boundary patch %ld boundary surface %ld boundary shell %ld", 
																 ++region_id, sign[1 - ind] * xyz_patch_id, sur_id2, shell_id);
									}
									else{
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld fixed vertex %ld rate 0 size %g tile %ld\n",
																 ++shell_id, sur_id1, cur_id1, -cur_id2, side_vertex_array[0], inclusion -> bmsz, tile_id);
										
										write_record("region %ld boundary surface %ld boundary shell %ld", ++region_id, sur_id2, shell_id);
									}
								}
								
								if(inclusion -> mat == 0)
									write_record(" hole tile %ld\n", tile_id);
								else
									write_record(" size %g tile %ld property %d\n", inclusion -> imsz, tile_id, inclusion -> mat);
								
								tile_region_array[tile_regions++] = region_id;
							}
							if(count == 2){

/* if ellipsoid intersects two sides, ellipsoid is split by plane passing through ellipsoid center and tile edge at which these sides intersect;
	 note that edge defined by intersection of the two sides should not be intersected

	 alternatively, similar approach as that used when count == 3 can be applied with the only difference that distance in the direction 
	 in which there is neither intersection nor touching is computed with respect to extreme points of the ellipsoid only
	 and distance in other directions are computed only from a single intersection ellipse or touching point;
	 NOT IMPLEMENTED */

								if(x_intersection == NO && x_touching == NO){
									if(x_count != 0)error_message(GENERAL_ERROR, "Unexpected situation");

/* vec1 and vec2 are computed so that:
	 normal.y > 0 for y_ind = 0
	 normal.y < 0 for y_ind = 1
	 normal.z < 0 for z_ind = 0
	 normal.z > 0 for z_ind = 1 */

									if(y_ind == z_ind){
										vec1.x = tile.corner[1].x - ellipsoid -> center.x;
										vec2.x = tile.corner[0].x - ellipsoid -> center.x;
									}
									else{
										vec1.x = tile.corner[0].x - ellipsoid -> center.x;
										vec2.x = tile.corner[1].x - ellipsoid -> center.x;
									}
									vec1.y = vec2.y = tile.corner[y_ind].y - ellipsoid -> center.y;
									vec1.z = vec2.z = tile.corner[z_ind].z - ellipsoid -> center.z;
								}
								else if(y_intersection == NO && y_touching == NO){
									if(y_count != 0)error_message(GENERAL_ERROR, "Unexpected situation");

/* vec1 and vec2 are computed so that:
	 normal.z > 0 for z_ind = 0
	 normal.z < 0 for z_ind = 1
	 normal.x < 0 for x_ind = 0
	 normal.x > 0 for x_ind = 1 */

									if(z_ind == x_ind){
										vec1.y = tile.corner[1].y - ellipsoid -> center.y;
										vec2.y = tile.corner[0].y - ellipsoid -> center.y;
									}
									else{
										vec1.y = tile.corner[0].y - ellipsoid -> center.y;
										vec2.y = tile.corner[1].y - ellipsoid -> center.y;
									}
									vec1.z = vec2.z = tile.corner[z_ind].z - ellipsoid -> center.z;
									vec1.x = vec2.x = tile.corner[x_ind].x - ellipsoid -> center.x;
								}
								else{
									if(z_count != 0)error_message(GENERAL_ERROR, "Unexpected situation");

/* vec1 and vec2 are computed so that:
	 normal.x > 0 for x_ind = 0
	 normal.x < 0 for x_ind = 1
	 normal.y < 0 for y_ind = 0
	 normal.y > 0 for y_ind = 1 */

									if(x_ind == y_ind){
										vec1.z = tile.corner[1].z - ellipsoid -> center.z;
										vec2.z = tile.corner[0].z - ellipsoid -> center.z;
									}
									else{
										vec1.z = tile.corner[0].z - ellipsoid -> center.z;
										vec2.z = tile.corner[1].z - ellipsoid -> center.z;
									}
									vec1.x = vec2.x = tile.corner[x_ind].x - ellipsoid -> center.x;
									vec1.y = vec2.y = tile.corner[y_ind].y - ellipsoid -> center.y;
								}
								
								cross_product(normal, vec1, vec2);
								size = 1.0 / sqrt(size_vec(normal));
								mul_vec(normal, size);
								
/* compute auxiliary intersection parallel with computed plane passing through ellipsoid */

								add_mul_vec(point, ellipsoid -> center, normal, 1.0, ellipsoid -> min * 0.9);               /* IMPORTANT: ensure that plane intersects ellipsoid */
								get_ellipsoid_plane_intersection(ellipsoid, &point, &normal, &xyz_ellipse);
								sub_vec(vec, xyz_ellipse.center, ellipsoid -> center);
								
								get_ellipsoid_plane_intersection(ellipsoid, &(ellipsoid -> center), &normal, &xyz_ellipse);
								get_ellipsoid_line_center_intersection(ellipsoid, &vec, &point1, &point2);

#ifdef DRAW_EXTREME_POINT_ELLIPSOID
								if(draw == YES){
									draw_point(&point1, 0, green);
									draw_point(&point2, 0, orange);
									ESIEventLoop(YES, "Extreme points on associated diameter");
								}
#endif

/* vec_mid is oriented as normal */

								copy_mul_vec(vec_max, xyz_ellipse.major, xyz_ellipse.max);
								copy_mul_vec(vec_min, xyz_ellipse.minor, xyz_ellipse.min);
								sub_vec(vec_mid, point1, ellipsoid -> center);
								
								mul_vec(vec_min, 2.0);
								mul_vec(vec_mid, 4.0);
								
								add_vec(center, ellipsoid -> center, shift_vec);            /* shift ellipsoid center */
								
								add_vec(point1, center, vec_max);
								sub_vec(point2, center, vec_max);
								add_vec(point3, point1, vec_min);
								add_vec(point4, point2, vec_min);
								sub_vec(point5, point1, vec_min);
								sub_vec(point6, point2, vec_min);

								write_record("vertex %ld xyz "FMT3" tile %ld\n", ver_id1 = ++vertex_id, point1.x, point1.y, point1.z, tile_id);
								write_record("vertex %ld xyz "FMT3" tile %ld\n", ver_id2 = ++vertex_id, point2.x, point2.y, point2.z, tile_id);
								
								write_record("curve %ld vertex %ld %ld order 4 tile %ld\n", cur_id1 = ++curve_id, ver_id1, ver_id2, tile_id);
								write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point3.x, point3.y, point3.z, 1.0 / 3.0);
								write_record("polygon 2 xyz "FMT3" weight "FMT1"\n", point4.x, point4.y, point4.z, 1.0 / 3.0);
								
								write_record("curve %ld vertex %ld %ld order 4 tile %ld\n", cur_id2 = ++curve_id, ver_id1, ver_id2, tile_id);
								write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, 1.0 / 3.0);
								write_record("polygon 2 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, 1.0 / 3.0);
								
								write_record("curve %ld vertex %ld %ld order 4 tile %ld\n", cur_id3 = ++curve_id, ver_id1, ver_id1, tile_id);
								write_record("polygon 1 poly 0 weight "FMT1"\n", 1.0 / 3.0);
								write_record("polygon 2 poly 3 weight "FMT1"\n", 1.0 / 3.0);
								
								write_record("curve %ld vertex %ld %ld order 4 tile %ld\n", cur_id4 = ++curve_id, ver_id2, ver_id2, tile_id);
								write_record("polygon 1 poly 0 weight "FMT1"\n", 1.0 / 3.0);
								write_record("polygon 2 poly 3 weight "FMT1"\n", 1.0 / 3.0);
								
								add_vec(point3, point3, vec_mid);
								add_vec(point4, point4, vec_mid);
								add_vec(point5, point5, vec_mid);
								add_vec(point6, point6, vec_mid);
								
								write_record("surface %ld curve %ld %ld %ld %ld size %g virtual\n", sur_id1 = ++surface_id, cur_id1, cur_id3, cur_id2, cur_id4, inclusion -> bmsz);
								write_record("polygon 1 1 xyz "FMT3" weight "FMT1"\n", point3.x, point3.y, point3.z, 1.0 / 9.0);
								write_record("polygon 2 1 xyz "FMT3" weight "FMT1"\n", point4.x, point4.y, point4.z, 1.0 / 9.0);
								write_record("polygon 1 2 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, 1.0 / 9.0);
								write_record("polygon 2 2 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, 1.0 / 9.0);
								
								mul_vec(vec_mid, 2.0);
								
								sub_vec(point3, point3, vec_mid);
								sub_vec(point4, point4, vec_mid);
								sub_vec(point5, point5, vec_mid);
								sub_vec(point6, point6, vec_mid);
								
								write_record("surface %ld curve %ld %ld %ld %ld size %g virtual\n", sur_id2 = ++surface_id, cur_id2, cur_id3, cur_id1, cur_id4, inclusion -> bmsz);
								write_record("polygon 1 1 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, 1.0 / 9.0);
								write_record("polygon 2 1 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, 1.0 / 9.0);
								write_record("polygon 1 2 xyz "FMT3" weight "FMT1"\n", point3.x, point3.y, point3.z, 1.0 / 9.0);
								write_record("polygon 2 2 xyz "FMT3" weight "FMT1"\n", point4.x, point4.y, point4.z, 1.0 / 9.0);
								
								if(x_intersection == NO && x_touching == NO){
									if(z_intersection == YES)
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld %ld %ld rate 0 size %g tile %ld\n",
																 shll_id1 = ++shell_id, sur_id1, cur_id1, -cur_id2, z_side_curve_array[0] * sign[z_ind], z_side_curve_array[1] * sign[z_ind], inclusion -> bmsz, tile_id);
									else /* z_touching == YES */
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld fixed vertex %ld rate 0 size %g tile %ld\n",
																 shll_id1 = ++shell_id, sur_id1, cur_id1, -cur_id2, side_vertex_array[ind_zz], inclusion -> bmsz, tile_id);
									
									if(y_intersection == YES)
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld %ld %ld rate 0 size %g tile %ld\n",
																 shll_id2 = ++shell_id, sur_id2, -cur_id1, cur_id2, y_side_curve_array[0] * sign[y_ind], y_side_curve_array[1] * sign[y_ind], inclusion -> bmsz, tile_id);
									else /* y_touching == YES */
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld fixed vertex %ld rate 0 size %g tile %ld\n",
																 shll_id2 = ++shell_id, sur_id2, -cur_id1, cur_id2, side_vertex_array[ind_yy], inclusion -> bmsz, tile_id);
									
									write_record("region %ld boundary shell %ld %ld", ++region_id, shll_id1, shll_id2);
									if(y_intersection == YES || z_intersection == YES){
										write_record(" boundary patch");
										if(y_intersection == YES)write_record(" %ld", sign[1 - y_ind] * y_patch_id);
										if(z_intersection == YES)write_record(" %ld", sign[1 - z_ind] * z_patch_id);
									}
								}
								else if(y_intersection == NO && y_touching == NO){
									if(x_intersection == YES)
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld %ld %ld rate 0 size %g tile %ld\n",
																 shll_id1 = ++shell_id, sur_id1, cur_id1, -cur_id2, x_side_curve_array[0] * sign[x_ind], x_side_curve_array[1] * sign[x_ind], inclusion -> bmsz, tile_id);
									else /* x_touching == YES */
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld fixed vertex %ld rate 0 size %g tile %ld\n",
																 shll_id1 = ++shell_id, sur_id1, cur_id1, -cur_id2, side_vertex_array[ind_xx], inclusion -> bmsz, tile_id);
									
									if(z_intersection == YES)
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld %ld %ld rate 0 size %g tile %ld\n",
																	 shll_id2 = ++shell_id, sur_id2, -cur_id1, cur_id2, z_side_curve_array[0] * sign[x_ind], z_side_curve_array[1] * sign[z_ind], inclusion -> bmsz, tile_id);
									else /* z_touching == YES */
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld fixed vertex %ld rate 0 size %g tile %ld\n",
																 shll_id2 = ++shell_id, sur_id2, -cur_id1, cur_id2, side_vertex_array[ind_zz], inclusion -> bmsz, tile_id);
									
									write_record("region %ld boundary shell %ld %ld", ++region_id, shll_id1, shll_id2);
									if(z_intersection == YES || x_intersection == YES){
										write_record(" boundary patch");
										if(z_intersection == YES)write_record(" %ld", sign[1 - z_ind] * z_patch_id);
										if(x_intersection == YES)write_record(" %ld", sign[1 - x_ind] * x_patch_id);
									}
								}
								else{
									if(y_intersection == YES)
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld %ld %ld rate 0 size %g tile %ld\n",
																 shll_id1 = ++shell_id, sur_id1, cur_id1, -cur_id2, y_side_curve_array[0] * sign[y_ind], y_side_curve_array[1] * sign[y_ind], inclusion -> bmsz, tile_id);
									else /* y_touching == YES */
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld fixed vertex %ld  rate 0 size %g tile %ld\n",
																 shll_id1 = ++shell_id, sur_id1, cur_id1, -cur_id2, side_vertex_array[ind_yy], inclusion -> bmsz, tile_id);
									
									if(x_intersection == YES)
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld %ld %ld rate 0 size %g tile %ld\n",
																 shll_id2 = ++shell_id, sur_id2, -cur_id1, cur_id2, x_side_curve_array[0] * sign[x_ind], x_side_curve_array[1] * sign[x_ind], inclusion -> bmsz, tile_id);
									else /* x_touching == YES */
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld fixed vertex %ld rate 0 size %g tile %ld\n",
																 shll_id2 = ++shell_id, sur_id2, -cur_id1, cur_id2, side_vertex_array[ind_xx], inclusion -> bmsz, tile_id);
									
									write_record("region %ld boundary shell %ld %ld", ++region_id, shll_id1, shll_id2);
									if(x_intersection == YES || y_intersection == YES){
										write_record(" boundary patch");
										if(x_intersection == YES)write_record(" %ld", sign[1 - x_ind] * x_patch_id);
										if(y_intersection == YES)write_record(" %ld", sign[1 - y_ind] * y_patch_id);
									}
								}
								
								if(inclusion -> mat == 0)
									write_record(" hole tile %ld\n", tile_id);
								else
									write_record(" size %g tile %ld property %d\n", inclusion -> imsz, tile_id, inclusion -> mat);
								
								tile_region_array[tile_regions++] = region_id;
							}
							if(count == 3){

/* if ellipsoid intersects all three sides, ellipsoid is split by plane perpendicular to that x|y|z direction 
	 in which the minimal distance of intersection ellipses or touching points from side perpendicular to the direction is maximal;

	 distance of intersection ellipses is already computed;
	 distance of touching points must be computed */

								if(x_intersection == NO){
									diff_xy = tile.corner[y_ind].y - x_extreme_point.y;
									diff_xz = tile.corner[z_ind].z - x_extreme_point.z;
								}
#if 0
								else{
									fprintf(stderr, "\nReusing:\n");
									fprintf(stderr, "diff_xy %20.13e\n", diff_xy);
									fprintf(stderr, "diff_xz %20.13e\n", diff_xz);
								}
#endif
								if(y_intersection == NO){
									diff_yz = tile.corner[z_ind].z - y_extreme_point.z;
									diff_yx = tile.corner[x_ind].x - y_extreme_point.x;
								}
#if 0
								else{
									fprintf(stderr, "\nReusing:\n");
									fprintf(stderr, "diff_yz %20.13e\n", diff_yz);
									fprintf(stderr, "diff_yx %20.13e\n", diff_yx);
								}
#endif
								if(z_intersection == NO){
									diff_zx = tile.corner[x_ind].x - z_extreme_point.x;
									diff_zy = tile.corner[y_ind].y - z_extreme_point.y;
								}
#if 0
								else{
									fprintf(stderr, "\nReusing:\n");
									fprintf(stderr, "diff_zx %20.13e\n", diff_zx);
									fprintf(stderr, "diff_zy %20.13e\n", diff_zy);
								}
#endif
								a_diff_xy = fabs(diff_xy);
								a_diff_xz = fabs(diff_xz);
								a_diff_yz = fabs(diff_yz);
								a_diff_yx = fabs(diff_yx);
								a_diff_zx = fabs(diff_zx);
								a_diff_zy = fabs(diff_zy);
								
								diff_x = (a_diff_yx < a_diff_zx) ? a_diff_yx : a_diff_zx;
								diff_y = (a_diff_zy < a_diff_xy) ? a_diff_zy : a_diff_xy;
								diff_z = (a_diff_xz < a_diff_yz) ? a_diff_xz : a_diff_yz;
								
/* it is useful to consider also distances of intersection ellipses or touching points from opposite extreme points 
	 which are inside tile with respect to particular tile side */
							
/* distance of opposite ellipse extreme point or touching point from tile side (absolute value) */

								if(x_intersection == YES){
									len_xy = 2.0 * fabs(x_ellipse.center.y - tile.corner[y_ind].y) - a_diff_xy;
									len_xz = 2.0 * fabs(x_ellipse.center.z - tile.corner[z_ind].z) - a_diff_xz;
								}
								else{
									len_xy = a_diff_xy;
									len_xz = a_diff_xz;
								}
								if(y_intersection == YES){
									len_yz = 2.0 * fabs(y_ellipse.center.z - tile.corner[z_ind].z) - a_diff_yz;
									len_yx = 2.0 * fabs(y_ellipse.center.x - tile.corner[x_ind].x) - a_diff_yx;
								}
								else{
									len_yz = a_diff_yz;
									len_yx = a_diff_yx;
								}
								if(z_intersection == YES){
									len_zx = 2.0 * fabs(z_ellipse.center.x - tile.corner[x_ind].x) - a_diff_zx;
									len_zy = 2.0 * fabs(z_ellipse.center.y - tile.corner[y_ind].y) - a_diff_zy;
								}
								else{
									len_zx = a_diff_zx;
									len_zy = a_diff_zy;
								}

/* distance of opposite ellipsoid extreme point from tile side (absolute value) */

								len_x = fabs(2.0 * ellipsoid -> center.x - x_extreme_point.x - tile.corner[x_ind].x);
								len_y = fabs(2.0 * ellipsoid -> center.y - y_extreme_point.y - tile.corner[y_ind].y);
								len_z = fabs(2.0 * ellipsoid -> center.z - z_extreme_point.z - tile.corner[z_ind].z);
								
/* I do rely on fact that len_yx < len_x > len_yz, etc */

								a_diff_yx = len_x - len_yx;
								a_diff_zx = len_x - len_zx;
								
								a_diff_zy = len_y - len_zy;
								a_diff_xy = len_y - len_xy;
								
								a_diff_xz = len_z - len_xz;
								a_diff_yz = len_z - len_yz;
								
								len_x = (a_diff_yx < a_diff_zx) ? a_diff_yx : a_diff_zx;
								len_y = (a_diff_zy < a_diff_xy) ? a_diff_zy : a_diff_xy;
								len_z = (a_diff_xz < a_diff_yz) ? a_diff_xz : a_diff_yz;
								
								opposite_x = opposite_y = opposite_z = NO;
								
								if(len_x > diff_x){
									diff_x = len_x;
									opposite_x = YES;
								}
								if(len_y > diff_y){
									diff_y = len_y;
									opposite_y = YES;
								}
								if(len_z > diff_z){
									diff_z = len_z;
									opposite_z = YES;
								}
								
/* process splitting plane */

								copy_vec(point, tile.center);
								
#ifdef SPECIAL_TESTING
#if SPLITTING_PLANE == XX
								diff_y = diff_z = diff_x * 0.99;         /* ==> splitting plane parallel with x-side */
								opposite_x = (logic)OPPOSITE_PLANE;
#endif
#if SPLITTING_PLANE == YY
								diff_z = diff_x = diff_y * 0.99;         /* ==> splitting plane parallel with y-side */
								opposite_y = (logic)OPPOSITE_PLANE;
#endif
#if SPLITTING_PLANE == ZZ
								diff_x = diff_y = diff_z * 0.99;         /* ==> splitting plane parallel with z-side */
								opposite_z = (logic)OPPOSITE_PLANE;
#endif
#endif
								if(diff_x >= diff_y && diff_x >= diff_z){     /* IMPORTANT: using >= is relevant !!! */
									side = X + 1;
									ind = x_ind;
									
									if((opposite_xyz = opposite_x) == NO)
										point.x = tile.corner[ind].x + sign[ind] * diff_x / 2.0;
									else
										point.x = 2.0 * ellipsoid -> center.x - x_extreme_point.x - sign[ind] * diff_x / 2.0;
									
									transform_point_from_global_to_local(&point, &pnt, &(ellipsoid -> center), &trans_g2l);
									transform_local_from_ellipsoid_to_sphere(&pnt, &pnt, &scale_e2s);
									
									get_ellipsoid_xyz_side_intersection(&xyz_ellipse, &pnt, &vec_x, &dir_2, &dir_c);
									copy_vec(xyz_ellipse.normal, dir_x);
								}
								else if(diff_y >= diff_z && diff_y >= diff_x){    /* IMPORTANT: using >= is relevant !!! */
									side = Y + 1;
									ind = y_ind;
									
									if((opposite_xyz = opposite_y) == NO)
										point.y = tile.corner[ind].y + sign[ind] * diff_y / 2.0;
									else
										point.y = 2.0 * ellipsoid -> center.y - y_extreme_point.y - sign[ind] * diff_y / 2.0;
									
									transform_point_from_global_to_local(&point, &pnt, &(ellipsoid -> center), &trans_g2l);
									transform_local_from_ellipsoid_to_sphere(&pnt, &pnt, &scale_e2s);
									
									get_ellipsoid_xyz_side_intersection(&xyz_ellipse, &pnt, &vec_y, &dir_3, &dir_a);
									copy_vec(xyz_ellipse.normal, dir_y);
								}
								else{
									side = Z + 1;
									ind = z_ind;
									
									if((opposite_xyz = opposite_z) == NO)
										point.z = tile.corner[ind].z + sign[ind] * diff_z / 2.0;
									else
										point.z = 2.0 * ellipsoid -> center.z - z_extreme_point.z - sign[ind] * diff_z / 2.0;
									
									transform_point_from_global_to_local(&point, &pnt, &(ellipsoid -> center), &trans_g2l);
									transform_local_from_ellipsoid_to_sphere(&pnt, &pnt, &scale_e2s);
									
									get_ellipsoid_xyz_side_intersection(&xyz_ellipse, &pnt, &vec_z, &dir_1, &dir_b);
									copy_vec(xyz_ellipse.normal, dir_z);
								}
								
#ifdef DRAW_ELLIPSE
								if(draw == YES){
									draw_ellipse(&xyz_ellipse, 0, orange);
									ESIEventLoop(YES, "Splitting ellipse\n\n");
								}
#endif
								copy_mul_vec(vec_max, xyz_ellipse.major, xyz_ellipse.max);
								copy_mul_vec(vec_min, xyz_ellipse.minor, xyz_ellipse.min);
								
								mul_vec(vec_min, 2.0);
								
								add_vec(center, xyz_ellipse.center, shift_vec);            /* shift ellipse center */
								
								add_vec(point1, center, vec_max);
								sub_vec(point2, center, vec_max);
								add_vec(point3, point1, vec_min);
								add_vec(point4, point2, vec_min);
								sub_vec(point5, point1, vec_min);
								sub_vec(point6, point2, vec_min);
								
								write_record("vertex %ld xyz "FMT3" tile %ld\n", ver_id1 = ++vertex_id, point1.x, point1.y, point1.z, tile_id);
								write_record("vertex %ld xyz "FMT3" tile %ld\n", ver_id2 = ++vertex_id, point2.x, point2.y, point2.z, tile_id);
								
								write_record("curve %ld vertex %ld %ld order 4 tile %ld\n", cur_id5 = ++curve_id, ver_id1, ver_id2, tile_id);
								write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point3.x, point3.y, point3.z, 1.0 / 3.0);
								write_record("polygon 2 xyz "FMT3" weight "FMT1"\n", point4.x, point4.y, point4.z, 1.0 / 3.0);
								
								write_record("curve %ld vertex %ld %ld order 4 tile %ld\n", cur_id6 = ++curve_id, ver_id1, ver_id2, tile_id);
								write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, 1.0 / 3.0);
								write_record("polygon 2 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, 1.0 / 3.0);

/* only the bigger part of the ellipsoid with respect to the splitting plane is modelled as shell with biquadratic bg surface;
	 the smaller part is modelled as shell with exact half-ellipsoid bg surface to avoid problems with singularity for rate close to sqrt(2) - 1 */

/* larger shell */

/* IMPORTANT: do not use splitting plane as base of biquadratic surface
	 shift the plane slightly in the direction out from ellipsoid center to eliminate singularity and shape deformation along base of biquadratic surface;
	 shift is accomplished by increasing the rate;
	 I do rely on fact that orthonormal base was already precomputed for particular side */

								if(side == X + 1){
									if(opposite_xyz == NO){
										sub_vec(vec, ellipsoid -> center, x_extreme_point);
									}
									else{
										sub_vec(vec, x_extreme_point, ellipsoid -> center);
									}
									
									rate_x = (ellipsoid -> center.x - point.x) / vec.x;
									rate = rate_x + (1.0 - rate_x) * INC_RATE;
									point.x = ellipsoid -> center.x - rate * vec.x;
									
									transform_point_from_global_to_local(&point, &pnt, &(ellipsoid -> center), &trans_g2l);
									transform_local_from_ellipsoid_to_sphere(&pnt, &pnt, &scale_e2s);
									
									get_ellipsoid_xyz_side_intersection(&xyz_ellipse, &pnt, &vec_x, &dir_2, &dir_c);
									copy_vec(xyz_ellipse.normal, dir_x);
								}
								else if(side == Y + 1){
									if(opposite_xyz == NO){
										sub_vec(vec, ellipsoid -> center, y_extreme_point);
									}
									else{
										sub_vec(vec, y_extreme_point, ellipsoid -> center);
									}
									
									rate_y = (ellipsoid -> center.y - point.y) / vec.y;
									rate = rate_y + (1.0 - rate_y) * INC_RATE;
									point.y = ellipsoid -> center.y - rate * vec.y;
									
									transform_point_from_global_to_local(&point, &pnt, &(ellipsoid -> center), &trans_g2l);
									transform_local_from_ellipsoid_to_sphere(&pnt, &pnt, &scale_e2s);
									
									get_ellipsoid_xyz_side_intersection(&xyz_ellipse, &pnt, &vec_y, &dir_3, &dir_a);
									copy_vec(xyz_ellipse.normal, dir_y);
								}
								else{
									if(opposite_xyz == NO){
										sub_vec(vec, ellipsoid -> center, z_extreme_point);
									}
									else{
										sub_vec(vec, z_extreme_point, ellipsoid -> center);
									}
									
									rate_z = (ellipsoid -> center.z - point.z) / vec.z;
									rate = rate_z + (1.0 - rate_z) * INC_RATE;
									point.z = ellipsoid -> center.z - rate * vec.z;
									
									transform_point_from_global_to_local(&point, &pnt, &(ellipsoid -> center), &trans_g2l);
									transform_local_from_ellipsoid_to_sphere(&pnt, &pnt, &scale_e2s);
									
									get_ellipsoid_xyz_side_intersection(&xyz_ellipse, &pnt, &vec_z, &dir_1, &dir_b);
									copy_vec(xyz_ellipse.normal, dir_z);
								}
								
								weight = -0.5 * ((2.0 + SQRT2) * rate + SQRT2);
								rate = 1.0 + (1.0 + SQRT2) * (1.0 + rate) / weight;
								add_mul_vec(point, ellipsoid -> center, vec, 1.0, rate);
								
#ifdef DRAW_ELLIPSE
								draw_point(&point, 0, orange);
								draw_ellipse(&xyz_ellipse, 0, orange);
								if(draw == YES)ESIEventLoop(YES, "Base ellipse of inexact ellipsoid + ctrl point\n\n");
#endif
								copy_mul_vec(vec_max, xyz_ellipse.major, xyz_ellipse.max);
								copy_mul_vec(vec_min, xyz_ellipse.minor, xyz_ellipse.min);
								
								add_vec(center, xyz_ellipse.center, shift_vec);            /* shift ellipse center */
								
								add_vec(point1, center, vec_max);
								sub_vec(point2, center, vec_max);
								add_vec(point3, center, vec_min);
								sub_vec(point4, center, vec_min);
								
								write_record("vertex %ld xyz "FMT3" virtual\n", ver_id1 = ++vertex_id, point1.x, point1.y, point1.z);
								write_record("vertex %ld xyz "FMT3" virtual\n", ver_id2 = ++vertex_id, point2.x, point2.y, point2.z);
								write_record("vertex %ld xyz "FMT3" virtual\n", ver_id3 = ++vertex_id, point3.x, point3.y, point3.z);
								write_record("vertex %ld xyz "FMT3" virtual\n", ver_id4 = ++vertex_id, point4.x, point4.y, point4.z);
								
								add_vec(point5, point1, vec_min);
								add_vec(point6, point2, vec_min);
								
								write_record("curve %ld vertex %ld %ld order 3 virtual\n", cur_id1 = ++curve_id, ver_id1, ver_id3);
								write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, SQRT1_2);
								write_record("curve %ld vertex %ld %ld order 3 virtual\n", cur_id2 = ++curve_id, ver_id3, ver_id2);
								write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, SQRT1_2);
								
								sub_vec(point5, point1, vec_min);
								sub_vec(point6, point2, vec_min);
								
								write_record("curve %ld vertex %ld %ld order 3 virtual\n", cur_id3 = ++curve_id, ver_id2, ver_id4);
								write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, SQRT1_2);
								write_record("curve %ld vertex %ld %ld order 3 virtual\n", cur_id4 = ++curve_id, ver_id4, ver_id1);
								write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, SQRT1_2);
								
								if((int)opposite_xyz == ind)
									write_record("surface %ld curve %ld %ld %ld %ld", ++surface_id, cur_id1, cur_id4, cur_id3, cur_id2);
								else
									write_record("surface %ld curve %ld %ld %ld %ld", ++surface_id, cur_id1, cur_id2, cur_id3, cur_id4);
#ifdef USE_VIRTUAL_ELLIPSOID
								write_record(" rate 0 virtual\n");
#else
								write_record(" virtual\n");
#endif

								add_vec(point, point, shift_vec);
								write_record("polygon 1 1 xyz "FMT3" weight "FMT1"\n", point.x, point.y, point.z, weight);
								
								if(inclusion -> support == NO){
									inclusion -> support = YES;
									
									write_record("ellipsoid %ld max "FMT1" min "FMT1" mid "FMT1" \\\n", inclusion -> id, ellipsoid -> max, ellipsoid -> min, ellipsoid -> mid);
									write_record("   major "FMT3" \\\n", ellipsoid -> major.x, ellipsoid -> major.y, ellipsoid -> major.z);
									write_record("   minor "FMT3" \\\n", ellipsoid -> minor.x, ellipsoid -> minor.y, ellipsoid -> minor.z);
									write_record("   middle "FMT3"\n", ellipsoid -> middle.x, ellipsoid -> middle.y, ellipsoid -> middle.z);
								}

#if 0
								if(opposite_xyz == NO){
									if(ind == 0)
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld", shll_id1 = ++shell_id, surface_id, cur_id5, -cur_id6);
									else
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld", shll_id1 = ++shell_id, surface_id, -cur_id5, cur_id6);
								}
								else{
									if(ind == 0)
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld", shll_id1 = ++shell_id, surface_id, -cur_id5, cur_id6);
									else
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld", shll_id1 = ++shell_id, surface_id, cur_id5, -cur_id6);
								}
#else
								if((int)opposite_xyz == ind)
									write_record("shell %ld bgsurface %ld boundary curve %ld %ld", shll_id1 = ++shell_id, surface_id, cur_id5, -cur_id6);
								else
									write_record("shell %ld bgsurface %ld boundary curve %ld %ld", shll_id1 = ++shell_id, surface_id, -cur_id5, cur_id6);
#endif
								
								if(side == X + 1){
									if(x_intersection == YES && opposite_x == YES)write_record(" %ld %ld", x_side_curve_array[0] * sign[x_ind], x_side_curve_array[1] * sign[x_ind]);
									if(y_intersection == YES)write_record(" %ld %ld", y_side_curve_array[0] * sign[y_ind], y_side_curve_array[1] * sign[y_ind]);
									if(z_intersection == YES)write_record(" %ld %ld", z_side_curve_array[0] * sign[z_ind], z_side_curve_array[1] * sign[z_ind]);

									if((x_touching == YES && opposite_x == YES) || y_touching == YES || z_touching == YES){
										write_record(" fixed vertex");
										if(x_touching == YES && opposite_x == YES)write_record(" %ld", side_vertex_array[ind_xx]);
										if(y_touching == YES)write_record(" %ld", side_vertex_array[ind_yy]);
										if(z_touching == YES)write_record(" %ld", side_vertex_array[ind_zz]);
									}
								}
								else if(side == Y + 1){
									if(y_intersection == YES && opposite_y == YES)write_record(" %ld %ld", y_side_curve_array[0] * sign[y_ind], y_side_curve_array[1] * sign[y_ind]);
									if(z_intersection == YES)write_record(" %ld %ld", z_side_curve_array[0] * sign[z_ind], z_side_curve_array[1] * sign[z_ind]);
									if(x_intersection == YES)write_record(" %ld %ld", x_side_curve_array[0] * sign[x_ind], x_side_curve_array[1] * sign[x_ind]);

									if((y_touching == YES && opposite_y == YES) || z_touching == YES || x_touching == YES){
										write_record(" fixed vertex");
										if(y_touching == YES && opposite_y == YES)write_record(" %ld", side_vertex_array[ind_yy]);
										if(z_touching == YES)write_record(" %ld", side_vertex_array[ind_zz]);
										if(x_touching == YES)write_record(" %ld", side_vertex_array[ind_xx]);
									}
								}
								else{
									if(z_intersection == YES && opposite_z == YES)write_record(" %ld %ld", z_side_curve_array[0] * sign[z_ind], z_side_curve_array[1] * sign[z_ind]);
									if(x_intersection == YES)write_record(" %ld %ld", x_side_curve_array[0] * sign[x_ind], x_side_curve_array[1] * sign[x_ind]);
									if(y_intersection == YES)write_record(" %ld %ld", y_side_curve_array[0] * sign[y_ind], y_side_curve_array[1] * sign[y_ind]);

									if((z_touching == YES && opposite_z == YES) || x_touching == YES || y_touching == YES){
										write_record(" fixed vertex");
										if(z_touching == YES && opposite_z == YES)write_record(" %ld", side_vertex_array[ind_zz]);
										if(x_touching == YES)write_record(" %ld", side_vertex_array[ind_xx]);
										if(y_touching == YES)write_record(" %ld", side_vertex_array[ind_yy]);
									}
								}

								add_vec(center, ellipsoid -> center, shift_vec);            /* shift ellipsoid center */
								
								write_record(" rate 0 size %g tile %ld ellipsoid %ld center "FMT3"\n", 
														 inclusion -> bmsz, tile_id, inclusion -> id, center.x, center.y, center.z);

/* smaller shell */

								if(side == X + 1){
									sub_vec(vec, x_extreme_point, ellipsoid -> center);
									point.x = ellipsoid -> center.x;
									
									transform_point_from_global_to_local(&point, &pnt, &(ellipsoid -> center), &trans_g2l);
									transform_local_from_ellipsoid_to_sphere(&pnt, &pnt, &scale_e2s);
									
									get_ellipsoid_xyz_side_intersection(&xyz_ellipse, &pnt, &vec_x, &dir_2, &dir_c);
									copy_vec(xyz_ellipse.normal, dir_x);
								}
								else if(side == Y + 1){
									sub_vec(vec, y_extreme_point, ellipsoid -> center);
									point.y = ellipsoid -> center.y;
									
									transform_point_from_global_to_local(&point, &pnt, &(ellipsoid -> center), &trans_g2l);
									transform_local_from_ellipsoid_to_sphere(&pnt, &pnt, &scale_e2s);
									
									get_ellipsoid_xyz_side_intersection(&xyz_ellipse, &pnt, &vec_y, &dir_3, &dir_a);
									copy_vec(xyz_ellipse.normal, dir_y);
								}
								else{
									sub_vec(vec, z_extreme_point, ellipsoid -> center);
									point.z = ellipsoid -> center.z;
									
									transform_point_from_global_to_local(&point, &pnt, &(ellipsoid -> center), &trans_g2l);
									transform_local_from_ellipsoid_to_sphere(&pnt, &pnt, &scale_e2s);
									
									get_ellipsoid_xyz_side_intersection(&xyz_ellipse, &pnt, &vec_z, &dir_1, &dir_b);
									copy_vec(xyz_ellipse.normal, dir_z);
								}
								
/* node: since x|y|z_extreme_point is relevant extreme point outside the tile, vec_mid is pointing outside the tile;
	 change its orientation for opposite spliting plane */

								if(opposite_xyz == YES)mul_vec(vec, -1.0);
								
								copy_mul_vec(vec_max, xyz_ellipse.major, xyz_ellipse.max);
								copy_mul_vec(vec_min, xyz_ellipse.minor, xyz_ellipse.min);
								copy_vec(vec_mid, vec);
								
								mul_vec(vec_min, 2.0);
								mul_vec(vec_mid, 4.0);
								
								add_vec(center, ellipsoid -> center, shift_vec);            /* shift ellipsoid center */
								
								add_vec(point1, center, vec_max);
								sub_vec(point2, center, vec_max);
								
								write_record("vertex %ld xyz "FMT3" virtual\n", ver_id1 = ++vertex_id, point1.x, point1.y, point1.z);
								write_record("vertex %ld xyz "FMT3" virtual\n", ver_id2 = ++vertex_id, point2.x, point2.y, point2.z);
								
								add_vec(point3, point1, vec_min);
								add_vec(point4, point2, vec_min);
								
								write_record("curve %ld vertex %ld %ld order 4 virtual\n", cur_id1 = ++curve_id, ver_id1, ver_id2);
								write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point3.x, point3.y, point3.z, 1.0 / 3.0);
								write_record("polygon 2 xyz "FMT3" weight "FMT1"\n", point4.x, point4.y, point4.z, 1.0 / 3.0);
								
								sub_vec(point5, point1, vec_min);
								sub_vec(point6, point2, vec_min);
								
								write_record("curve %ld vertex %ld %ld order 4 virtual\n", cur_id2 = ++curve_id, ver_id1, ver_id2);
								write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, 1.0 / 3.0);
								write_record("polygon 2 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, 1.0 / 3.0);
								
								write_record("curve %ld vertex %ld %ld order 4 virtual\n", cur_id3 = ++curve_id, ver_id1, ver_id1);
								write_record("polygon 1 poly 0 weight "FMT1"\n", 1.0 / 3.0);
								write_record("polygon 2 poly 3 weight "FMT1"\n", 1.0 / 3.0);
								
								write_record("curve %ld vertex %ld %ld order 4 virtual\n", cur_id4 = ++curve_id, ver_id2, ver_id2);
								write_record("polygon 1 poly 0 weight "FMT1"\n", 1.0 / 3.0);
								write_record("polygon 2 poly 3 weight "FMT1"\n", 1.0 / 3.0);
								
								add_vec(point3, point3, vec_mid);
								add_vec(point4, point4, vec_mid);
								add_vec(point5, point5, vec_mid);
								add_vec(point6, point6, vec_mid);
								
								if((int)opposite_xyz == ind){
									write_record("surface %ld curve %ld %ld %ld %ld size %g virtual\n", sur_id1 = ++surface_id, cur_id1, cur_id4, cur_id2, cur_id3, inclusion -> bmsz);
									write_record("polygon 1 1 xyz "FMT3" weight "FMT1"\n", point4.x, point4.y, point4.z, 1.0 / 9.0);
									write_record("polygon 2 1 xyz "FMT3" weight "FMT1"\n", point3.x, point3.y, point3.z, 1.0 / 9.0);
									write_record("polygon 1 2 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, 1.0 / 9.0);
									write_record("polygon 2 2 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, 1.0 / 9.0);
								}
								else{
									write_record("surface %ld curve %ld %ld %ld %ld size %g virtual\n", sur_id1 = ++surface_id, cur_id1, cur_id3, cur_id2, cur_id4, inclusion -> bmsz);
									write_record("polygon 1 1 xyz "FMT3" weight "FMT1"\n", point3.x, point3.y, point3.z, 1.0 / 9.0);
									write_record("polygon 2 1 xyz "FMT3" weight "FMT1"\n", point4.x, point4.y, point4.z, 1.0 / 9.0);
									write_record("polygon 1 2 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, 1.0 / 9.0);
									write_record("polygon 2 2 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, 1.0 / 9.0);
								}
#if 0
								if(opposite_xyz == NO){
									if(ind == 0)
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld", shll_id2 = ++shell_id, surface_id, -cur_id5, cur_id6);
									else
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld", shll_id2 = ++shell_id, surface_id, cur_id5, -cur_id6);
								}
								else{
									if(ind == 0)
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld", shll_id2 = ++shell_id, surface_id, cur_id5, -cur_id6);
									else
										write_record("shell %ld bgsurface %ld boundary curve %ld %ld", shll_id2 = ++shell_id, surface_id, -cur_id5, cur_id6);
								}
#else
								if((int)opposite_xyz == ind)
									write_record("shell %ld bgsurface %ld boundary curve %ld %ld", shll_id2 = ++shell_id, surface_id, -cur_id5, cur_id6);
								else
									write_record("shell %ld bgsurface %ld boundary curve %ld %ld", shll_id2 = ++shell_id, surface_id, cur_id5, -cur_id6);
#endif
								
								if(opposite_xyz == NO){
									if(side == X + 1){
										if(x_intersection == YES)
											write_record(" %ld %ld", x_side_curve_array[0] * sign[x_ind], x_side_curve_array[1] * sign[x_ind]);
										else
											write_record(" fixed vertex %ld", side_vertex_array[ind_xx]);
									}
									else if(side == Y + 1){
										if(y_intersection == YES)
											write_record(" %ld %ld", y_side_curve_array[0] * sign[y_ind], y_side_curve_array[1] * sign[y_ind]);
										else
											write_record(" fixed vertex %ld", side_vertex_array[ind_yy]);
									}
									else{
										if(z_intersection == YES)
											write_record(" %ld %ld", z_side_curve_array[0] * sign[z_ind], z_side_curve_array[1] * sign[z_ind]);
										else
											write_record(" fixed vertex %ld", side_vertex_array[ind_zz]);
									}
								}
						
								write_record(" rate 0 size %g tile %ld\n", inclusion -> bmsz, tile_id);

								write_record("region %ld boundary shell %ld %ld", ++region_id, shll_id1, shll_id2);
								if(x_intersection == YES || y_intersection == YES || z_intersection == YES){
									write_record(" boundary patch");
									if(x_intersection == YES)write_record(" %ld", sign[1 - x_ind] * x_patch_id);
									if(y_intersection == YES)write_record(" %ld", sign[1 - y_ind] * y_patch_id);
									if(z_intersection == YES)write_record(" %ld", sign[1 - z_ind] * z_patch_id);
								}
									
								if(inclusion -> mat == 0)
									write_record(" hole tile %ld\n", tile_id);
								else
									write_record(" size %g tile %ld property %d\n", inclusion -> imsz, tile_id, inclusion -> mat);
								
								tile_region_array[tile_regions++] = region_id;
							}
						}

#ifdef USE_VIRTUAL_ELLIPSOID

/* virtual complete ellipsoid to control curvature based density */

						write_record("\n");

						copy_mul_vec(vec_max, ellipsoid -> major, ellipsoid -> max);
						copy_mul_vec(vec_min, ellipsoid -> minor, ellipsoid -> min);
						copy_mul_vec(vec_mid, ellipsoid -> middle, ellipsoid -> mid);
						
						mul_vec(vec_min, 4.0);
						mul_vec(vec_mid, 2.0);

						add_vec(center, ellipsoid -> center, shift_vec);            /* shift ellipsoid center */
						
						add_vec(point1, center, vec_max);
						sub_vec(point2, center, vec_max);
					
						write_record("vertex %ld xyz "FMT3" virtual\n", ver_id1 = ++vertex_id, point1.x, point1.y, point1.z);
						write_record("vertex %ld xyz "FMT3" virtual\n", ver_id2 = ++vertex_id, point2.x, point2.y, point2.z);
						
						add_vec(point3, point1, vec_mid);
						add_vec(point4, point2, vec_mid);
						
						write_record("curve %ld vertex %ld %ld order 4 virtual\n", cur_id1 = ++curve_id, ver_id1, ver_id2);
						write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point3.x, point3.y, point3.z, 1.0 / 3.0);
						write_record("polygon 2 xyz "FMT3" weight "FMT1"\n", point4.x, point4.y, point4.z, 1.0 / 3.0);
						
						sub_vec(point5, point1, vec_mid);
						sub_vec(point6, point2, vec_mid);
						
						write_record("curve %ld vertex %ld %ld order 4 virtual\n", cur_id2 = ++curve_id, ver_id1, ver_id2);
						write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, 1.0 / 3.0);
						write_record("polygon 2 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, 1.0 / 3.0);
						
						write_record("curve %ld vertex %ld %ld order 4 virtual\n", cur_id3 = ++curve_id, ver_id1, ver_id1);
						write_record("polygon 1 poly 0 weight "FMT1"\n", 1.0 / 3.0);
						write_record("polygon 2 poly 3 weight "FMT1"\n", 1.0 / 3.0);
						
						write_record("curve %ld vertex %ld %ld order 4 virtual\n", cur_id4 = ++curve_id, ver_id2, ver_id2);
						write_record("polygon 1 poly 0 weight "FMT1"\n", 1.0 / 3.0);
						write_record("polygon 2 poly 3 weight "FMT1"\n", 1.0 / 3.0);
						
						add_vec(point3, point3, vec_min);
						add_vec(point4, point4, vec_min);
						add_vec(point5, point5, vec_min);
						add_vec(point6, point6, vec_min);
						
						write_record("surface %ld curve %ld %ld %ld %ld virtual\n", ++surface_id, cur_id1, cur_id3, cur_id2, cur_id4);
						write_record("polygon 1 1 xyz "FMT3" weight "FMT1"\n", point3.x, point3.y, point3.z, 1.0 / 9.0);
						write_record("polygon 2 1 xyz "FMT3" weight "FMT1"\n", point4.x, point4.y, point4.z, 1.0 / 9.0);
						write_record("polygon 1 2 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, 1.0 / 9.0);
						write_record("polygon 2 2 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, 1.0 / 9.0);

						mul_vec(vec_min, 2.0);

						sub_vec(point3, point3, vec_min);
						sub_vec(point4, point4, vec_min);
						sub_vec(point5, point5, vec_min);
						sub_vec(point6, point6, vec_min);
						
						write_record("surface %ld curve %ld %ld %ld %ld virtual\n", ++surface_id, cur_id2, cur_id3, cur_id1, cur_id4);
						write_record("polygon 1 1 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, 1.0 / 9.0);
						write_record("polygon 2 1 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, 1.0 / 9.0);
						write_record("polygon 1 2 xyz "FMT3" weight "FMT1"\n", point3.x, point3.y, point3.z, 1.0 / 9.0);
						write_record("polygon 2 2 xyz "FMT3" weight "FMT1"\n", point4.x, point4.y, point4.z, 1.0 / 9.0);
#endif
					}
					else{

/* processing octant not containing ellipsoid center or containing ellipsoid center on boundary;
	 bg surface for the shell is half of the ellipsoid with base parallel to a particular tile side */

						if(x_patch_id != 0 || y_patch_id != 0 || z_patch_id != 0){

/* I do rely on fact that othonormal base was already precomputed for particular side */

							if(side == X + 1){
								ind = x_ind;
								get_ellipsoid_xyz_side_intersection(&xyz_ellipse, &origin, &vec_x, &dir_2, &dir_c);
								copy_vec(xyz_ellipse.normal, dir_x);
								sub_vec(vec, x_extreme_point, ellipsoid -> center);
							}
							else if(side == Y + 1){
								ind = y_ind;
								get_ellipsoid_xyz_side_intersection(&xyz_ellipse, &origin, &vec_y, &dir_3, &dir_a);
								copy_vec(xyz_ellipse.normal, dir_y);
								sub_vec(vec, y_extreme_point, ellipsoid -> center);
							}
							else{
								ind = z_ind;
								get_ellipsoid_xyz_side_intersection(&xyz_ellipse, &origin, &vec_z, &dir_1, &dir_b);
								copy_vec(xyz_ellipse.normal, dir_z);
								sub_vec(vec, z_extreme_point, ellipsoid -> center);
							}
#ifdef DRAW_ELLIPSE
							draw_ellipse(&xyz_ellipse, 0, orange);
							if(draw == YES)ESIEventLoop(YES, "Base ellipse of exact half-ellipsoid\n\n");
#endif
							copy_mul_vec(vec_max, xyz_ellipse.major, xyz_ellipse.max);
							copy_mul_vec(vec_min, xyz_ellipse.minor, xyz_ellipse.min);
						
							mul_vec(vec_min, 2.0);
							mul_vec(vec, 4.0);

							add_vec(center, ellipsoid -> center, shift_vec);            /* shift ellipsoid center */
						
							add_vec(point1, center, vec_max);
							sub_vec(point2, center, vec_max);
						
							write_record("vertex %ld xyz "FMT3" virtual\n", ver_id1 = ++vertex_id, point1.x, point1.y, point1.z);
							write_record("vertex %ld xyz "FMT3" virtual\n", ver_id2 = ++vertex_id, point2.x, point2.y, point2.z);
							
							add_vec(point3, point1, vec_min);
							add_vec(point4, point2, vec_min);
							
							write_record("curve %ld vertex %ld %ld order 4 virtual\n", cur_id1 = ++curve_id, ver_id1, ver_id2);
							write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point3.x, point3.y, point3.z, 1.0 / 3.0);
							write_record("polygon 2 xyz "FMT3" weight "FMT1"\n", point4.x, point4.y, point4.z, 1.0 / 3.0);
							
							sub_vec(point5, point1, vec_min);
							sub_vec(point6, point2, vec_min);
							
							write_record("curve %ld vertex %ld %ld order 4 virtual\n", cur_id2 = ++curve_id, ver_id1, ver_id2);
							write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, 1.0 / 3.0);
							write_record("polygon 2 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, 1.0 / 3.0);
							
							write_record("curve %ld vertex %ld %ld order 4 virtual\n", cur_id3 = ++curve_id, ver_id1, ver_id1);
							write_record("polygon 1 poly 0 weight "FMT1"\n", 1.0 / 3.0);
							write_record("polygon 2 poly 3 weight "FMT1"\n", 1.0 / 3.0);
							
							write_record("curve %ld vertex %ld %ld order 4 virtual\n", cur_id4 = ++curve_id, ver_id2, ver_id2);
							write_record("polygon 1 poly 0 weight "FMT1"\n", 1.0 / 3.0);
							write_record("polygon 2 poly 3 weight "FMT1"\n", 1.0 / 3.0);
							
							add_vec(point3, point3, vec);
							add_vec(point4, point4, vec);
							add_vec(point5, point5, vec);
							add_vec(point6, point6, vec);
							
							if(ind == 0){
								write_record("surface %ld curve %ld %ld %ld %ld virtual\n", ++surface_id, cur_id1, cur_id3, cur_id2, cur_id4);
								write_record("polygon 1 1 xyz "FMT3" weight "FMT1"\n", point3.x, point3.y, point3.z, 1.0 / 9.0);
								write_record("polygon 2 1 xyz "FMT3" weight "FMT1"\n", point4.x, point4.y, point4.z, 1.0 / 9.0);
								write_record("polygon 1 2 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, 1.0 / 9.0);
								write_record("polygon 2 2 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, 1.0 / 9.0);
							}
							else{
								write_record("surface %ld curve %ld %ld %ld %ld virtual\n", ++surface_id, cur_id2, cur_id3, cur_id1, cur_id4);
								write_record("polygon 1 1 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, 1.0 / 9.0);
								write_record("polygon 2 1 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, 1.0 / 9.0);
								write_record("polygon 1 2 xyz "FMT3" weight "FMT1"\n", point3.x, point3.y, point3.z, 1.0 / 9.0);
								write_record("polygon 2 2 xyz "FMT3" weight "FMT1"\n", point4.x, point4.y, point4.z, 1.0 / 9.0);
							}
							
							write_record("shell %ld bgsurface %ld boundary curve", ++shell_id, surface_id);
							
							for(m = 0; m < x_side_curves; m++)write_record(" %ld", sign[x_ind] * x_side_curve_array[m]);
							for(m = 0; m < y_side_curves; m++)write_record(" %ld", sign[y_ind] * y_side_curve_array[m]);
							for(m = 0; m < z_side_curves; m++)write_record(" %ld", sign[z_ind] * z_side_curve_array[m]);
							
							if(side_vertices != 0){
								write_record(" fixed vertex");
								
								for(m = 0; m < side_vertices; m++)write_record(" %ld", side_vertex_array[m]);
							}
							
							write_record(" rate 0 size %g tile %ld\n", inclusion -> bmsz, tile_id);
							
							write_record("region %ld boundary patch", ++region_id);
							if(x_patch_id != 0)write_record(" %ld", sign[1 - x_ind] * x_patch_id);
							if(y_patch_id != 0)write_record(" %ld", sign[1 - y_ind] * y_patch_id);
							if(z_patch_id != 0)write_record(" %ld", sign[1 - z_ind] * z_patch_id);

							if(inclusion -> mat == 0)
								write_record(" boundary shell %ld hole tile %ld\n", shell_id, tile_id);
							else
								write_record(" boundary shell %ld size %g tile %ld property %d\n", shell_id, inclusion -> imsz, tile_id, inclusion -> mat);
							
							tile_region_array[tile_regions++] = region_id;
						}
					}
				}
				else{   /* ellipsoid is inside tile */
					copy_mul_vec(vec_max, ellipsoid -> major, ellipsoid -> max);
					copy_mul_vec(vec_min, ellipsoid -> minor, ellipsoid -> min);
					copy_mul_vec(vec_mid, ellipsoid -> middle, ellipsoid -> mid);
					
					mul_vec(vec_min, 4.0);
					mul_vec(vec_mid, 2.0);

					add_vec(center, ellipsoid -> center, shift_vec);            /* shift ellipsoid center */
						
					add_vec(point1, center, vec_max);
					sub_vec(point2, center, vec_max);

					write_record("vertex %ld xyz "FMT3" tile %ld\n", ver_id1 = ++vertex_id, point1.x, point1.y, point1.z, tile_id);
					write_record("vertex %ld xyz "FMT3" tile %ld\n", ver_id2 = ++vertex_id, point2.x, point2.y, point2.z, tile_id);
					
					add_vec(point3, point1, vec_mid);
					add_vec(point4, point2, vec_mid);
						
					write_record("curve %ld vertex %ld %ld order 4 tile %ld\n", cur_id1 = ++curve_id, ver_id1, ver_id2, tile_id);
					write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point3.x, point3.y, point3.z, 1.0 / 3.0);
					write_record("polygon 2 xyz "FMT3" weight "FMT1"\n", point4.x, point4.y, point4.z, 1.0 / 3.0);
						
					sub_vec(point5, point1, vec_mid);
					sub_vec(point6, point2, vec_mid);
						
					write_record("curve %ld vertex %ld %ld order 4 tile %ld\n", cur_id2 = ++curve_id, ver_id1, ver_id2, tile_id);
					write_record("polygon 1 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, 1.0 / 3.0);
					write_record("polygon 2 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, 1.0 / 3.0);
					
					write_record("curve %ld vertex %ld %ld order 4 tile %ld\n", cur_id3 = ++curve_id, ver_id1, ver_id1, tile_id);
					write_record("polygon 1 poly 0 weight "FMT1"\n", 1.0 / 3.0);
					write_record("polygon 2 poly 3 weight "FMT1"\n", 1.0 / 3.0);
						
					write_record("curve %ld vertex %ld %ld order 4 tile %ld\n", cur_id4 = ++curve_id, ver_id2, ver_id2, tile_id);
					write_record("polygon 1 poly 0 weight "FMT1"\n", 1.0 / 3.0);
					write_record("polygon 2 poly 3 weight "FMT1"\n", 1.0 / 3.0);
						
					sub_vec(point3, point3, vec_min);
					sub_vec(point4, point4, vec_min);
					sub_vec(point5, point5, vec_min);
					sub_vec(point6, point6, vec_min);
						
					write_record("surface %ld curve %ld %ld %ld %ld size %g tile %ld\n", sur_id1 = ++surface_id, cur_id1, cur_id3, cur_id2, cur_id4, inclusion -> bmsz, tile_id);
					write_record("polygon 1 1 xyz "FMT3" weight "FMT1"\n", point3.x, point3.y, point3.z, 1.0 / 9.0);
					write_record("polygon 2 1 xyz "FMT3" weight "FMT1"\n", point4.x, point4.y, point4.z, 1.0 / 9.0);
					write_record("polygon 1 2 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, 1.0 / 9.0);
					write_record("polygon 2 2 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, 1.0 / 9.0);

					mul_vec(vec_min, 2.0);

					add_vec(point3, point3, vec_min);
					add_vec(point4, point4, vec_min);
					add_vec(point5, point5, vec_min);
					add_vec(point6, point6, vec_min);
						
					write_record("surface %ld curve %ld %ld %ld %ld size %g tile %ld\n", sur_id2 = ++surface_id, cur_id2, cur_id3, cur_id1, cur_id4, inclusion -> bmsz, tile_id);
					write_record("polygon 1 1 xyz "FMT3" weight "FMT1"\n", point5.x, point5.y, point5.z, 1.0 / 9.0);
					write_record("polygon 2 1 xyz "FMT3" weight "FMT1"\n", point6.x, point6.y, point6.z, 1.0 / 9.0);
					write_record("polygon 1 2 xyz "FMT3" weight "FMT1"\n", point3.x, point3.y, point3.z, 1.0 / 9.0);
					write_record("polygon 2 2 xyz "FMT3" weight "FMT1"\n", point4.x, point4.y, point4.z, 1.0 / 9.0);

					if(inclusion -> mat == 0)
						write_record("region %ld boundary surface %ld %ld hole tile %ld\n", ++region_id, sur_id1, sur_id2, tile_id);
					else
						write_record("region %ld boundary surface %ld %ld size %g tile %ld property %d\n", ++region_id, sur_id1, sur_id2, inclusion -> imsz, tile_id, inclusion -> mat);

					tile_region_array[tile_regions++] = region_id;
				}

/* zero id of vertices at intersection points */

				if(x_count != 0){
					for(m = 0; m < x_count; m++)x_edge_is -> point[m].id = 0;
				}
				if(y_count != 0){
					for(m = 0; m < y_count; m++)y_edge_is -> point[m].id = 0;
				}
				if(z_count != 0){
					for(m = 0; m < z_count; m++)z_edge_is -> point[m].id = 0;
				}
			}
		}

/* check for "compatibility" of sides with nonzero code with prototype side;
	 alarm is issued instead of error to see the geometry in t3d;
	 note that t3d is likely to not process the whole input because of the incompatibility */

		edge_curves[0] = y_edge_curves_1[0] + y_edge_curves_2[0];   /* x-z- */
		edge_curves[1] = y_edge_curves_1[1] + y_edge_curves_2[1];   /* x-z+ */
		edge_curves[2] = z_edge_curves_1[0] + z_edge_curves_2[0];   /* x-y- */
		edge_curves[3] = z_edge_curves_1[2] + z_edge_curves_2[2];   /* x-y+ */
		edge_curves[4] = y_edge_curves_1[2] + y_edge_curves_2[2];   /* x+z- */
		edge_curves[5] = y_edge_curves_1[3] + y_edge_curves_2[3];   /* x+z+ */
		edge_curves[6] = z_edge_curves_1[1] + z_edge_curves_2[1];   /* x+y- */
		edge_curves[7] = z_edge_curves_1[3] + z_edge_curves_2[3];   /* x+y+ */

		for(m = 0; m < 2; m++){
			if(tile.code[2 * X + m] != 0){
				code = &code_array[tile.code[2 * X + m] - 1];
				if(code -> curves[0] == 0){                /* not yet processed ==> store prototype counts */
					code -> patches = x_side_patches[m];
					code -> vertices = x_side_vertices[m];
					for(j = 0; j < 4; j++)code -> curves[j] = edge_curves[m * 4 + j];
				}
				else{
					if(code -> patches != x_side_patches[m])
						error_message(ALARM, "%s-side of tile %ld is not compatible (code %ld) with %s-side of tile %ld",
													Facespec[2 * X + m], tile.id, code -> id, facespec[code -> side], code -> tile_id_orig);
					if(code -> vertices != x_side_vertices[m])
						error_message(ALARM, "%s-side of tile %ld is not compatible w(code %ld) ith %s-side of tile %ld",
													Facespec[2 * X + m], tile.id, code -> id, facespec[code -> side], code -> tile_id_orig);
					for(j = 0; j < 4; j++){
						if(code -> curves[j] != edge_curves[m * 4 + j])
							error_message(ALARM, "%s-side of tile %ld is not compatible (code %ld) with %s-side of tile %ld",
														Facespec[2 * X + m], tile.id, code -> id, facespec[code -> side], code -> tile_id_orig);
					}
				}
			}
		}					

		edge_curves[0] = z_edge_curves_1[0] + z_edge_curves_2[0];   /* y-x- */
		edge_curves[1] = z_edge_curves_1[1] + z_edge_curves_2[1];   /* y-x+ */
		edge_curves[2] = x_edge_curves_1[0] + x_edge_curves_2[0];   /* y-z- */
		edge_curves[3] = x_edge_curves_1[2] + x_edge_curves_2[2];   /* y-z+ */
		edge_curves[4] = z_edge_curves_1[2] + z_edge_curves_2[2];   /* y+x- */
		edge_curves[5] = z_edge_curves_1[3] + z_edge_curves_2[3];   /* y+x+ */
		edge_curves[6] = x_edge_curves_1[1] + x_edge_curves_2[1];   /* y+z- */
		edge_curves[7] = x_edge_curves_1[3] + x_edge_curves_2[3];   /* y+z+ */

		for(m = 0; m < 2; m++){
			if(tile.code[2 * Y + m] != 0){
				code = &code_array[tile.code[2 * Y + m] - 1];
				if(code -> curves[0] == 0){                /* not yet processed ==> store prototype counts */
					code -> patches = y_side_patches[m];
					code -> vertices = y_side_vertices[m];
					for(j = 0; j < 4; j++)code -> curves[j] = edge_curves[m * 4 + j];
				}
				else{
					if(code -> patches != y_side_patches[m])
						error_message(ALARM, "%s-side of tile %ld is not compatible (code %ld) with %s-side of tile %ld",
													Facespec[2 * Y + m], tile.id, code -> id, facespec[code -> side], code -> tile_id_orig);
					if(code -> vertices != y_side_vertices[m])
						error_message(ALARM, "%s-side of tile %ld is not compatible (code %ld) with %s-side of tile %ld",
													Facespec[2 * Y + m], tile.id, code -> id, facespec[code -> side], code -> tile_id_orig);
					for(j = 0; j < 4; j++){
						if(code -> curves[j] != edge_curves[m * 4 + j])
							error_message(ALARM, "%s-side of tile %ld is not compatible (code %ld) with %s-side of tile %ld",
														Facespec[2 * Y + m], tile.id, code -> id, facespec[code -> side], code -> tile_id_orig);
					}
				}
			}
		}					

		edge_curves[0] = x_edge_curves_1[0] + x_edge_curves_2[0];   /* z-y- */
		edge_curves[1] = x_edge_curves_1[1] + x_edge_curves_2[1];   /* z-y+ */
		edge_curves[2] = y_edge_curves_1[0] + y_edge_curves_2[0];   /* z-x- */
		edge_curves[3] = y_edge_curves_1[2] + y_edge_curves_2[2];   /* z-x+ */
		edge_curves[4] = x_edge_curves_1[2] + x_edge_curves_2[2];   /* z+y- */
		edge_curves[5] = x_edge_curves_1[3] + x_edge_curves_2[3];   /* z+y+ */
		edge_curves[6] = y_edge_curves_1[1] + y_edge_curves_2[1];   /* z+x- */
		edge_curves[7] = y_edge_curves_1[3] + y_edge_curves_2[3];   /* z+x+ */

		for(m = 0; m < 2; m++){
			if(tile.code[2 * Z + m] != 0){
				code = &code_array[tile.code[2 * Z + m] - 1];
				if(code -> curves[0] == 0){                /* not yet processed ==> store prototype counts */
					code -> patches = z_side_patches[m];
					code -> vertices = z_side_vertices[m];
					for(j = 0; j < 4; j++)code -> curves[j] = edge_curves[m * 4 + j];
				}
				else{
					if(code -> patches != z_side_patches[m])
						error_message(ALARM, "%s-side of tile %ld is not compatible (code %ld) with %s-side of tile %ld",
													Facespec[2 * Z + m], tile.id, code -> id, facespec[code -> side], code -> tile_id_orig);
					if(code -> vertices != z_side_vertices[m])
						error_message(ALARM, "%s-side of tile %ld is not compatible (code %ld) with %s-side of tile %ld",
													Facespec[2 * Z + m], tile.id, code -> id, facespec[code -> side], code -> tile_id_orig);
					for(j = 0; j < 4; j++){
						if(code -> curves[j] != edge_curves[m * 4 + j])
							error_message(ALARM, "%s-side of tile %ld is not compatible (code %ld) with %s-side of tile %ld",
														Facespec[2 * Z + m], tile.id, code -> id, facespec[code -> side], code -> tile_id_orig);
					}
				}
			}
		}					

		write_record("\n# matrix\n\n");

/* generate remaining segments on tile edges */

		for(m = 0; m < 4; m++){
			if(x_edge_curves_2[m] > max_edge_curves_2[X])error_message(GENERAL_ERROR, "Unexpected situation");
			sort_id_coord(x_edge_vertices[m], x_edge_vertex_id_array[m], x_edge_vertex_data_array[m]);

			zz_ind = (m & 2) ? 1 : 0;
			yy_ind = (m & 1) ? 1 : 0;

			ver_id1 = origin_id + (1 << Z) * zz_ind + (1 << Y) * yy_ind;
			ver_id2 = ver_id1 + (1 << X);

			count = 0;
			status = END;
			for(n = 0; n < x_edge_vertices[m]; n++){
				ver_id = x_edge_vertex_id_array[m][n];
				if((status = x_edge_vertex_data_array[m][n] -> status) != END){
#ifdef DEBUG
					if(status == START){
						if(n != x_edge_vertices[m] - 1){
							if(x_edge_vertex_data_array[m][n + 1] -> status != END)error_message(GENERAL_ERROR, "Unexpected situation");
						}
					}
#endif
					write_record("curve %ld vertex %ld %ld tile %ld\n", ++curve_id, ver_id1, ver_id, tile_id);
					x_edge_curve_2_array[m][count++] = curve_id;
						
					if(count > x_edge_curves_2[m])error_message(GENERAL_ERROR, "Unexpected situation");
				}
#ifdef DEBUG
				else{
					if(n != 0){
						if(x_edge_vertex_data_array[m][n - 1] -> status != START)error_message(GENERAL_ERROR, "Unexpected situation");
					}
				}
#endif
				ver_id1 = ver_id;
			}
			if(status != START){
				write_record("curve %ld vertex %ld %ld tile %ld\n", ++curve_id, ver_id1, ver_id2, tile_id);
				x_edge_curve_2_array[m][count++] = curve_id;
			}
#ifdef DEBUG
			if(count != x_edge_curves_2[m])error_message(GENERAL_ERROR, "Unexpected situation");
#endif

			if(y_edge_curves_2[m] > max_edge_curves_2[Y])error_message(GENERAL_ERROR, "Unexpected situation");
			sort_id_coord(y_edge_vertices[m], y_edge_vertex_id_array[m], y_edge_vertex_data_array[m]);

			xx_ind = (m & 2) ? 1 : 0;
			zz_ind = (m & 1) ? 1 : 0;

			ver_id1 = origin_id + (1 << X) * xx_ind + (1 << Z) * zz_ind;
			ver_id2 = ver_id1 + (1 << Y);

			count = 0;
			status = END;
			for(n = 0; n < y_edge_vertices[m]; n++){
				ver_id = y_edge_vertex_id_array[m][n];
				if((status = y_edge_vertex_data_array[m][n] -> status) != END){
#ifdef DEBUG
					if(status == START){
						if(n != y_edge_vertices[m] - 1){
							if(y_edge_vertex_data_array[m][n + 1] -> status != END)error_message(GENERAL_ERROR, "Unexpected situation");
						}
					}
#endif
					write_record("curve %ld vertex %ld %ld tile %ld\n", ++curve_id, ver_id1, ver_id, tile_id);
					y_edge_curve_2_array[m][count++] = curve_id;
						
					if(count > y_edge_curves_2[m])error_message(GENERAL_ERROR, "Unexpected situation");
				}
#ifdef DEBUG
				else{
					if(n != 0){
						if(y_edge_vertex_data_array[m][n - 1] -> status != START)error_message(GENERAL_ERROR, "Unexpected situation");
					}
				}
#endif
				ver_id1 = ver_id;
			}
			if(status != START){
				write_record("curve %ld vertex %ld %ld tile %ld\n", ++curve_id, ver_id1, ver_id2, tile_id);
				y_edge_curve_2_array[m][count++] = curve_id;
			}
#ifdef DEBUG
			if(count != y_edge_curves_2[m])error_message(GENERAL_ERROR, "Unexpected situation");
#endif

			if(z_edge_curves_2[m] > max_edge_curves_2[Z])error_message(GENERAL_ERROR, "Unexpected situation");
			sort_id_coord(z_edge_vertices[m], z_edge_vertex_id_array[m], z_edge_vertex_data_array[m]);

			yy_ind = (m & 2) ? 1 : 0;
			xx_ind = (m & 1) ? 1 : 0;

			ver_id1 = origin_id + (1 << Y) * yy_ind + (1 << X) * xx_ind;
			ver_id2 = ver_id1 + (1 << Z);

			count = 0;
			status = END;
			for(n = 0; n < z_edge_vertices[m]; n++){
				ver_id = z_edge_vertex_id_array[m][n];
				if((status = z_edge_vertex_data_array[m][n] -> status) != END){
#ifdef DEBUG
					if(status == START){
						if(n != z_edge_vertices[m] - 1){
							if(z_edge_vertex_data_array[m][n + 1] -> status != END)error_message(GENERAL_ERROR, "Unexpected situation");
						}
					}
#endif
					write_record("curve %ld vertex %ld %ld tile %ld\n", ++curve_id, ver_id1, ver_id, tile_id);
					z_edge_curve_2_array[m][count++] = curve_id;
						
					if(count > z_edge_curves_2[m])error_message(GENERAL_ERROR, "Unexpected situation");
				}
#ifdef DEBUG
				else{
					if(n != 0){
						if(z_edge_vertex_data_array[m][n - 1] -> status != START)error_message(GENERAL_ERROR, "Unexpected situation");
					}
				}
#endif
				ver_id1 = ver_id;
			}
			if(status != START){
				write_record("curve %ld vertex %ld %ld tile %ld\n", ++curve_id, ver_id1, ver_id2, tile_id);
				z_edge_curve_2_array[m][count++] = curve_id;
			}
#ifdef DEBUG
			if(count != z_edge_curves_2[m])error_message(GENERAL_ERROR, "Unexpected situation");
#endif
		}

/* generate tile side patches;

	 IMPORTANT: since the side codes are processed (when searching for the first occurrence of particular code) according to face ordering, 
	 tile side patches (at least the opposite ones) must be generated in the same order as is the face numbering (XN, XP, YN, YP, ZN, ZP) !!! */

		write_record("\n");

		edge_curve_sign[0] = edge_curve_sign[4] = 1;
		edge_curve_sign[1] = edge_curve_sign[5] = 1;
		edge_curve_sign[2] = edge_curve_sign[6] = -1;
		edge_curve_sign[3] = edge_curve_sign[7] = -1;

/* caution: x|y|z_edge_curve_1|2_array must be stored in edge_curve_array to not violate edge_curve_sign (defined above);
	 note that x|y|z_edge_curve_2_array (corresponding to real boundary of tile side) is stored before x|y|z_edge_curve_1_array (corresponding to 
	 fictitious boundary of tile side) */

		ptch_id1 = tile_id * 10 + XN + 1;
		write_record("patch %ld normal 1.0 0.0 0.0 boundary curve", ptch_id1);

		edge_curve_array[0] = y_edge_curve_2_array[0];
		edge_curve_array[1] = z_edge_curve_2_array[2];
		edge_curve_array[2] = y_edge_curve_2_array[1]; 
		edge_curve_array[3] = z_edge_curve_2_array[0];
		edge_curve_array[4] = y_edge_curve_1_array[0];
		edge_curve_array[5] = z_edge_curve_1_array[2];
		edge_curve_array[6] = y_edge_curve_1_array[1];
		edge_curve_array[7] = z_edge_curve_1_array[0];

		edge_curves[0] = y_edge_curves_2[0];
		edge_curves[1] = z_edge_curves_2[2];
		edge_curves[2] = y_edge_curves_2[1]; 
		edge_curves[3] = z_edge_curves_2[0];
		edge_curves[4] = y_edge_curves_1[0];
		edge_curves[5] = z_edge_curves_1[2];
		edge_curves[6] = y_edge_curves_1[1];
		edge_curves[7] = z_edge_curves_1[0];
		
		output_patch_boundary(edge_curves, edge_curve_array, edge_curve_sign, x_side_patches[0], x_side_patch_array[0], x_side_vertices[0], x_side_vertex_array[0]);

		if(tile.code[XN] == 0)
			*mirror_spec = '\0';
		else{
			code = &code_array[tile.code[XN] - 1];
			if(code -> tile_id == tile_id && code -> side == XN)
				*mirror_spec = '\0';
			else{
#ifdef USE_SIMPLE
				sprintf(mirror_spec, "mirror %ld simple", code -> tile_id * 10 + code -> side + 1);
#else
				sprintf(mirror_spec, "mirror %ld", code -> tile_id * 10 + code -> side + 1);
#endif
			}
		}

#ifdef USE_SIDE
		write_record(" size %g tile %ld side %d %s\n", tile.bmsz, tile_id, XN + 1, mirror_spec);
#else
		write_record(" size %g tile %ld %s\n", tile.bmsz, tile_id, mirror_spec);
#endif

		ptch_id2 = tile_id * 10 + XP + 1;
		write_record("patch %ld normal 1.0 0.0 0.0 boundary curve", ptch_id2);

		edge_curve_array[0] = y_edge_curve_2_array[2];
		edge_curve_array[1] = z_edge_curve_2_array[3];
		edge_curve_array[2] = y_edge_curve_2_array[3]; 
		edge_curve_array[3] = z_edge_curve_2_array[1];
		edge_curve_array[4] = y_edge_curve_1_array[2];
		edge_curve_array[5] = z_edge_curve_1_array[3];
		edge_curve_array[6] = y_edge_curve_1_array[3];
		edge_curve_array[7] = z_edge_curve_1_array[1];

		edge_curves[0] = y_edge_curves_2[2];
		edge_curves[1] = z_edge_curves_2[3];
		edge_curves[2] = y_edge_curves_2[3]; 
		edge_curves[3] = z_edge_curves_2[1];
		edge_curves[4] = y_edge_curves_1[2];
		edge_curves[5] = z_edge_curves_1[3];
		edge_curves[6] = y_edge_curves_1[3];
		edge_curves[7] = z_edge_curves_1[1];

		output_patch_boundary(edge_curves, edge_curve_array, edge_curve_sign, x_side_patches[1], x_side_patch_array[1], x_side_vertices[1], x_side_vertex_array[1]);

		if(tile.code[XP] == 0)
			*mirror_spec = '\0';
		else{
			code = &code_array[tile.code[XP] - 1];
			if(code -> tile_id == tile_id && code -> side == XP)
				*mirror_spec = '\0';
			else{
#ifdef USE_SIMPLE
				sprintf(mirror_spec, "mirror %ld simple", code -> tile_id * 10 + code -> side + 1);
#else
				sprintf(mirror_spec, "mirror %ld", code -> tile_id * 10 + code -> side + 1);
#endif
			}
		}

#ifdef USE_SIDE
		write_record(" size %g tile %ld side %d %s\n", tile.bmsz, tile_id, XP + 1, mirror_spec);
#else
		write_record(" size %g tile %ld %s\n", tile.bmsz, tile_id, mirror_spec);
#endif

		ptch_id3 = tile_id * 10 + YN + 1;
		write_record("patch %ld normal 0.0 1.0 0.0 boundary curve", ptch_id3);

		edge_curve_array[0] = z_edge_curve_2_array[0];
		edge_curve_array[1] = x_edge_curve_2_array[2];
		edge_curve_array[2] = z_edge_curve_2_array[1]; 
		edge_curve_array[3] = x_edge_curve_2_array[0];
		edge_curve_array[4] = z_edge_curve_1_array[0];
		edge_curve_array[5] = x_edge_curve_1_array[2];
		edge_curve_array[6] = z_edge_curve_1_array[1];
		edge_curve_array[7] = x_edge_curve_1_array[0];

		edge_curves[0] = z_edge_curves_2[0];
		edge_curves[1] = x_edge_curves_2[2];
		edge_curves[2] = z_edge_curves_2[1]; 
		edge_curves[3] = x_edge_curves_2[0];
		edge_curves[4] = z_edge_curves_1[0];
		edge_curves[5] = x_edge_curves_1[2];
		edge_curves[6] = z_edge_curves_1[1];
		edge_curves[7] = x_edge_curves_1[0];
		
		output_patch_boundary(edge_curves, edge_curve_array, edge_curve_sign, y_side_patches[0], y_side_patch_array[0], y_side_vertices[0], y_side_vertex_array[0]);

		if(tile.code[YN] == 0)
			*mirror_spec = '\0';
		else{
			code = &code_array[tile.code[YN] - 1];
			if(code -> tile_id == tile_id && code -> side == YN)
				*mirror_spec = '\0';
			else{
#ifdef USE_SIMPLE
				sprintf(mirror_spec, "mirror %ld simple", code -> tile_id * 10 + code -> side + 1);
#else
				sprintf(mirror_spec, "mirror %ld", code -> tile_id * 10 + code -> side + 1);
#endif
			}
		}

#ifdef USE_SIDE
		write_record(" size %g tile %ld side %d %s\n", tile.bmsz, tile_id, YN + 1, mirror_spec);
#else
		write_record(" size %g tile %ld %s\n", tile.bmsz, tile_id, mirror_spec);
#endif

		ptch_id4 = tile_id * 10 + YP + 1;
		write_record("patch %ld normal 0.0 1.0 0.0 boundary curve", ptch_id4);

		edge_curve_array[0] = z_edge_curve_2_array[2];
		edge_curve_array[1] = x_edge_curve_2_array[3];
		edge_curve_array[2] = z_edge_curve_2_array[3]; 
		edge_curve_array[3] = x_edge_curve_2_array[1];
		edge_curve_array[4] = z_edge_curve_1_array[2];
		edge_curve_array[5] = x_edge_curve_1_array[3];
		edge_curve_array[6] = z_edge_curve_1_array[3];
		edge_curve_array[7] = x_edge_curve_1_array[1];

		edge_curves[0] = z_edge_curves_2[2];
		edge_curves[1] = x_edge_curves_2[3];
		edge_curves[2] = z_edge_curves_2[3]; 
		edge_curves[3] = x_edge_curves_2[1];
		edge_curves[4] = z_edge_curves_1[2];
		edge_curves[5] = x_edge_curves_1[3];
		edge_curves[6] = z_edge_curves_1[3];
		edge_curves[7] = x_edge_curves_1[1];

		output_patch_boundary(edge_curves, edge_curve_array, edge_curve_sign, y_side_patches[1], y_side_patch_array[1], y_side_vertices[1], y_side_vertex_array[1]);

		if(tile.code[YP] == 0)
			*mirror_spec = '\0';
		else{
			code = &code_array[tile.code[YP] - 1];
			if(code -> tile_id == tile_id && code -> side == YP)
				*mirror_spec = '\0';
			else{
#ifdef USE_SIMPLE
				sprintf(mirror_spec, "mirror %ld simple", code -> tile_id * 10 + code -> side + 1);
#else
				sprintf(mirror_spec, "mirror %ld", code -> tile_id * 10 + code -> side + 1);
#endif
			}
		}

#ifdef USE_SIDE
		write_record(" size %g tile %ld side %d %s\n", tile.bmsz, tile_id, YP + 1, mirror_spec);
#else
		write_record(" size %g tile %ld %s\n", tile.bmsz, tile_id, mirror_spec);
#endif

		ptch_id5 = tile_id * 10 + ZN + 1;
		write_record("patch %ld normal 0.0 0.0 1.0 boundary curve", ptch_id5);

		edge_curve_array[0] = x_edge_curve_2_array[0];
		edge_curve_array[1] = y_edge_curve_2_array[2];
		edge_curve_array[2] = x_edge_curve_2_array[1]; 
		edge_curve_array[3] = y_edge_curve_2_array[0];
		edge_curve_array[4] = x_edge_curve_1_array[0];
		edge_curve_array[5] = y_edge_curve_1_array[2];
		edge_curve_array[6] = x_edge_curve_1_array[1];
		edge_curve_array[7] = y_edge_curve_1_array[0];

		edge_curves[0] = x_edge_curves_2[0];
		edge_curves[1] = y_edge_curves_2[2];
		edge_curves[2] = x_edge_curves_2[1]; 
		edge_curves[3] = y_edge_curves_2[0];
		edge_curves[4] = x_edge_curves_1[0];
		edge_curves[5] = y_edge_curves_1[2];
		edge_curves[6] = x_edge_curves_1[1];
		edge_curves[7] = y_edge_curves_1[0];
		
		output_patch_boundary(edge_curves, edge_curve_array, edge_curve_sign, z_side_patches[0], z_side_patch_array[0], z_side_vertices[0], z_side_vertex_array[0]);

		if(tile.code[ZN] == 0)
			*mirror_spec = '\0';
		else{
			code = &code_array[tile.code[ZN] - 1];
			if(code -> tile_id == tile_id && code -> side == ZN)
				*mirror_spec = '\0';
			else{
#ifdef USE_SIMPLE
				sprintf(mirror_spec, "mirror %ld simple", code -> tile_id * 10 + code -> side + 1);
#else
				sprintf(mirror_spec, "mirror %ld", code -> tile_id * 10 + code -> side + 1);
#endif
			}
		}

#ifdef USE_SIDE
		write_record(" size %g tile %ld side %d %s\n", tile.bmsz, tile_id, ZN + 1, mirror_spec);
#else
		write_record(" size %g tile %ld %s\n", tile.bmsz, tile_id, mirror_spec);
#endif

		ptch_id6 = tile_id * 10 + ZP + 1;
		write_record("patch %ld normal 0.0 0.0 1.0 boundary curve", ptch_id6);

		edge_curve_array[0] = x_edge_curve_2_array[2];
		edge_curve_array[1] = y_edge_curve_2_array[3];
		edge_curve_array[2] = x_edge_curve_2_array[3]; 
		edge_curve_array[3] = y_edge_curve_2_array[1];
		edge_curve_array[4] = x_edge_curve_1_array[2];
		edge_curve_array[5] = y_edge_curve_1_array[3];
		edge_curve_array[6] = x_edge_curve_1_array[3];
		edge_curve_array[7] = y_edge_curve_1_array[1];

		edge_curves[0] = x_edge_curves_2[2];
		edge_curves[1] = y_edge_curves_2[3];
		edge_curves[2] = x_edge_curves_2[3]; 
		edge_curves[3] = y_edge_curves_2[1];
		edge_curves[4] = x_edge_curves_1[2];
		edge_curves[5] = y_edge_curves_1[3];
		edge_curves[6] = x_edge_curves_1[3];
		edge_curves[7] = y_edge_curves_1[1];

		output_patch_boundary(edge_curves, edge_curve_array, edge_curve_sign, z_side_patches[1], z_side_patch_array[1], z_side_vertices[1], z_side_vertex_array[1]);

		if(tile.code[ZP] == 0)
			*mirror_spec = '\0';
		else{
			code = &code_array[tile.code[ZP] - 1];
			if(code -> tile_id == tile_id && code -> side == ZP)
				*mirror_spec = '\0';
			else{
#ifdef USE_SIMPLE
				sprintf(mirror_spec, "mirror %ld simple", code -> tile_id * 10 + code -> side + 1);
#else
				sprintf(mirror_spec, "mirror %ld", code -> tile_id * 10 + code -> side + 1);
#endif
			}
		}

#ifdef USE_SIDE
		write_record(" size %g tile %ld side %d %s\n", tile.bmsz, tile_id, ZP + 1, mirror_spec);
#else
		write_record(" size %g tile %ld %s\n", tile.bmsz, tile_id, mirror_spec);
#endif

		write_record("\n");

		side_patch_sign[0] = side_patch_sign[1] = side_patch_sign[2] = -1;
		side_patch_sign[3] = side_patch_sign[4] = side_patch_sign[5] = 1;

/* caution: x|y|z_side_patch_array must be stored in side_patch_array to not violate side_patch_sign (defined above) */

		write_record("region %ld boundary patch %ld %ld %ld %ld %ld %ld", ++region_id, -ptch_id1, ptch_id2, -ptch_id3, ptch_id4, -ptch_id5, ptch_id6);
		
		side_patches[0] = x_side_patches[0];
		side_patches[1] = y_side_patches[0];
		side_patches[2] = z_side_patches[0];
		side_patches[3] = x_side_patches[1];
		side_patches[4] = y_side_patches[1];
		side_patches[5] = z_side_patches[1];

		side_patch_array[0] = x_side_patch_array[0];
		side_patch_array[1] = y_side_patch_array[0];
		side_patch_array[2] = z_side_patch_array[0];
		side_patch_array[3] = x_side_patch_array[1];
		side_patch_array[4] = y_side_patch_array[1];
		side_patch_array[5] = z_side_patch_array[1];

		output_region_boundary(side_patches, side_patch_array, side_patch_sign, tile_regions, tile_region_array);

		write_record(" size %g tile %ld property %d\n", tile.imsz, tile_id, tile.mat);

		if(tile.points != 0)write_record("\n");

		for(j = 0; j < tile.points; j++){
			kk = (j <= 3) ? j : 3;
			get_next_relevant_record("Missing control point definition");

			if(sscanf(line_buffer, "%lf %lf %lf %lf", &(point.x), &(point.y), &(point.z), &msz) != 4)
				error_message(INPUT_ERROR, "Invalid record - definition of the %ld%s ctrl point of particular tile %ld", j + 1, suffix[kk], tile.id);        /* caution: use tile.id */

			if(point.x < tile.corner[0].x - epsilon || point.x > tile.corner[1].x + epsilon ||
				 point.y < tile.corner[0].y - epsilon || point.y > tile.corner[1].y + epsilon ||
				 point.z < tile.corner[0].z - epsilon || point.z > tile.corner[1].z + epsilon)
				error_message(INPUT_ERROR, "Invalid coordinates of the %ld%s control point of particular tile %ld",  
											j + 1, suffix[kk], tile.id);        /* caution: use tile.id */
	
			if(msz < 0.0)error_message(INPUT_ERROR, "Invalid mesh size of the %ld%s ctrl point of particular tile %ld: MSZ = %e (0 < MSZ)", 
																 j + 1, suffix[k], tile.id, msz);        /* caution: use tile.id */

			write_record("vertex %ld virtual xyz "FMT3" size %g\n", ++vertex_id, point.x, point.y, point.z, msz);
		}

		if(i != tiles - 1){
#if SETUP == GRID_3D || SETUP == GRID_2D
			if(++row == grid_size){
				tile_shift_x = min_sz * 0.5 - tile.center.x;
				row = 0;
				if(++column == grid_size){
					tile_shift_y = min_sz * 0.5 - tile.center.y;
					column = 0;
#if SETUP == GRID_3D
					if(++layer == grid_size)
						error_message(GENERAL_ERROR, "Unexpected situation");
#endif
#if SETUP == GRID_2D
					if(++layer == 1)
						error_message(GENERAL_ERROR, "Unexpected situation");
#endif
					tile_shift_z += offset;
				}
				else
					tile_shift_y += offset;
			}
			else
				tile_shift_x += offset;
#endif
#if SETUP == DIAGONAL
			tile_shift_x += offset;
			tile_shift_y += offset;
			tile_shift_z += offset;
#endif
		}
	}

/* CAUTION: patches are not numbered continuously !!!
	 using the largest patch id is contraproductive as too large array would have to be allocated by t3d (see unique patch numbering)
	 resulting possibly in memory allocation failure */

	write_record("\n");
	write_record("# Optionally, add to command line -n %ld %ld %ld %ld %ld %ld %ld\n", vertex_id, curve_id, surface_id, 0, shell_id, region_id, interface_id);
	write_record("# or uncomment the following records and move them to the start of this file\n");
	write_record("#\n");
	if(vertex_id != 0)write_record("#vertices %ld\n", vertex_id);
	if(curve_id != 0)write_record("#curves %ld\n", curve_id);
	if(surface_id != 0)write_record("#surfaces %ld\n", surface_id);
	if(shell_id != 0)write_record("#shells %ld\n", shell_id);
	if(region_id != 0)write_record("#regions %ld\n", region_id);
	if(interface_id != 0)write_record("#interfaces %ld\n", interface_id);

	if(tiles == 1)
		fprintf(stdout, "\n1 tile successfully processed\n\n");
	else
		fprintf(stdout, "\n%ld tiles successfully processed\n\n", tiles);
		
#ifdef ELIXIR
	if(draw == YES)ESIEventLoop(YES, "");
#endif

#ifdef IS_TESTING
	if(((int)CYCLE_COORD) % 3 == 0)write_record("\nview_setup view normal 3 2 1 view vup 0 0 1\n");    /* default setting */
	if(((int)CYCLE_COORD) % 3 == 1)write_record("\nview_setup view normal 1 3 2 view vup 1 0 0\n");
	if(((int)CYCLE_COORD) % 3 == 2)write_record("\nview_setup view normal 2 1 3 view vup 0 1 0\n");

	write_record("\nview_model layer vertex_num on layer vertex_virtual on view zoom 1.1\n");     /* check for overlaping vertices */
#endif

	fclose(t3d_in_file);
	fclose(geom_file);

	return(0);
}



/* YES is returned if inclusion is intersecting or touching the tile, otherwise NO is returned;
	 note that it is safe to return YES even if the no intersection or touching occurs */

static logic
get_inclusion_tile_intersection(void)
{
	point_rec pnt_1, pnt_2, *pnt_extreme = NULL, *p1 = NULL, *p2 = NULL;
	vector_rec vec, vector, vec_tmp;
	vector_rec normal1, normal2;
	double size, val, shift_pos, shift_neg, shift, xyz_shift, eps;
	double sz, product;
	double a_diff_xy, a_diff_xz;
	double a_diff_yz, a_diff_yx;
	double a_diff_zx, a_diff_zy;
	double a_x_diff, a_y_diff, a_z_diff;
	double diff_pos, diff_1_pos, diff_2_pos;
	double diff_neg, diff_1_neg, diff_2_neg;
	double sz_x, sz_y, sz_z, coord;
	int i, ind, count, run, cycle, cycles, shifting_strategy, subdivision, phase;
	logic x_shifting, y_shifting, z_shifting, shifting, ret_val, halving;
	logic corner_x_between, corner_y_between, corner_z_between;
	is_point_rec *is_point = NULL;
	int cnt_x, cnt_y, cnt_z, cnt_xx, cnt_yy, cnt_zz;
	
	ellipsoid = &(inclusion -> ellipsoid);

	x_intersection = y_intersection = z_intersection = NO;
	x_touching = y_touching = z_touching = NO;

	x_count = y_count = z_count = 0;

	total_shift.x = total_shift.y = total_shift.z = 0.0;

/* compute extreme points in GCS;

	 they are computed from the fact that gradient at extreme point is parallel to direction in which extreme is investigated;
	 the equality between the gradient and the direction is expressed in LCS;
	 for example when searching for extreme in z-direction:
	 {2x/max^2;2y/min^2;2z/mid^2}=dir_3*value;
	 this is equivalent to
	 {x/max^2;y/min^2;z/mid^2}=dir_3*val;
	 this implies
	 x=dir_3.x*val*max^2;
	 y=dir_3.y*val*min^2;
	 z=dir_3.z*val*mid^2;
	 assuming
	 vec_z.x=dir_3.x*max;
	 vec_z.y=dir_3.y*min;
	 vec_z.z=dir_3.z*mid;
	 one can write
	 x=vec_z.x*max*val
	 y=vec_z.y*min*val
	 z=vec_z.z*mid*val
	 val is computed from equation of the ellipsoid
	 x^2/max^2+y^2/min^2+z^2/mid^2=1;
	 (vec_z.x)^2*val^2+(vec_z.y)^2*val^2+(vec_z.z)^2*val^2=1;
	 val=sqrt(1.0/((vec_z.x)^2+(vec_z.y)^2+(vec_z.z)^2)) */

	copy_vec(trans_g2l.x, ellipsoid -> major);
	copy_vec(trans_g2l.y, ellipsoid -> minor);
	copy_vec(trans_g2l.z, ellipsoid -> middle);
	
	get_inverted_rotation_transformation(&trans_g2l, &trans_l2g);

	transform_vector_from_global_to_local(&dir_x, &dir_1, &trans_g2l);
	transform_vector_from_global_to_local(&dir_y, &dir_2, &trans_g2l);
	transform_vector_from_global_to_local(&dir_z, &dir_3, &trans_g2l);
	
	vec.x = ellipsoid -> max;
	vec.y = ellipsoid -> min;
	vec.z = ellipsoid -> mid;
	
	comp_product(vector, vec, vec);
	
	comp_product(vec_x, dir_1, vec);
	comp_product(vec_y, dir_2, vec);
	comp_product(vec_z, dir_3, vec);
	
	val = 1.0 / sqrt(size_vec(vec_x));
	copy_mul_vec(vec_tmp, vector, val);
	comp_product(pnt_1, dir_1, vec_tmp);
	
	transform_point_from_local_to_global(&pnt_1, &ext_point_x[1], &(ellipsoid -> center), &trans_l2g);
	
#if 1 /* more efficient */
	sub_vec(vec_tmp, ellipsoid -> center, ext_point_x[1]);
	add_vec(ext_point_x[0], ellipsoid -> center, vec_tmp);
#else
	copy_mul_vec(pnt_2, pnt_1, -1.0);
	transform_point_from_local_to_global(&pnt_2, &ext_point_x[0], &(ellipsoid -> center), &trans_l2g);
#endif
	
	val = 1.0 / sqrt(size_vec(vec_y));
	copy_mul_vec(vec_tmp, vector, val);
	comp_product(pnt_1, dir_2, vec_tmp);
	
	transform_point_from_local_to_global(&pnt_1, &ext_point_y[1], &(ellipsoid -> center), &trans_l2g);
	
#if 1 /* more efficient */
	sub_vec(vec_tmp, ellipsoid -> center, ext_point_y[1]);
	add_vec(ext_point_y[0], ellipsoid -> center, vec_tmp);
#else
	copy_mul_vec(pnt_2, pnt_1, -1.0);
	transform_point_from_local_to_global(&pnt_2, &ext_point_y[0], &(ellipsoid -> center), &trans_l2g);
#endif
	
	val = 1.0 / sqrt(size_vec(vec_z));
	copy_mul_vec(vec_tmp, vector, val);
	comp_product(pnt_1, dir_3, vec_tmp);
	
	transform_point_from_local_to_global(&pnt_1, &ext_point_z[1], &(ellipsoid -> center), &trans_l2g);

#if 1 /* more efficient */
	sub_vec(vec_tmp, ellipsoid -> center, ext_point_z[1]);
	add_vec(ext_point_z[0], ellipsoid -> center, vec_tmp);
#else
	copy_mul_vec(pnt_2, pnt_1, -1.0);
	transform_point_from_local_to_global(&pnt_2, &ext_point_z[0], &(ellipsoid -> center), &trans_l2g);
#endif
	
#ifdef DRAW_EXTREME_POINT_ELLIPSOID
	for(i = 0; i < 2; i++){
		draw_point(&ext_point_x[i], 0, green);
		draw_point(&ext_point_y[i], 0, cyan);
		draw_point(&ext_point_z[i], 0, magenta);
	}
#endif

/* resolve intersections with tile sides */

	if(ext_point_x[0].x - tile.corner[0].x > epsilon && tile.corner[1].x - ext_point_x[1].x > epsilon &&
		 ext_point_y[0].y - tile.corner[0].y > epsilon && tile.corner[1].y - ext_point_y[1].y > epsilon &&
		 ext_point_z[0].z - tile.corner[0].z > epsilon && tile.corner[1].z - ext_point_z[1].z > epsilon)return(NO); /* inclusion is inside tile */
	
	if(ext_point_x[0].x < tile.corner[0].x + epsilon || ext_point_x[1].x > tile.corner[1].x - epsilon)x_intersection = YES;
	if(ext_point_y[0].y < tile.corner[0].y + epsilon || ext_point_y[1].y > tile.corner[1].y - epsilon)y_intersection = YES;
	if(ext_point_z[0].z < tile.corner[0].z + epsilon || ext_point_z[1].z > tile.corner[1].z - epsilon)z_intersection = YES;
	
/* note that ind_x|y|z should not change due to eventual ellipsoid shifting */

	if(ellipsoid -> center.x < tile.center.x)           /* XN side */
		ind_x = 0;
	else                                                /* XP side */
		ind_x = 1;
	if(ellipsoid -> center.y < tile.center.y)           /* YN side */
		ind_y = 0;
	else                                                /* YP side */
		ind_y = 1;
	if(ellipsoid -> center.z < tile.center.z)           /* ZN side */
		ind_z = 0;
	else                                                /* ZP side */
		ind_z = 1;
	
	shift_vec.x = shift_vec.y = shift_vec.z = 0.0;
	
	diff_x = diff_y = diff_z = max_sz;    /* diffs of ellipsoid extreme points */
	
/* diff_x = x-coordinate of relevant tile side - x-coordinate of relevant extreme point;
	 diff_y = y-coordinate of relevant tile side - y-coordinate of relevant extreme point;
	 diff_z = z-coordinate of relevant tile side - z-coordinate of relevant extreme point;

   diff_x|y|z are computed in resolve_ellipsoid_x|y|z_side_touching;
	 note that x|y|z_intersection is set to NO if touching is identified;
	 note that relevant x|y|z_count (equal to 1) for the case when inclusion is touching a side on tile edge 
	 is also computed in resolve_ellipsoid_x|y|z_side_touching */

	if(x_intersection == YES)resolve_ellipsoid_x_side_touching();
	if(y_intersection == YES)resolve_ellipsoid_y_side_touching();
	if(z_intersection == YES)resolve_ellipsoid_z_side_touching();
	
	size = size_vec(shift_vec);
	if(size != 0.0){

#ifdef DEBUG_PRINT
#if 0 /* obsolete */

/* this is not correct;
	 x|y|z_count != 0 does not imply that there was a shift to resolve edge touching */

		if(x_count == 0 && y_count == 0 && z_count == 0)
			fprintf(stderr, "Shifting inclusion %ld by [%e %e %e] due to side touching\n", inclusion -> id, shift_vec.x, shift_vec.y, shift_vec.z);
		else
			fprintf(stderr, "Shifting inclusion %ld by [%e %e %e] due to side and edge touching\n", inclusion -> id, shift_vec.x, shift_vec.y, shift_vec.z);
#else
		if((x_touching == NO && shift_vec.x != 0.0) || (y_touching == NO && shift_vec.y != 0.0) || (z_touching == NO && shift_vec.z != 0.0))
			fprintf(stderr, "Shifting inclusion %ld by [%e %e %e] due to side and edge touching\n", inclusion -> id, shift_vec.x, shift_vec.y, shift_vec.z);
		else
			fprintf(stderr, "Shifting inclusion %ld by [%e %e %e] due to side touching\n", inclusion -> id, shift_vec.x, shift_vec.y, shift_vec.z);
#endif
#endif		
#ifdef DRAW_EXTREME_POINT_ELLIPSOID
		for(i = 0; i < 2; i++){
			draw_point(&ext_point_x[i], 0, bgcolor);
			draw_point(&ext_point_y[i], 0, bgcolor);
			draw_point(&ext_point_z[i], 0, bgcolor);
		}
#endif
		
		copy_vec(total_shift, shift_vec);
		add_vec(ellipsoid -> center, ellipsoid -> center, shift_vec);
		for(i = 0; i < 2; i++){
			add_vec(ext_point_x[i], ext_point_x[i], shift_vec);
			add_vec(ext_point_y[i], ext_point_y[i], shift_vec);
			add_vec(ext_point_z[i], ext_point_z[i], shift_vec);
		}
		diff_x -= shift_vec.x;
		diff_y -= shift_vec.y;
		diff_z -= shift_vec.z;
		
#ifdef DRAW_EXTREME_POINT_ELLIPSOID
		for(i = 0; i < 2; i++){
			draw_point(&ext_point_x[i], 0, green);
			draw_point(&ext_point_y[i], 0, cyan);
			draw_point(&ext_point_z[i], 0, magenta);
		}
#endif
	}
	
	if(x_touching == YES)setup_ellipse_touching_point(&x_ellipse, &ext_point_x[ind_x], &dir_x);
	if(y_touching == YES)setup_ellipse_touching_point(&y_ellipse, &ext_point_y[ind_y], &dir_y);
	if(z_touching == YES)setup_ellipse_touching_point(&z_ellipse, &ext_point_z[ind_z], &dir_z);
	
	if(x_intersection == NO && y_intersection == NO && z_intersection == NO){

/* there is no need to resolve intersection with tile edges if there in no intersection with tile sides */

		ret_val = NO;
		if(x_touching == YES){
			if((inclusion -> face_is[X] = (face_is_rec *)calloc(1, sizeof(face_is_rec))) == NULL)
				error_message(MEMORY_ERROR, "Memory allocation error");
			memcpy(&(inclusion -> face_is[X] -> ellipse), &x_ellipse, sizeof(ellipse_rec));
			ret_val = YES;
		}
		if(y_touching == YES){
			if((inclusion -> face_is[Y] = (face_is_rec *)calloc(1, sizeof(face_is_rec))) == NULL)
				error_message(MEMORY_ERROR, "Memory allocation error");
			memcpy(&(inclusion -> face_is[Y] -> ellipse), &y_ellipse, sizeof(ellipse_rec));
			ret_val = YES;
		}
		if(z_touching == YES){
			if((inclusion -> face_is[Z] = (face_is_rec *)calloc(1, sizeof(face_is_rec))) == NULL)
				error_message(MEMORY_ERROR, "Memory allocation error");
			memcpy(&(inclusion -> face_is[Z] -> ellipse), &z_ellipse, sizeof(ellipse_rec));
			ret_val = YES;
		}
		
/* note: face_is[X|Y|Z] -> side[0|1] is resolved after return;
	 it is therefore important to return YES if inclusion is touching tile */

		return(ret_val);
	}

	radius = ellipsoid -> max;     /* sphere radius */
	//radius = ellipsoid -> min;     /* sphere radius */
	//radius = ellipsoid -> mid;     /* sphere radius */
	
	scale_e2s.x = radius / ellipsoid -> max;
	scale_e2s.y = radius / ellipsoid -> min;
	scale_e2s.z = radius / ellipsoid -> mid;
	
	get_inverted_scale_transformation(&scale_e2s, &scale_s2e);
	
/* precompute tile corner points in LCS of ellipsoid;
	 this must be done before call to get_ellipsoid_x|y|z_side_intersection */

	for(i = 0; i < 2; i++){
		transform_point_from_global_to_local(&(tile.corner[i]), &corner_point[i], &(ellipsoid -> center), &trans_g2l);
		transform_local_from_ellipsoid_to_sphere(&corner_point[i], &corner_point[i], &scale_e2s);
	}

/* if intersection with only one tile side is detected I precompute complete orthonormal base of that transformed tile side only
	 because no further shift will be applied (thus intersections with other sides will not occur);
	 otherwise I precompute complete orthonormal bases of all transformed tile sides no matter whether intersection or touching was identified
	 because intersection may appear after shifting */

	transform_local_from_ellipsoid_to_sphere(&dir_1, &dir_1, &scale_e2s);
	transform_local_from_ellipsoid_to_sphere(&dir_2, &dir_2, &scale_e2s);
	transform_local_from_ellipsoid_to_sphere(&dir_3, &dir_3, &scale_e2s);
	
	size = sqrt(size_vec(dir_1));
	div_vec(dir_1, size);
	
	size = sqrt(size_vec(dir_2));
	div_vec(dir_2, size);
	
	size = sqrt(size_vec(dir_3));
	div_vec(dir_3, size);
	
	count = 0;
	if(x_intersection == YES)count++;
	if(y_intersection == YES)count++;
	if(z_intersection == YES)count++;
	
	if(count == 1){

/* if there is intersection with one side only, there can be intersection with edges
	 only if a side perpendicular to intersected side is touched on the edge;
	 this implies that the code below corresponding to x|y|z_count != 0 is always processed only once
	 although it occurs twice (for different x|y|z_intersection or x|y|z_touching) */

		if(x_count != 0 || y_count != 0 || z_count != 0){            /* one or two counts can be equal to 1 */
			if((pnt_extreme = (point_rec *)calloc(4, sizeof(point_rec))) == NULL)
				error_message(MEMORY_ERROR, "Memory allocation error");
#ifdef DEBUG

/* WARNING: if only two extreme points are relevant, the other 2 will be itialized to tile center */

			for(i = 0; i < 4; i++){
				copy_vec(pnt_extreme[i], tile.center);
			}
#else

/* WARNING: if only two extreme points are relevant, the other 2 will remain unitialized !!! */

#endif
		}

		if(x_intersection == YES || x_touching == YES){
			if(x_intersection == YES){
				cross_product(vec_x, dir_2, dir_3);
				size = sqrt(size_vec(vec_x));
				div_vec(vec_x, size);
				
				cross_product(dir_c, vec_x, dir_2);
			
				get_ellipsoid_x_side_intersection();

				if(y_count != 0){

/* compute 2 relevant extreme points of intersection ellipse */

					copy_vec(pnt_extreme[2 * XZ + ind_z], ext_point_z[ind_z]);
					sub_vec(vec_tmp, x_ellipse.center, ext_point_z[ind_z]);
					add_vec(pnt_extreme[2 * XZ + 1 - ind_z], x_ellipse.center, vec_tmp);
				}
				if(z_count != 0){

/* compute 2 relevant extreme points of intersection ellipse */

					copy_vec(pnt_extreme[2 * XY + ind_y], ext_point_y[ind_y]);
					sub_vec(vec_tmp, x_ellipse.center, ext_point_y[ind_y]);
					add_vec(pnt_extreme[2 * XY + 1 - ind_y], x_ellipse.center, vec_tmp);
				}
			}
			else{
				setup_ellipse_touching_point(&x_ellipse, &ext_point_x[ind_x], &dir_x);

				if(y_count != 0 && z_count != 0)inclusion -> corner = BND;

				if(y_count != 0){
					if((inclusion -> edge_is[Y] = (edge_is_rec *)malloc(sizeof(edge_is_rec))) == NULL)
						error_message(MEMORY_ERROR, "Memory allocation error");
					inclusion -> edge_is[Y] -> count = 1;
					if((inclusion -> edge_is[Y] -> point = (is_point_rec *)calloc(1, sizeof(is_point_rec))) == NULL)
						error_message(MEMORY_ERROR, "Memory allocation error");
					is_point = &(inclusion -> edge_is[Y] -> point[0]);
					is_point -> data.coord = ext_point_x[ind_x].y;

					if(inclusion -> corner == BND)
						inclusion -> edge_is[Y] -> side = INDEF;
					else
						inclusion -> edge_is[Y] -> side = (is_point -> data.coord > tile.corner[ind_y].y) ? POS : NEG;
				}
				if(z_count != 0){
					if((inclusion -> edge_is[Z] = (edge_is_rec *)malloc(sizeof(edge_is_rec))) == NULL)
						error_message(MEMORY_ERROR, "Memory allocation error");
					inclusion -> edge_is[Z] -> count = 1;
					if((inclusion -> edge_is[Z] -> point = (is_point_rec *)calloc(1, sizeof(is_point_rec))) == NULL)
						error_message(MEMORY_ERROR, "Memory allocation error");
					is_point = &(inclusion -> edge_is[Z] -> point[0]);
					is_point -> data.coord = ext_point_x[ind_x].z;

					if(inclusion -> corner == BND)
						inclusion -> edge_is[Z] -> side = INDEF;
					else
						inclusion -> edge_is[Z] -> side = (is_point -> data.coord > tile.corner[ind_z].z) ? POS : NEG;
				}
			}

			if((inclusion -> face_is[X] = (face_is_rec *)calloc(1, sizeof(face_is_rec))) == NULL)
				error_message(MEMORY_ERROR, "Memory allocation error");
			memcpy(&(inclusion -> face_is[X] -> ellipse), &x_ellipse, sizeof(ellipse_rec));
			inclusion -> face_is[X] -> extreme_pnt = pnt_extreme;
		}

		if(y_intersection == YES || y_touching == YES){
			if(y_intersection == YES){
				cross_product(vec_y, dir_3, dir_1);
				size = sqrt(size_vec(vec_y));
				div_vec(vec_y, size);
				
				cross_product(dir_a, vec_y, dir_3);
				
				get_ellipsoid_y_side_intersection();

				if(z_count != 0){

/* compute 2 relevant extreme points of intersection ellipse */

					copy_vec(pnt_extreme[2 * YX + ind_x], ext_point_x[ind_x]);
					sub_vec(vec_tmp, y_ellipse.center, ext_point_x[ind_x]);
					add_vec(pnt_extreme[2 * YX + 1 - ind_x], y_ellipse.center, vec_tmp);
				}
				if(x_count != 0){

/* compute 2 relevant extreme points of intersection ellipse */

					copy_vec(pnt_extreme[2 * YZ + ind_z], ext_point_z[ind_z]);
					sub_vec(vec_tmp, y_ellipse.center, ext_point_z[ind_z]);
					add_vec(pnt_extreme[2 * YZ + 1 - ind_z], y_ellipse.center, vec_tmp);
				}
			}
			else{
				setup_ellipse_touching_point(&y_ellipse, &ext_point_y[ind_y], &dir_y);

				if(z_count != 0 && x_count != 0)inclusion -> corner = BND;

				if(z_count != 0){
					if((inclusion -> edge_is[Z] = (edge_is_rec *)malloc(sizeof(edge_is_rec))) == NULL)
						error_message(MEMORY_ERROR, "Memory allocation error");
					inclusion -> edge_is[Z] -> count = 1;
					if((inclusion -> edge_is[Z] -> point = (is_point_rec *)calloc(1, sizeof(is_point_rec))) == NULL)
						error_message(MEMORY_ERROR, "Memory allocation error");
					is_point = &(inclusion -> edge_is[Z] -> point[0]);
					is_point -> data.coord = ext_point_y[ind_y].z;

					if(inclusion -> corner == BND)
						inclusion -> edge_is[Z] -> side = INDEF;
					else
						inclusion -> edge_is[Z] -> side = (is_point -> data.coord > tile.corner[ind_z].z) ? POS : NEG;
				}
				if(x_count != 0){
					if((inclusion -> edge_is[X] = (edge_is_rec *)malloc(sizeof(edge_is_rec))) == NULL)
						error_message(MEMORY_ERROR, "Memory allocation error");
					inclusion -> edge_is[X] -> count = 1;
					if((inclusion -> edge_is[X] -> point = (is_point_rec *)calloc(1, sizeof(is_point_rec))) == NULL)
						error_message(MEMORY_ERROR, "Memory allocation error");
					is_point = &(inclusion -> edge_is[X] -> point[0]);
					is_point -> data.coord = ext_point_y[ind_y].x;

					if(inclusion -> corner == BND)
						inclusion -> edge_is[X] -> side = INDEF;
					else
						inclusion -> edge_is[X] -> side = (is_point -> data.coord > tile.corner[ind_x].x) ? POS : NEG;
				}
			}

			if((inclusion -> face_is[Y] = (face_is_rec *)calloc(1, sizeof(face_is_rec))) == NULL)
				error_message(MEMORY_ERROR, "Memory allocation error");
			memcpy(&(inclusion -> face_is[Y] -> ellipse), &y_ellipse, sizeof(ellipse_rec));
			inclusion -> face_is[Y] -> extreme_pnt = pnt_extreme;
		}

		if(z_intersection == YES || z_touching == YES){
			if(z_intersection == YES){
				cross_product(vec_z, dir_1, dir_2);
				size = sqrt(size_vec(vec_z));
				div_vec(vec_z, size);
				
				cross_product(dir_b, vec_z, dir_1);
			
				get_ellipsoid_z_side_intersection();

				if(x_count != 0){

/* compute 2 relevant extreme points of intersection ellipse */

					copy_vec(pnt_extreme[2 * ZY + ind_y], ext_point_y[ind_y]);
					sub_vec(vec_tmp, z_ellipse.center, ext_point_y[ind_y]);
					add_vec(pnt_extreme[2 * ZY + 1 - ind_y], z_ellipse.center, vec_tmp);
				}
				if(y_count != 0){

/* compute 2 relevant extreme points of intersection ellipse */

					copy_vec(pnt_extreme[2 * ZX + ind_x], ext_point_x[ind_x]);
					sub_vec(vec_tmp, z_ellipse.center, ext_point_x[ind_x]);
					add_vec(pnt_extreme[2 * ZX + 1 - ind_x], z_ellipse.center, vec_tmp);
				}
			}
			else{
				setup_ellipse_touching_point(&z_ellipse, &ext_point_z[ind_z], &dir_z);

				if(x_count != 0 && y_count != 0)inclusion -> corner = BND;

				if(x_count != 0){
					if((inclusion -> edge_is[X] = (edge_is_rec *)malloc(sizeof(edge_is_rec))) == NULL)
						error_message(MEMORY_ERROR, "Memory allocation error");
					inclusion -> edge_is[X] -> count = 1;
					if((inclusion -> edge_is[X] -> point = (is_point_rec *)calloc(1, sizeof(is_point_rec))) == NULL)
						error_message(MEMORY_ERROR, "Memory allocation error");
					is_point = &(inclusion -> edge_is[X] -> point[0]);
					is_point -> data.coord = ext_point_z[ind_z].x;

					if(inclusion -> corner == BND)
						inclusion -> edge_is[X] -> side = INDEF;
					else
						inclusion -> edge_is[X] -> side = (is_point -> data.coord > tile.corner[ind_x].x) ? POS : NEG;
				}
				if(y_count != 0){
					if((inclusion -> edge_is[Y] = (edge_is_rec *)malloc(sizeof(edge_is_rec))) == NULL)
						error_message(MEMORY_ERROR, "Memory allocation error");
					inclusion -> edge_is[Y] -> count = 1;
					if((inclusion -> edge_is[Y] -> point = (is_point_rec *)calloc(1, sizeof(is_point_rec))) == NULL)
						error_message(MEMORY_ERROR, "Memory allocation error");
					is_point = &(inclusion -> edge_is[Y] -> point[0]);
					is_point -> data.coord = ext_point_z[ind_z].y;

					if(inclusion -> corner == BND)
						inclusion -> edge_is[Y] -> side = INDEF;
					else
						inclusion -> edge_is[Y] -> side = (is_point -> data.coord > tile.corner[ind_y].y) ? POS : NEG;
				}
			}

			if((inclusion -> face_is[Z] = (face_is_rec *)calloc(1, sizeof(face_is_rec))) == NULL)
				error_message(MEMORY_ERROR, "Memory allocation error");
			memcpy(&(inclusion -> face_is[Z] -> ellipse), &z_ellipse, sizeof(ellipse_rec));
			inclusion -> face_is[Z] -> extreme_pnt = pnt_extreme;
		}
		
/* note: face_is[X|Y|Z] -> side[0|1] is resolved after return */
		
		return(YES);    /* inclusion intersecting just one side (==> no edge is intersected) */
	}
	
	x_count = y_count = z_count = 0;         /* clear values computed in resolve_ellipsoid_x|y|z_side_touching !!! */

	cross_product(vec_x, dir_2, dir_3);
	size = sqrt(size_vec(vec_x));
	div_vec(vec_x, size);
	
	cross_product(vec_y, dir_3, dir_1);
	size = sqrt(size_vec(vec_y));
	div_vec(vec_y, size);
	
	cross_product(vec_z, dir_1, dir_2);
	size = sqrt(size_vec(vec_z));
	div_vec(vec_z, size);
	
	cross_product(dir_a, vec_y, dir_3);
	cross_product(dir_b, vec_z, dir_1);
	cross_product(dir_c, vec_x, dir_2);

/* IMPORTANT: do not modify the above computed orthonormal bases !!! */

/* precompute transformation matrices of intersection ellipses (in get_ellipsoid_x|y|z_side_intersection);
	 they do not change if ellipsoid is shifted;

	 currently transformation matrices are precomputed only for existing intersections;
	 consider their evaluation for fictitious Cartesian planes passing through ellipsoid center;
	 although this approach is computationally more demanding because fictitious intersections need to be computed
	 it is probably computationally more stable because evaluation of intersection passing through ellipsoid center 
	 is better conditioned then evaluation of intersection close to extreme point;
	 since intersection close to extreme point (touching) is avoided this should be not really an issue;
	 note however that fictitious intersection could be possibly used to simplify evaluation of real intersection 
	 (only center and principal radii need to be computed) */

	x_trans = y_trans = z_trans = NO;    /* IMPORTANT: must be initialized before first call to get_ellipsoid_x|y|z_side_intersection */

	if(x_intersection == YES)get_ellipsoid_x_side_intersection();
	if(y_intersection == YES)get_ellipsoid_y_side_intersection();
	if(z_intersection == YES)get_ellipsoid_z_side_intersection();
	
#ifdef DRAW_ELLIPSE
	if(draw == YES)ESIEventLoop(YES, "Intersection ellipses\n\n");
#endif

/* get relevant extreme points in GCS of intersection ellipse if inclusion intersection or touching with other side(s) happens;
	 if center of intersection ellipse has smaller coordinate than perpendicular side 
	 then extreme point with larger coordinate is searched for (and vice versa) */

	diff_xy = diff_xz = diff_yz = diff_yx = diff_zx = diff_zy = max_sz;        /* diffs of ellipse extreme points */

/* diff_xy = y-coordinate of relevant tile side - y-coordinate of relevant extreme point of intersection x_ellipse;
	 diff_xz = z-coordinate of relevant tile side - z-coordinate of relevant extreme point of intersection x_ellipse;
	 diff_yz = z-coordinate of relevant tile side - z-coordinate of relevant extreme point of intersection y_ellipse;
	 diff_yx = x-coordinate of relevant tile side - x-coordinate of relevant extreme point of intersection y_ellipse;
	 diff_zx = x-coordinate of relevant tile side - x-coordinate of relevant extreme point of intersection z_ellipse;
	 diff_zy = y-coordinate of relevant tile side - y-coordinate of relevant extreme point of intersection z_ellipse;

   note: diffs are computed in resolve_ellipsoid_yz|zx|xy_side_touching */

	x_shifting = y_shifting = z_shifting = YES;        /* enforce computation of diffs for all intersecting ellipses */
	
	shifting_strategy = 0;                             /* choose initial shifting strategy */

#ifndef DETOUCH
	if(shifting_strategy == TOUCHING)shifting_strategy++;        /* skip TOUCHING strategy */
#endif

	if(shifting_strategy < 0 || shifting_strategy >= MAX_STRATEGY)error_message(GENERAL_ERROR, "Unexpected situation");
	switch(shifting_strategy){
	case TOUCHING:
	case OPPOSITE:
	case MAXIMAL:
		cycles = 3;
		break;
	case TOTAL:
	case MINIMAL:
		cycles = 6;
		break;
	default:
		error_message(GENERAL_ERROR, "Unexpected situation");
		break;
	}

	run = 0;
	cycle = 0;
	eps = epsilon;
	subdivision = 0;

	while(YES){
		
/* get intersection ellipse extreme points */

		if(x_intersection == YES && x_shifting == YES){
			if(y_intersection == YES || z_intersection == YES || y_touching == YES || z_touching == YES)resolve_ellipsoid_yz_edge_touching(NO, NO);
		}
		if(y_intersection == YES && y_shifting == YES){
			if(z_intersection == YES || x_intersection == YES || z_touching == YES || x_touching == YES)resolve_ellipsoid_zx_edge_touching(NO, NO);
		}
		if(z_intersection == YES && z_shifting == YES){
			if(x_intersection == YES || y_intersection == YES || x_touching == YES || y_touching == YES)resolve_ellipsoid_xy_edge_touching(NO, NO);
		}

#if 0
		fprintf(stderr, "\ndiff_x %20.13e\n", diff_x);
		fprintf(stderr, "diff_yx %20.13e\n", diff_yx);
		fprintf(stderr, "diff_zx %20.13e\n\n", diff_zx);
		fprintf(stderr, "diff_y %20.13e\n", diff_y);
		fprintf(stderr, "diff_zy %20.13e\n", diff_zy);
		fprintf(stderr, "diff_xy %20.13e\n\n", diff_xy);
		fprintf(stderr, "diff_z %20.13e\n", diff_z);
		fprintf(stderr, "diff_xz %20.13e\n", diff_xz);
		fprintf(stderr, "diff_yz %20.13e\n\n", diff_yz);
#endif

#ifdef DRAW_EXTREME_POINT_ELLIPSE
		if(draw == YES)ESIEventLoop(YES, "Ellipse extreme points (relevant only)\n\n");
#endif

/* try to calculate (possibly minimal) shift so that neither of extreme points is closer to side than eps;

	 shift in particular direction influences intersection ellipse in perpendicular plane
	 thus diffs in other 2 perpendicular directions need to be recomputed only for ellipse in perpendicular plane;
	 individual directions are processed separately in staggered manner;

   note that if intersection ellipse with a tile side is touching tile edge 
	 then intersection ellipse with other side sharing the same edge must be also touching the edge */

		x_shifting = y_shifting = z_shifting = shifting = NO;

/* caution: a_diff_?? used for choosing initial shifting direction is not updated when shifting is applied;
	 therefore a_diff_?? is not declared globally, although it may be used in main when handling special intersection with all 3 sides */

		a_diff_yx = fabs(diff_yx);
		a_diff_zx = fabs(diff_zx);
		a_diff_zy = fabs(diff_zy);
		a_diff_xy = fabs(diff_xy);
		a_diff_xz = fabs(diff_xz);
		a_diff_yz = fabs(diff_yz);

		cnt_x = cnt_y = cnt_z = 0;

/* note: diff_x|y|z should not violate proximity criteria */

		if(a_diff_yx < eps && a_diff_yx > min_epsilon)cnt_x++;
		if(a_diff_zx < eps && a_diff_zx > min_epsilon)cnt_x++;
		if(a_diff_zy < eps && a_diff_zy > min_epsilon)cnt_y++;
		if(a_diff_xy < eps && a_diff_xy > min_epsilon)cnt_y++;
		if(a_diff_xz < eps && a_diff_xz > min_epsilon)cnt_z++;
		if(a_diff_yz < eps && a_diff_yz > min_epsilon)cnt_z++;
		
		if(cnt_x != 0 || cnt_y != 0 || cnt_z != 0){
			if(no_shift == YES)error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);

			cnt_xx = cnt_yy = cnt_zz = 0;

/* select shifting direction to start with;
	 generally choose that direction where only one of diff_ab is violated
	 and simultaneously there is no zero diff_cb;
	 currently magnitude of diffs is not considered;
	 note that shifting strategy applies to shifting in particular direction only !!! */
		
			if(a_diff_yx < eps && a_diff_yx > min_epsilon && a_diff_zx > min_epsilon)cnt_xx++;
			if(a_diff_zx < eps && a_diff_zx > min_epsilon && a_diff_yx > min_epsilon)cnt_xx++;
			if(a_diff_zy < eps && a_diff_zy > min_epsilon && a_diff_xy > min_epsilon)cnt_yy++;
			if(a_diff_xy < eps && a_diff_xy > min_epsilon && a_diff_zy > min_epsilon)cnt_yy++;
			if(a_diff_xz < eps && a_diff_xz > min_epsilon && a_diff_xz > min_epsilon)cnt_zz++;
			if(a_diff_yz < eps && a_diff_yz > min_epsilon && a_diff_zx > min_epsilon)cnt_zz++;

			if(cnt_xx == 1 || cnt_yy == 1 || cnt_zz == 1){
				if(cnt_xx == 1)
					run = X;
				else if(cnt_yy == 1)
					run = Y;
				else if(cnt_zz == 1)
					run = Z;
			}
			else{
				if(cnt_x == 1 || cnt_y == 1 || cnt_z == 1){
					if(cnt_xx == 1)
						run = X;
					else if(cnt_y == 1)
						run = Y;
					else if(cnt_z == 1)
						run = Z;
				}
			}
			
			count = 0;
			while(count++ != 3){         /* check x, y, z <==> count = 3 */
				switch(run % 3){
				case X:
					if(cnt_x != 0){
						if(get_ellipsoid_shift(diff_x, diff_yx, diff_zx, &shift_pos, &shift_neg, eps) == YES){
							x_shifting = shifting = YES;
							xyz_shift = total_shift.x;
							if(shifting_strategy == TOUCHING){
								diff_1_pos = diff_yx - shift_pos;
								diff_2_pos = diff_zx - shift_pos;
								diff_1_neg = diff_yx + shift_neg;
								diff_2_neg = diff_zx + shift_neg;
							}
							break;
						}
					}
					run++;
					continue;
				case Y:
					if(cnt_y != 0){
						if(get_ellipsoid_shift(diff_y, diff_zy, diff_xy, &shift_pos, &shift_neg, eps) == YES){
							y_shifting = shifting = YES;
							xyz_shift = total_shift.y;
							if(shifting_strategy == TOUCHING){
								diff_1_pos = diff_zy - shift_pos;
								diff_2_pos = diff_xy - shift_pos;
								diff_1_neg = diff_zy + shift_neg;
								diff_2_neg = diff_xy + shift_neg;
							}
							break;
						}
					}
					run++;
					continue;
				case Z:
					if(cnt_z != 0){
						if(get_ellipsoid_shift(diff_z, diff_xz, diff_yz, &shift_pos, &shift_neg, eps) == YES){
							z_shifting = shifting = YES;
							xyz_shift = total_shift.z;
							if(shifting_strategy == TOUCHING){
								diff_1_pos = diff_xz - shift_pos;
								diff_2_pos = diff_yz - shift_pos;
								diff_1_neg = diff_xz + shift_neg;
								diff_2_neg = diff_yz + shift_neg;
							}
							break;
						}
					}
					run++;
					continue;
				default:
					break;
				}
				
				if(shifting == YES)break;
			}
		}

		if(shifting == NO)break;
		
		switch(shifting_strategy){
		case TOUCHING:
			diff_1_pos = fabs(diff_1_pos);
			diff_2_pos = fabs(diff_2_pos);
			diff_1_neg = fabs(diff_1_neg);
			diff_2_neg = fabs(diff_2_neg);
			diff_pos = min_val(diff_1_pos, diff_2_pos);
			diff_neg = min_val(diff_1_neg, diff_2_neg);
			if(diff_pos < min_epsilon){
				shift = shift_pos;
				break;
			}
			else if(diff_neg < min_epsilon){
				shift = -shift_neg;
				break;
			}

/* if TOUCHING cannot be applied, TOTAL is applied */

		case TOTAL:
			if(fabs(xyz_shift + shift_pos) <= fabs(xyz_shift - shift_neg))
				shift = shift_pos;
			else
				shift = -shift_neg;
			break;
		case OPPOSITE:
			if(xyz_shift > 0)
				shift = -shift_neg;
			else
				shift = shift_pos;
			break;
		case MINIMAL:
			if((shift = shift_pos) > shift_neg)shift = -shift_neg;
			break;
		case MAXIMAL:
			if((shift = shift_pos) <= shift_neg)shift = -shift_neg;
			break;
		default:
			break;
		}

		if(cycle++ == cycles){

/* apply cyclic change of shifting strategy with eps subdivision when completing the whole cycle;
	 number of cycles should be positive multiple of 3 (3, 6, 9, ....) because there are 3 coordinates to check */

			cycle = 0;
			if(++shifting_strategy == MAX_STRATEGY){
				shifting_strategy = 0;

#ifndef DETOUCH
				if(shifting_strategy == TOUCHING)shifting_strategy++;    /* skip TOUCHING strategy */
#endif

				eps /= 2.0;
				subdivision++;
				if(subdivision == max_subdivisions || eps <= min_epsilon * 10.0)error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);

/* prevent shifting using previous eps */
				
				x_shifting = y_shifting = z_shifting = NO;

/* enforce recomputing shift in the same direction for new epsilon */

				run--;
			}

			switch(shifting_strategy){
			case TOUCHING:
			case OPPOSITE:
			case MAXIMAL:
				cycles = 3;
				break;
			case TOTAL:
			case MINIMAL:
				cycles = 6;
				break;
			default:
				break;
			}
		}
		
		if(xyz_shift + shift <= MULTIPLE_EPS * epsilon){
#ifdef DRAW_ELLIPSE
			if(x_intersection == YES)draw_ellipse(&x_ellipse, 0, bgcolor);
			if(y_intersection == YES)draw_ellipse(&y_ellipse, 0, bgcolor);
			if(z_intersection == YES)draw_ellipse(&z_ellipse, 0, bgcolor);
#endif
#ifdef DRAW_EXTREME_POINT_ELLIPSOID
			for(i = 0; i < 2; i++){
				draw_point(&ext_point_x[i], 0, bgcolor);
				draw_point(&ext_point_y[i], 0, bgcolor);
				draw_point(&ext_point_z[i], 0, bgcolor);
			}
#endif
#ifdef DRAW_EXTREME_POINT_ELLIPSE
			if(x_intersection == YES){
				if(y_intersection == YES || y_touching == YES)draw_point(&extreme_pnt_xy, 0, bgcolor);
				if(z_intersection == YES || z_touching == YES)draw_point(&extreme_pnt_xz, 0, bgcolor);
			}
			if(y_intersection == YES){
				if(z_intersection == YES || z_touching == YES)draw_point(&extreme_pnt_yz, 0, bgcolor);
				if(x_intersection == YES || x_touching == YES)draw_point(&extreme_pnt_yx, 0, bgcolor);
			}
			if(z_intersection == YES){
				if(x_intersection == YES || x_touching == YES)draw_point(&extreme_pnt_zx, 0, bgcolor);
				if(y_intersection == YES || y_touching == YES)draw_point(&extreme_pnt_zy, 0, bgcolor);
			}
#endif

#ifdef PARAMETRIC_ROUTINES
/* caution: ctrl points are not shifted */
#endif

/* shift ellipsoid center at first */

			if(x_shifting == YES)ellipsoid -> center.x += shift;
			if(y_shifting == YES)ellipsoid -> center.y += shift;
			if(z_shifting == YES)ellipsoid -> center.z += shift;
			
/* update tile corner points in LCS of ellipsoid after center of ellipsoid was shifted;
	 this must be done before call to get_ellipsoid_x|y|z_side_intersection */

			for(i = 0; i < 2; i++){
				transform_point_from_global_to_local(&(tile.corner[i]), &corner_point[i], &(ellipsoid -> center), &trans_g2l);
				transform_local_from_ellipsoid_to_sphere(&corner_point[i], &corner_point[i], &scale_e2s);
			}

			if(x_shifting == YES){
#ifdef DEBUG_PRINT
				if(fabs(diff_x) > min_epsilon && fabs(diff_x) < eps)fprintf(stderr, "Violation: diff_x = %e\n", diff_x);
				if(fabs(diff_yx) > min_epsilon && fabs(diff_yx) < eps)fprintf(stderr, "Violation: diff_yx = %e\n", diff_yx);
				if(fabs(diff_zx) > min_epsilon && fabs(diff_zx) < eps)fprintf(stderr, "Violation: diff_zx = %e\n", diff_zx);
				
				fprintf(stderr, "X-shifting inclusion %ld by %e due to edge touching\n", inclusion -> id, shift);
#endif
				
				total_shift.x += shift;
				
				for(i = 0; i < 2; i++){
					ext_point_x[i].x += shift;
					ext_point_y[i].x += shift;
					ext_point_z[i].x += shift;
				}
				
				diff_x -= shift;
				if(y_intersection == YES || y_touching == YES){
					y_ellipse.center.x += shift;
					diff_yx -= shift;
				}
				if(y_intersection == YES){

/* caution: only relevant ellipse extreme points are to be handled */

					if(z_intersection == YES || z_touching == YES){
						extreme_pnt_yz.x += shift;
#ifdef DRAW_EXTREME_POINT_ELLIPSE
						draw_point(&extreme_pnt_yz, 0, cyan);
#endif
					}
					if(x_intersection == YES || x_touching == YES){
						extreme_pnt_yx.x += shift;
#ifdef DRAW_EXTREME_POINT_ELLIPSE
						draw_point(&extreme_pnt_yx, 0, cyan);
#endif
					}
				}
				if(z_intersection == YES || z_touching == YES){
					z_ellipse.center.x += shift;
					diff_zx -= shift;
				}
				if(z_intersection == YES){

/* caution: only relevant ellipse extreme points are to be handled */

					if(x_intersection == YES || x_touching == YES){
						extreme_pnt_zx.x += shift;
#ifdef DRAW_EXTREME_POINT_ELLIPSE
						draw_point(&extreme_pnt_zx, 0, magenta);
#endif
					}
					if(y_intersection == YES || y_touching == YES){
						extreme_pnt_zy.x += shift;
#ifdef DRAW_EXTREME_POINT_ELLIPSE
						draw_point(&extreme_pnt_zy, 0, magenta);
#endif
					}
				}
			
/* update touching and intersection */
			
				x_intersection = x_touching = NO;
				if(ext_point_x[0].x < tile.corner[0].x + eps || ext_point_x[1].x > tile.corner[1].x - eps){
					if(fabs(diff_x) < min_epsilon)
						x_touching = YES;
					else{
						x_intersection = YES;
						get_ellipsoid_x_side_intersection();
					}
				}
			}
			
			if(y_shifting == YES){
#ifdef DEBUG_PRINT
				if(fabs(diff_y) > min_epsilon && fabs(diff_y) < eps)fprintf(stderr, "Violation: diff_y = %e\n", diff_y);
				if(fabs(diff_zy) > min_epsilon && fabs(diff_zy) < eps)fprintf(stderr, "Violation: diff_zy = %e\n", diff_zy);
				if(fabs(diff_xy) > min_epsilon && fabs(diff_xy) < eps)fprintf(stderr, "Violation: diff_xy = %e\n", diff_xy);
				
				fprintf(stderr, "Y-shifting inclusion %ld by %e due to edge touching\n", inclusion -> id, shift);
#endif

				total_shift.y += shift;
				
				for(i = 0; i < 2; i++){
					ext_point_x[i].y += shift;
					ext_point_y[i].y += shift;
					ext_point_z[i].y += shift;
				}
				
				diff_y -= shift;
				if(z_intersection == YES || z_touching == YES){
					z_ellipse.center.y += shift;
					diff_zy -= shift;
				}
				if(z_intersection == YES){
					
/* caution: only relevant ellipse extreme points are to be handled */

					if(x_intersection == YES || x_touching == YES){
						extreme_pnt_zx.y += shift;
#ifdef DRAW_EXTREME_POINT_ELLIPSE
						draw_point(&extreme_pnt_zx, 0, magenta);
#endif
					}
					if(y_intersection == YES || y_touching == YES){
						extreme_pnt_zy.y += shift;
#ifdef DRAW_EXTREME_POINT_ELLIPSE
						draw_point(&extreme_pnt_zy, 0, magenta);
#endif
					}
				}
				if(x_intersection == YES || x_touching == YES){
					x_ellipse.center.y += shift;
					diff_xy -= shift;
				}
				if(x_intersection == YES){

/* caution: only relevant ellipse extreme points are to be handled */

					if(y_intersection == YES || y_touching == YES){
						extreme_pnt_xy.y += shift;
#ifdef DRAW_EXTREME_POINT_ELLIPSE
						draw_point(&extreme_pnt_xy, 0, green);
#endif
					}
					if(z_intersection == YES || z_touching == YES){
						extreme_pnt_xz.y += shift;
#ifdef DRAW_EXTREME_POINT_ELLIPSE
						draw_point(&extreme_pnt_xz, 0, green);
#endif
					}
				}
			
/* update touching and intersection */
			
				y_intersection = y_touching = NO;
				if(ext_point_y[0].y < tile.corner[0].y + eps || ext_point_y[1].y > tile.corner[1].y - eps){
					if(fabs(diff_y) < min_epsilon)
						y_touching = YES;
					else{
						y_intersection = YES;
						get_ellipsoid_y_side_intersection();
					}
				}
			}
			
			if(z_shifting == YES){
#ifdef DEBUG_PRINT
				if(fabs(diff_z) > min_epsilon && fabs(diff_z) < eps)fprintf(stderr, "Violation: diff_z = %e\n", diff_z);
				if(fabs(diff_xz) > min_epsilon && fabs(diff_xz) < eps)fprintf(stderr, "Violation: diff_xz = %e\n", diff_xz);
				if(fabs(diff_yz) > min_epsilon && fabs(diff_yz) < eps)fprintf(stderr, "Violation: diff_yz = %e\n", diff_yz);
				
				fprintf(stderr, "Z-shifting inclusion %ld by %e due to edge touching\n", inclusion -> id, shift);
#endif
				
				total_shift.z += shift;
				
				for(i = 0; i < 2; i++){
					ext_point_x[i].z += shift;
					ext_point_y[i].z += shift;
					ext_point_z[i].z += shift;
				}
				
				diff_z -= shift;
				if(x_intersection == YES || x_touching == YES){
					x_ellipse.center.z += shift;
					diff_xz -= shift;
				}
				if(x_intersection == YES){
					
/* caution: only relevant ellipse extreme points are to be handled */

					if(y_intersection == YES || y_touching == YES){
						extreme_pnt_xy.z += shift;
#ifdef DRAW_EXTREME_POINT_ELLIPSE
						draw_point(&extreme_pnt_xy, 0, green);
#endif
					}
					if(z_intersection == YES || z_touching == YES){
						extreme_pnt_xz.z += shift;
#ifdef DRAW_EXTREME_POINT_ELLIPSE
						draw_point(&extreme_pnt_xz, 0, green);
#endif
					}
				}
				if(y_intersection == YES || y_touching == YES){
					y_ellipse.center.z += shift;
					diff_yz -= shift;
				}
				if(y_intersection == YES){

/* caution: only relevant ellipse extreme points are to be handled */

					if(z_intersection == YES || z_touching == YES){
						extreme_pnt_yz.z += shift;
#ifdef DRAW_EXTREME_POINT_ELLIPSE
						draw_point(&extreme_pnt_yz, 0, cyan);
#endif
					}
					if(x_intersection == YES || x_touching == YES){
						extreme_pnt_yx.z += shift;
#ifdef DRAW_EXTREME_POINT_ELLIPSE
						draw_point(&extreme_pnt_yx, 0, cyan);
#endif
					}
				}
				
/* update touching and intersection */
			
				z_intersection = z_touching = NO;
				if(ext_point_z[0].z < tile.corner[0].z + eps || ext_point_z[1].z > tile.corner[1].z - eps){
					if(fabs(diff_z) < min_epsilon)
						z_touching = YES;
					else{
						z_intersection = YES;
						get_ellipsoid_z_side_intersection();
					}
				}
			}
			
#ifdef DRAW_ELLIPSE
			if(x_intersection == YES && x_shifting == NO)draw_ellipse(&x_ellipse, 0, green);
			if(y_intersection == YES && y_shifting == NO)draw_ellipse(&y_ellipse, 0, cyan);
			if(z_intersection == YES && z_shifting == NO)draw_ellipse(&z_ellipse, 0, magenta);
#endif
#ifdef DRAW_EXTREME_POINT_ELLIPSOID
			for(i = 0; i < 2; i++){
				draw_point(&ext_point_x[i], 0, green);
				draw_point(&ext_point_y[i], 0, cyan);
				draw_point(&ext_point_z[i], 0, magenta);
			}
#endif
#ifdef DEBUG_DRAW
			if(draw == YES)ESIEventLoop(YES, "Shifting\n\n");
#endif
		}
	}
		
/* resolve intersections with tile edges */

	x_diff = y_diff = z_diff = max_sz;     /* diffs of ellipsoid intersections with tile edges */

#if 0 /* DO NOT USE - just to check that current intersection ellipsis are still correct */
	{
		if(x_intersection == YES)get_ellipsoid_x_side_intersection();
		if(y_intersection == YES)get_ellipsoid_y_side_intersection();
		if(z_intersection == YES)get_ellipsoid_z_side_intersection();
	}
#endif

/* x_diff = x-coordinate of relevant tile side (corner) - x-coordinate of the closest ellipsoid intersection with relevant tile x-edge;
	 y_diff = y-coordinate of relevant tile side (corner) - y-coordinate of the closest ellipsoid intersection with relevant tile y-edge;
	 z_diff = z-coordinate of relevant tile side (corner) - z-coordinate of the closest ellipsoid intersection with relevant tile z-edge;

   note: diffs, counts and coords are computed in get_ellipsoid_yz|zx|xy_edge_intersection */

	x_count = y_count = z_count = 0;

	x_axis = y_axis = z_axis = NO;

/* note: x_touching, y_touching and z_touching is not explicitly handled;
	 for example if x_touching == YES, then intersection with edges y and z can happen only if ext_point_x is on edge y and/or z;
	 but in such a case y_intersection and/or z_intersection equals to YES and intersection is thus resolved
	 when processing y_intersection and/or z_intersection */

	if(x_intersection == YES)get_ellipsoid_yz_edge_intersection(YES);
	if(y_intersection == YES)get_ellipsoid_zx_edge_intersection(YES);
	if(z_intersection == YES)get_ellipsoid_xy_edge_intersection(YES);

/* IMPORTANT: computed edge intersections may differ from exact intersections of ellipsoid with tile edges
	 because if the intersection ellipse extreme point is within min_epsilon from edge,
	 then just 1 intersection is considered, despite the fact that in reality there can be
	 0 or 1 or 2 intersections (see is_x_21.geom) */

#if 0
	for(i = 0; i < x_count; i++)fprintf(stderr, "x[%d]: %20.13e\n", i, x_coord[i]);
	for(i = 0; i < y_count; i++)fprintf(stderr, "y[%d]: %20.13e\n", i, y_coord[i]);
	for(i = 0; i < z_count; i++)fprintf(stderr, "z[%d]: %20.13e\n", i, z_coord[i]);
	fprintf(stderr, "\n");
#endif

#if 0 /* DO NOT USE - just to check intersections with edges */
	{
		point_rec point[2], corner;
		int i, cnt;
		
		corner.x = tile.corner[ind_x].x;
		corner.y = tile.corner[ind_y].y;
		corner.z = tile.corner[ind_z].z;

		cnt = get_ellipsoid_line_intersection(ellipsoid, &corner, &dir_x, point);
		for(i = 0; i < cnt; i++)fprintf(stderr, "x[%d]: %20.13e %20.13e %20.13e\n", i, point[i].x, point[i].y, point[i].z); 
		cnt = get_ellipsoid_line_intersection(ellipsoid, &corner, &dir_y, point);
		for(i = 0; i < cnt; i++)fprintf(stderr, "y[%d]: %20.13e %20.13e %20.13e\n", i, point[i].x, point[i].y, point[i].z); 
		cnt = get_ellipsoid_line_intersection(ellipsoid, &corner, &dir_z, point);
		for(i = 0; i < cnt; i++)fprintf(stderr, "z[%d]: %20.13e %20.13e %20.13e\n", i, point[i].x, point[i].y, point[i].z); 
		fprintf(stderr, "\n");
	}
#endif
	
	a_x_diff = fabs(x_diff);
	a_y_diff = fabs(y_diff);
	a_z_diff = fabs(z_diff);
	
	shifting = NO;
	if(a_x_diff < eps || a_y_diff < eps || a_z_diff < eps){
		
/* there should not happen the case when no tile edge is intersected
	 otherwise all diffs will be equal to tile size > eps */

#ifdef DEBUG
		if(x_count + y_count + z_count == 0)error_message(GENERAL_ERROR, "Unexpected situation");
#endif
		
/* there should not happen the case when only one tile edge is intersected only once
	 because intersection ellipses with sides sharing the edge are not allowed to be closer to other to edges than eps */

#ifdef DEBUG
		if(x_count + y_count + z_count == 1)error_message(GENERAL_ERROR, "Unexpected situation");
#endif

/* there should not happen the case when only one tile edge is intersected (twice) 
	 no matter whether two or three tile sides are intersected;
	 if two tile sides are intersected then these sides are sharing the edge;
	 since intersection ellipses with these sides are not allowed to be closer to other to edges than eps,
	 intersection of ellipsoid with edge cannot be closer to tile corner than eps;
	 if three tile sides are intersected, then intersection ellipse with either of sides sharing edge 
	 must have extreme point in the direction of edge closer (in the direction of edge) to tile corner
	 than intersection of ellipsoid with edge;

	 there should also not happen the case when only two tile edges are intersected only once and more than
	 one tile side is intersected;
	 in such a case side sharing the two edges and at least one more side are intersected;
	 but itersection ellipse with the side(s) not sharing both edges cannot be closer to the third edge than epsilon
	 (otherwise the third edge would be intersected);

	 there should also not happen the case when only two tile edges are intersected only once and only the side
	 sharing these two edges is intersected;
	 in such a case the intersection points are extreme points of the ellipsoid and thus touching points of the ellipsoid;
	 but touching points cannot be closer to edges than eps */

#ifdef DEBUG
		if(x_count + y_count + z_count == 2)error_message(GENERAL_ERROR, "Unexpected situation");
#else
		if(x_count + y_count + z_count <= 2)error_message(GENERAL_ERROR, "Unexpected situation");   /* check all above cases together */
#endif

		if((a_x_diff < min_epsilon || x_count == 0) && 
			 (a_y_diff < min_epsilon || y_count == 0) && 
			 (a_z_diff < min_epsilon || z_count == 0)){
			
/* shift relevant intersection to tile corner but do not shift the inclusion;
	 check at first that such shift will not cause that intersections of the same edge will get closer to each other than min_epsilon */

			corner_x_between = corner_y_between = corner_z_between = NO;
			if(x_count == 2 && shifting == NO){
				if(x_coord[0] < tile.corner[ind_x].x && x_coord[1] > tile.corner[ind_x].x){
					corner_x_between = YES;
					if(x_coord[1] - x_coord[0] - a_x_diff < min_epsilon)shifting = YES;
				}
			}
			if(y_count == 2 && shifting == NO){
				if(y_coord[0] < tile.corner[ind_y].y && y_coord[1] > tile.corner[ind_y].y){
					corner_y_between = YES;
					if(y_coord[1] - y_coord[0] - a_y_diff < min_epsilon)shifting = YES;
				}
			}
			if(z_count == 2 && shifting == NO){
				if(z_coord[0] < tile.corner[ind_z].z && z_coord[1] > tile.corner[ind_z].z){
					corner_z_between = YES;
					if(z_coord[1] - z_coord[0] - a_z_diff < min_epsilon)shifting = YES;
				}
			}
			
			if(shifting == NO){
				inclusion -> corner = BND;
				
				if(x_count != 0){
					ind = 0;
					if(x_count == 2){
						
/* sign of diff determines intersection to be shifted */

						if(corner_x_between == YES){
							if(x_diff < 0.0)ind = 1;
						}
						else{
							if(x_diff > 0.0)ind = 1;
						}
					}
					x_coord[ind] += x_diff;
				}
				if(y_count != 0){
					ind = 0;
					if(y_count == 2){
						
/* sign of diff determines intersection to be shifted */

						if(corner_y_between == YES){
							if(y_diff < 0.0)ind = 1;
						}
						else{
							if(y_diff > 0.0)ind = 1;
						}
					}
					y_coord[ind] += y_diff;
				}
				if(z_count != 0){
					ind = 0;
					if(z_count == 2){
						
/* sign of diff determines intersection to be shifted */

						if(corner_z_between == YES){
							if(z_diff < 0.0)ind = 1;
						}
						else{
							if(z_diff > 0.0)ind = 1;
						}
					}
					z_coord[ind] += z_diff;
				}
			}
		}
		else
			shifting = YES;
	}
	
	if(shifting == YES){
		if(no_shift == YES)error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);

/* it would be generally necessary to run again the loop performing different shifts while checking for violation of proximity criteria;
	 this is currently considered too complex;
	 therefore several shifts are processed and checked against proximity criteria without attempting to fix their violation through other shift;
	 currently only the shifts satisfying simultaneously the following criteria are considered:
	 - only shift towards the corner !!!
	   this means that shift is always added, never subtracted;
		 IMPORTANT: as a conseqeunce tile corner is always on inclusion boundary !!!
	 - only shift along tile edges violating proximity criteria;
	 - only shift not violating maximal shift */

/* shifting failure often occurs if there is only 1 intersection with an axis (while there are 2 intersections
	 with other two axes) and when diff along that axis does not violate proximity criteria;
	 however shift along this axes may resolve the problem;
	 therefore the above criteria are weakend so that also shift along tile edges with only one intersection point
	 and not violating proximity criteria is performed */

		x_shifting = y_shifting = z_shifting = NO;
		sz_x = sz_y = sz_z = max_sz;
		count = 0;
		if((a_x_diff < eps || x_count == 1) && x_diff != 0.0){
			if((total_shift.x += x_diff) <= MULTIPLE_EPS * epsilon){
				sz_x = size_vec(total_shift);
				x_shifting = YES;
				count++;
			}
			total_shift.x -= x_diff;
		}
		if((a_y_diff < eps || y_count == 1) && y_diff != 0.0){
			if((total_shift.y += y_diff) <= MULTIPLE_EPS * epsilon){
				sz_y = size_vec(total_shift);
				y_shifting = YES;
				count++;
			}
			total_shift.y -= y_diff;
		}
		if((a_z_diff < eps || z_count == 1) && z_diff != 0.0){
			if((total_shift.z += z_diff) <= MULTIPLE_EPS * epsilon){
				sz_z = size_vec(total_shift);
				z_shifting = YES;
				count++;
			}
			total_shift.z -= z_diff;
		}
		
/* tentative shifts are investigated in the order of increasing total shift;
	 IMPORTANT: when particular shift is checked without success, opposite shift is performed to restore original state;
	 to prevent redundant work, the opposite shift is performed only up to that phase of original shift
	 at which proximity violation was detected */

/* note: the above strategy of ordering shifts in order of increasing total shift may not be efficient
	 if some shifts are smaller than min_epsilon;
	 then is it probably better to start with smallest shift larger than min_epsilon (see is_x_21.geom) */

		shift_vec.x = shift_vec.y = shift_vec.z = 0.0;
		
		while(count--){
			if(x_shifting == YES){
				if(sz_x < sz_y && sz_x < sz_z){
#ifdef DEBUG_PRINT
					fprintf(stderr, "X-shifting inclusion %ld by %e due to corner touching\n", inclusion -> id, x_diff);
#endif
					shift_vec.x = x_diff;
					if(check_ellipsoid_shift_x_corner(x_diff, eps, &phase) == YES){
						x_diff = 0.0;
						shifting = NO;
						break;
					}
					
#ifdef DEBUG_PRINT
					fprintf(stderr, "Back x-shifting inclusion %ld by %e due to corner touching\n", inclusion -> id, -x_diff);
#endif
					ellipsoid_back_shift_x_corner(x_diff, phase);               /* do not change sign of shift */
					shift_vec.x = 0.0;
					sz_x = max_sz;                             /* prevent sz_x to be minimal for next attempts */
					continue;
				}
			}
			if(y_shifting == YES){
				if(sz_y < sz_z && sz_y < sz_x){
#ifdef DEBUG_PRINT
					fprintf(stderr, "Y-shifting inclusion %ld by %e due to corner touching\n", inclusion -> id, y_diff);
#endif
					shift_vec.y = y_diff;
					if(check_ellipsoid_shift_y_corner(y_diff, eps, &phase) == YES){
						y_diff = 0.0;
						shifting = NO;
						break;
					}
					
#ifdef DEBUG_PRINT
					fprintf(stderr, "Back y-shifting inclusion %ld by %e due to corner touching\n", inclusion -> id, -y_diff);
#endif
					ellipsoid_back_shift_y_corner(y_diff, phase);               /* do not change sign of shift */
					shift_vec.y = 0.0;
					sz_y = max_sz;                             /* prevent sz_y to be minimal for next attempts */
					continue;
				}
			}
			if(z_shifting == YES){
				if(sz_z < sz_x && sz_z < sz_y){
#ifdef DEBUG_PRINT
					fprintf(stderr, "Z-shifting inclusion %ld by %e due to corner touching\n", inclusion -> id, z_diff);
#endif
					shift_vec.z = z_diff;
					if(check_ellipsoid_shift_z_corner(z_diff, eps, &phase) == YES){
						z_diff = 0.0;
						shifting = NO;
						break;
					}

#ifdef DEBUG_PRINT
					fprintf(stderr, "Back z-shifting inclusion %ld by %e due to corner touching\n", inclusion -> id, -z_diff);
#endif
					ellipsoid_back_shift_z_corner(z_diff, phase);               /* do not change sign of shift */
					shift_vec.z = 0.0;
					sz_z = max_sz;                             /* prevent sz_z to be minimal for next attempts */
					continue;
				}
			}
		}
		
		inclusion -> corner = BND;                  /* shifting only towards corner !!! */

		if(shifting == YES)error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);
	}

	if(shifting == NO){
		if(inclusion -> corner == OUT){
			if(x_count == 2 && y_count == 2 && z_count == 2){
				
/* check that tile corner is between tile edges intersections
	 results should be same no matter which tile edge is considered */

				coord = tile.corner[ind_x].x;
				if(x_coord[0] < coord && x_coord[1] > coord){
					inclusion -> corner = IN;
					
#ifdef DEBUG
					coord = tile.corner[ind_y].y;
					if(y_coord[0] > coord || y_coord[1] < coord)error_message(GENERAL_ERROR, "Unexpected situation");
					coord = tile.corner[ind_z].z;
					if(z_coord[0] > coord || z_coord[1] < coord)error_message(GENERAL_ERROR, "Unexpected situation");
#endif
				}
			}
		}
	}
	
	ret_val = NO;

	if(x_touching == YES || x_intersection == YES){
		if((inclusion -> face_is[X] = (face_is_rec *)calloc(1, sizeof(face_is_rec))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
		memcpy(&(inclusion -> face_is[X] -> ellipse), &x_ellipse, sizeof(ellipse_rec));
		
/* compute x_ellipse extreme points;
	 note: I do not reuse already computed x_ellipse extreme points because only relevant (if intersection or touching of tile side occurs
	 and closer to tile edge) are available */

		if(x_intersection == YES){
			if(y_count != 0 || z_count != 0){
				if((pnt_extreme = (point_rec *)calloc(4, sizeof(point_rec))) == NULL)
					error_message(MEMORY_ERROR, "Memory allocation error");
				inclusion -> face_is[X] -> extreme_pnt = pnt_extreme;
				
				resolve_ellipsoid_yz_edge_touching(YES, NO);
				copy_vec(pnt_extreme[2 * XY + 0], extreme_pnt_xy);
				copy_vec(pnt_extreme[2 * XZ + 0], extreme_pnt_xz);
#if 0
				resolve_ellipsoid_yz_edge_touching(NO, YES);
				copy_vec(pnt_extreme[2 * XY + 1], extreme_pnt_xy);
				copy_vec(pnt_extreme[2 * XZ + 1], extreme_pnt_xz);
#else /* more efficient */
				sub_vec(vec_tmp, x_ellipse.center, extreme_pnt_xy);
				add_vec(pnt_extreme[2 * XY + 1], x_ellipse.center, vec_tmp);
				sub_vec(vec_tmp, x_ellipse.center, extreme_pnt_xz);
				add_vec(pnt_extreme[2 * XZ + 1], x_ellipse.center, vec_tmp);
#ifdef DRAW_EXTREME_POINT_ELLIPSE
				draw_point(&pnt_extreme[2 * XY + 1], 0, green);
				draw_point(&pnt_extreme[2 * XZ + 1], 0, green);
#endif
#endif
				inclusion -> face_is[X] -> weight = get_ellipse_weight(&x_ellipse, &pnt_extreme[2 * XY + 0], &pnt_extreme[2 * XZ + 1], &x_trans_g2l);
			}
		}

		ret_val = YES;
	}
	if(y_touching == YES || y_intersection == YES){
		if((inclusion -> face_is[Y] = (face_is_rec *)calloc(1, sizeof(face_is_rec))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
		memcpy(&(inclusion -> face_is[Y] -> ellipse), &y_ellipse, sizeof(ellipse_rec));

/* compute y_ellipse extreme points;
	 note: I do not reuse already computed y_ellipse extreme points because only relevant (if intersection or touching of tile side occurs
	 and closer to tile edge) are available */

		if(y_intersection == YES){
			if(z_count != 0 || x_count != 0){
				if((pnt_extreme = (point_rec *)calloc(4, sizeof(point_rec))) == NULL)
					error_message(MEMORY_ERROR, "Memory allocation error");
				inclusion -> face_is[Y] -> extreme_pnt = pnt_extreme;
				
				resolve_ellipsoid_zx_edge_touching(YES, NO);
				copy_vec(pnt_extreme[2 * YZ + 0], extreme_pnt_yz);
				copy_vec(pnt_extreme[2 * YX + 0], extreme_pnt_yx);
#if 0
				resolve_ellipsoid_zx_edge_touching(NO, YES);
				copy_vec(pnt_extreme[2 * YZ + 1], extreme_pnt_yz);
				copy_vec(pnt_extreme[2 * YX + 1], extreme_pnt_yx);
#else /* more efficient */
				sub_vec(vec_tmp, y_ellipse.center, extreme_pnt_yz);
				add_vec(pnt_extreme[2 * YZ + 1], y_ellipse.center, vec_tmp);
				sub_vec(vec_tmp, y_ellipse.center, extreme_pnt_yx);
				add_vec(pnt_extreme[2 * YX + 1], y_ellipse.center, vec_tmp);
#ifdef DRAW_EXTREME_POINT_ELLIPSE
				draw_point(&pnt_extreme[2 * YZ + 1], 0, cyan);
				draw_point(&pnt_extreme[2 * YX + 1], 0, cyan);
#endif
#endif
						
				inclusion -> face_is[Y] -> weight = get_ellipse_weight(&y_ellipse, &pnt_extreme[2 * YZ + 0], &pnt_extreme[2 * YX + 1], &y_trans_g2l);
			}
		}
		
		ret_val = YES;
	}
	if(z_touching == YES || z_intersection == YES){
		if((inclusion -> face_is[Z] = (face_is_rec *)calloc(1, sizeof(face_is_rec))) == NULL)
			error_message(MEMORY_ERROR, "Memory allocation error");
		memcpy(&(inclusion -> face_is[Z] -> ellipse), &z_ellipse, sizeof(ellipse_rec));

/* compute z_ellipse extreme points;
	 note: I do not reuse already computed z_ellipse extreme points because only relevant (if intersection or touching of tile side occurs
	 and closer to tile edge) are available */

		if(z_intersection == YES){
			if(x_count != 0 || y_count != 0){
				if((pnt_extreme = (point_rec *)calloc(4, sizeof(point_rec))) == NULL)
					error_message(MEMORY_ERROR, "Memory allocation error");
				inclusion -> face_is[Z] -> extreme_pnt = pnt_extreme;
				
				resolve_ellipsoid_xy_edge_touching(YES, NO);
				copy_vec(pnt_extreme[2 * ZX + 0], extreme_pnt_zx);
				copy_vec(pnt_extreme[2 * ZY + 0], extreme_pnt_zy);
#if 0
				resolve_ellipsoid_xy_edge_touching(NO, YES);
				copy_vec(pnt_extreme[2 * ZX + 1], extreme_pnt_zx);
				copy_vec(pnt_extreme[2 * ZY + 1], extreme_pnt_zy);
#else /* more efficient */
				sub_vec(vec_tmp, z_ellipse.center, extreme_pnt_zx);
				add_vec(pnt_extreme[2 * ZX + 1], z_ellipse.center, vec_tmp);
				sub_vec(vec_tmp, z_ellipse.center, extreme_pnt_zy);
				add_vec(pnt_extreme[2 * ZY + 1], z_ellipse.center, vec_tmp);
#ifdef DRAW_EXTREME_POINT_ELLIPSE
				draw_point(&pnt_extreme[2 * ZX + 1], 0, magenta);
				draw_point(&pnt_extreme[2 * ZY + 1], 0, magenta);
#endif
#endif
						
				inclusion -> face_is[Z] -> weight = get_ellipse_weight(&z_ellipse, &pnt_extreme[2 * ZX + 0], &pnt_extreme[2 * ZY + 1], &z_trans_g2l);
			}
		}
		
		ret_val = YES;
	}
	
	if(ret_val == YES){
		if(x_count != 0){
			if((inclusion -> edge_is[X] = (edge_is_rec *)malloc(sizeof(edge_is_rec))) == NULL)
				error_message(MEMORY_ERROR, "Memory allocation error");
			inclusion -> edge_is[X] -> count = x_count;
			if((inclusion -> edge_is[X] -> point = (is_point_rec *)calloc(x_count, sizeof(is_point_rec))) == NULL)
				error_message(MEMORY_ERROR, "Memory allocation error");
			for(i = 0; i < x_count; i++){
				is_point = &(inclusion -> edge_is[X] -> point[i]);
				is_point -> data.coord = x_coord[i];
			}
			
/* resolve position of intersection points with respect to tile corner;
	 resolve status of intersection points;
	 resolve ellipse edge halving */

			if(x_count == 1){
				inclusion -> edge_is[X] -> point[0].data.status = TOUCH;
				
				if(inclusion -> corner == BND)
					inclusion -> edge_is[X] -> side = INDEF;
				else
					inclusion -> edge_is[X] -> side = (x_coord[0] > tile.corner[ind_x].x) ? POS : NEG;
			}
			else{
				inclusion -> edge_is[X] -> point[0].data.status = START;
				inclusion -> edge_is[X] -> point[1].data.status = END;
				
/* o----x-----x------------o   ==> POS
	 o---------------x----x--o   ==> NEG
   -----x---o----x----------   ==> BOTH

	 o - corner
	 x - edge_is */

				if(x_coord[0] > tile.corner[ind_x].x - min_epsilon * 10.0)
					inclusion -> edge_is[X] -> side = POS;
				else if(x_coord[1] < tile.corner[ind_x].x + min_epsilon * 10.0)
					inclusion -> edge_is[X] -> side = NEG;
				else{
#ifdef DEBUG
					if(inclusion -> corner != IN || y_count != 2 || z_count != 2)error_message(GENERAL_ERROR, "Unexpected situation");
#endif
					inclusion -> edge_is[X] -> side = BOTH;

/* IMPORTANT: update coordinate of edge_is point outside tile to fall inside tile by appropriate periodic repetition;
	 this is necessary for coordinate-based ordering */

					inclusion -> edge_is[X] -> point[ind_x].data.coord += tile.size.x * sign[ind_x];
				}

				pnt_1.x = x_coord[0];
				pnt_2.x = x_coord[1];
				pnt_1.y = pnt_2.y = tile.corner[ind_y].y;
				pnt_1.z = pnt_2.z = tile.corner[ind_z].z;

				get_ellipse_normal_point(&y_ellipse, &pnt_1, &normal1);
				get_ellipse_normal_point(&y_ellipse, &pnt_2, &normal2);

				if(-dot_product(normal1, normal2) > ZERO_ANGLE)inclusion -> face_is[Y] -> edge_half[YX] = YES;

				get_ellipse_normal_point(&z_ellipse, &pnt_1, &normal1);
				get_ellipse_normal_point(&z_ellipse, &pnt_2, &normal2);

				if(-dot_product(normal1, normal2) > ZERO_ANGLE)inclusion -> face_is[Z] -> edge_half[ZX] = YES;
			}
		}
		if(y_count != 0){
			if((inclusion -> edge_is[Y] = (edge_is_rec *)malloc(sizeof(edge_is_rec))) == NULL)
				error_message(MEMORY_ERROR, "Memory allocation error");
			inclusion -> edge_is[Y] -> count = y_count;
			if((inclusion -> edge_is[Y] -> point = (is_point_rec *)calloc(y_count, sizeof(is_point_rec))) == NULL)
				error_message(MEMORY_ERROR, "Memory allocation error");
			for(i = 0; i < y_count; i++){
				is_point = &(inclusion -> edge_is[Y] -> point[i]);
				is_point -> data.coord = y_coord[i];
			}
				
/* resolve position of intersection points with respect to tile corner;
	 resolve status of intersection points;
	 resolve ellipse edge halving */

			if(y_count == 1){
				inclusion -> edge_is[Y] -> point[0].data.status = TOUCH;
				
				if(inclusion -> corner == BND)
					inclusion -> edge_is[Y] -> side = INDEF;
				else
					inclusion -> edge_is[Y] -> side = (y_coord[0] > tile.corner[ind_y].y) ? POS : NEG;
			}
			else{
				inclusion -> edge_is[Y] -> point[0].data.status = START;
				inclusion -> edge_is[Y] -> point[1].data.status = END;
				
				if(y_coord[0] > tile.corner[ind_y].y - min_epsilon * 10.0)
					inclusion -> edge_is[Y] -> side = POS;
				else if(y_coord[1] < tile.corner[ind_y].y + min_epsilon * 10.0)
					inclusion -> edge_is[Y] -> side = NEG;
				else{
#ifdef DEBUG
					if(inclusion -> corner != IN || z_count != 2 || x_count != 2)error_message(GENERAL_ERROR, "Unexpected situation");
#endif
					inclusion -> edge_is[Y] -> side = BOTH;

/* IMPORTANT: update coordinate of edge_is point outside tile to fall inside tile by appropriate periodic repetition;
	 this is necessary for coordinate-based ordering */

					inclusion -> edge_is[Y] -> point[ind_y].data.coord += tile.size.y * sign[ind_y];
				}

				pnt_1.y = y_coord[0];
				pnt_2.y = y_coord[1];
				pnt_1.x = pnt_2.x = tile.corner[ind_x].x;
				pnt_1.z = pnt_2.z = tile.corner[ind_z].z;

				get_ellipse_normal_point(&z_ellipse, &pnt_1, &normal1);
				get_ellipse_normal_point(&z_ellipse, &pnt_2, &normal2);

				if(-dot_product(normal1, normal2) > ZERO_ANGLE)inclusion -> face_is[Z] -> edge_half[ZY] = YES;

				get_ellipse_normal_point(&x_ellipse, &pnt_1, &normal1);
				get_ellipse_normal_point(&x_ellipse, &pnt_2, &normal2);

				if(-dot_product(normal1, normal2) > ZERO_ANGLE)inclusion -> face_is[X] -> edge_half[XY] = YES;
			}
		}
		if(z_count != 0){
			if((inclusion -> edge_is[Z] = (edge_is_rec *)malloc(sizeof(edge_is_rec))) == NULL)
				error_message(MEMORY_ERROR, "Memory allocation error");
			inclusion -> edge_is[Z] -> count = z_count;
			if((inclusion -> edge_is[Z] -> point = (is_point_rec *)calloc(z_count, sizeof(is_point_rec))) == NULL)
				error_message(MEMORY_ERROR, "Memory allocation error");
			for(i = 0; i < z_count; i++){
				is_point = &(inclusion -> edge_is[Z] -> point[i]);
				is_point -> data.coord = z_coord[i];
			}

/* resolve position of intersection points with respect to tile corner;
	 resolve status of intersection points;
	 resolve ellipse edge halving */
			
			if(z_count == 1){
				inclusion -> edge_is[Z] -> point[0].data.status = TOUCH;
				
				if(inclusion -> corner == BND)
					inclusion -> edge_is[Z] -> side = INDEF;
				else
					inclusion -> edge_is[Z] -> side = (z_coord[0] > tile.corner[ind_z].z) ? POS : NEG;
			}
			else{
				inclusion -> edge_is[Z] -> point[0].data.status = START;
				inclusion -> edge_is[Z] -> point[1].data.status = END;
				
				if(z_coord[0] > tile.corner[ind_z].z - min_epsilon * 10.0)
					inclusion -> edge_is[Z] -> side = POS;
				else if(z_coord[1] < tile.corner[ind_z].z + min_epsilon * 10.0)
					inclusion -> edge_is[Z] -> side = NEG;
				else{
#ifdef DEBUG
					if(inclusion -> corner != IN || x_count != 2 || y_count != 2)error_message(GENERAL_ERROR, "Unexpected situation");
#endif
					inclusion -> edge_is[Z] -> side = BOTH;

/* IMPORTANT: update coordinate of edge_is point outside tile to fall inside tile by appropriate periodic repetition;
	 this is necessary for coordinate-based ordering */

					inclusion -> edge_is[Z] -> point[ind_z].data.coord += tile.size.z * sign[ind_z];
				}

				pnt_1.z = z_coord[0];
				pnt_2.z = z_coord[1];
				pnt_1.x = pnt_2.x = tile.corner[ind_x].x;
				pnt_1.y = pnt_2.y = tile.corner[ind_y].y;

				get_ellipse_normal_point(&x_ellipse, &pnt_1, &normal1);
				get_ellipse_normal_point(&x_ellipse, &pnt_2, &normal2);

				if(-dot_product(normal1, normal2) > ZERO_ANGLE)inclusion -> face_is[X] -> edge_half[XZ] = YES;

				get_ellipse_normal_point(&y_ellipse, &pnt_1, &normal1);
				get_ellipse_normal_point(&y_ellipse, &pnt_2, &normal2);

				if(-dot_product(normal1, normal2) > ZERO_ANGLE)inclusion -> face_is[Y] -> edge_half[YZ] = YES;
			}
		}

/* resolve cross halving;
	 IMPORTANT: I do rely on fact that status (POS, NEG, BOTH) of intersection points is already resolved
	 and that inclusion -> corner (OUT, IN, BND) is already resolved */

/* note: intersection points checked against cross halving are related to relative position of center of intersection ellipse
	 (NOT center of ellipsoid) with respect to tile corner */

		if(y_count != 0 && z_count != 0){
			halving = NO;

			pnt_1.x = pnt_2.x = tile.corner[ind_x].x;
			pnt_1.z = tile.corner[ind_z].z;
			pnt_2.y = tile.corner[ind_y].y;

			if(y_count == 2 && z_count == 2){
				if(inclusion -> corner == BND){
					pnt_1.y = y_coord[2 - inclusion -> edge_is[Y] -> side];
					pnt_2.z = z_coord[2 - inclusion -> edge_is[Z] -> side];
				}
				else{
					if(x_ellipse.center.y > tile.corner[ind_y].y)
						pnt_1.y = y_coord[1];
					else
						pnt_1.y = y_coord[0];
					if(x_ellipse.center.z > tile.corner[ind_z].z)
						pnt_2.z = z_coord[1];
					else
						pnt_2.z = z_coord[0];
				}
					
				get_ellipse_normal_point(&x_ellipse, &pnt_1, &normal1);
				get_ellipse_normal_point(&x_ellipse, &pnt_2, &normal2);

				if(-dot_product(normal1, normal2) > ZERO_ANGLE)halving = YES;
			}
			else{
				if(y_count == 2 && z_count == 1){
					pnt_1.y = y_coord[2 - inclusion -> edge_is[Y] -> side];
					pnt_2.z = z_coord[0];
				
					get_ellipse_normal_point(&x_ellipse, &pnt_1, &normal1);
					get_ellipse_normal_point(&x_ellipse, &pnt_2, &normal2);

					if(-dot_product(normal1, normal2) > ZERO_ANGLE)halving = YES;
				}
				else if(y_count == 1 && z_count == 2){
					pnt_1.y = y_coord[0];
					pnt_2.z = z_coord[2 - inclusion -> edge_is[Z] -> side];
				
					get_ellipse_normal_point(&x_ellipse, &pnt_1, &normal1);
					get_ellipse_normal_point(&x_ellipse, &pnt_2, &normal2);

					if(-dot_product(normal1, normal2) > ZERO_ANGLE)halving = YES;
				}
			}

			if((inclusion -> face_is[X] -> cross_half = halving) == YES){

/* choose from directions extreme_y, extreme_z, min, max which is most perpendicular to connection pnt_1 and pnt_2 */

				val = 1.0;
				sub_vec(vec, pnt_2, pnt_1);
				sz = sqrt(size_vec(vec));
				div_vec(vec, sz);

				p1 = &(inclusion -> face_is[X] -> extreme_pnt[2 * XY + 0]);
				p2 = &(inclusion -> face_is[X] -> extreme_pnt[2 * XY + 1]);
				sub_vec(vec_tmp, *p2, *p1);
				sz = sqrt(size_vec(vec_tmp));
				div_vec(vec_tmp, sz);

				product = dot_product(vec, vec_tmp);
				product = fabs(product);
				if(product < val){
					ind = Y;       /* halving with respect to y-extreme point */
					val = product;
				}

				p1 = &(inclusion -> face_is[X] -> extreme_pnt[2 * XZ + 0]);
				p2 = &(inclusion -> face_is[X] -> extreme_pnt[2 * XZ + 1]);
				sub_vec(vec_tmp, *p2, *p1);
				sz = sqrt(size_vec(vec_tmp));
				div_vec(vec_tmp, sz);

				product = dot_product(vec, vec_tmp);
				product = fabs(product);
				if(product < val){
					ind = Z;       /* halving with respect to z-extreme point */
					val = product;
				}

#ifdef MIN_MAX_HALVING
				product = dot_product(vec, x_ellipse -> minor);
				product = fabs(product);
				if(product < val){
					ind = MIN;       /* halving with respect to min diameter */
					val = product;
				}

				product = dot_product(vec, x_ellipse -> major);
				product = fabs(product);
				if(product < val){
					ind = MAX;       /* halving with respect to max diameter */
					val = product;
				}
#endif
				inclusion -> face_is[X] -> cross_halving = ind;
			}
		}

		if(z_count != 0 && x_count != 0){
			halving = NO;

			pnt_1.y = pnt_2.y = tile.corner[ind_y].y;
			pnt_1.x = tile.corner[ind_x].x;
			pnt_2.z = tile.corner[ind_z].z;

			if(z_count == 2 && x_count == 2){
				if(inclusion -> corner == BND){
					pnt_1.z = z_coord[2 - inclusion -> edge_is[Z] -> side];
					pnt_2.x = x_coord[2 - inclusion -> edge_is[X] -> side];
				}
				else{
					if(y_ellipse.center.z > tile.corner[ind_z].z)
						pnt_1.z = z_coord[1];
					else
						pnt_1.z = z_coord[0];
					if(y_ellipse.center.x > tile.corner[ind_x].x)
						pnt_2.x = x_coord[1];
					else
						pnt_2.x = x_coord[0];
				}
					
				get_ellipse_normal_point(&y_ellipse, &pnt_1, &normal1);
				get_ellipse_normal_point(&y_ellipse, &pnt_2, &normal2);

				if(-dot_product(normal1, normal2) > ZERO_ANGLE)halving = YES;
			}
			else{
				if(z_count == 2 && x_count == 1){
					pnt_1.z = z_coord[2 - inclusion -> edge_is[Z] -> side];
					pnt_2.x = x_coord[0];
				
					get_ellipse_normal_point(&y_ellipse, &pnt_1, &normal1);
					get_ellipse_normal_point(&y_ellipse, &pnt_2, &normal2);

					if(-dot_product(normal1, normal2) > ZERO_ANGLE)halving = YES;
				}
				else if(z_count == 1 && x_count == 2){
					pnt_1.z = z_coord[0];
					pnt_2.x = x_coord[2 - inclusion -> edge_is[X] -> side];
				
					get_ellipse_normal_point(&y_ellipse, &pnt_1, &normal1);
					get_ellipse_normal_point(&y_ellipse, &pnt_2, &normal2);

					if(-dot_product(normal1, normal2) > ZERO_ANGLE)halving = YES;
				}
			}

			if((inclusion -> face_is[Y] -> cross_half = halving) == YES){

/* choose from directions extreme_z, extreme_x, min, max which is most perpendicular to connection pnt_1 and pnt_2 */

				val = 1.0;
				sub_vec(vec, pnt_2, pnt_1);
				sz = sqrt(size_vec(vec));
				div_vec(vec, sz);

				p1 = &(inclusion -> face_is[Y] -> extreme_pnt[2 * YZ + 0]);
				p2 = &(inclusion -> face_is[Y] -> extreme_pnt[2 * YZ + 1]);
				sub_vec(vec_tmp, *p2, *p1);
				sz = sqrt(size_vec(vec_tmp));
				div_vec(vec_tmp, sz);

				product = dot_product(vec, vec_tmp);
				product = fabs(product);
				if(product < val){
					ind = Z;       /* halving with respect to z-extreme point */
					val = product;
				}

				p1 = &(inclusion -> face_is[Y] -> extreme_pnt[2 * YX + 0]);
				p2 = &(inclusion -> face_is[Y] -> extreme_pnt[2 * YX + 1]);
				sub_vec(vec_tmp, *p2, *p1);
				sz = sqrt(size_vec(vec_tmp));
				div_vec(vec_tmp, sz);

				product = dot_product(vec, vec_tmp);
				product = fabs(product);
				if(product < val){
					ind = X;       /* halving with respect to x-extreme point */
					val = product;
				}

#ifdef MIN_MAX_HALVING
				product = dot_product(vec, y_ellipse -> minor);
				product = fabs(product);
				if(product < val){
					ind = MIN;       /* halving with respect to min diameter */
					val = product;
				}

				product = dot_product(vec, y_ellipse -> major);
				product = fabs(product);
				if(product < val){
					ind = MAX;       /* halving with respect to max diameter */
					val = product;
				}
#endif
				inclusion -> face_is[Y] -> cross_halving = ind;
			}
		}

		if(x_count != 0 && y_count != 0){
			halving = NO;

			pnt_1.z = pnt_2.z = tile.corner[ind_z].z;
			pnt_1.y = tile.corner[ind_y].y;
			pnt_2.x = tile.corner[ind_x].x;

			if(x_count == 2 && y_count == 2){
				if(inclusion -> corner == BND){
					pnt_1.x = x_coord[2 - inclusion -> edge_is[X] -> side];
					pnt_2.y = y_coord[2 - inclusion -> edge_is[Y] -> side];
				}
				else{
					if(z_ellipse.center.x > tile.corner[ind_x].x)
						pnt_1.x = x_coord[1];
					else
						pnt_1.x = x_coord[0];
					if(z_ellipse.center.y > tile.corner[ind_y].y)
						pnt_2.y = y_coord[1];
					else
						pnt_2.y = y_coord[0];
				}
					
				get_ellipse_normal_point(&z_ellipse, &pnt_1, &normal1);
				get_ellipse_normal_point(&z_ellipse, &pnt_2, &normal2);

				if(-dot_product(normal1, normal2) > ZERO_ANGLE)halving = YES;
			}
			else{
				if(x_count == 2 && y_count == 1){
					pnt_1.x = x_coord[2 - inclusion -> edge_is[X] -> side];
					pnt_2.y = y_coord[0];
				
					get_ellipse_normal_point(&z_ellipse, &pnt_1, &normal1);
					get_ellipse_normal_point(&z_ellipse, &pnt_2, &normal2);

					if(-dot_product(normal1, normal2) > ZERO_ANGLE)halving = YES;
				}
				else if(x_count == 1 && y_count == 2){
					pnt_1.x = x_coord[0];
					pnt_2.y = y_coord[2 - inclusion -> edge_is[Y] -> side];
				
					get_ellipse_normal_point(&z_ellipse, &pnt_1, &normal1);
					get_ellipse_normal_point(&z_ellipse, &pnt_2, &normal2);

					if(-dot_product(normal1, normal2) > ZERO_ANGLE)halving = YES;
				}
			}

			if((inclusion -> face_is[Z] -> cross_half = halving) == YES){

/* choose from directions extreme_x, extreme_y, min, max which is most perpendicular to connection pnt_1 and pnt_2 */

				val = 1.0;
				sub_vec(vec, pnt_2, pnt_1);
				sz = sqrt(size_vec(vec));
				div_vec(vec, sz);

				p1 = &(inclusion -> face_is[Z] -> extreme_pnt[2 * ZX + 0]);
				p2 = &(inclusion -> face_is[Z] -> extreme_pnt[2 * ZX + 1]);
				sub_vec(vec_tmp, *p2, *p1);
				sz = sqrt(size_vec(vec_tmp));
				div_vec(vec_tmp, sz);

				product = dot_product(vec, vec_tmp);
				product = fabs(product);
				if(product < val){
					ind = X;       /* halving with respect to x-extreme point */
					val = product;
				}

				p1 = &(inclusion -> face_is[Z] -> extreme_pnt[2 * ZY + 0]);
				p2 = &(inclusion -> face_is[Z] -> extreme_pnt[2 * ZY + 1]);
				sub_vec(vec_tmp, *p2, *p1);
				sz = sqrt(size_vec(vec_tmp));
				div_vec(vec_tmp, sz);

				product = dot_product(vec, vec_tmp);
				product = fabs(product);
				if(product < val){
					ind = Y;       /* halving with respect to y-extrme point */
					val = product;
				}

#ifdef MIN_MAX_HALVING
				product = dot_product(vec, z_ellipse -> minor);
				product = fabs(product);
				if(product < val){
					ind = MIN;       /* halving with respect to min diameter */
					val = product;
				}

				product = dot_product(vec, z_ellipse -> major);
				product = fabs(product);
				if(product < val){
					ind = MAX;       /* halving with respect to max diameter */
					val = product;
				}
#endif
				inclusion -> face_is[Z] -> cross_halving = ind;
			}
		}

#ifdef DRAW_EXTREME_POINT_ELLIPSE
		if(draw == YES){
			if(x_count != 0 || y_count != 0 || z_count != 0)ESIEventLoop(YES, "Ellipse extreme points (all for ellipse crossing edge)\n\n");
		}
#endif

	}

	if(eps != epsilon)error_message(WARNING, "Epsilon temporarily reduced to %e when processing inclusion %ld", inclusion -> id);

/* note: face_is[X|Y|Z] -> side[0|1] is resolved after return */
	
	return(ret_val);
}		



/* compute distance of relevant ellipsoid extreme point in x direction from tile side perpendicular to x direction
	 and if closer to that side than epsilon while not touching it, shiftting yielding touching or minimal shifting
	 inside tile is computed;
	 resolve simultaneously touching of y- and z-edges;

	 IMPORTANT: I do rely on fact that inclusion shifting due to edge touching does not interfere
	 with shifting due to side touching (this is ensured by min > epsilon * 10);
	 this implies that I can safely use shift_vec.x|y|z = val instead of shift_vec.x|y|z += val */

static void
resolve_ellipsoid_x_side_touching(void)
{
	double a_diff_x, a_xy_diff, a_xz_diff;
	double xy_diff, xz_diff;

	diff_x = tile.corner[ind_x].x - ext_point_x[ind_x].x;
	a_diff_x = fabs(diff_x);
	if(a_diff_x >= epsilon)return;

	if(no_shift == YES)error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);

	x_intersection = NO;

#ifdef OVERDETOUCH

/* note: also simple detouch is handled here */

	if(ind_x == 0)
		shift_vec.x = diff_x + epsilon;
	else
		shift_vec.x = diff_x - epsilon;

	if(fabs(shift_vec.x) <= MULTIPLE_EPS * epsilon)return;
	shift_vec.x = 0.0;

	error_message(WARNING, "Overdetouch of inclusion %ld in x-direction suppressed due to MULTIPLE_EPS (%f)", inclusion -> id, MULTIPLE_EPS);
#endif
	
#ifdef DETOUCH
	if(a_diff_x >= min_epsilon){
		if(ind_x == 0){
			if(diff_x < 0.0){
				shift_vec.x = diff_x + epsilon;
				return;
			}
		}
		else{
			if(diff_x > 0.0){
				shift_vec.x = diff_x - epsilon;
				return;
			}
		}
	}
#endif

	x_touching = YES;
	shift_vec.x = diff_x;

	xy_diff = tile.corner[ind_y].y - ext_point_x[ind_x].y;
	a_xy_diff = fabs(xy_diff);
	if(a_xy_diff < epsilon){
		shift_vec.y = xy_diff;
		z_count = 1;
	}
	xz_diff = tile.corner[ind_z].z - ext_point_x[ind_x].z;
	a_xz_diff = fabs(xz_diff);
	if(a_xz_diff < epsilon){
		shift_vec.z = xz_diff;
		y_count = 1;
	}
}


/* compute distance of relevant ellipsoid extreme point in y direction from tile side perpendicular to y direction
	 and if closer to that side than epsilon while not touching it, shiftting yielding touching or minimal shifting
	 inside tile is computed;
	 resolve simultaneously touching of z- and x-edges;

	 IMPORTANT: I do rely on fact that inclusion shifting due to edge touching does not interfere
	 with shifting due to side touching (this is ensured by min > epsilon * 10);
	 this implies that I can safely use shift_vec.x|y|z = val instead of shift_vec.x|y|z += val */

static void
resolve_ellipsoid_y_side_touching(void)
{
	double a_diff_y, a_yz_diff, a_yx_diff;
	double yz_diff, yx_diff;

	diff_y = tile.corner[ind_y].y - ext_point_y[ind_y].y;
	a_diff_y = fabs(diff_y);
	if(a_diff_y >= epsilon)return;

	if(no_shift == YES)error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);

	y_intersection = NO;

#ifdef OVERDETOUCH

/* note: also simple detouch is handled here */

	if(ind_y == 0)
		shift_vec.y = diff_y + epsilon;
	else
		shift_vec.y = diff_y - epsilon;

	if(fabs(shift_vec.y) <= MULTIPLE_EPS * epsilon)return;
	shift_vec.y = 0.0;

	error_message(WARNING, "Overdetouch of inclusion %ld in y-direction suppressed due to MULTIPLE_EPS (%f)", inclusion -> id, MULTIPLE_EPS);
#endif
	
#ifdef DETOUCH
	if(a_diff_y >= min_epsilon){
		if(ind_y == 0){
			if(diff_y < 0.0){
				shift_vec.y = diff_y + epsilon;
				return;
			}
		}
		else{
			if(diff_y > 0.0){
				shift_vec.y = diff_y - epsilon;
				return;
			}
		}
	}
#endif

	y_touching = YES;
	shift_vec.y = diff_y;

	yz_diff = tile.corner[ind_z].z - ext_point_y[ind_y].z;
	a_yz_diff = fabs(yz_diff);
	if(a_yz_diff < epsilon){
		shift_vec.z = yz_diff;
		x_count = 1;
	}
	yx_diff = tile.corner[ind_x].x - ext_point_y[ind_y].x;
	a_yx_diff = fabs(yx_diff);
	if(a_yx_diff < epsilon){
		shift_vec.x = yx_diff;
		z_count = 1;
	}
}



/* compute distance of relevant ellipsoid extreme point in z direction from tile side perpendicular to z direction
	 and if closer to that side than epsilon while not touching it, shiftting yielding touching or minimal shifting
	 inside tile is computed;
	 resolve simultaneously touching of x- and y-edges;

	 IMPORTANT: I do rely on fact that inclusion shifting due to edge touching does not interfere
	 with shifting due to side touching (this is ensured by min > epsilon * 10);
	 this implies that I can safely use shift_vec.x|y|z = val instead of shift_vec.x|y|z += val */

static void
resolve_ellipsoid_z_side_touching(void)
{
	double a_diff_z, a_zx_diff, a_zy_diff;
	double zx_diff, zy_diff;

	diff_z = tile.corner[ind_z].z - ext_point_z[ind_z].z;
	a_diff_z = fabs(diff_z);
	if(a_diff_z >= epsilon)return;

	if(no_shift == YES)error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);

	z_intersection = NO;

#ifdef OVERDETOUCH

/* note: also simple detouch is handled here */

	if(ind_z == 0)
		shift_vec.z = diff_z + epsilon;
	else
		shift_vec.z = diff_z - epsilon;

	if(fabs(shift_vec.z) <= MULTIPLE_EPS * epsilon)return;
	shift_vec.z = 0.0;

	error_message(WARNING, "Overdetouch of inclusion %ld in z-direction suppressed due to MULTIPLE_EPS (%f)", inclusion -> id, MULTIPLE_EPS);
#endif
	
#ifdef DETOUCH
	if(a_diff_z >= min_epsilon){
		if(ind_z == 0){
			if(diff_z < 0.0){
				shift_vec.z = diff_z + epsilon;
				return;
			}
		}
		else{
			if(diff_z > 0.0){
				shift_vec.z = diff_z - epsilon;
				return;
			}
		}
	}
#endif

	z_touching = YES;
	shift_vec.z = diff_z;

	zx_diff = tile.corner[ind_x].x - ext_point_z[ind_z].x;
	a_zx_diff = fabs(zx_diff);
	if(a_zx_diff < epsilon){
		shift_vec.x = zx_diff;
		y_count = 1;
	}
	zy_diff = tile.corner[ind_y].y - ext_point_z[ind_z].y;
	a_zy_diff = fabs(zy_diff);
	if(a_zy_diff < epsilon){
		shift_vec.y = zy_diff;
		x_count = 1;
	}
}



static void
setup_ellipse_touching_point(ellipse_rec *ellipse, point_rec *point, vector_rec *dir)
{
	copy_vec(ellipse -> center, *point);

	ellipse -> max = ellipse -> min = 0.0;

	ellipse -> major.x = ellipse -> major.y = ellipse -> major.z = 0.0;
	ellipse -> minor.x = ellipse -> minor.y = ellipse -> minor.z = 0.0;
	//ellipse -> normal.x = ellipse -> normal.y = ellipse -> normal.z = 0.0;
	copy_vec(ellipse -> normal, *dir);

#ifdef PARAMETRIC_ROUTINES
	copy_vec(ellipse -> pnt0, *point);
	copy_vec(ellipse -> pnt1, *point);
	copy_vec(ellipse -> pnt2, *point);
#endif

#ifdef DRAW_TOUCHING_POINT_ELLIPSOID
	draw_point(&(ellipse -> center), 0, orange);
#endif
}		



static void
get_ellipsoid_x_side_intersection(void)
{
	get_ellipsoid_xyz_side_intersection(&x_ellipse, &corner_point[ind_x], &vec_x, &dir_2, &dir_c);
	copy_vec(x_ellipse.normal, dir_x);

	if(x_trans == NO){
		x_trans = YES;

		copy_vec(x_trans_g2l.x, x_ellipse.major);
		copy_vec(x_trans_g2l.y, x_ellipse.minor);
		copy_vec(x_trans_g2l.z, x_ellipse.normal);
				
		get_inverted_rotation_transformation(&x_trans_g2l, &x_trans_l2g);
	}

#ifdef DRAW_ELLIPSE
	draw_ellipse(&x_ellipse, 0, green);
#endif
}


static void
get_ellipsoid_y_side_intersection(void)
{
	get_ellipsoid_xyz_side_intersection(&y_ellipse, &corner_point[ind_y], &vec_y, &dir_3, &dir_a);
	copy_vec(y_ellipse.normal, dir_y);

	if(y_trans == NO){
		y_trans = YES;

		copy_vec(y_trans_g2l.x, y_ellipse.major);
		copy_vec(y_trans_g2l.y, y_ellipse.minor);
		copy_vec(y_trans_g2l.z, y_ellipse.normal);
				
		get_inverted_rotation_transformation(&y_trans_g2l, &y_trans_l2g);
	}

#ifdef DRAW_ELLIPSE
	draw_ellipse(&y_ellipse, 0, cyan);
#endif
}


static void
get_ellipsoid_z_side_intersection(void)
{
	get_ellipsoid_xyz_side_intersection(&z_ellipse, &corner_point[ind_z], &vec_z, &dir_1, &dir_b);
	copy_vec(z_ellipse.normal, dir_z);

	if(z_trans == NO){
		z_trans = YES;

		copy_vec(z_trans_g2l.x, z_ellipse.major);
		copy_vec(z_trans_g2l.y, z_ellipse.minor);
		copy_vec(z_trans_g2l.z, z_ellipse.normal);
				
		get_inverted_rotation_transformation(&z_trans_g2l, &z_trans_l2g);
	}

#ifdef DRAW_ELLIPSE
	draw_ellipse(&z_ellipse, 0, magenta);
#endif
}



/* note: tan_1, tan_2 and nor form orthonormal base of transformed Cartesian plane */

/* CAUTION: ellipse normal is not computed here !!!
	 note that nor is normal in space of sphere and cannot be used as ellipse normal;
	 note also that it is redundant to compute ellipse normal as major x minor, because the normal is known in advance;
	 it is therefore only neccessary to store the normal (probably in the caller) */

static void
get_ellipsoid_xyz_side_intersection(ellipse_rec *ellipse, point_rec *point, vector_rec *nor, vector_rec *tan_1, vector_rec *tan_2)
{
	point_rec center;
	vector_rec vector_1, vector_2;
	double dist, rad, nom, den, alpha, ca, sa, sz_1, sz_2, tmp;

/* resolve intersection with plane in space of sphere ==> circle */

	dist = dot_product(*nor, *point);
	copy_mul_vec(center, *nor, dist);              /* circle center */
	rad = sqrt(radius * radius - dist * dist);     /* circle radius */

/* search for two perpendicular diameters of circle which remain mutually perpendicular after s2e transformation */

	comp_product(vector_1, *tan_1, scale_s2e);
	comp_product(vector_2, *tan_2, scale_s2e);

	nom = 2.0 * dot_product(vector_1, vector_2);
	//den = dot_product(vector_1, vector_1) - dot_product(vector_2, vector_2);
	den = size_vec(vector_1) - size_vec(vector_2);
	if(den != 0.0){
				
/* if nom/den positively (negatively) overflows atan should return PI/2 (-PI/2) */

		alpha = atan(nom/den) / 2.0;
		ca = cos(alpha) * rad;
		sa = sin(alpha) * rad;
	}
	else{
		if(nom != 0.0){
			//alpha = PI_4;
			ca = sa = SQRT1_2 * rad;
		}
		else{
					
/* nom = den = 0.0 (nom/den = NaN) ==> ellipse is circle ==> alpha is arbitrary */

			//alpha = 0.0;
			ca = rad;
			sa = 0.0;
		}
	}
			
	add_mul_vec(vector_1, *tan_1, *tan_2, ca, sa);
	add_mul_vec(vector_2, *tan_1, *tan_2, -sa, ca); 
			
/* transform circle from space of sphere to space of ellipsoid	*/

	transform_local_from_sphere_to_ellipsoid(&center, &center, &scale_s2e);
	transform_local_from_sphere_to_ellipsoid(&vector_1, &vector_1, &scale_s2e);
	transform_local_from_sphere_to_ellipsoid(&vector_2, &vector_2, &scale_s2e);
		
	sz_1 = size_vec(vector_1);
	sz_2 = size_vec(vector_2);
		
	if(sz_1 < sz_2){
		swap(vector_1.x, vector_2.x, tmp);
		swap(vector_1.y, vector_2.y, tmp);
		swap(vector_1.z, vector_2.z, tmp);
		
		mul_vec(vector_2, -1.0);            /* ensure that vec_x, vector_1 and vector_2 is righthanded CS */
	}

/* transform from local to global */

	transform_point_from_local_to_global(&center, &(ellipse -> center), &(ellipsoid -> center), &trans_l2g);
	transform_vector_from_local_to_global(&vector_1, &(ellipse -> major), &trans_l2g);
	transform_vector_from_local_to_global(&vector_2, &(ellipse -> minor), &trans_l2g);
	
#ifdef PARAMETRIC_ROUTINES
	add_vec(ellipse -> pnt0, ellipse -> center, ellipse -> major);
	add_vec(ellipse -> pnt1, ellipse -> pnt0, ellipse -> minor);
	add_vec(ellipse -> pnt2, ellipse -> center, ellipse -> minor);
#endif			

	ellipse -> max = sqrt(size_vec(ellipse -> major));
	ellipse -> min = sqrt(size_vec(ellipse -> minor));

	if(ellipse -> max < ellipse -> min){
		if(ellipse -> min - ellipse -> max > EPSILON)error_message(GENERAL_ERROR, "Unexpected situation");
		ellipse -> max = ellipse -> min = (ellipse -> max + ellipse -> min) / 2.0;
	}

/* CAUTION: I do not handle the case when intersection ellipse is too small although the ellipsoid excceds tile side by more than epsilon */

	if(ellipse -> min <= min_epsilon)error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);

	div_vec(ellipse -> major, ellipse -> max);
	div_vec(ellipse -> minor, ellipse -> min);
}


							
/* compute relevant extreme points and corresponding diffs of x_ellipse;
	 sign of diffs determines the orientation of shift to enforce touching */

/* if min xor max is set to YES, then min xor max extreme points in both directions are computed no matter whether intersection or touching occurs */

static void
resolve_ellipsoid_yz_edge_touching(logic min, logic max)
{
	vector_rec vec, vec_tmp, vector, vector_y, vector_z, vector_b, vector_c;
	point_rec pnt;
	double val;

#ifdef DEBUG
	if(min == YES && max == YES)error_message(GENERAL_ERROR, "Unexpected situation");
#endif

	vec.x = x_ellipse.max;
	vec.y = x_ellipse.min;
	vec.z = 0.0;
				
	comp_product(vector, vec, vec);

	if(y_intersection == YES || y_touching == YES || min == YES || max == YES){
		transform_vector_from_global_to_local(&dir_y, &vector_b, &x_trans_g2l);

		comp_product(vector_y, vector_b, vec);
		val = 1.0 / sqrt(size_vec(vector_y));
		if(min == NO && max == NO){
			if(x_ellipse.center.y > tile.corner[ind_y].y)val = -val;
		}
		else{
			if(min == YES)val = -val;
		}
		copy_mul_vec(vec_tmp, vector, val);
		comp_product(pnt, vector_b, vec_tmp);
		
		transform_point_from_local_to_global(&pnt, &extreme_pnt_xy, &(x_ellipse.center), &x_trans_l2g);
		diff_xy = tile.corner[ind_y].y - extreme_pnt_xy.y;
		
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		draw_point(&extreme_pnt_xy, 0, green);
#endif
	}
				
	if(z_intersection == YES || z_touching == YES || min == YES || max == YES){
		transform_vector_from_global_to_local(&dir_z, &vector_c, &x_trans_g2l);
		
		comp_product(vector_z, vector_c, vec);
		val = 1.0 / sqrt(size_vec(vector_z));
		if(min == NO && max == NO){
			if(x_ellipse.center.z > tile.corner[ind_z].z)val = -val;
		}
		else{
			if(min == YES)val = -val;
		}
		copy_mul_vec(vec_tmp, vector, val);
		comp_product(pnt, vector_c, vec_tmp);
		
		transform_point_from_local_to_global(&pnt, &extreme_pnt_xz, &(x_ellipse.center), &x_trans_l2g);
		diff_xz = tile.corner[ind_z].z - extreme_pnt_xz.z;
		
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		draw_point(&extreme_pnt_xz, 0, green);
#endif
	}
}


/* compute relevant extreme points and corresponding diffs of y_ellipse;
	 sign of diffs determines the orientation of shift to enforce touching */

/* if min xor max is set to YES, then min xor max extreme points in both directions are computed no matter whether intersection or touching occurs */

static void
resolve_ellipsoid_zx_edge_touching(logic min, logic max)
{
	vector_rec vec, vec_tmp, vector, vector_z, vector_x, vector_c, vector_a;
	point_rec pnt;
	double val;

#ifdef DEBUG
	if(min == YES && max == YES)error_message(GENERAL_ERROR, "Unexpected situation");
#endif

	vec.x = y_ellipse.max;
	vec.y = y_ellipse.min;
	vec.z = 0.0;
	
	comp_product(vector, vec, vec);

	if(z_intersection == YES || z_touching == YES || min == YES || max == YES){
		transform_vector_from_global_to_local(&dir_z, &vector_c, &y_trans_g2l);
		
		comp_product(vector_z, vector_c, vec);
		val = 1.0 / sqrt(size_vec(vector_z));
		if(min == NO && max == NO){
			if(y_ellipse.center.z > tile.corner[ind_z].z)val = -val;
		}
		else{
			if(min == YES)val = -val;
		}
		copy_mul_vec(vec_tmp, vector, val);
		comp_product(pnt, vector_c, vec_tmp);
		
		transform_point_from_local_to_global(&pnt, &extreme_pnt_yz, &(y_ellipse.center), &y_trans_l2g);
		diff_yz = tile.corner[ind_z].z - extreme_pnt_yz.z;
		
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		draw_point(&extreme_pnt_yz, 0, cyan);
#endif
	}
	
	if(x_intersection == YES || x_touching == YES || min == YES || max == YES){
		transform_vector_from_global_to_local(&dir_x, &vector_a, &y_trans_g2l);
		
		comp_product(vector_x, vector_a, vec);
		val = 1.0 / sqrt(size_vec(vector_x));
		if(min == NO && max == NO){
			if(y_ellipse.center.x > tile.corner[ind_x].x)val = -val;
		}
		else{
			if(min == YES)val = -val;
		}
		copy_mul_vec(vec_tmp, vector, val);
		comp_product(pnt, vector_a, vec_tmp);
		
		transform_point_from_local_to_global(&pnt, &extreme_pnt_yx, &(y_ellipse.center), &y_trans_l2g);
		diff_yx = tile.corner[ind_x].x - extreme_pnt_yx.x;
		
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		draw_point(&extreme_pnt_yx, 0, cyan);
#endif
	}
}



/* compute relevant extreme points and corresponding diffs of z_ellipse;
	 sign of diffs determines the orientation of shift to enforce touching */

/* if min xor max is set to YES, then min xor max extreme points in both directions are computed no matter whether intersection or touching occurs */

static void
resolve_ellipsoid_xy_edge_touching(logic min, logic max)
{
	vector_rec vec, vec_tmp, vector, vector_x, vector_y, vector_a, vector_b;
	point_rec pnt;
	double val;

#ifdef DEBUG
	if(min == YES && max == YES)error_message(GENERAL_ERROR, "Unexpected situation");
#endif

	vec.x = z_ellipse.max;
	vec.y = z_ellipse.min;
	vec.z = 0.0;
	
	comp_product(vector, vec, vec);

	if(x_intersection == YES || x_touching == YES || min == YES || max == YES){
		transform_vector_from_global_to_local(&dir_x, &vector_a, &z_trans_g2l);
		
		comp_product(vector_x, vector_a, vec);
		val = 1.0 / sqrt(size_vec(vector_x));
		if(min == NO && max == NO){
			if(z_ellipse.center.x > tile.corner[ind_x].x)val = -val;
		}
		else{
			if(min == YES)val = -val;
		}
		copy_mul_vec(vec_tmp, vector, val);
		comp_product(pnt, vector_a, vec_tmp);
		
		transform_point_from_local_to_global(&pnt, &extreme_pnt_zx, &(z_ellipse.center), &z_trans_l2g);
		diff_zx = tile.corner[ind_x].x - extreme_pnt_zx.x;
		
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		draw_point(&extreme_pnt_zx, 0, magenta);
#endif
	}
				
	if(y_intersection == YES || y_touching == YES || min == YES || max == YES){
		transform_vector_from_global_to_local(&dir_y, &vector_b, &z_trans_g2l);
		
		comp_product(vector_y, vector_b, vec);
		val = 1.0 / sqrt(size_vec(vector_y));
		if(min == NO && max == NO){
			if(z_ellipse.center.y > tile.corner[ind_y].y)val = -val;
		}
		else{
			if(min == YES)val = -val;
		}
		copy_mul_vec(vec_tmp, vector, val);
		comp_product(pnt, vector_b, vec_tmp);
		
		transform_point_from_local_to_global(&pnt, &extreme_pnt_zy, &(z_ellipse.center), &z_trans_l2g);
		diff_zy = tile.corner[ind_y].y - extreme_pnt_zy.y;
		
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		draw_point(&extreme_pnt_zy, 0, magenta);
#endif
	}
}



/* calculate minimal positive and negative shifts of ellipsoid in particular direction
	 so that neither ellipsoid nor intersection ellipses with tile sides parallel with the direction 
	 are closer than eps to tile side perpendicular to the direction while not touching it */

static logic
get_ellipsoid_shift(double diff_0, double diff_1, double diff_2, double *shift_pos, double *shift_neg, double eps)
{
	double a_diff_0, a_diff_1, a_diff_2, df_1, df_2, tmp;
	logic swap_diff = NO, sign_change = NO;

#ifdef DEBUG
	double diff;
#endif

/* diff_0 - diff in given direction of ellipsoid
	 diff_1 - diff in given direction of intersection ellipse with 1st tile side parallel with direction
	 diff_2 - diff in given direction of intersection ellipse with 2nd tile side parallel with direction


	 coord - <-------x-------> coord +

                o                               point with positive diff
							        o                         point with negative diff
           <-------                             negative shift increasing diff
                    ------->                    positive shift decreasing diff

   x - position of tile side
   o - extreme point (of intersection ellipse or ellipsoid)

	 sign of diff determines the orientation of shift to enforce touching;

	 IMPORTANT: note the difference:
	 diff  -= pos_shift; diff  += neg_shift;
	 shift += pos_shift; shift -= neg_shift; */

	a_diff_0 = fabs(diff_0);
	a_diff_1 = fabs(diff_1);
	a_diff_2 = fabs(diff_2);

#ifdef DEBUG
	if(a_diff_0 < eps && a_diff_0 >= min_epsilon)error_message(GENERAL_ERROR, "Unexpected situation");
#endif

#if 0 /* obsolete - this is now handled by cnt_x|y|z in get_inclusion_tile_intersection */
	if(a_diff_1 >= eps && a_diff_2 >= eps)return(NO);
	if(a_diff_1 < min_epsilon && a_diff_2 < min_epsilon)return(NO);
#endif

/* reorder diff_1 and diff_2 so that a_diff_1 <= a_diff_2 */

	if(a_diff_2 < a_diff_1){
		swap_diff = YES;
		
		swap(a_diff_1, a_diff_2, tmp);
		swap(diff_1, diff_2, tmp);
	}

/* note that it should always hold
	 diff_0 <= min(diff_1, diff_2) or diff_0 >= max(diff_1, diff_2) 

	 but diff_1 or diff_2 (only diff_2 after reodering) can be equal to max_sz if relevant intersection does not occur;
	 this must be accounted for */

	if(diff_2 == max_sz){
		if(diff_1 < diff_0)diff_2 = -diff_2;
	}

	df_1 = min_val(diff_1, diff_2);
	df_2 = max_val(diff_1, diff_2);

	if(diff_0 > df_1 && diff_0 < df_2){
		if(fabs(diff_1 - diff_2) >= min_epsilon)error_message(GENERAL_ERROR, "Unexpected situation");

/* if the condition is not satisfied within min_epsilon
	 setup shifts so that both positive and negative shift yields the same shifting;
	 magnitude and orientation is chosen so that there is no proximity criteria violation after shifting */

		if(a_diff_0 < min_epsilon){
			*shift_pos = diff_0;
			*shift_neg = -diff_0;
		}
		else{
			if(diff_0 > 0.0){
				*shift_pos = -min_epsilon;
				*shift_neg = min_epsilon;
			}
			else{
				*shift_pos = min_epsilon;
				*shift_neg = -min_epsilon;
			}
		}
		return(YES);
	}
	
	if(a_diff_1 < min_epsilon && a_diff_2 >= eps)return(NO);
			
/* ensure that diff_2 >= 0.0 */

	if(diff_2 < 0.0){
		sign_change = YES;
		
		diff_0 = -diff_0;
		diff_1 = -diff_1;
		diff_2 = -diff_2;
	}

#ifdef DEBUG
	*shift_pos = *shift_neg = -1.0;
#endif

	if(a_diff_1 < eps && a_diff_2 < eps){
		if(a_diff_0 < min_epsilon){    /* ==> diff_1 * diff_2 >= 0.0 <==> sign(diff_1) = sign(diff_2) */

/*
   -------------  --            /|\   diff < 0
                   |             |    
                   | eps         |    positive shift
                   |             |
   ------0------  --            ---   diff = 0
                   |             |
      1            | eps         |    negative shift
            2      |             |
   -------------  --            \|/   diff > 0
*/

			*shift_neg = eps - diff_0;

/* caution: positive shift by eps must take into account that a_diff_0 may be != 0.0 */

			df_1 = diff_1 - (eps + diff_0);
			df_2 = diff_2 - (eps + diff_0);
			if(fabs(df_1) < min_epsilon && fabs(df_2) < min_epsilon)/* diff_1 ~ eps && diff_2 ~ eps */
				*shift_pos = eps + diff_0;
			else
				*shift_pos = eps + diff_2;
		}
		else{
			if(diff_0 > 0.0){   /* <==> diff_0 >= eps */
				if(diff_1 < 0.0){

/*
   -------------  --            /|\   diff < 0
                   |             |    
      1            | eps         |    positive shift
                   |             |
   -------------  --            ---   diff = 0
                   |             |
                   | eps         |    negative shift
            2      |             |
   ------0------  --            \|/   diff > 0
         0
         0
*/

					if(-diff_1 > eps - diff_2)
						*shift_neg = -diff_1;
					else
						*shift_neg = eps - diff_1;
					if(eps + diff_1 <= diff_2){
						if(diff_0 - eps >= diff_2)
							*shift_pos = diff_2;
						else if(diff_0 >= diff_2 + eps)
							*shift_pos = diff_0;
						else
							*shift_pos = diff_0 + eps;
					}
					else{
						if(diff_0 - eps >= diff_2 + eps)
							*shift_pos = diff_2 + eps;
						else if(diff_0 >= diff_2 + eps)
							*shift_pos = diff_0;
						else
							*shift_pos = diff_0 + eps;
					}
				}
				else{
					*shift_neg = eps - diff_1;
					if(diff_1 > diff_2 - min_epsilon){    /* diff_1 ~ diff_2 */

/*
   -------------  --            /|\   diff < 0
                   |             |    
                   | eps         |    positive shift
                   |             |
   -------------  --            ---   diff = 0
                   |             |
      1     2      | eps         |    negative shift
                   |             |
   ------0------  --            \|/   diff > 0
         0
         0
*/
						
						if(diff_0 - eps >= (diff_1 + diff_2) / 2.0)
							*shift_pos = (diff_1 + diff_2) / 2.0;
						else if(diff_0 - eps >= eps + diff_2)    /* this is the same as if diff_1 !~ diff_2 */
							*shift_pos = eps + diff_2;
						else if(diff_0 >= eps + diff_2)
							*shift_pos = diff_0;
						else
							*shift_pos = diff_0 + eps;
					}
					else{    /* diff_1 !~ diff_2 */

/*
   -------------  --            /|\   diff < 0
                   |             |    
                   | eps         |    positive shift
                   |             |
   -------------  --            ---   diff = 0
                   |             |
      1            | eps         |    negative shift
            2      |             |
   ------0------  --            \|/   diff > 0
         0
         0
*/

						if(diff_0 - eps >= eps + diff_2)
							*shift_pos = eps + diff_2;
						else if(diff_0 >= eps + diff_2)
							*shift_pos = diff_0;
						else
							*shift_pos = diff_0 + eps;
					}
				}
			}
			else{   /* diff_0 < 0 <==> -diff_0 >= eps */
				if(diff_1 < 0.0){

/*
         0
         0
   ------0------  --            /|\   diff < 0
                   |             |    
      1            | eps         |    positive shift
                   |             |
   -------------  --            ---   diff = 0
                   |             |
                   | eps         |    negative shift
            2      |             |
   -------------  --            \|/   diff > 0
*/

					if(diff_2 >= eps + diff_1)
						*shift_pos = diff_2;
					else
						*shift_pos = diff_2 + eps;
					if(-diff_1 >= eps - diff_2){
						if(-diff_0 - eps >= -diff_1)
							*shift_neg = -diff_1;
						else if(-diff_0 - eps >= eps - diff_1)
							*shift_neg = eps - diff_1;
						else if(-diff_0 >= eps - diff_1)
							*shift_neg = -diff_0;
						else
							*shift_neg = -diff_0 + eps;
					}
					else{
						if(-diff_0 - eps >= eps - diff_1)
							*shift_neg = eps - diff_1;
						else if(-diff_0 >= eps - diff_1)
							*shift_neg = -diff_0;
						else
							*shift_neg = -diff_0 + eps;
					}
				}
				else{

/*
         0
         0
   ------0------  --            /|\   diff < 0
                   |             |    
                   | eps         |    positive shift
                   |             |
   -------------  --            ---   diff = 0
                   |             |
      1            | eps         |    negative shift
            2      |             |
   -------------  --            \|/   diff > 0
*/

					if(-diff_0 - eps >= eps - diff_1)
						*shift_neg = eps - diff_1;
					else
						*shift_neg = -diff_0;
					if(diff_1 > diff_2 - min_epsilon)    /* diff_1 ~ diff_2 */
						*shift_pos = (diff_1 + diff_2) / 2.0;
					else
						*shift_pos = eps + diff_2;
				}
			}
		}
	}
	else{
		if(a_diff_0 < min_epsilon){    /* ==> diff_1 * diff_2 >= 0.0 <==> sign(diff_1) = sign(diff_2) */
			*shift_neg = eps - diff_0;
			if(diff_1 > diff_2 - min_epsilon){    /* diff_1 ~ diff_2 */

/*
   -------------  --            /|\   diff < 0
                   |             |    
                   | eps         |    positive shift
                   |             |
   ------0------  --            ---   diff = 0
                   |             |
                   | eps         |    negative shift
                   |             |
   ---1-----2---  --            \|/   diff > 0
            2
            2
*/

				if((diff_1 + diff_2) / 2.0 >= eps + diff_0)
					*shift_pos = (diff_1 + diff_2) / 2.0;
				else if(diff_2 - eps >= eps + diff_1)    /* this is the same as if diff_1 !~ diff_2 */
					*shift_pos = eps + diff_1;
				else if(diff_2 >= eps + diff_1)
					*shift_pos = diff_2;
				else
					*shift_pos = diff_2 + eps;
			}
			else{    /* diff_1 !~ diff_2 */

/*
   -------------  --            /|\   diff < 0
                   |             |    
                   | eps         |    positive shift
                   |             |
   ------0------  --            ---   diff = 0
                   |             |
      1            | eps         |    negative shift
                   |             |
   ---------2---  --            \|/   diff > 0
            2
            2
*/

				if(diff_2 - eps >= eps + diff_1)
					*shift_pos = eps + diff_1;
				else if(diff_2 >= eps + diff_1)
					*shift_pos = diff_2;
				else
					*shift_pos = diff_2 + eps;
			}
		}
		else{
			if(diff_0 > 0.0){   /* <==> diff_0 >= diff_2 */
				if(diff_1 < 0.0){
						
/*
   -------------  --            /|\   diff < 0
                   |             |    
         1         | eps         |    positive shift
                   |             |
   -------------  --            ---   diff = 0
                   |             |
                   | eps         |    negative shift
                   |             |
   ---2---------  --            \|/   diff > 0
      2
      2-----0
            0
*/

					*shift_neg = -diff_1;
					if(diff_2 - eps >= eps + diff_1)
						*shift_pos = eps + diff_1;
					else if(diff_0 - diff_2 >= eps)
						*shift_pos = diff_2;
					else if(diff_0 >= diff_2 + eps)
						*shift_pos = diff_0;
					else
						*shift_pos = diff_0 + eps;
				}
				else{
						
/*
   -------------  --            /|\   diff < 0
                   |             |    
                   | eps         |    positive shift
                   |             |
   -------------  --            ---   diff = 0
                   |             |
         1         | eps         |    negative shift
                   |             |
   ---2---------  --            \|/   diff > 0
      2
      2-----0
            0
*/

					*shift_neg = eps - diff_1;
					if(diff_2 - eps >= diff_1)
						*shift_pos = diff_1;
					else if(diff_0 - eps >= diff_2 && diff_2 >= eps + diff_1)
						*shift_pos = diff_2;
					else if(diff_0 >= diff_2 + eps)
						*shift_pos = diff_0;
					else
						*shift_pos = diff_0 + eps;
				}
			}
			else{   /* <==> diff_0 <= -eps */
				if(diff_1 > 0.0){

/*
            0
            0
   ---------0---  --            /|\   diff < 0
                   |             |    
                   | eps         |    positive shift
                   |             |
   -------------  --            ---   diff = 0
                   |             |
         1         | eps         |    negative shift
                   |             |
   ---2---------  --            \|/   diff > 0
      2
      2
*/

					if(-diff_0 - eps >= eps - diff_1)
						*shift_neg = eps - diff_1;
					else
						*shift_neg = -diff_0;
					if(diff_2 - eps >+ diff_1)
						*shift_pos = diff_1;
					else if(diff_2 - eps >= eps + diff_1)
						*shift_pos = eps + diff_1;
					else if(diff_2 >= eps + diff_1)
						*shift_pos = diff_2;
					else
						*shift_pos = diff_2 + eps;
				}
				else{

/*
             0
             0
   ----------0--  --            /|\   diff < 0
                   |             |    
         1         | eps         |    positive shift
                   |             |
   -------------  --            ---   diff = 0
                   |             |
                   | eps         |    negative shift
                   |             |
   ---2---------  --            \|/   diff > 0
      2
      2
*/

					if(-diff_0 - eps >= -diff_1)
						*shift_neg = -diff_1;
					else if(-diff_0 >= eps - diff_1)
						*shift_neg = -diff_0;
					else
						*shift_neg = -diff_0 + eps;
					if(diff_2 - eps >= eps + diff_1)
						*shift_pos = eps + diff_1;
					else if(diff_2 >= eps + diff_1)
						*shift_pos = diff_2;
					else
						*shift_pos = diff_2 + eps;
				}
			}
		}
	}

	if(sign_change == YES){
		swap(*shift_neg, *shift_pos, tmp);

#ifdef DEBUG
		diff_0 = -diff_0;
		diff_1 = -diff_1;
		diff_2 = -diff_2;
#endif
	}		

#ifdef DEBUG
	if(*shift_pos < 0.0 || *shift_neg < 0.0)error_message(GENERAL_ERROR, "Unexpected situation");

/* check positive shift */

	diff = fabs(diff_0 - *shift_pos);
	if(diff < eps && diff >= min_epsilon)error_message(GENERAL_ERROR, "Unexpected situation");
	diff = fabs(diff_1 - *shift_pos);
	if(diff < eps && diff >= min_epsilon)error_message(GENERAL_ERROR, "Unexpected situation");
	diff = fabs(diff_2 - *shift_pos);
	if(diff < eps && diff >= min_epsilon)error_message(GENERAL_ERROR, "Unexpected situation");

/* check negative shift */

	diff = fabs(diff_0 + *shift_neg);
	if(diff < eps && diff >= min_epsilon)error_message(GENERAL_ERROR, "Unexpected situation");
	diff = fabs(diff_1 + *shift_neg);
	if(diff < eps && diff >= min_epsilon)error_message(GENERAL_ERROR, "Unexpected situation");
	diff = fabs(diff_2 + *shift_neg);
	if(diff < eps && diff >= min_epsilon)error_message(GENERAL_ERROR, "Unexpected situation");
#endif

	return(YES);
}


/* compute intersection points and corresponding diffs (if check == YES) of x_ellipse with y and z edges;
	 if 2 intersections occur, diff is related to intersection point closer to corner;
	 sign of diffs determines the orientation of shift to enforce passing through corner;
	 global variables y_axis and z_axis are used to prevent duplicity of computation */

static void
get_ellipsoid_yz_edge_intersection(logic check)
{
	point_rec xyz_point[2];

#if 0
	vector_rec vec, vec_tmp, vector, vector_y, vector_z, vector_b, vector_c;
	point_rec pnt;
	double val;
#ifndef DRAW_INTERSECTING_POINT
	point_rec pnt_xyz;
#endif
#endif
#ifdef DRAW_INTERSECTING_POINT
	point_rec pnt_xyz;
	int i;
#endif

	if(y_axis == NO){
		y_axis = YES;
		if(fabs(diff_xz) < min_epsilon){
			y_count = 1;
		
			if(z_touching == YES)                                                  /* reuse ellipsoid extreme point */
				y_coord[0] = ext_point_z[ind_z].y;
			else{
				if(z_intersection == NO)error_message(GENERAL_ERROR, "Unexpected situation");
#if 1
				y_coord[0] = extreme_pnt_xz.y;                                       /* reuse ellipse extreme point */
#else 

/* compute extreme point - duplicity of code in resolve_ellipsoid_yz_edge_touching */

				vec.x = x_ellipse.max;
				vec.y = x_ellipse.min;
				vec.z = 0.0;
				
				comp_product(vector, vec, vec);
				
				transform_vector_from_global_to_local(&dir_z, &vector_c, &x_trans_g2l);
				
				comp_product(vector_z, vector_c, vec);
				val = 1.0 / sqrt(size_vec(vector_z));
				if(x_ellipse.center.z > tile.corner[ind_z].y)val = -val;
				copy_mul_vec(vec_tmp, vector, val);
				comp_product(pnt, vector_c, vec_tmp);
				
				transform_point_from_local_to_global(&pnt, &pnt_xyz, &(x_ellipse.center), &x_trans_l2g);
				y_coord[0] = pnt_xyz.y;
#endif
			}
			
			if(check == YES)y_diff = tile.corner[ind_y].y - y_coord[0];
		}
		else{
			if(z_intersection == YES){
				if((y_count = get_ellipse_line_intersection(&x_ellipse, &(tile.corner[ind_z]), &dir_y, xyz_point, NO)) != 0){
#ifdef DEBUG
					if(y_count == 1)error_message(GENERAL_ERROR, "Unexpected situation");
#endif
					if(fabs(xyz_point[0].y - xyz_point[1].y) <= 2.0 * min_epsilon)
						error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);
					
					if(xyz_point[0].y < xyz_point[1].y){
						y_coord[0] = xyz_point[0].y;
						y_coord[1] = xyz_point[1].y;
					}
					else{
						y_coord[0] = xyz_point[1].y;
						y_coord[1] = xyz_point[0].y;
					}
					
					if(check == YES){
						if(y_coord[0] >= tile.corner[ind_y].y)
							y_diff = tile.corner[ind_y].y - y_coord[0];
						else if(y_coord[1] <= tile.corner[ind_y].y)
							y_diff = tile.corner[ind_y].y - y_coord[1];
						else if(y_coord[1] - tile.corner[ind_y].y <= tile.corner[ind_y].y - y_coord[0])
							y_diff = tile.corner[ind_y].y - y_coord[1];
						else
							y_diff = tile.corner[ind_y].y - y_coord[0];
					}
				}
			}
		}

#ifdef DRAW_INTERSECTING_POINT
		pnt_xyz.z = tile.corner[ind_z].z;
		pnt_xyz.x = tile.corner[ind_x].x;
		for(i = 0; i < y_count; i++){
			pnt_xyz.y = y_coord[i];
			draw_point(&pnt_xyz, 0, cyan);
		}
#endif
	}

	if(z_axis == NO){
		z_axis = YES;
		if(fabs(diff_xy) < min_epsilon){
			z_count = 1;
			
			if(y_touching == YES)                                                  /* reuse ellipsoid extreme point */
				z_coord[0] = ext_point_y[ind_y].z;
			else{
				if(y_intersection == NO)error_message(GENERAL_ERROR, "Unexpected situation");
#if 1
				z_coord[0] = extreme_pnt_xy.z;                                       /* reuse ellipse extreme point */
#else 

/* compute extreme point - duplicity of code in resolve_ellipsoid_yz_edge_touching */

				vec.x = x_ellipse.max;
				vec.y = x_ellipse.min;
				vec.z = 0.0;
				
				comp_product(vector, vec, vec);
				
				transform_vector_from_global_to_local(&dir_y, &vector_b, &x_trans_g2l);
				
				comp_product(vector_y, vector_b, vec);
				val = 1.0 / sqrt(size_vec(vector_y));
				if(x_ellipse.center.y > tile.corner[ind_y].y)val = -val;
				copy_mul_vec(vec_tmp, vector, val);
				comp_product(pnt, vector_b, vec_tmp);
				
				transform_point_from_local_to_global(&pnt, &pnt_xyz, &(x_ellipse.center), &x_trans_l2g);
				z_coord[0] = pnt_xyz.z;
#endif
			}
			
			if(check == YES)z_diff = tile.corner[ind_z].z - z_coord[0];
		}
		else{
			if(y_intersection == YES){
				if((z_count = get_ellipse_line_intersection(&x_ellipse, &(tile.corner[ind_y]), &dir_z, xyz_point, NO)) != 0){
#ifdef DEBUG
					if(z_count == 1)error_message(GENERAL_ERROR, "Unexpected situation");
#endif
					if(fabs(xyz_point[0].z - xyz_point[1].z) <= 2.0 * min_epsilon)
						error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);
					
					if(xyz_point[0].z < xyz_point[1].z){
						z_coord[0] = xyz_point[0].z;
						z_coord[1] = xyz_point[1].z;
					}
					else{
						z_coord[0] = xyz_point[1].z;
						z_coord[1] = xyz_point[0].z;
					}
					
					if(check == YES){
						if(z_coord[0] >= tile.corner[ind_z].z)
							z_diff = tile.corner[ind_z].z - z_coord[0];
						else if(z_coord[1] <= tile.corner[ind_z].z)
							z_diff = tile.corner[ind_z].z - z_coord[1];
						else if(z_coord[1] - tile.corner[ind_z].z <= tile.corner[ind_z].z - z_coord[0])
							z_diff = tile.corner[ind_z].z - z_coord[1];
						else
							z_diff = tile.corner[ind_z].z - z_coord[0];
					}
				}
			}
		}

#ifdef DRAW_INTERSECTING_POINT
		pnt_xyz.x = tile.corner[ind_x].x;
		pnt_xyz.y = tile.corner[ind_y].y;
		for(i = 0; i < z_count; i++){
			pnt_xyz.z = z_coord[i];
			draw_point(&pnt_xyz, 0, magenta);
		}
#endif
	}
}


/* compute intersection points and corresponding diffs (if check == YES) of y_ellipse with z and x edges;
	 if 2 intersections occur, diff is related to intersection point closer to corner;
	 sign of diffs determines the orientation of shift to enforce passing through corner;
	 global variables z_axis and x_axis are used to prevent duplicity of computation */

static void
get_ellipsoid_zx_edge_intersection(logic check)
{
	point_rec xyz_point[2];

#if 0
	vector_rec vec, vec_tmp, vector, vector_z, vector_x, vector_c, vector_a;
	point_rec pnt;
	double val;
#ifndef DRAW_INTERSECTING_POINT
	point_rec pnt_xyz;
#endif
#endif
#ifdef DRAW_INTERSECTING_POINT
	point_rec pnt_xyz;
	int i;
#endif

	if(z_axis == NO){
		z_axis = YES;
		if(fabs(diff_yx) < min_epsilon){
			z_count = 1;
			
			if(x_touching == YES)                                                  /* reuse ellipsoid extreme point */
				z_coord[0] = ext_point_x[ind_x].z;
			else{
				if(x_intersection == NO)error_message(GENERAL_ERROR, "Unexpected situation");
#if 1
				z_coord[0] = extreme_pnt_yx.z;                                       /* reuse ellipse extreme point */
#else 

/* compute extreme point - duplicity of code in resolve_ellipsoid_zx_edge_touching */

				vec.x = y_ellipse.max;
				vec.y = y_ellipse.min;
				vec.z = 0.0;
				
				comp_product(vector, vec, vec);
				
				transform_vector_from_global_to_local(&dir_x, &vector_a, &y_trans_g2l);
				
				comp_product(vector_x, vector_a, vec);
				val = 1.0 / sqrt(size_vec(vector_x));
				if(y_ellipse.center.x > tile.corner[ind_x].z)val = -val;
				copy_mul_vec(vec_tmp, vector, val);
				comp_product(pnt, vector_a, vec_tmp);
				
				transform_point_from_local_to_global(&pnt, &pnt_xyz, &(y_ellipse.center), &y_trans_l2g);
				z_coord[0] = pnt_xyz.z;
#endif
			}
			
			if(check == YES)z_diff = tile.corner[ind_z].z - z_coord[0];
		}
		else{
			if(x_intersection == YES){
				if((z_count = get_ellipse_line_intersection(&y_ellipse, &(tile.corner[ind_x]), &dir_z, xyz_point, NO)) != 0){
#ifdef DEBUG
					if(z_count == 1)error_message(GENERAL_ERROR, "Unexpected situation");
#endif
					if(fabs(xyz_point[0].z - xyz_point[1].z) <= 2.0 * min_epsilon)
						error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);
					
					if(xyz_point[0].z < xyz_point[1].z){
						z_coord[0] = xyz_point[0].z;
						z_coord[1] = xyz_point[1].z;
					}
					else{
						z_coord[0] = xyz_point[1].z;
						z_coord[1] = xyz_point[0].z;
					}
					
					if(check == YES){
						if(z_coord[0] >= tile.corner[ind_z].z)
							z_diff = tile.corner[ind_z].z - z_coord[0];
						else if(z_coord[1] <= tile.corner[ind_z].z)
							z_diff = tile.corner[ind_z].z - z_coord[1];
						else if(z_coord[1] - tile.corner[ind_z].z <= tile.corner[ind_z].z - z_coord[0])
							z_diff = tile.corner[ind_z].z - z_coord[1];
						else
							z_diff = tile.corner[ind_z].z - z_coord[0];
					}
				}
			}
		}

#ifdef DRAW_INTERSECTING_POINT
		pnt_xyz.x = tile.corner[ind_x].x;
		pnt_xyz.y = tile.corner[ind_y].y;
		for(i = 0; i < z_count; i++){
			pnt_xyz.z = z_coord[i];
			draw_point(&pnt_xyz, 0, magenta);
		}
#endif
	}

	if(x_axis == NO){
		x_axis = YES;
		if(fabs(diff_yz) < min_epsilon){
			x_count = 1;
			
			if(z_touching == YES)                                                  /* reuse ellipsoid extreme point */
				x_coord[0] = ext_point_z[ind_z].x;
			else{
				if(z_intersection == NO)error_message(GENERAL_ERROR, "Unexpected situation");
#if 1
				x_coord[0] = extreme_pnt_yz.x;                                       /* reuse ellipse extreme point */
#else 

/* compute extreme point - duplicity of code in resolve_ellipsoid_zx_edge_touching */

				vec.x = y_ellipse.max;
				vec.y = y_ellipse.min;
				vec.z = 0.0;
				
				comp_product(vector, vec, vec);
				
				transform_vector_from_global_to_local(&dir_z, &vector_c, &y_trans_g2l);
				
				comp_product(vector_z, vector_c, vec);
				val = 1.0 / sqrt(size_vec(vector_z));
				if(y_ellipse.center.z > tile.corner[ind_z].z)val = -val;
				copy_mul_vec(vec_tmp, vector, val);
				comp_product(pnt, vector_c, vec_tmp);
				
				transform_point_from_local_to_global(&pnt, &pnt_xyz, &(y_ellipse.center), &y_trans_l2g);
				x_coord[0] = pnt_xyz.x;
#endif
			}
			
			if(check == YES)x_diff = tile.corner[ind_x].x - x_coord[0];
		}
		else{
			if(z_intersection == YES){
				if((x_count = get_ellipse_line_intersection(&y_ellipse, &(tile.corner[ind_z]), &dir_x, xyz_point, NO)) != 0){
#ifdef DEBUG
					if(x_count == 1)error_message(GENERAL_ERROR, "Unexpected situation");
#endif
					if(fabs(xyz_point[0].x - xyz_point[1].x) <= 2.0 * min_epsilon)
						error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);
					
					if(xyz_point[0].x < xyz_point[1].x){
						x_coord[0] = xyz_point[0].x;
						x_coord[1] = xyz_point[1].x;
					}
					else{
						x_coord[0] = xyz_point[1].x;
						x_coord[1] = xyz_point[0].x;
					}
					
					if(check == YES){
						if(x_coord[0] >= tile.corner[ind_x].x)
							x_diff = tile.corner[ind_x].x - x_coord[0];
						else if(x_coord[1] <= tile.corner[ind_x].x)
							x_diff = tile.corner[ind_x].x - x_coord[1];
						else if(x_coord[1] - tile.corner[ind_x].x <= tile.corner[ind_x].x - x_coord[0])
							x_diff = tile.corner[ind_x].x - x_coord[1];
						else
							x_diff = tile.corner[ind_x].x - x_coord[0];
					}
				}
			}
		}

#ifdef DRAW_INTERSECTING_POINT
		pnt_xyz.y = tile.corner[ind_y].y;
		pnt_xyz.z = tile.corner[ind_z].z;
		for(i = 0; i < x_count; i++){
			pnt_xyz.x = x_coord[i];
			draw_point(&pnt_xyz, 0, green);
		}
#endif
	}
}


/* compute intersection points and corresponding diffs (if check == YES) of z_ellipse with x and y edges;
	 if 2 intersections occur, diff is related to intersection point closer to corner;
	 sign of diffs determines the orientation of shift to enforce passing through corner;
	 global variables x_axis and y_axis are used to prevent duplicity of computation */

static void
get_ellipsoid_xy_edge_intersection(logic check)
{
	point_rec xyz_point[2];

#if 0
	vector_rec vec, vec_tmp, vector, vector_x, vector_y, vector_a, vector_b;
	point_rec pnt;
	double val;
#ifndef DRAW_INTERSECTING_POINT
	point_rec pnt_xyz;
#endif
#endif
#ifdef DRAW_INTERSECTING_POINT
	point_rec pnt_xyz;
	int i;
#endif

	if(x_axis == NO){
		x_axis = YES;
		if(fabs(diff_zy) < min_epsilon){
			x_count = 1;
			
			if(y_touching == YES)                                                  /* reuse ellipsoid extreme point */
				x_coord[0] = ext_point_y[ind_y].x;
			else{
				if(y_intersection == NO)error_message(GENERAL_ERROR, "Unexpected situation");
#if 1
				x_coord[0] = extreme_pnt_zy.x;                                       /* reuse ellipse extreme point */
#else 

/* compute extreme point - duplicity of code in resolve_ellipsoid_xy_edge_touching */

				vec.x = z_ellipse.max;
				vec.y = z_ellipse.min;
				vec.z = 0.0;
				
				comp_product(vector, vec, vec);
				
				transform_vector_from_global_to_local(&dir_y, &vector_b, &z_trans_g2l);
				
				comp_product(vector_y, vector_b, vec);
				val = 1.0 / sqrt(size_vec(vector_y));
				if(z_ellipse.center.y > tile.corner[ind_y].x)val = -val;
				copy_mul_vec(vec_tmp, vector, val);
				comp_product(pnt, vector_b, vec_tmp);
				
				transform_point_from_local_to_global(&pnt, &pnt_xyz, &(z_ellipse.center), &z_trans_l2g);
				x_coord[0] = pnt_xyz.x;
#endif
			}
			
			if(check == YES)x_diff = tile.corner[ind_x].x - x_coord[0];
		}
		else{
			if(y_intersection == YES){
				if((x_count = get_ellipse_line_intersection(&z_ellipse, &(tile.corner[ind_y]), &dir_x, xyz_point, NO)) != 0){
#ifdef DEBUG
					if(x_count == 1)error_message(GENERAL_ERROR, "Unexpected situation");
#endif
					if(fabs(xyz_point[0].x - xyz_point[1].x) <= 2.0 * min_epsilon)
						error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);
					
					if(xyz_point[0].x < xyz_point[1].x){
						x_coord[0] = xyz_point[0].x;
						x_coord[1] = xyz_point[1].x;
					}
					else{
						x_coord[0] = xyz_point[1].x;
						x_coord[1] = xyz_point[0].x;
					}
					
					if(check == YES){
						if(x_coord[0] >= tile.corner[ind_x].x)
							x_diff = tile.corner[ind_x].x - x_coord[0];
						else if(x_coord[1] <= tile.corner[ind_x].x)
							x_diff = tile.corner[ind_x].x - x_coord[1];
						else if(x_coord[1] - tile.corner[ind_x].x <= tile.corner[ind_x].x - x_coord[0])
							x_diff = tile.corner[ind_x].x - x_coord[1];
						else
							x_diff = tile.corner[ind_x].x - x_coord[0];
					}
				}
			}
		}

#ifdef DRAW_INTERSECTING_POINT
		pnt_xyz.y = tile.corner[ind_y].y;
		pnt_xyz.z = tile.corner[ind_z].z;
		for(i = 0; i < x_count; i++){
			pnt_xyz.x = x_coord[i];
			draw_point(&pnt_xyz, 0, green);
		}
#endif

	}

	if(y_axis == NO){
		y_axis = YES;
		if(fabs(diff_zx) < min_epsilon){
			y_count = 1;
			
			if(x_touching == YES)                                                  /* reuse ellipsoid extreme point */
				y_coord[0] = ext_point_x[ind_x].y;
			else{
				if(x_intersection == NO)error_message(GENERAL_ERROR, "Unexpected situation");
#if 1
				y_coord[0] = extreme_pnt_zx.y;                                       /* reuse ellipse extreme point */
#else 

/* compute extreme point - duplicity of code in resolve_ellipsoid_xy_edge_touching */

				vec.x = z_ellipse.max;
				vec.y = z_ellipse.min;
				vec.z = 0.0;
				
				comp_product(vector, vec, vec);
				
				transform_vector_from_global_to_local(&dir_x, &vector_a, &z_trans_g2l);
				
				comp_product(vector_x, vector_a, vec);
				val = 1.0 / sqrt(size_vec(vector_x));
				if(z_ellipse.center.x > tile.corner[ind_x].x)val = -val;
				copy_mul_vec(vec_tmp, vector, val);
				comp_product(pnt, vector_a, vec_tmp);
				
				transform_point_from_local_to_global(&pnt, &pnt_xyz, &(z_ellipse.center), &z_trans_l2g);
				y_coord[0] = pnt_xyz.y;
#endif
			}
			
			if(check == YES)y_diff = tile.corner[ind_y].y - y_coord[0];
		}
		else{
			if(x_intersection == YES){
				if((y_count = get_ellipse_line_intersection(&z_ellipse, &(tile.corner[ind_x]), &dir_y, xyz_point, NO)) != 0){
#ifdef DEBUG
					if(y_count == 1)error_message(GENERAL_ERROR, "Unexpected situation");
#endif
					if(fabs(xyz_point[0].y - xyz_point[1].y) <= 2.0 * min_epsilon)
						error_message(GENERAL_ERROR, "Inappropriate position or shape of inclusion %ld", inclusion -> id);
					
					if(xyz_point[0].y < xyz_point[1].y){
						y_coord[0] = xyz_point[0].y;
						y_coord[1] = xyz_point[1].y;
					}
					else{
						y_coord[0] = xyz_point[1].y;
						y_coord[1] = xyz_point[0].y;
					}
					
					if(check == YES){
						if(y_coord[0] >= tile.corner[ind_y].y)
							y_diff = tile.corner[ind_y].y - y_coord[0];
						else if(y_coord[1] <= tile.corner[ind_y].y)
							y_diff = tile.corner[ind_y].y - y_coord[1];
						else if(y_coord[1] - tile.corner[ind_y].y <= tile.corner[ind_y].y - y_coord[0])
							y_diff = tile.corner[ind_y].y - y_coord[1];
						else
							y_diff = tile.corner[ind_y].y - y_coord[0];
					}
				}
			}
		}

#ifdef DRAW_INTERSECTING_POINT
		pnt_xyz.z = tile.corner[ind_z].z;
		pnt_xyz.x = tile.corner[ind_x].x;
		for(i = 0; i < y_count; i++){
			pnt_xyz.y = y_coord[i];
			draw_point(&pnt_xyz, 0, cyan);
		}
#endif
	}
}


/* shift ellipsoid in x direction by shift and check (if eps != 0.0) for violation of proximity criteria;
	 if violation is detected, shifting is immediately interrupted;
	 variable phase is used to perform only relevant operations when back shifting */

#ifdef PARAMETRIC_ROUTINES
/* caution: ctrl points are not shifted */
#endif

static logic
check_ellipsoid_shift_x_corner(double shift, double eps, int *phase)
{
	double a_diff;
	logic check = NO;
	int i;

#ifdef DRAW_INTERSECTING_POINT
	point_rec pnt_xyz;
#endif

	if(eps != 0.0){
		check = YES;
		*phase = 0;
	}

/* IMPORTANT: do not update x_diff !!!
	 it is used for eventual back shift */

	ellipsoid -> center.x += shift;

	for(i = 0; i < 2; i++){

#ifdef DRAW_EXTREME_POINT_ELLIPSOID
		draw_point(&ext_point_x[i], 0, bgcolor);
		draw_point(&ext_point_y[i], 0, bgcolor);
		draw_point(&ext_point_z[i], 0, bgcolor);
#endif

		ext_point_x[i].x += shift;
		ext_point_y[i].x += shift;
		ext_point_z[i].x += shift;

#ifdef DRAW_EXTREME_POINT_ELLIPSOID
		draw_point(&ext_point_x[i], 0, green);
		draw_point(&ext_point_y[i], 0, cyan);
		draw_point(&ext_point_z[i], 0, magenta);
#endif

	}

	diff_x -= shift;

	if(y_intersection == YES){

#ifdef DRAW_ELLIPSE
		draw_ellipse(&y_ellipse, 0, bgcolor);
#endif
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		draw_point(&extreme_pnt_yx, 0, bgcolor);
		draw_point(&extreme_pnt_yz, 0, bgcolor);
#endif

		y_ellipse.center.x += shift;
		extreme_pnt_yx.x += shift;
		extreme_pnt_yz.x += shift;
		diff_yx -= shift;
		diff_yz -= shift;

#ifdef DRAW_ELLIPSE
		draw_ellipse(&y_ellipse, 0, cyan);
#endif
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		draw_point(&extreme_pnt_yx, 0, cyan);
		draw_point(&extreme_pnt_yz, 0, cyan);
#endif

	}
	if(y_touching == YES)y_ellipse.center.x += shift;

	if(z_intersection == YES){

#ifdef DRAW_ELLIPSE
		draw_ellipse(&z_ellipse, 0, bgcolor);
#endif
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		draw_point(&extreme_pnt_zy, 0, bgcolor);
		draw_point(&extreme_pnt_zx, 0, bgcolor);
#endif

		z_ellipse.center.x += shift;
		extreme_pnt_zy.x += shift;
		extreme_pnt_zx.x += shift;
		diff_zy -= shift;
		diff_zx -= shift;

#ifdef DRAW_ELLIPSE
		draw_ellipse(&z_ellipse, 0, magenta);
#endif
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		draw_point(&extreme_pnt_zy, 0, magenta);
		draw_point(&extreme_pnt_zx, 0, magenta);
#endif

	}
	if(z_touching == YES)z_ellipse.center.x += shift;

	if(check == YES){
		(*phase)++;

#ifdef DEBUG
		if(fabs(x_diff - shift) > min_epsilon)error_message(GENERAL_ERROR, "Unexpected situation");
#endif

		a_diff = fabs(diff_yx);
		if(a_diff < eps && a_diff > min_epsilon)return(NO);

		a_diff = fabs(diff_zx);
		if(a_diff < eps && a_diff > min_epsilon)return(NO);

		a_diff = fabs(diff_x);        /* IMPORTANT: this fabs must be called as last as it is used bellow !!! */
		if(a_diff < eps && a_diff > min_epsilon)return(NO);
	}
	else{
		if(*phase == 1)return(YES);
		a_diff = fabs(diff_x);        /* IMPORTANT: this fabs must be called as last as it is used bellow !!! */
	}

/* update tile corner points in LCS of ellipsoid after center of ellipsoid was shifted;
	 this must be done before call to get_ellipsoid_x_side_intersection */

	for(i = 0; i < 2; i++){
		transform_point_from_global_to_local(&(tile.corner[i]), &corner_point[i], &(ellipsoid -> center), &trans_g2l);
		transform_local_from_ellipsoid_to_sphere(&corner_point[i], &corner_point[i], &scale_e2s);
	}

/* resolve intersections with tile sides */

#ifdef DRAW_ELLIPSE
	if(x_intersection == YES)draw_ellipse(&x_ellipse, 0, bgcolor);
#endif
#ifdef DRAW_EXTREME_POINT_ELLIPSE
	if(x_intersection == YES){
		draw_point(&extreme_pnt_xy, 0, bgcolor);
		draw_point(&extreme_pnt_xz, 0, bgcolor);
	}
#endif

	x_touching = x_intersection = NO;

	diff_xy = diff_xz = max_sz;

	if(a_diff <= min_epsilon){
		x_touching = YES;
		setup_ellipse_touching_point(&x_ellipse, &ext_point_x[ind_x], &dir_x);
	}
	else{
		if((diff_x > 0.0 && ind_x == 0) || (diff_x < 0.0 && ind_x == 1)){
			x_intersection = YES;
			get_ellipsoid_x_side_intersection();

/* get intersection ellipse extreme points */

			if(y_intersection == YES || z_intersection == YES || y_touching == YES || z_touching == YES)resolve_ellipsoid_yz_edge_touching(NO, NO);
		}
	}
								
	if(check == YES){
		(*phase)++;

		a_diff = fabs(diff_xy);
		if(a_diff < eps && a_diff > min_epsilon)return(NO);

		a_diff = fabs(diff_xz);
		if(a_diff < eps && a_diff > min_epsilon)return(NO);

		y_diff = z_diff = max_sz;
	}
	else{
		if(*phase == 2)return(YES);
	}
		
/* resolve intersections with tile edges */

#ifdef DRAW_INTERSECTING_POINT
	pnt_xyz.y = tile.corner[ind_y].y;
	pnt_xyz.z = tile.corner[ind_z].z;

	for(i = 0; i < x_count; i++){
		pnt_xyz.x = x_coord[i];
		draw_point(&pnt_xyz, 0, bgcolor);
	}
#endif

	for(i = 0; i < x_count; i++)x_coord[i] += shift;

#ifdef DRAW_INTERSECTING_POINT
	for(i = 0; i < x_count; i++){
		pnt_xyz.x = x_coord[i];
		draw_point(&pnt_xyz, 0, green);
	}
#endif

	if(x_intersection == YES){

#ifdef DRAW_INTERSECTING_POINT
		pnt_xyz.x = tile.corner[ind_x].x;

		pnt_xyz.z = tile.corner[ind_z].z;
		for(i = 0; i < y_count; i++){
			pnt_xyz.y = y_coord[i];
			draw_point(&pnt_xyz, 0, bgcolor);
		}

		pnt_xyz.y = tile.corner[ind_y].y;
		for(i = 0; i < z_count; i++){
			pnt_xyz.z = z_coord[i];
			draw_point(&pnt_xyz, 0, bgcolor);
		}
#endif

		y_axis = z_axis = NO;
		y_count = z_count = 0;
		get_ellipsoid_yz_edge_intersection(check);
	}

#if 0 /* DO NOT USE - just to check intersections with edges */
	{
		point_rec point[2], corner;
		int i, cnt;
		
		corner.x = tile.corner[ind_x].x;
		corner.y = tile.corner[ind_y].y;
		corner.z = tile.corner[ind_z].z;

		cnt = get_ellipsoid_line_intersection(ellipsoid, &corner, &dir_x, point);
		for(i = 0; i < cnt; i++)fprintf(stderr, "x[%d]: %20.13e %20.13e %20.13e\n", i, point[i].x, point[i].y, point[i].z); 
		cnt = get_ellipsoid_line_intersection(ellipsoid, &corner, &dir_y, point);
		for(i = 0; i < cnt; i++)fprintf(stderr, "y[%d]: %20.13e %20.13e %20.13e\n", i, point[i].x, point[i].y, point[i].z); 
		cnt = get_ellipsoid_line_intersection(ellipsoid, &corner, &dir_z, point);
		for(i = 0; i < cnt; i++)fprintf(stderr, "z[%d]: %20.13e %20.13e %20.13e\n", i, point[i].x, point[i].y, point[i].z); 
	}
#endif

	if(check == YES){
		(*phase)++;

		a_diff = fabs(y_diff);
		if(a_diff < eps && a_diff > min_epsilon)return(NO);

		a_diff = fabs(z_diff);
		if(a_diff < eps && a_diff > min_epsilon)return(NO);
	}
	else{
		if(*phase == 3)return(YES);
	}

	return(YES);
}


/* shift ellipsoid in y direction by shift and check (if eps != 0.0) for violation of proximity criteria;
	 if violation is detected, shifting is immediately interrupted;
	 variable phase is used to perform only relevant operations when back shifting */

#ifdef PARAMETRIC_ROUTINES
/* caution: ctrl points are not shifted */
#endif

static logic
check_ellipsoid_shift_y_corner(double shift, double eps, int *phase)
{
	double a_diff;
	logic check = NO;
	int i;

#ifdef DRAW_INTERSECTING_POINT
	point_rec pnt_xyz;
#endif

	if(eps != 0.0){
		check = YES;
		*phase = 0;
	}

/* IMPORTANT: do not update y_diff !!!
	 it is used for eventual back shift */
	
	ellipsoid -> center.y += shift;

	for(i = 0; i < 2; i++){

#ifdef DRAW_EXTREME_POINT_ELLIPSOID
		draw_point(&ext_point_x[i], 0, bgcolor);
		draw_point(&ext_point_y[i], 0, bgcolor);
		draw_point(&ext_point_z[i], 0, bgcolor);
#endif

		ext_point_x[i].y += shift;
		ext_point_y[i].y += shift;
		ext_point_z[i].y += shift;

#ifdef DRAW_EXTREME_POINT_ELLIPSOID
		draw_point(&ext_point_x[i], 0, green);
		draw_point(&ext_point_y[i], 0, cyan);
		draw_point(&ext_point_z[i], 0, magenta);
#endif

	}

	diff_y -= shift;

	if(z_intersection == YES){

#ifdef DRAW_ELLIPSE
		draw_ellipse(&z_ellipse, 0, bgcolor);
#endif
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		draw_point(&extreme_pnt_zy, 0, bgcolor);
		draw_point(&extreme_pnt_zx, 0, bgcolor);
#endif

		z_ellipse.center.y += shift;
		extreme_pnt_zy.y += shift;
		extreme_pnt_zx.y += shift;
		diff_zy -= shift;
		diff_zx -= shift;

#ifdef DRAW_ELLIPSE
		draw_ellipse(&z_ellipse, 0, magenta);
#endif
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		draw_point(&extreme_pnt_zy, 0, magenta);
		draw_point(&extreme_pnt_zx, 0, magenta);
#endif

	}
	if(z_touching == YES)z_ellipse.center.y += shift;

	if(x_intersection == YES){

#ifdef DRAW_ELLIPSE
		draw_ellipse(&x_ellipse, 0, bgcolor);
#endif
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		draw_point(&extreme_pnt_xz, 0, bgcolor);
		draw_point(&extreme_pnt_xy, 0, bgcolor);
#endif

		x_ellipse.center.y += shift;
		extreme_pnt_xz.y += shift;
		extreme_pnt_xy.y += shift;
		diff_xz -= shift;
		diff_xy -= shift;

#ifdef DRAW_ELLIPSE
		draw_ellipse(&x_ellipse, 0, green);
#endif
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		draw_point(&extreme_pnt_xz, 0, green);
		draw_point(&extreme_pnt_xy, 0, green);
#endif

	}
	if(x_touching == YES)x_ellipse.center.y += shift;

	if(check == YES){
		(*phase)++;

#ifdef DEBUG
		if(fabs(y_diff - shift) > min_epsilon)error_message(GENERAL_ERROR, "Unexpected situation");
#endif

		a_diff = fabs(diff_zy);
		if(a_diff < eps && a_diff > min_epsilon)return(NO);

		a_diff = fabs(diff_xy);
		if(a_diff < eps && a_diff > min_epsilon)return(NO);

		a_diff = fabs(diff_y);        /* IMPORTANT: this fabs must be called as last as it is used bellow !!! */
		if(a_diff < eps && a_diff > min_epsilon)return(NO);
	}
	else{
		if(*phase == 1)return(YES);
		a_diff = fabs(diff_y);        /* IMPORTANT: this fabs must be called as last as it is used bellow !!! */
	}

/* update tile corner points in LCS of ellipsoid after center of ellipsoid was shifted;
	 this must be done before call to get_ellipsoid_y_side_intersection */

	for(i = 0; i < 2; i++){
		transform_point_from_global_to_local(&(tile.corner[i]), &corner_point[i], &(ellipsoid -> center), &trans_g2l);
		transform_local_from_ellipsoid_to_sphere(&corner_point[i], &corner_point[i], &scale_e2s);
	}

/* resolve intersections with tile sides */

#ifdef DRAW_ELLIPSE
	if(y_intersection == YES)draw_ellipse(&y_ellipse, 0, bgcolor);
#endif
#ifdef DRAW_EXTREME_POINT_ELLIPSE
	if(y_intersection == YES){
		draw_point(&extreme_pnt_yz, 0, bgcolor);
		draw_point(&extreme_pnt_yx, 0, bgcolor);
	}
#endif

	y_touching = y_intersection = NO;

	diff_yz = diff_yx = max_sz;

	if(a_diff <= min_epsilon){
		y_touching = YES;
		setup_ellipse_touching_point(&y_ellipse, &ext_point_y[ind_y], &dir_y);
	}
	else{
		if((diff_y > 0.0 && ind_y == 0) || (diff_y < 0.0 && ind_y == 1)){
			y_intersection = YES;
			get_ellipsoid_y_side_intersection();

/* get intersection ellipse extreme points */

			if(z_intersection == YES || x_intersection == YES || z_touching == YES || x_touching == YES)resolve_ellipsoid_zx_edge_touching(NO, NO);
		}
	}
								
	if(check == YES){
		(*phase)++;

		a_diff = fabs(diff_yz);
		if(a_diff < eps && a_diff > min_epsilon)return(NO);

		a_diff = fabs(diff_yx);
		if(a_diff < eps && a_diff > min_epsilon)return(NO);

		z_diff = x_diff = max_sz;
	}
	else{
		if(*phase == 2)return(YES);
	}
		
/* resolve intersections with tile edges */

#ifdef DRAW_INTERSECTING_POINT
	pnt_xyz.z = tile.corner[ind_z].z;
	pnt_xyz.x = tile.corner[ind_x].x;

	for(i = 0; i < y_count; i++){
		pnt_xyz.y = y_coord[i];
		draw_point(&pnt_xyz, 0, bgcolor);
	}
#endif

	for(i = 0; i < y_count; i++)y_coord[i] += shift;

#ifdef DRAW_INTERSECTING_POINT
	for(i = 0; i < y_count; i++){
		pnt_xyz.y = y_coord[i];
		draw_point(&pnt_xyz, 0, cyan);
	}
#endif

	if(y_intersection == YES){

#ifdef DRAW_INTERSECTING_POINT
		pnt_xyz.y = tile.corner[ind_y].y;

		pnt_xyz.x = tile.corner[ind_x].x;
		for(i = 0; i < z_count; i++){
			pnt_xyz.z = z_coord[i];
			draw_point(&pnt_xyz, 0, bgcolor);
		}

		pnt_xyz.z = tile.corner[ind_z].z;
		for(i = 0; i < x_count; i++){
			pnt_xyz.x = x_coord[i];
			draw_point(&pnt_xyz, 0, bgcolor);
		}
#endif

		z_axis = x_axis = NO;
		z_count = x_count = 0;
		get_ellipsoid_zx_edge_intersection(check);
	}

#if 0 /* DO NOT USE - just to check intersections with edges */
	{
		point_rec point[2], corner;
		int i, cnt;
		
		corner.x = tile.corner[ind_x].x;
		corner.y = tile.corner[ind_y].y;
		corner.z = tile.corner[ind_z].z;

		cnt = get_ellipsoid_line_intersection(ellipsoid, &corner, &dir_x, point);
		for(i = 0; i < cnt; i++)fprintf(stderr, "x[%d]: %20.13e %20.13e %20.13e\n", i, point[i].x, point[i].y, point[i].z); 
		cnt = get_ellipsoid_line_intersection(ellipsoid, &corner, &dir_y, point);
		for(i = 0; i < cnt; i++)fprintf(stderr, "y[%d]: %20.13e %20.13e %20.13e\n", i, point[i].x, point[i].y, point[i].z); 
		cnt = get_ellipsoid_line_intersection(ellipsoid, &corner, &dir_z, point);
		for(i = 0; i < cnt; i++)fprintf(stderr, "z[%d]: %20.13e %20.13e %20.13e\n", i, point[i].x, point[i].y, point[i].z); 
	}
#endif

	if(check == YES){
		(*phase)++;

		a_diff = fabs(z_diff);
		if(a_diff < eps && a_diff > min_epsilon)return(NO);

		a_diff = fabs(x_diff);
		if(a_diff < eps && a_diff > min_epsilon)return(NO);
	}
	else{
		if(*phase == 3)return(YES);
	}

	return(YES);
}


/* shift ellipsoid in z direction by shift and check (if eps != 0.0) for violation of proximity criteria;
	 if violation is detected, shifting is immediately interrupted;
	 variable phase is used to perform only relevant operations when back shifting */

#ifdef PARAMETRIC_ROUTINES
/* caution: ctrl points are not shifted */
#endif

static logic
check_ellipsoid_shift_z_corner(double shift, double eps, int *phase)
{
	double a_diff;
	logic check = NO;
	int i;

#ifdef DRAW_INTERSECTING_POINT
	point_rec pnt_xyz;
#endif

	if(eps != 0.0){
		check = YES;
		*phase = 0;
	}

/* IMPORTANT: do not update z_diff !!!
	 it is used for eventual back shift */
	
	ellipsoid -> center.z += shift;

	for(i = 0; i < 2; i++){

#ifdef DRAW_EXTREME_POINT_ELLIPSOID
		draw_point(&ext_point_x[i], 0, bgcolor);
		draw_point(&ext_point_y[i], 0, bgcolor);
		draw_point(&ext_point_z[i], 0, bgcolor);
#endif

		ext_point_x[i].z += shift;
		ext_point_y[i].z += shift;
		ext_point_z[i].z += shift;

#ifdef DRAW_EXTREME_POINT_ELLIPSOID
		draw_point(&ext_point_x[i], 0, green);
		draw_point(&ext_point_y[i], 0, cyan);
		draw_point(&ext_point_z[i], 0, magenta);
#endif

	}

	diff_z -= shift;

	if(x_intersection == YES){

#ifdef DRAW_ELLIPSE
		draw_ellipse(&x_ellipse, 0, bgcolor);
#endif
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		draw_point(&extreme_pnt_xz, 0, bgcolor);
		draw_point(&extreme_pnt_xy, 0, bgcolor);
#endif

		x_ellipse.center.z += shift;
		extreme_pnt_xz.z += shift;
		extreme_pnt_xy.z += shift;
		diff_xz -= shift;
		diff_xy -= shift;

#ifdef DRAW_ELLIPSE
		draw_ellipse(&x_ellipse, 0, green);
#endif
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		draw_point(&extreme_pnt_xz, 0, green);
		draw_point(&extreme_pnt_xy, 0, green);
#endif

	}
	if(x_touching == YES)x_ellipse.center.z += shift;

	if(y_intersection == YES){

#ifdef DRAW_ELLIPSE
		draw_ellipse(&y_ellipse, 0, bgcolor);
#endif
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		draw_point(&extreme_pnt_yx, 0, bgcolor);
		draw_point(&extreme_pnt_yz, 0, bgcolor);
#endif

		y_ellipse.center.z += shift;
		extreme_pnt_yx.z += shift;
		extreme_pnt_yz.z += shift;
		diff_yx -= shift;
		diff_yz -= shift;

#ifdef DRAW_ELLIPSE
		draw_ellipse(&y_ellipse, 0, cyan);
#endif
#ifdef DRAW_EXTREME_POINT_ELLIPSE
		draw_point(&extreme_pnt_yx, 0, cyan);
		draw_point(&extreme_pnt_yz, 0, cyan);
#endif

	}
	if(y_touching == YES)y_ellipse.center.z += shift;

	if(check == YES){
		(*phase)++;

#ifdef DEBUG
		if(fabs(z_diff - shift) > min_epsilon)error_message(GENERAL_ERROR, "Unexpected situation");
#endif

		a_diff = fabs(diff_xz);
		if(a_diff < eps && a_diff > min_epsilon)return(NO);

		a_diff = fabs(diff_yz);
		if(a_diff < eps && a_diff > min_epsilon)return(NO);

		a_diff = fabs(diff_z);        /* IMPORTANT: this fabs must be called as last as it is used bellow !!! */
		if(a_diff < eps && a_diff > min_epsilon)return(NO);
	}
	else{
		if(*phase == 1)return(YES);
		a_diff = fabs(diff_z);        /* IMPORTANT: this fabs must be called as last as it is used bellow !!! */
	}

/* update tile corner points in LCS of ellipsoid after center of ellipsoid was shifted;
	 this must be done before call to get_ellipsoid_z_side_intersection */

	for(i = 0; i < 2; i++){
		transform_point_from_global_to_local(&(tile.corner[i]), &corner_point[i], &(ellipsoid -> center), &trans_g2l);
		transform_local_from_ellipsoid_to_sphere(&corner_point[i], &corner_point[i], &scale_e2s);
	}

/* resolve intersections with tile sides */

#ifdef DRAW_ELLIPSE
	if(z_intersection == YES)draw_ellipse(&z_ellipse, 0, bgcolor);
#endif
#ifdef DRAW_EXTREME_POINT_ELLIPSE
	if(z_intersection == YES){
		draw_point(&extreme_pnt_zx, 0, bgcolor);
		draw_point(&extreme_pnt_zy, 0, bgcolor);
	}
#endif

	z_touching = z_intersection = NO;

	diff_zx = diff_zy = max_sz;

	if(a_diff <= min_epsilon){
		y_touching = YES;
		setup_ellipse_touching_point(&z_ellipse, &ext_point_z[ind_z], &dir_z);
	}
	else{
		if((diff_z > 0.0 && ind_z == 0) || (diff_z < 0.0 && ind_z == 1)){
			z_intersection = YES;
			get_ellipsoid_z_side_intersection();

/* get intersection ellipse extreme points */

			if(x_intersection == YES || y_intersection == YES || x_touching == YES || y_touching == YES)resolve_ellipsoid_xy_edge_touching(NO, NO);
		}
	}
								
	if(check == YES){
		(*phase)++;

		a_diff = fabs(diff_zx);
		if(a_diff < eps && a_diff > min_epsilon)return(NO);

		a_diff = fabs(diff_zy);
		if(a_diff < eps && a_diff > min_epsilon)return(NO);

		x_diff = y_diff = max_sz;
	}
	else{
		if(*phase == 2)return(YES);
	}
		
/* resolve intersections with tile edges */

#ifdef DRAW_INTERSECTING_POINT
	pnt_xyz.x = tile.corner[ind_x].x;
	pnt_xyz.y = tile.corner[ind_y].y;

	for(i = 0; i < z_count; i++){
		pnt_xyz.z = z_coord[i];
		draw_point(&pnt_xyz, 0, bgcolor);
	}
#endif

	for(i = 0; i < z_count; i++)z_coord[i] += shift;

#ifdef DRAW_INTERSECTING_POINT
	for(i = 0; i < z_count; i++){
		pnt_xyz.z = z_coord[i];
		draw_point(&pnt_xyz, 0, magenta);
	}
#endif

	if(z_intersection == YES){

#ifdef DRAW_INTERSECTING_POINT
		pnt_xyz.z = tile.corner[ind_z].z;

		pnt_xyz.y = tile.corner[ind_y].y;
		for(i = 0; i < x_count; i++){
			pnt_xyz.x = x_coord[i];
			draw_point(&pnt_xyz, 0, bgcolor);
		}

		pnt_xyz.x = tile.corner[ind_x].x;
		for(i = 0; i < y_count; i++){
			pnt_xyz.y = y_coord[i];
			draw_point(&pnt_xyz, 0, bgcolor);
		}
#endif

		x_axis = y_axis = NO;
		x_count = y_count = 0;
		get_ellipsoid_xy_edge_intersection(check);
	}

#if 0 /* DO NOT USE - just to check intersections with edges */
	{
		point_rec point[2], corner;
		int i, cnt;
		
		corner.x = tile.corner[ind_x].x;
		corner.y = tile.corner[ind_y].y;
		corner.z = tile.corner[ind_z].z;

		cnt = get_ellipsoid_line_intersection(ellipsoid, &corner, &dir_x, point);
		for(i = 0; i < cnt; i++)fprintf(stderr, "x[%d]: %20.13e %20.13e %20.13e\n", i, point[i].x, point[i].y, point[i].z); 
		cnt = get_ellipsoid_line_intersection(ellipsoid, &corner, &dir_y, point);
		for(i = 0; i < cnt; i++)fprintf(stderr, "y[%d]: %20.13e %20.13e %20.13e\n", i, point[i].x, point[i].y, point[i].z); 
		cnt = get_ellipsoid_line_intersection(ellipsoid, &corner, &dir_z, point);
		for(i = 0; i < cnt; i++)fprintf(stderr, "z[%d]: %20.13e %20.13e %20.13e\n", i, point[i].x, point[i].y, point[i].z); 
	}
#endif

	if(check == YES){
		(*phase)++;

		a_diff = fabs(x_diff);
		if(a_diff < eps && a_diff > min_epsilon)return(NO);

		a_diff = fabs(y_diff);
		if(a_diff < eps && a_diff > min_epsilon)return(NO);
	}
	else{
		if(*phase == 3)return(YES);
	}

	return(YES);
}


/* back shifting in x direction */

static void
ellipsoid_back_shift_x_corner(double shift, int phase)
{
	check_ellipsoid_shift_x_corner(-shift, 0.0, &phase);   /* zero must be passed as eps */
}


/* back shifting in y direction */

static void
ellipsoid_back_shift_y_corner(double shift, int phase)
{
	check_ellipsoid_shift_y_corner(-shift, 0.0, &phase);   /* zero must be passed as eps */
}


/* back shifting in z direction */

static void
ellipsoid_back_shift_z_corner(double shift, int phase)
{
	check_ellipsoid_shift_z_corner(-shift, 0.0, &phase);   /* zero must be passed as eps */
}


/* points point1 and point2 are expected to lie on the ellipse in adjacent quadrants */

static double
get_ellipse_weight(ellipse_rec *ellipse, point_rec *point1, point_rec *point2, trans_matrix *trans_g2l)
{
	point_rec pnt1, pnt2;
	double weight, cosine, Radius, radius, x1, y1, x2, y2;

#ifdef DEBUG
	double val1, val2;
#endif

	Radius = ellipse -> max;
	radius = ellipse -> min;

	transform_point_from_global_to_local(point1, &pnt1, &(ellipse -> center), trans_g2l);
	transform_point_from_global_to_local(point2, &pnt2, &(ellipse -> center), trans_g2l);

	x1 = pnt1.x / Radius;
	y1 = pnt1.y / radius;

	x2 = pnt2.x / Radius;
	y2 = pnt2.y / radius;

#ifdef DEBUG
	if(fabs(pnt1.z) > 1.0e-10 || fabs(pnt1.z) > 1.0e-10)error_message(GENERAL_ERROR, "Unexpected situation");

	val1 = x1 * x1 + y1 * y1;
	val2 = x2 * x2 + y2 * y2;

	if(fabs(1.0 - val1) > 1.0e-10 || fabs(1.0 - val2) > 1.0e-10)error_message(GENERAL_ERROR, "Unexpected situation");

	if(x1 * x2 < 0.0 && y1 * y2 < 0.0)error_message(GENERAL_ERROR, "Unexpected situation");
	if(x1 * x2 > 0.0 && y1 * y2 > 0.0)error_message(GENERAL_ERROR, "Unexpected situation");
#endif

#if 0
	if(x1 * x2 >= 0.0)
	  cosine = x1 * x2 - fabs(y1 * y2);
	else
	  cosine = y1 * y2 - fabs(x1 * x2);
#else

/* I do rely on fact that points are in adjacent quadrants;
	 this implies
	 if(x1 * x2 >= 0.0) ==> -fabs(y1 * y2) = y1 * y2
	 if(x1 * x2 <= 0.0) ==> -fabs(x1 * x2) = x1 * x2 */

	cosine = x1 * x2 + y1 * y2;
#endif

	weight = sqrt((1.0 + cosine) / 2.0);

	return(weight);
}


#ifdef CHECK_OVERLAP

static void
check_active_inclusion_overlap(ellipsoid_rec *ellipsoid1, int per_x, int per_y, int per_z, long i)
{
	inclusion_rec *inclusion = NULL;
	ellipsoid_rec *ellipsoid2 = NULL;
	double per_shift_x, per_shift_y, per_shift_z;
	int j, k, x_per, y_per, z_per;

#ifndef ALLOW_OVERLAP
	logic shift;

	shift = inclusion_array[i].shift;
#endif

	for(j = 0; j < i; j++){
		inclusion = &inclusion_array[j];
		ellipsoid2 = &(inclusion -> ellipsoid);

#ifndef ALLOW_OVERLAP

/* there is no need to check for intersection if neither of inclusions has been shifted and inclusions are not allowed to overlap in reference tile */

		if(shift == NO){
			if(inclusion -> shift == NO)continue;
		}
#endif

		for(k = 0; k < 8; k++){
			if(inclusion -> active[k] == NO)continue;
			
			if(shifting_x[k] == 0){
				per_shift_x = 0.0;
				x_per = 0;
			}
			else{
				x_per = 1;
				if(ellipsoid2 -> center.x < tile.center.x)          /* XN side */
					per_shift_x = tile.size.x;
				else                                                /* XP side */
					per_shift_x = -tile.size.x;
			}
			if(shifting_y[k] == 0){
				per_shift_y = 0.0;
				y_per = 0;
			}
			else{
				y_per = 1;
				if(ellipsoid2 -> center.y < tile.center.y)          /* YN side */
					per_shift_y = tile.size.y;
				else                                                /* YP side */
					per_shift_y = -tile.size.y;
			}
			if(shifting_z[k] == 0){
				per_shift_z = 0.0;
				z_per = 0;
			}
			else{
				z_per = 1;
				if(ellipsoid2 -> center.z < tile.center.z)          /* ZN side */
					per_shift_z = tile.size.z;
				else                                                /* ZP side */
					per_shift_z = -tile.size.z;
			}
			
			shift_ellipsoid(*ellipsoid2, per_shift_x, per_shift_y, per_shift_z);
			if(check_ellipsoid_ellipsoid_overlap(ellipsoid1, ellipsoid2, epsilon) == YES)
				error_message(GENERAL_ERROR, "Inclusion %ld (%d %d %d) intersects inclusion %ld (%d %d %d)", i + 1, per_x, per_y, per_z, j + 1, x_per, y_per, z_per);
			shift_ellipsoid(*ellipsoid2, -per_shift_x, -per_shift_y, -per_shift_z);
		}
	}
}
		
static void
check_inclusion_periodic_overlap(ellipsoid_rec *ellipsoid1, int per_x, int per_y, int per_z, long i)
{
	inclusion_rec *inclusion = NULL;
	ellipsoid_rec *ellipsoid2 = NULL;
	double per_shift_x, per_shift_y, per_shift_z;
	int j;

	for(j = 0; j < i; j++){
		inclusion = &inclusion_array[j];
		ellipsoid2 = &(inclusion -> ellipsoid);

/* primary instance */

		if(check_ellipsoid_ellipsoid_overlap(ellipsoid1, ellipsoid2, epsilon) == YES)
			error_message(GENERAL_ERROR, "Inclusion %ld (%d %d %d) intersects inclusion %ld (0 0 0)", i + 1, per_x, per_y, per_z, j + 1);

/* secondary instances */

		if(inclusion -> face_is[X] == NULL)
			per_shift_x = 0.0;
		else{
			if(ellipsoid2 -> center.x < tile.center.x)          /* XN side */
				per_shift_x = tile.size.x;
			else                                                /* XP side */
				per_shift_x = -tile.size.x;

			shift_ellipsoid(*ellipsoid2, per_shift_x, 0.0, 0.0);
			if(check_ellipsoid_ellipsoid_overlap(ellipsoid1, ellipsoid2, epsilon) == YES)
				error_message(GENERAL_ERROR, "Inclusion %ld (%d %d %d) intersects inclusion %ld (1 0 0)", i + 1, per_x, per_y, per_z, j + 1);
			shift_ellipsoid(*ellipsoid2, -per_shift_x, 0.0, 0.0);
		}
		if(inclusion -> face_is[Y] == NULL)
			per_shift_y = 0.0;
		else{
			if(ellipsoid2 -> center.y < tile.center.y)          /* YN side */
				per_shift_y = tile.size.y;
			else                                                /* YP side */
				per_shift_y = -tile.size.y;

			shift_ellipsoid(*ellipsoid2, 0.0, per_shift_y, 0.0);
			if(check_ellipsoid_ellipsoid_overlap(ellipsoid1, ellipsoid2, epsilon) == YES)
				error_message(GENERAL_ERROR, "Inclusion %ld (%d %d %d) intersects inclusion %ld (0 1 0)", i + 1, per_x, per_y, per_z, j + 1);
			shift_ellipsoid(*ellipsoid2, 0.0, -per_shift_y, 0.0);
		}
		if(inclusion -> face_is[Z] == NULL)
			per_shift_z = 0.0;
		else{
			if(ellipsoid2 -> center.z < tile.center.z)          /* ZN side */
				per_shift_z = tile.size.z;
			else                                                /* ZP side */
				per_shift_z = -tile.size.z;

			shift_ellipsoid(*ellipsoid2, 0.0, 0.0, per_shift_z);
			if(check_ellipsoid_ellipsoid_overlap(ellipsoid1, ellipsoid2, epsilon) == YES)
				error_message(GENERAL_ERROR, "Inclusion %ld (%d %d %d) intersects inclusion %ld (0 0 1)", i + 1, per_x, per_y, per_z, j + 1);
			shift_ellipsoid(*ellipsoid2, 0.0, 0.0, -per_shift_z);
		}
			
		if(per_shift_x != 0.0 && per_shift_y != 0.0){
			shift_ellipsoid(*ellipsoid2, per_shift_x, per_shift_y, 0.0);
			if(check_ellipsoid_ellipsoid_overlap(ellipsoid1, ellipsoid2, epsilon) == YES)
				error_message(GENERAL_ERROR, "Inclusion %ld (%d %d %d) intersects inclusion %ld (1 1 0)", i + 1, per_x, per_y, per_z, j + 1);
			shift_ellipsoid(*ellipsoid2, -per_shift_x, -per_shift_y, 0.0);
		}
		if(per_shift_y != 0.0 && per_shift_z != 0.0){
			shift_ellipsoid(*ellipsoid2, 0.0, per_shift_y, per_shift_z);
			if(check_ellipsoid_ellipsoid_overlap(ellipsoid1, ellipsoid2, epsilon) == YES)
				error_message(GENERAL_ERROR, "Inclusion %ld (%d %d %d) intersects inclusion %ld (0 1 1)", i + 1, per_x, per_y, per_z, j + 1);
			shift_ellipsoid(*ellipsoid2, 0.0, -per_shift_y, -per_shift_z);
		}
		if(per_shift_z != 0.0 && per_shift_x != 0.0){
			shift_ellipsoid(*ellipsoid2, per_shift_x, 0.0, per_shift_z);
			if(check_ellipsoid_ellipsoid_overlap(ellipsoid1, ellipsoid2, epsilon) == YES)
				error_message(GENERAL_ERROR, "Inclusion %ld (%d %d %d) intersects inclusion %ld (1 0 1)", i + 1, per_x, per_y, per_z, j + 1);
			shift_ellipsoid(*ellipsoid2, -per_shift_x, 0.0, -per_shift_z);
		}

		if(per_shift_x != 0.0 && per_shift_y != 0.0 && per_shift_z != 0.0){
			if(inclusion_array[j].corner == IN){
				shift_ellipsoid(*ellipsoid2, per_shift_x, per_shift_y, per_shift_z);
				if(check_ellipsoid_ellipsoid_overlap(ellipsoid1, ellipsoid2, epsilon) == YES)
					error_message(GENERAL_ERROR, "Inclusion %ld (%d %d %d) intersects inclusion %ld (1 1 1)", i + 1, per_x, per_y, per_z, j + 1);
				shift_ellipsoid(*ellipsoid2, -per_shift_x, -per_shift_y, -per_shift_z);
			}
		}
	}
}

#endif


/* ellipsoids greater/smaller by positive/negative eps */

static logic
check_ellipsoid_ellipsoid_overlap(ellipsoid_rec *ellipsoid1, ellipsoid_rec *ellipsoid2, double eps)
{
	point_rec pnt1, pnt2, p1, p2;
	double max, max1, max2;
	double dist, dist1, dist2, value;

#if 1
	vector_rec diff;
#else /* obsolete */
	double u1, uu1, uuu1, u2, uu2, uuu2, delta_u = EPSILON_U;
	double v1, vv1, vvv1, v2, vv2, vvv2, delta_v = EPSILON_V;
	int o1, oo1, ooo1, o2, oo2, ooo2;
#endif

	if(ellipsoid1 == ellipsoid2)return(NO);

	max1 = ellipsoid1 -> max + eps;
	max2 = ellipsoid2 -> max + eps;
	max = max1 + max2;

/* intersection of circumsribed spheres */
	dist = dist_point(ellipsoid1 -> center, ellipsoid2 -> center);
	if(dist > max * max)return(NO);

/* center of ellipsoid2 inside ellipsoid1 */
	if(check_point_inside_ellipsoid(&(ellipsoid2 -> center), ellipsoid1, eps, &value) == YES)return(YES);    /* accounts for eps */

/* center of ellipsoid1 inside ellipsoid2 */
	if(check_point_inside_ellipsoid(&(ellipsoid1 -> center), ellipsoid2, eps, &value) == YES)return(YES);    /* accounts for eps */

/* intersection of sphere circumscribed to ellipsoid1 with ellipsoid2 */
#if 1
	project_point_to_ellipsoid(&(ellipsoid1 -> center), ellipsoid2, &pnt2);                                  /* exact */
#else /* obsolete */
	o2 = project_point_to_ellipsoid_uv_octant(&(ellipsoid1 -> center), ellipsoid2, &pnt2, &u2, &v2);         /* exact */
#endif
	dist1 = dist_point(ellipsoid1 -> center, pnt2);
	if(dist1 > max1 * max1)return(NO);                                                                       /* accounts for eps */

/* closest point on ellipsoid2 to center of ellipsoid1 inside ellipsoid1 */
	if(check_point_inside_ellipsoid(&pnt2, ellipsoid1, eps, &value) == YES)return(YES);                      /* accounts for eps */

/* intersection of sphere circumscribed to ellipsoid2 with ellipsoid1 */
#if 1
	project_point_to_ellipsoid(&(ellipsoid2 -> center), ellipsoid1, &pnt1);                                  /* exact */
#else /* obsolete */
	o1 = project_point_to_ellipsoid_uv_octant(&(ellipsoid2 -> center), ellipsoid1, &pnt1, &u1, &v1);         /* exact */
#endif
	dist2 = dist_point(ellipsoid2 -> center, pnt1);
	if(dist2 > max2 * max2)return(NO);                                                                       /* accounts for eps */
	
/* closest point on ellipsoid1 to center of ellipsoid2 inside ellipsoid2 */
	if(check_point_inside_ellipsoid(&pnt1, ellipsoid2, eps, &value) == YES)return(YES);                      /* accounts for eps */

#if 1
/* find closest points starting from pnt1 */
	while(YES){
		copy_vec(diff, pnt1);
		project_point_to_ellipsoid(&pnt1, ellipsoid2, &p2);                                                    /* exact */
		project_point_to_ellipsoid(&p2, ellipsoid1, &pnt1);                                                    /* exact */
		sub_vec(diff, diff, pnt1);
		if(size_vec(diff) < EPSILON * EPSILON)break;
	}

/* find closest points starting from pnt2 */
	while(YES){
		copy_vec(diff, pnt2);
		project_point_to_ellipsoid(&pnt2, ellipsoid1, &p1);                                                    /* exact */
		project_point_to_ellipsoid(&p1, ellipsoid2, &pnt2);                                                    /* exact */
		sub_vec(diff, diff, pnt2);
		if(size_vec(diff) < EPSILON * EPSILON)break;
	}

	aver_point(pnt1, pnt1, p1);
	aver_point(pnt2, pnt2, p2);
#else /* obsolete */

/* the iterations should not oscillate between octants because the sphere tests have been already done */

/* find closest points starting from pnt1 */
	while(YES){
		ooo2 = project_point_to_ellipsoid_uv_octant(&pnt1, ellipsoid2, &p2, &uuu2, &vvv2);                     /* exact */
		oo1 = project_point_to_ellipsoid_uv_octant(&p2, ellipsoid1, &pnt1, &uu1, &vv1);                        /* exact */
		if(fabs(u1 - uu1) < delta_u && fabs(v1 - vv1) < delta_v)break;
		u1 = uu1;
		v1 = vv1;
	}

/* find closest points starting from pnt2 */
	while(YES){
		ooo1 = project_point_to_ellipsoid_uv_octant(&pnt2, ellipsoid1, &p1, &uuu1, &vvv1);                     /* exact */
		oo2 = project_point_to_ellipsoid_uv_octant(&p1, ellipsoid2, &pnt2, &uu2, &vv2);                        /* exact */
		if(fabs(u2 - uu2) < delta_u && fabs(v2 - vv2) < delta_v)break;
		u2 = uu2;
		v2 = vv2;
	}

/* there is a problem that intersections of ellipsoids may be in different octants;
	 in this case averaging of parameters makes no sense but it is clear that ellipsoids do intersect */

	if(oo1 != ooo1 || oo2 != ooo2)return(YES);

	u1 = (uu1 + uuu1) / 2.0;
	v1 = (vv1 + vvv1) / 2.0;
	u2 = (uu2 + uuu2) / 2.0;
	v2 = (vv2 + vvv2) / 2.0;
	get_ellipsoid_point_uv_octant(ellipsoid1, oo1, u1, v1, &pnt1);                                           /* exact */
	get_ellipsoid_point_uv_octant(ellipsoid2, oo2, u2, v2, &pnt2);                                           /* exact */
#endif

/* check pnt1 inside ellipsoid2 */
	if(check_point_inside_ellipsoid(&pnt1, ellipsoid2, eps, &value) == YES)return(YES);                      /* accounts for eps */
	
/* check pnt2 inside ellipsoid1 */
	if(check_point_inside_ellipsoid(&pnt2, ellipsoid1, eps, &value) == YES)return(YES);                      /* accounts for eps */

	return(NO);
} 


/* square of distance of non-overlapping ellipsoids */

static double
get_ellipsoid_ellipsoid_square_distance(ellipsoid_rec *ellipsoid1, ellipsoid_rec *ellipsoid2)
{
	double dist = 0.0;

	return(dist);
}


/* ellipsoid greater/smaller by positive/negative eps */

static logic
check_point_inside_ellipsoid(point_rec *point, ellipsoid_rec *ellipsoid, double eps, double *value)
{
	point_rec pnt;
	trans_matrix trans;
	double x_rate, y_rate, z_rate;

	copy_vec(trans.x, ellipsoid -> major);
	copy_vec(trans.y, ellipsoid -> minor);
	copy_vec(trans.z, ellipsoid -> middle);

	transform_point_from_global_to_local(point, &pnt, &(ellipsoid -> center), &trans);

	x_rate = pnt.x / (ellipsoid -> max + eps);
	y_rate = pnt.y / (ellipsoid -> min + eps);
	z_rate = pnt.z / (ellipsoid -> mid + eps);

	*value = x_rate * x_rate + y_rate * y_rate + z_rate * z_rate - 1.0;
	if(*value <= 0.0)return(YES);

	return(NO);
}



/* distance from the closest touching plane is calculated;
	 touching point is not returned as there may be two closest points */

static double
get_ellipsoid_plane_distance(ellipsoid_rec *ellipsoid, point_rec *point, vector_rec *normal)
{
	trans_matrix trans;
	vector_rec nor, vec, vector, vec_tmp;
	point_rec point_1, point_2, pnt_1 /*, pnt_2 */;
	double dist, dist_1, dist_2, val;

#ifdef DEBUG
	double size;

	size = sqrt(size_vec(*normal));
	if(fabs(size - 1.0) > EPSILON)
		error_message(GENERAL_ERROR, "Normal not normalized");
#endif

	copy_vec(trans.x, ellipsoid -> major);
	copy_vec(trans.y, ellipsoid -> minor);
	copy_vec(trans.z, ellipsoid -> middle);

	transform_vector_from_global_to_local(normal, &nor, &trans);

	invert_rotation_transformation(&trans);

	vec.x = ellipsoid -> max;
	vec.y = ellipsoid -> min;
	vec.z = ellipsoid -> mid;

	comp_product(vector, vec, vec);

	comp_product(vec_tmp, nor, vec);
	val = 1.0 / sqrt(size_vec(vec_tmp));
	copy_mul_vec(vec_tmp, vector, val);
	comp_product(pnt_1, nor, vec_tmp);

	transform_point_from_local_to_global(&pnt_1, &point_1, &(ellipsoid -> center), &trans);

#if 1 /* more efficient */
	sub_vec(vector, ellipsoid -> center, point_1);
	add_vec(point_2, ellipsoid -> center, vector);
#else
	copy_mul_vec(pnt_2, pnt_1, -1.0);
	transform_point_from_local_to_global(&pnt_2, &point_2, &(ellipsoid -> center), &trans);
#endif

	sub_vec(vec, *point, point_1);
	dist_1 = fabs(dot_product(vec, *normal));

	sub_vec(vec, *point, point_2);
	dist_2 = fabs(dot_product(vec, *normal));

	dist = min_val(dist_1, dist_2);
	return(dist);
}


/* distance from the closest touching line is calculated;
	 touching point is not returned as there may be infinite number of closest points */

#define MAXIMIZE_ONLY            /* always major axis is computed */
#define NEWTON                   /* use Newton method to compute der = 0.0 */
#define AREA_BASED_APPROACH      /* area defined by any two associated diameters is constant;
																		size of cross product of any two associated diameters is constant */
#ifndef AREA_BASED_APPROACH
//#define CIRCLE_BASED_APPROACH    /* tranformation to circle - less accurate approach unless higher precision for der is prescribed ==> not recommended */
#endif

static double
get_ellipsoid_line_square_distance(ellipsoid_rec *ellipsoid, point_rec *point, vector_rec *dir)
{
	point_rec pnt, pnt1, pnt2, proj;
	vector_rec normal, vec, vec0, vec1, vec2;
	ellipse_rec ellipse;
	double dist, angle, size, size2, t0, t1, t2;
	double d00, d11, d22, d01, d12, d20, B0, B1, B2, dB0, dB1, dB2, R0, R1, R2, dR0, dR1, dR2, sum_B, sum_dB, sum_B2;
	double der = 1.0, weight = SQRT1_2;

#ifdef NEWTON
	double ddB0, ddB1, ddB2, ddR0, ddR1, ddR2, sum_ddB, /* sum_B4, */ dder, d0, d1, d2;
	int count = 0;
#endif
#ifdef AREA_BASED_APPROACH   
	double area;
#else
#ifdef CIRCLE_BASED_APPROACH
	double u0, u2, v0, v2;
#else
	double tmp;
#endif
#endif
#if 0 /* obsolete */
	int oct;
	double t, u, v;
#endif

#ifdef DEBUG
	size = sqrt(size_vec(*dir));
	if(fabs(size - 1.0) > EPSILON)
		error_message(GENERAL_ERROR, "Direction not normalized");
#endif

	get_ellipsoid_line_center_intersection(ellipsoid, dir, &pnt1, &pnt2);
#if 1
	get_ellipsoid_normal_point(ellipsoid, &pnt1, &normal);
#else /* obsolete */
	oct = project_point_to_ellipsoid_uv_octant(&pnt1, ellipsoid, &pnt, &u, &v);
	get_ellipsoid_normal_uv_octant(ellipsoid, oct, u, v, &normal);
#endif
	get_ellipsoid_plane_intersection(ellipsoid, &(ellipsoid -> center), &normal, &ellipse);

#if 0 /* THIS IS WRONG */

/* find intersection of line with plane of ellipse;
	 get closest point on ellipse to that intersection;
	 get the distance and take from it component perpendicular to line */

	get_plane_line_intersection(&(ellipse.center), &ellipse.normal, point, dir, &pnt);
#if 1
	project_point_to_ellipse(&pnt, &ellipse, &proj);
#else /* obsolete */
	project_point_to_ellipse_t_quadrant(&pnt, &ellipse, &proj, &t, YES);           /* handle also special case !!! */
#endif

	sub_vec(vec, proj, pnt);
	size = dot_product(*dir, vec);
	dist = size_vec(vec);
	dist -= size * size;
#endif

/* project ellipse to plane perpendicular to line */

	dist = -dot_product(ellipse.major, *dir);
	copy_mul_vec(vec, *dir, dist);
	add_vec(ellipse.major, ellipse.major, vec);
	size = sqrt(size_vec(ellipse.major));
	div_vec(ellipse.major, size);
	ellipse.max *= size;

	dist = -dot_product(ellipse.minor, *dir);
	copy_mul_vec(vec, *dir, dist);
	add_vec(ellipse.minor, ellipse.minor, vec);
	size = sqrt(size_vec(ellipse.minor));
	div_vec(ellipse.minor, size);
	ellipse.min *= size;

/* IMPORTANT: projected axes are generally not principal axes any more;
	 they are axes of associated diameters;
	 this implies that standard projection routine cannot be applied;

	 it seems rather difficult to convert the ellipse given by associated diamteres to canonical form
	 because its quadratic form, spectral decomposition of which defines principal axes) is not known !!!

	 there exist set of three non-algebraic non-linear equation for alpha, a^2 and b^2 defining rotation of
	 one of associated diameters to principal axis and major and minor radii (one equation ensures that
	 vector of one associated diameter is perpendicular to ellipse tangent at other associated diameter, 
	 the other two equations ensures that both associated diameters are on the ellipse), but it is
	 not clear how to solve the set;

	 alternatively, approach minimizing x or y component of ellipse tangent can be applied;
	 this requires however to solve a cubic equation;
	 moreover, after finding the extreme point, coordinate system id updated so that x or y axis is passing through this point
	 and the process is repeated;

	 I am using approach maximising or minimising (depending on whether angle between associated diameters
	 is acute or obtuse) the distance of point on ellipse seqment from its center;
	 note that this task cannot be solved analytically because the obtained equation (derivation of the distance)
	 is rational and hard to solve;
	 therefore interval halving approach is adopted */

#if 0
	//copy_vec(ellipse.normal, *dir);          /* +dir or -dir ? orientation is not clear */
	cross_product(ellipse.normal, ellipse.major, ellipse.minor);
	
#ifdef PARAMETRIC_ROUTINES
#if 0 /* project ctrl points */
	sub_vec(vec, ellipse.pnt0, ellipse.center);
	dist = -dot_product(vec, *dir);
	copy_mul_vec(vec, *dir, dist);
	add_vec(ellipse.pnt0, ellipse.pnt0, vec);
	
	sub_vec(vec, ellipse.pnt1, ellipse.center);
	dist = -dot_product(vec, *dir);
	copy_mul_vec(vec, *dir, dist);
	add_vec(ellipse.pnt1, ellipse.pnt1, vec);
	
	sub_vec(vec, ellipse.pnt2, ellipse.center);
	dist = -dot_product(vec, *dir);
	copy_mul_vec(vec, *dir, dist);
	add_vec(ellipse.pnt2, ellipse.pnt2, vec);
#else /* recalculate ctrl points */
	copy_mul_vec(vec0, ellipse.major, ellipse.max);
	copy_mul_vec(vec2, ellipse.minor, ellipse.min);
	add_vec(vec1, vec0, vec2);

	add_vec(ellipse.pnt0, ellipse.center, vec0);
	add_vec(ellipse.pnt1, ellipse.center, vec1);
	add_vec(ellipse.pnt2, ellipse.center, vec2);
#endif
#endif
#endif

	angle = dot_product(ellipse.major, ellipse.minor);
	if(fabs(angle) > 1.0e-10){

/* if angle > 0.0 ==> maximize distance
	 if angle < 0.0 ==> minimize distance */

#ifdef MAXIMIZE_ONLY
		if(angle < 0.0){

/* swap minor to enforce maximization */

			mul_vec(ellipse.minor, -1.0);
			angle *= -1.0;
		}
#endif

		copy_mul_vec(vec0, ellipse.major, ellipse.max);
		copy_mul_vec(vec2, ellipse.minor, ellipse.min);
		add_vec(vec1, vec0, vec2);

		d00 = dot_product(vec0, vec0);
		d11 = dot_product(vec1, vec1);
		d22 = dot_product(vec2, vec2);

		d01 = dot_product(vec0, vec1);
		d12 = dot_product(vec1, vec2);
		d20 = dot_product(vec2, vec0);

		t0 = 0.0;
		t2 = 1.0;

#ifdef NEWTON
		ddB0 = 2.0;
		ddB1 = -4.0 * weight;
		ddB2 = 2.0;

/* search for appropriate starting point;
	 note that der is generally not monotonic over <0;1>;
	 this implies that end points are generally not suitable as starting points (see graph_1.dat);

	 feasible region for starting point is interval with negative/positive dder when maximizing/minimizing the distance;
	 the problem is however, that dt = der/dder may shift t outside of this feasible region;
	 note that it is dangerous to search for point at which der is almost zero, as such starting point may converge
	 to maximum or minimum outside interval (for example when associated diameters are almost perpendicular - see graph_2.dat);
	 thus I am searching directly for almost maximal/minimal distance */

		d0 = d00;
		d2 = d22;

		while(++count){
			t1 = (t0 + t2) / 2.0;
			if(count == 4)break;    /* count >= 3 */

			B0 = B_0(t1);
			B1 = B_1(t1) * weight;
			B2 = B_2(t1);
			
			dB0 = dB_0(t1);
			dB1 = dB_1(t1) * weight;
			dB2 = dB_2(t1);

			sum_B = B0 + B1 + B2;
			
			R0 = B0 / sum_B;
			R1 = B1 / sum_B;
			R2 = B2 / sum_B;

			d1 = R0 * R0 * d00 + R1 * R1 * d11 + R2 * R2 * d22 + 2.0 * (R0 * R1 * d01 + R1 * R2 * d12 + R2 * R0 * d20);

#ifndef MAXIMIZE_ONLY
			if(angle > 0.0){
#endif
				if(d1 > d0 && d1 > d2){
					break;
				}
				else{
					if(d1 > d0){
						t0 = t1;
						d0 = d1;
					}
					else{
						t2 = t1;
						d2 = d1;
					}
				}
#ifndef MAXIMIZE_ONLY
			}
			else{
				if(d1 < d0 && d1 < d2){
					break;
				}
				else{
					if(d1 < d0){
						t0 = t1;
						d0 = d1;
					}
					else{
						t2 = t1;
						d2 = d1;
					}
				}
			}
#endif
		}

		while(fabs(der) > 1.0e-10){
			B0 = B_0(t1);
			B1 = B_1(t1) * weight;
			B2 = B_2(t1);
			
			dB0 = dB_0(t1);
			dB1 = dB_1(t1) * weight;
			dB2 = dB_2(t1);

			sum_B = B0 + B1 + B2;
			sum_dB = dB0 + dB1 + dB2;
			sum_ddB = ddB0 + ddB1 + ddB2;
			
			sum_B2 = sum_B * sum_B;
			
			R0 = B0 / sum_B;
			R1 = B1 / sum_B;
			R2 = B2 / sum_B;

			dR0 = (dB0 * sum_B - B0 * sum_dB) / sum_B2;
			dR1 = (dB1 * sum_B - B1 * sum_dB) / sum_B2;
			dR2 = (dB2 * sum_B - B2 * sum_dB) / sum_B2;

#if 0 /* less efficient */
			sum_B4 = sum_B2 * sum_B2;

			ddR0 = ((ddB0 * sum_B - B0 * sum_ddB) * sum_B2 - 2.0 * (dB0 * sum_B - B0 * sum_dB) * sum_B * sum_dB) / sum_B4;
			ddR1 = ((ddB1 * sum_B - B1 * sum_ddB) * sum_B2 - 2.0 * (dB1 * sum_B - B1 * sum_dB) * sum_B * sum_dB) / sum_B4;
			ddR2 = ((ddB2 * sum_B - B2 * sum_ddB) * sum_B2 - 2.0 * (dB2 * sum_B - B2 * sum_dB) * sum_B * sum_dB) / sum_B4;
#else /* more efficient */
			ddR0 = ((ddB0 * sum_B - B0 * sum_ddB) - 2.0 * dR0 * sum_B * sum_dB) / sum_B2;
			ddR1 = ((ddB1 * sum_B - B1 * sum_ddB) - 2.0 * dR1 * sum_B * sum_dB) / sum_B2;
			ddR2 = ((ddB2 * sum_B - B2 * sum_ddB) - 2.0 * dR2 * sum_B * sum_dB) / sum_B2;
#endif

/* compute 1st and 2nd derivative of square of distance of point at t1 from ellipse center */			

			der = 2.0 * (R0 * dR0 * d00 + R1 * dR1 * d11 + R2 * dR2 * d22 + (R0 * dR1 + dR0 * R1) * d01 + (R1 * dR2 + dR1 * R2) * d12 + (R2 * dR0 + dR2 * R0) * d20);
			dder = 2.0 * ((dR0 * dR0 + R0 * ddR0) * d00 + (dR1 * dR1 + R1 * ddR1) * d11 + (dR2 * dR2 + R2 * ddR2) * d22 + (2.0 * dR0 * dR1 + R0 * ddR1 + ddR0 * R1) * d01 + (2.0 * dR1 * dR2 + R1 * ddR2 + ddR1 * R2) * d12 + (2.0 * dR2 * dR0 + R2 * ddR0 + ddR2 * R0) * d20);

			t1 -= der/dder;
			if(t1 < 0.0)
				t1 = 0.0;
			else if(t1 > 1.0)
				t1 = 1.0;
		}
#else /* NEWTON */
		while(fabs(der) > 1.0e-10){
			t1 = (t0 + t2) / 2.0;

			B0 = B_0(t1);
			B1 = B_1(t1) * weight;
			B2 = B_2(t1);
			
			dB0 = dB_0(t1);
			dB1 = dB_1(t1) * weight;
			dB2 = dB_2(t1);

			sum_B = B0 + B1 + B2;
			sum_dB = dB0 + dB1 + dB2;
			
			sum_B2 = sum_B * sum_B;

			R0 = B0 / sum_B;
			R1 = B1 / sum_B;
			R2 = B2 / sum_B;
			
			dR0 = (dB0 * sum_B - B0 * sum_dB) / sum_B2;
			dR1 = (dB1 * sum_B - B1 * sum_dB) / sum_B2;
			dR2 = (dB2 * sum_B - B2 * sum_dB) / sum_B2;
			
/* compute derivative of square of distance of point at t1 from ellipse center */			

			der = 2.0 * (R0 * dR0 * d00 + R1 * dR1 * d11 + R2 * dR2 * d22 + (R0 * dR1 + dR0 * R1) * d01 + (R1 * dR2 + dR1 * R2) * d12 + (R2 * dR0 + dR2 * R0) * d20);

#ifndef MAXIMIZE_ONLY
			if(angle > 0.0){
#endif
				if(der > 0.0)
					t0 = t1;
				else
					t2 = t1;
#ifndef MAXIMIZE_ONLY
			}
			else{
				if(der < 0.0)
					t0 = t1;
				else
					t2 = t1;
			}
#endif
		}
#endif /* NEWTON */

		copy_vec(ellipse.normal, *dir);

#ifdef AREA_BASED_APPROACH
#if 0 /* less efficient */
		cross_product(vec, vec0, vec2);
		area = sqrt(size_vec(vec));
#else /* more efficient */
		area = sqrt(d00 * d22 - d20 * d20);
#endif
#endif

#ifndef MAXIMIZE_ONLY
		if(angle > 0.0){
#endif
			ellipse.major.x = R0 * vec0.x + R1 * vec1.x + R2 * vec2.x;
			ellipse.major.y = R0 * vec0.y + R1 * vec1.y + R2 * vec2.y;
			ellipse.major.z = R0 * vec0.z + R1 * vec1.z + R2 * vec2.z;

#if 1 /* no difference in accuracy */ /* cheaper */
			size2 = size_vec(ellipse.major);
#else                                 /* more expensive */
			size2 = R0 * R0 * d00 + R1 * R1 * d11 + R2 * R2 * d22 + 2.0 * (R0 * R1 * d01 + R1 * R2 * d12 + R2 * R0 * d20);
#endif

			size = sqrt(size2);
			div_vec(ellipse.major, size);
			ellipse.max = size;
			
#ifdef AREA_BASED_APPROACH
			cross_product(ellipse.minor, *dir, ellipse.major);

#if 0 /* no need to normalize */		
			tmp = sqrt(size_vec(ellipse.minor));
			div_vec(ellipse.minor, tmp);
#endif

			ellipse.min = area / ellipse.max;
#else /* AREA_BASED_APPROACH */
#ifdef CIRCLE_BASED_APPROACH /* less accurate */
			cross_product(ellipse.minor, *dir, ellipse.major);
			
#if 0 /* no need to normalize */		
			tmp = sqrt(size_vec(ellipse.minor));
			div_vec(ellipse.minor, tmp);
#endif

			u0 = dot_product(vec0, ellipse.major);
			u2 = dot_product(vec2, ellipse.major);

/* avoid problem with zero division or negative sqrt */

			if(fabs(u0) < fabs(u2)){
				u = u0;
				v = dot_product(vec0, ellipse.minor);
			}
			else{
				u = u2;
				v = dot_product(vec2, ellipse.minor);
			}

			ellipse.min = size * fabs(v) / sqrt(size2 - u * u);
#else /* CIRCLE_BASED_APPROACH *//* more accurate */
			t1 = 1.0 - t1;
			mul_vec(ellipse.minor, -1.0);

			//copy_mul_vec(vec0, ellipse.major, ellipse.max);    /* does not change */
			copy_mul_vec(vec2, ellipse.minor, ellipse.min);
			add_vec(vec1, vec0, vec2);

			//d00 = dot_product(vec0, vec0);    /* does not change */
			d11 = dot_product(vec1, vec1);
			//d22 = dot_product(vec2, vec2);    /* does not change */

			d01 = dot_product(vec0, vec1);
			d12 = dot_product(vec1, vec2);
			//d20 = dot_product(vec2, vec0);    /* changes sign */
			d20 = -d20;

/* B0(1-t)=B2(t)
	 B1(1-t)=B1(t) ==> swap B0 and B2 ==> swap R0 and R2
	 B2(1-t)=B0(t) */

			swap(R0, R2, tmp);
			size2 = R0 * R0 * d00 + R1 * R1 * d11 + R2 * R2 * d22 + 2.0 * (R0 * R1 * d01 + R1 * R2 * d12 + R2 * R0 * d20);

			cross_product(ellipse.minor, *dir, ellipse.major);
			ellipse.min = sqrt(size2);
#endif /* CIRCLE_BASED_APPROACH */
#endif /* AREA_BASED_APPROACH */
#ifndef MAXIMIZE_ONLY
		}
		else{
			ellipse.minor.x = R0 * vec0.x + R1 * vec1.x + R2 * vec2.x;
			ellipse.minor.y = R0 * vec0.y + R1 * vec1.y + R2 * vec2.y;
			ellipse.minor.z = R0 * vec0.z + R1 * vec1.z + R2 * vec2.z;
			
#if 1 /* no difference in accuracy */	/* cheaper */
			size2 = size_vec(ellipse.minor);
#else                                 /* more expensive */
			size2 = R0 * R0 * d00 + R1 * R1 * d11 + R2 * R2 * d22 + 2.0 * (R0 * R1 * d01 + R1 * R2 * d12 + R2 * R0 * d20);
#endif

			size = sqrt(size2);
			div_vec(ellipse.minor, size);
			ellipse.min = size;

#ifdef AREA_BASED_APPROACH
			cross_product(ellipse.major, ellipse.minor, *dir);

#if 0 /* no need to normalize */	
			tmp = sqrt(size_vec(ellipse.major));
			div_vec(ellipse.major, tmp);
#endif

			ellipse.max = area / ellipse.min;
#else /* AREA_BASED_APPROACH */
#ifdef CIRCLE_BASED_APPROACH /* less accurate */
			cross_product(ellipse.major, ellipse.minor, *dir);
			
#if 0 /* no need to normalize */	
			tmp = sqrt(size_vec(ellipse.major));
			div_vec(ellipse.major, tmp);
#endif

			v0 = dot_product(vec0, ellipse.minor);
			v2 = dot_product(vec2, ellipse.minor);

/* avoid problem with zero division or negative sqrt */

			if(fabs(v0) < fabs(v2)){
				u = dot_product(vec0, ellipse.major);
				v = v0;
			}
			else{
				u = dot_product(vec2, ellipse.major);
				v = v2;
			}
			
			ellipse.max = size * fabs(u) / sqrt(size2 - v * v);
#else /* CIRCLE_BASED_APPROACH *//* more accurate */
			t1 = 1.0 - t1;
			mul_vec(ellipse.major, -1.0);
			
			copy_mul_vec(vec0, ellipse.major, ellipse.max);
			//copy_mul_vec(vec2, ellipse.minor, ellipse.min);    /* does not change */
			add_vec(vec1, vec0, vec2);

			//d00 = dot_product(vec0, vec0);    /* does not change */
			d11 = dot_product(vec1, vec1);
			//d22 = dot_product(vec2, vec2);    /* does not change */

			d01 = dot_product(vec0, vec1);
			d12 = dot_product(vec1, vec2);
			//d20 = dot_product(vec2, vec0);    /* changes sign */
			d20 = -d20;

/* B0(1-t)=B2(t)
	 B1(1-t)=B1(t) ==> swap B0 and B2 ==> swap R0 and R2
	 B2(1-t)=B0(t) */

			swap(R0, R2, tmp);
			size2 = R0 * R0 * d00 + R1 * R1 * d11 + R2 * R2 * d22 + 2.0 * (R0 * R1 * d01 + R1 * R2 * d12 + R2 * R0 * d20);

			cross_product(ellipse.major, ellipse.minor, *dir);
			ellipse.max = sqrt(size2);
#endif /* CIRCLE_BASED_APPROACH */
#endif /* AREA_BASED_APPROACH */

		}
#endif /* MAXIMIZE_ONLY */
	}
	else{
		//copy_vec(ellipse.normal, *dir);          /* +dir or -dir ? orientation is not clear */
		cross_product(ellipse.normal, ellipse.major, ellipse.minor);
	}

	if(ellipse.max < ellipse.min){
		if(ellipse.min - ellipse.max > EPSILON)error_message(GENERAL_ERROR, "Unexpected situation");
		ellipse.max = ellipse.min = (ellipse.max + ellipse.min) / 2.0;
	}

#ifdef PARAMETRIC_ROUTINES

/* recalculate ctrl points */

	copy_mul_vec(vec0, ellipse.major, ellipse.max);
	copy_mul_vec(vec2, ellipse.minor, ellipse.min);
	add_vec(vec1, vec0, vec2);

	add_vec(ellipse.pnt0, ellipse.center, vec0);
	add_vec(ellipse.pnt1, ellipse.center, vec1);
	add_vec(ellipse.pnt2, ellipse.center, vec2);
#endif

#ifdef DRAW_ELLIPSE
	draw_ellipse(&ellipse, 0, blue);
#endif

	get_plane_line_intersection(&(ellipse.center), dir, point, dir, &pnt);
#if 1
	project_point_to_ellipse(&pnt, &ellipse, &proj);
#else /* obsolete */
	project_point_to_ellipse_t_quadrant(&pnt, &ellipse, &proj, &t, YES);           /* handle also special case !!! */
#endif
	dist = dist_point(pnt, proj);

	return(dist);
}


static logic
get_ellipsoid_plane_intersection(ellipsoid_rec *ellipsoid, point_rec *point, vector_rec *normal, ellipse_rec *ellipse)
{
	trans_matrix trans_g2l, trans_l2g;
	vector_rec scale_e2s, scale_s2e;
	vector_rec nor, vec_1, vec_2, vec_a, vec_b;
	point_rec pnt, center;
	double max, min, dist, size, radius, rad, val, nom, den, alpha, ca, sa, tmp, sz_a, sz_b;

#ifdef DEBUG
	size = sqrt(size_vec(*normal));
	if(fabs(size - 1.0) > EPSILON)
		error_message(GENERAL_ERROR, "Normal not normalized");
#endif

	copy_vec(trans_g2l.x, ellipsoid -> major);
	copy_vec(trans_g2l.y, ellipsoid -> minor);
	copy_vec(trans_g2l.z, ellipsoid -> middle);

	get_inverted_rotation_transformation(&trans_g2l, &trans_l2g);

	radius = ellipsoid -> max;     /* sphere radius */
	//radius = ellipsoid -> min;     /* sphere radius */
	//radius = ellipsoid -> mid;     /* sphere radius */

	scale_e2s.x = radius / ellipsoid -> max;
	scale_e2s.y = radius / ellipsoid -> min;
	scale_e2s.z = radius / ellipsoid -> mid;

	get_inverted_scale_transformation(&scale_e2s, &scale_s2e);

	transform_vector_from_global_to_local(normal, &nor, &trans_g2l);
	transform_point_from_global_to_local(point, &pnt, &(ellipsoid -> center), &trans_g2l);

	transform_local_from_ellipsoid_to_sphere(&pnt, &pnt, &scale_e2s);

/* construct two perpendicular vectors so that vec_1 x vec_2 = nor */

	max = 0.0;
	min = 1.0;
	val = fabs(nor.x);
	if(val > max)max = val;
	if(val < min)min = val;

	if((max - min) / max < 0.00001){    /* all components of the same size (actually, I do mind only if all components are the same) */
		vec_1.x = -nor.z;
		vec_1.y = 0.0;
		vec_1.z = nor.x;
	}
	else{
		vec_2.x = nor.y;
		vec_2.y = nor.z;
		vec_2.z = nor.x;
		cross_product(vec_1, vec_2, nor);
	}

	size = sqrt(size_vec(vec_1));
	div_vec(vec_1, size);

	cross_product(vec_2, nor, vec_1);

/* transform these vectors from ellipsoid to sphere */

	transform_local_from_ellipsoid_to_sphere(&vec_1, &vec_1, &scale_e2s);
	transform_local_from_ellipsoid_to_sphere(&vec_2, &vec_2, &scale_e2s);

	size = sqrt(size_vec(vec_1));
	div_vec(vec_1, size);

	size = sqrt(size_vec(vec_2));
	div_vec(vec_2, size);
	
/* compute plane normal in space of sphere */

	cross_product(nor, vec_1, vec_2);

	size = sqrt(size_vec(nor));
	div_vec(nor, size);

/* ensure vec_1 x vec_2 = nor in space of sphere */

	cross_product(vec_2, nor, vec_1);

/* resolve intersection with plane in space of sphere ==> circle */

	dist = dot_product(nor, pnt);
	if(fabs(dist) > radius)return(NO);

	if((radius - fabs(dist)) / radius < EPSILON){
		if(dist > 0.0)
			dist = radius;
		else
			dist = -radius;

		copy_mul_vec(center, nor, dist);               /* circle center */

/* transform circle from space of sphere to space of ellipsoid	*/

		transform_local_from_sphere_to_ellipsoid(&center, &center, &scale_s2e);

/* transform from local to global */

		transform_point_from_local_to_global(&center, &(ellipse -> center), &(ellipsoid -> center), &trans_l2g);

		setup_ellipse_touching_point(ellipse, &(ellipse -> center), normal);

#ifdef DRAW_TOUCHING_POINT_ELLIPSOID
		draw_point(&(ellipse -> center), 0, orange);
#endif
	}
	else{
		copy_mul_vec(center, nor, dist);               /* circle center */
		rad = sqrt(radius * radius - dist * dist);     /* circle radius */

/* search for two perpendicular diameters of circle which remain mutually perpendicular after s2e transformation */

		comp_product(vec_a, vec_1, scale_s2e);
		comp_product(vec_b, vec_2, scale_s2e);
	
		nom = 2.0 * dot_product(vec_a, vec_b);
		//den = dot_product(vec_a, vec_a) - dot_product(vec_b, vec_b);
		den = size_vec(vec_a) - size_vec(vec_b);
		if(den != 0.0){
			
/* if nom/den positively (negatively) overflows atan should return PI/2 (-PI/2) */

			alpha = atan(nom/den) / 2.0;
			ca = cos(alpha) * rad;
			sa = sin(alpha) * rad;
		}
		else{
			if(nom != 0.0){
				//alpha = PI_4;
				ca = sa = SQRT1_2 * rad;
			}
			else{
				
/* nom = den = 0.0 (nom/den = NaN) ==> ellipse is circle ==> alpha is arbitrary */

				//alpha = 0.0;
				ca = rad;
				sa = 0.0;
			}
		}
		
		add_mul_vec(vec_a, vec_1, vec_2, ca, sa);
		add_mul_vec(vec_b, vec_1, vec_2, -sa, ca); 

/* transform circle from space of sphere to space of ellipsoid	*/

		transform_local_from_sphere_to_ellipsoid(&center, &center, &scale_s2e);
		transform_local_from_sphere_to_ellipsoid(&vec_a, &vec_a, &scale_s2e);
		transform_local_from_sphere_to_ellipsoid(&vec_b, &vec_b, &scale_s2e);
		
		sz_a = size_vec(vec_a);
		sz_b = size_vec(vec_b);
		
		if(sz_a < sz_b){
			swap(vec_a.x, vec_b.x, tmp);
			swap(vec_a.y, vec_b.y, tmp);
			swap(vec_a.z, vec_b.z, tmp);
			
			mul_vec(vec_b, -1.0);              /* ensure that vec_a, vec_b and nor is righthanded CS */
		}

/* transform from local to global */

		transform_point_from_local_to_global(&center, &(ellipse -> center), &(ellipsoid -> center), &trans_l2g);
		transform_vector_from_local_to_global(&vec_a, &(ellipse -> major), &trans_l2g);
		transform_vector_from_local_to_global(&vec_b, &(ellipse -> minor), &trans_l2g);
		
#ifdef PARAMETRIC_ROUTINES
		add_vec(ellipse -> pnt0, ellipse -> center, ellipse -> major);
		add_vec(ellipse -> pnt2, ellipse -> center, ellipse -> minor);
		add_vec(ellipse -> pnt1, ellipse -> pnt0, ellipse -> minor);
#endif

		ellipse -> max = sqrt(size_vec(ellipse -> major));
		ellipse -> min = sqrt(size_vec(ellipse -> minor));

		if(ellipse -> max < ellipse -> min){
			if(ellipse -> min - ellipse -> max > EPSILON)error_message(GENERAL_ERROR, "Unexpected situation");
			ellipse -> max = ellipse -> min = (ellipse -> max + ellipse -> min) / 2.0;
		}

		div_vec(ellipse -> major, ellipse -> max);
		div_vec(ellipse -> minor, ellipse -> min);
		copy_vec(ellipse -> normal, *normal);

#ifdef DRAW_ELLIPSE
		draw_ellipse(ellipse, 0, yellow);
#endif
	}

	return(YES);
}


static int
get_ellipsoid_line_intersection(ellipsoid_rec *ellipsoid, point_rec *point, vector_rec *dir, point_rec *intersection)
{
	point_rec pnt, pnt1, pnt2;
	vector_rec direction;
	trans_matrix trans;
	double a0, a1, a2, a0x, a0y, a0z, a1x, a1y, a1z, a2x, a2y, a2z;
	double D, max, min, mid, t, t1, t2, norm;

	norm = size_vec(*dir);
	if(norm < EPSILON * EPSILON)error_message(GENERAL_ERROR, "Zero length direction");

	copy_vec(trans.x, ellipsoid -> major);
	copy_vec(trans.y, ellipsoid -> minor);
	copy_vec(trans.z, ellipsoid -> middle);

	transform_vector_from_global_to_local(dir, &direction, &trans);
	transform_point_from_global_to_local(point, &pnt, &(ellipsoid -> center), &trans);

	max = ellipsoid -> max;
	min = ellipsoid -> min;
	mid = ellipsoid -> mid;

	a2x = direction.x / max;
	a2y = direction.y / min;
	a2z = direction.z / mid;
	a2 = a2x * a2x + a2y * a2y + a2z * a2z;

	a0x = pnt.x / max;
	a0y = pnt.y / min;
	a0z = pnt.z / mid;
	a0 = a0x * a0x + a0y *a0y + a0z * a0z - 1.0;

	a1x = a2x * a0x;
	a1y = a2y * a0y;
	a1z = a2z * a0z;
	a1 = a1x + a1y + a1z;

	D = a1 * a1 - a0 * a2;
	
	if(D < 0.0)return(0);

	invert_rotation_transformation(&trans);

	if(D == 0.0){
		t = -a1 / a2;

		pnt.x += t * direction.x;
		pnt.y += t * direction.y;
		pnt.z += t * direction.z;

		transform_point_from_local_to_global(&pnt, &intersection[0], &(ellipsoid -> center), &trans);
		return(1);
	}

	D = sqrt(D);

	t1 = -(a1 - D) / a2;
	t2 = -(a1 + D) / a2;

	pnt1.x = pnt.x + t1 * direction.x;
	pnt1.y = pnt.y + t1 * direction.y;
	pnt1.z = pnt.z + t1 * direction.z;

	transform_point_from_local_to_global(&pnt1, &intersection[0], &(ellipsoid -> center), &trans);

	pnt2.x = pnt.x + t2 * direction.x;
	pnt2.y = pnt.y + t2 * direction.y;
	pnt2.z = pnt.z + t2 * direction.z;

	transform_point_from_local_to_global(&pnt2, &intersection[1], &(ellipsoid -> center), &trans);

	return(2);
}



static void
get_ellipsoid_line_center_intersection(ellipsoid_rec *ellipsoid, vector_rec *dir, point_rec *pnt1, point_rec *pnt2)
{
	trans_matrix trans;
	vector_rec direction;
	double a, b, c, t, norm;

	norm = size_vec(*dir);
	if(norm < EPSILON * EPSILON)error_message(GENERAL_ERROR, "Zero length direction");

	copy_vec(trans.x, ellipsoid -> major);
	copy_vec(trans.y, ellipsoid -> minor);
	copy_vec(trans.z, ellipsoid -> middle);

	transform_vector_from_global_to_local(dir, &direction, &trans);

	a = direction.x / ellipsoid -> max;
	b = direction.y / ellipsoid -> min;
	c = direction.z / ellipsoid -> mid;

	t = sqrt(1.0 / (a * a + b * b + c * c));
	
	invert_rotation_transformation(&trans);

	mul_vec(direction, t);
	transform_point_from_local_to_global(&direction, pnt1, &(ellipsoid -> center), &trans);

	mul_vec(direction, -1.0);
	transform_point_from_local_to_global(&direction, pnt2, &(ellipsoid -> center), &trans);
}


static logic
get_plane_plane_intersection(point_rec *pnt1, vector_rec *normal1, point_rec *pnt2, vector_rec *normal2, 
														 point_rec *point, vector_rec *dir)
{
	vector_rec vec2;
	double product, norm;

#ifdef DEBUG
	double size;

	size = sqrt(size_vec(*normal1));
	if(fabs(size - 1.0) > EPSILON)
		error_message(GENERAL_ERROR, "Normal1 not normalized");
	size = sqrt(size_vec(*normal2));
	if(fabs(size - 1.0) > EPSILON)
		error_message(GENERAL_ERROR, "Normal2 not normalized");
#endif

	product = dot_product(*normal1, *normal2);
	if(fabs(fabs(product) - 1.0) < EPSILON)return(NO);
	
	cross_product(*dir, *normal1, *normal2);
	norm = sqrt(size_vec(*dir));
	div_vec(*dir, norm);
	
	cross_product(vec2, *normal2, *dir);
	get_plane_line_intersection(pnt1, normal1, pnt2, &vec2, point);

	return(YES);
}


static logic
get_plane_line_intersection(point_rec *point, vector_rec *normal, point_rec *pnt, vector_rec *dir, point_rec *intersection)
{
	vector_rec vec;
	double product, t;
	
#ifdef DEBUG
	double size;

	size = sqrt(size_vec(*normal));
	if(fabs(size - 1.0) > EPSILON)
		error_message(GENERAL_ERROR, "Normal not normalized");
	size = sqrt(size_vec(*dir));
	if(fabs(size - 1.0) > EPSILON)
		error_message(GENERAL_ERROR, "Direction not normalized");
#endif

	product = dot_product(*normal, *dir);
	if(fabs(product) < EPSILON)return(NO);
	
	//t = (dot_product(*normal, *point) - dot_product(*normal, *pnt)) / product;
	sub_vec(vec, *point, *pnt);
	t = dot_product(*normal, vec) / product;
	
	intersection -> x = pnt -> x + t * dir -> x;
	intersection -> y = pnt -> y + t * dir -> y;
	intersection -> z = pnt -> z + t * dir -> z;

	return(YES);
}


static double
project_point_to_plane(point_rec *pnt, point_rec *point, vector_rec *normal, point_rec *projection)
{
	vector_rec vec;
	double product;
	
#ifdef DEBUG
	double size;

	size = sqrt(size_vec(*normal));
	if(fabs(size - 1.0) > EPSILON)
		error_message(GENERAL_ERROR, "Normal not normalized");
#endif

	sub_vec(vec, *point, *pnt);
	product = dot_product(*normal, vec);

	projection -> x = pnt -> x + product * normal -> x;
	projection -> y = pnt -> y + product * normal -> y;
	projection -> z = pnt -> z + product * normal -> z;

	return(product);
}


static double
project_point_to_line(point_rec *pnt, point_rec *point, vector_rec *dir, point_rec *projection)
{
	vector_rec vec;
	double product, dist;

#ifdef DEBUG
	double size;

	size = sqrt(size_vec(*dir));
	if(fabs(size - 1.0) > EPSILON)
		error_message(GENERAL_ERROR, "Direction not normalized");
#endif

	sub_vec(vec, *pnt, *point);
	product = dot_product(*dir, vec);
	
	projection -> x = point -> x + product * dir -> x;
	projection -> y = point -> y + product * dir -> y;
	projection -> z = point -> z + product * dir -> z;

	dist = dist_point(*pnt, *projection);
	return(sqrt(dist));
}


static void
project_point_to_ellipsoid(point_rec *point, ellipsoid_rec *ellipsoid, point_rec *pnt)
{
	point_rec p, pp;
	vector_rec vec, grad;
	trans_matrix trans;
	double max, min, mid, max2, min2, mid2, max4, min4, mid4, xp2, yp2, zp2;
	double val, rate, k, dk = 1.0, fk, dfk;
	double a, b, c, d, e, f, g;
	logic skip = NO;
	int iter = 0;

	copy_vec(trans.x, ellipsoid -> major);
	copy_vec(trans.y, ellipsoid -> minor);
	copy_vec(trans.z, ellipsoid -> middle);

	transform_point_from_global_to_local(point, &p, &(ellipsoid -> center), &trans);

	xp2 = p.x * p.x;
	yp2 = p.y * p.y;
	zp2 = p.z * p.z;

	max = ellipsoid -> max;
	min = ellipsoid -> min;
	mid = ellipsoid -> mid;

#ifdef DEBUG
	if(max < min || max < mid)error_message(GENERAL_ERROR, "Unexpected situation");		
	if(min > max || min > mid)error_message(GENERAL_ERROR, "Unexpected situation");		
#endif

	max2 = max * max;
	min2 = min * min;
	mid2 = mid * mid;

	max4 = max2 * max2;
	min4 = min2 * min2;
	mid4 = mid2 * mid2;

	a = 64.0;
	b = 64.0 * (max2 + min2 + mid2);
	c = 16.0 * (max4 + min4 + mid4 - max2 * xp2 - min2 * yp2 - mid2 * zp2) + 64.0 * (max2 * mid2 + min2 * mid2 + max2 * min2);
	d = 16.0 * (max4 * (min2 + mid2) + min4 * (max2 + mid2) + mid4 * (max2 + min2) - xp2 * max2 * (mid2 + min2) - yp2 * min2 * (max2 + mid2) - zp2 * mid2 * (max2 + min2)) + 64.0 * max2 * min2 * mid2;
	e = 4.0 * (max4 * min4 + max4 * mid4 + min4 * mid4 - xp2 * max2 * (min4 + mid4) - yp2 * min2 * (max4 + mid4) - zp2 * mid2 * (max4 + min4)) + 16.0 * max2 * min2 * mid2 * (max2 + min2 + mid2 - xp2 - yp2 - zp2); 
	f = 4.0 * max2 * min2 * mid2 * (max2 * mid2 + min2 * mid2 + max2 * min2 - xp2 * (min2 + mid2) - yp2 * (max2 + mid2) - zp2 * (max2 + min2));
	g = max4 * min4 * mid4 - max2 * min2 * mid2 * (xp2 * min2 * mid2 + yp2 * max2 * mid2 + zp2 * max2 * min2);

/* get radial projection */

	val = xp2 / max2 + yp2 / min2 + zp2 / mid2;

	if(val < 1.0e-10){       /* point is in center */
		rate = 0.0;
		k = -min2 / 2.0;
		skip = YES;
	}
	else{
		rate = sqrt(1.0 / val);
		if(rate <= 1.0){       /* point is on or outside ellipsoid */
			copy_mul_vec(pp, p, rate);

			grad.x = 2.0 * pp.x / max2;
			grad.y = 2.0 * pp.y / min2;
			grad.z = 2.0 * pp.z / mid2;

			sub_vec(vec, pp, p);

			k = sqrt(size_vec(vec) / size_vec(grad));
		}
		else{                  /* point is inside ellipsoid */
			k = 0.0;

#ifdef DEBUG
			if(max < min || max < mid)error_message(GENERAL_ERROR, "Unexpected situation");		
#endif

			if(p.y == 0.0 && p.z == 0.0){          /* point is on major axis */
				skip = YES;
				if(abs(p.x) <= max * (1.0 - min2 / max2))
					k = -min2 / 2.0;
				else
					k = -max * (max - abs(p.x)) / 2.0;
			}
			else if(p.x == 0.0 && p.y == 0.0){     /* point is on middle axis */
				skip = YES;
				if(abs(p.z) <= mid * (1.0 - min2 / mid2))
					k = -min2 / 2.0;
				else
					k = -mid * (mid - abs(p.z)) / 2.0;
			}
			else if(p.x == 0.0 && p.z == 0.0){     /* point is on minor axis */
				skip = YES;
				k = -min * (min - abs(p.y)) / 2.0;
			}
		}
	}

	if(skip == NO){
		dk = 1.0;
		while(fabs(dk) > 1.0e-10){
			iter++;
			fk = a * k * k * k * k * k * k + b * k * k * k * k * k + c * k * k * k * k + d * k * k * k + e * k * k + f * k + g;
			dfk = 6.0 * a * k * k * k * k * k + 5.0 * b * k * k * k * k+ 4.0 * c * k * k * k + 3.0 * d * k * k + 2.0 * e * k + f;
			dk = fk / dfk;
			k -= dk;

			if(iter == MAX_ITER)error_message(GENERAL_ERROR, "Projection to ellipsoid does not converge");
		}
	}

	if(rate <= 1.0){
		if(rate == 0.0){
			p.x = 0.0;
			p.y = min;
			p.z = 0.0;
		}
		else{
			p.x *= max2 / (max2 + 2.0 * k);
			p.y *= min2 / (min2 + 2.0 * k);
			p.z *= mid2 / (mid2 + 2.0 * k);
		}
	}
	else{
		p.x *= max2 / (max2 + 2.0 * k);
		if(p.z == 0.0){
			if(p.x > max)
				p.y = 0.0;
			else 
				p.y = sqrt(1.0 - p.x * p.x / max2) * min;
		}
		else{
			p.z *= mid2 / (mid2 + 2.0 * k);
			if(p.y != 0.0)
				p.y *= min2 / (min2 + 2.0 * k);
			else{
				val = 1.0 - p.x * p.x / max2 - p.z * p.z / mid2;
				if(val >= 0.0)p.y = sqrt(val) * min;
			}
		}
	}

	invert_rotation_transformation(&trans);
	transform_point_from_local_to_global(&p, pnt, &(ellipsoid -> center), &trans);
}


static void
get_ellipsoid_normal_point(ellipsoid_rec *ellipsoid, point_rec *point, vector_rec *normal)
{
	point_rec pnt;
	vector_rec nor;
	trans_matrix trans;
	double size;

	copy_vec(trans.x, ellipsoid -> major);
	copy_vec(trans.y, ellipsoid -> minor);
	copy_vec(trans.z, ellipsoid -> middle);

	transform_point_from_global_to_local(point, &pnt, &(ellipsoid -> center), &trans);

	nor.x = 2.0 * pnt.x / (ellipsoid -> max * ellipsoid -> max);
	nor.y = 2.0 * pnt.y / (ellipsoid -> min * ellipsoid -> min);
	nor.z = 2.0 * pnt.z / (ellipsoid -> mid * ellipsoid -> mid);

	invert_rotation_transformation(&trans);

	transform_vector_from_local_to_global(&nor, normal, &trans);

	size = sqrt(size_vec(*normal));
	div_vec(*normal, size);
}


#ifdef PARAMETRIC_ROUTINES
		
static int
project_point_to_ellipsoid_uv_octant(point_rec *point, ellipsoid_rec *ellipsoid, point_rec *pnt, double *u_par, double *v_par)
{
	point_rec point_spec, pnt_spec, p;
	vector_rec grad_u, grad_v, vec, normal;
	trans_matrix trans;
	double xx, yy, zz, dist, norm, rate, rate1 = 0.25, rate2 = 1.5, delta_u = EPSILON_U, delta_v = EPSILON_V;
	double u = 0.5, v = 0.5, du, dv, duu, dvv, ddu, ddv, duv, dvu, d_u = 0.0, d_v = 0.0, max_du = 0.25, max_dv = 0.25;
	double norm_u, norm_v, det_x, det_y, det_z, abs_det_x, abs_det_y, abs_det_z, delta_x, delta_y, delta_z, dd;
	int sign = 0, sign_x = 0, sign_y = 0, sign_z = 0, iter = 0, octant = 0;

	copy_vec(trans.x, ellipsoid -> major);
	copy_vec(trans.y, ellipsoid -> minor);
	copy_vec(trans.z, ellipsoid -> middle);

/* move point to the first octant */
	transform_point_from_global_to_local(point, &p, &(ellipsoid -> center), &trans);

	xx = p.x;
	yy = p.y;
	zz = p.z;

	if(xx == 0.0 && yy == 0.0 && zz == 0.0)error_message(GENERAL_ERROR, "Point is in center of ellipsoid"); 
	if(ellipsoid -> max != ellipsoid -> min && ellipsoid -> max != ellipsoid -> mid){
		if(yy == 0.0 && zz == 0.0){
			rate = ellipsoid -> min / ellipsoid -> max;
			if(fabs(xx) < ellipsoid -> max * (1.0 - rate * rate))
				error_message(GENERAL_ERROR, "Point lies on major axis of ellipsoid");
		}
		if(yy == 0.0 && xx == 0.0){
			rate = ellipsoid -> min / ellipsoid -> mid;
			if(fabs(zz) < ellipsoid -> mid * (1.0 - rate * rate))
				error_message(GENERAL_ERROR, "Point lies on middle axis of ellipsoid");
		}
	}

	p.x = fabs(xx);
	p.y = fabs(yy);
	p.z = fabs(zz);

	if(xx != 0.0)sign_x = (xx > 0.0) ? 1 : -1;
	if(yy != 0.0)sign_y = (yy > 0.0) ? 1 : -1;
	if(zz != 0.0)sign_z = (zz > 0.0) ? 1 : -1;

	if(sign_x > 0){
		if(sign_y > 0){
			if(sign_z > 0)
				octant = 1;
			else
				octant = 5;
		}
		else{
			if(sign_z > 0)
				octant = 4;
			else
				octant = 8;
		}
	}
	else{
		if(sign_y > 0){
			if(sign_z > 0)
				octant = 2;
			else
				octant = 6;
		}
		else{
			if(sign_z > 0)
				octant = 3;
			else
				octant = 7;
		}
	}

	if(sign_y == 0 && sign_z == 0)sign = sign_x;
	if(sign_z == 0 && sign_x == 0)sign = sign_y;
	if(sign_x == 0 && sign_y == 0)sign = sign_z;

	if(sign != 0){
		if(sign > 0)
			octant = 1;
		else
			octant = 7;
	}

	invert_rotation_transformation(&trans);
	transform_point_from_local_to_global(&p, &point_spec, &(ellipsoid -> center), &trans);

/* make projection (in the first octant) */
	while(YES){
		get_ellipsoid_point_uv(ellipsoid, u, v, &pnt_spec);

		norm_u = sqrt(get_ellipsoid_u_tangent_uv(ellipsoid, u, v, &grad_u));
		norm_v = sqrt(get_ellipsoid_v_tangent_uv(ellipsoid, u, v, &grad_v));

		if(norm_v < EPSILON && u < delta_u / 2.0){
			if(d_u < delta_u)break;

			d_u -= (delta_u / 2.0 - u);
			u = delta_u / 2.0;

			norm_u = sqrt(get_ellipsoid_u_tangent_uv(ellipsoid, u, v, &grad_u));
			norm_v = sqrt(get_ellipsoid_v_tangent_uv(ellipsoid, u, v, &grad_v));
		}

		cross_product(normal, grad_u, grad_v);

		abs_det_x = fabs(det_x = normal.x);
		abs_det_y = fabs(det_y = normal.y);
		abs_det_z = fabs(det_z = normal.z);

		norm = det_x * det_x + det_y * det_y + det_z * det_z;

		sub_vec(vec, point_spec, pnt_spec);
		dist = dot_product(vec, normal) / norm;

		delta_x = point_spec.x - dist * normal.x - pnt_spec.x;
		delta_y = point_spec.y - dist * normal.y - pnt_spec.y;
		delta_z = point_spec.z - dist * normal.z - pnt_spec.z;

		if(abs_det_x > abs_det_y && abs_det_x > abs_det_z){
			du = (+ grad_v.z * delta_y - grad_v.y * delta_z) / det_x;
			dv = (- grad_u.z * delta_y + grad_u.y * delta_z) / det_x;
		}
		else if(abs_det_y > abs_det_z){
			du = (+ grad_v.x * delta_z - grad_v.z * delta_x) / det_y;
			dv = (- grad_u.x * delta_z + grad_u.z * delta_x) / det_y;
		}
		else{
			du = (+ grad_v.y * delta_x - grad_v.x * delta_y) / det_z;
			dv = (- grad_u.y * delta_x + grad_u.x * delta_y) / det_z;
		}

		if(fabs(du) < delta_u && fabs(dv) < delta_v)break;

		ddu = fabs(du);
		ddv = fabs(dv);

		if(ddu > max_du || ddv > max_dv){
			ddu /= max_du;
			ddv /= max_dv;
			dd = (ddu > ddv) ? ddu : ddv;
			du /= dd;
			dv /= dd;
		}

		if(iter != 0){
			dd = du * d_u + dv * d_v;
			if(dd < 0.0){
				ddu = du + rate1 * d_u;
				ddv = dv + rate1 * d_v;
				if(d_u * ddu + d_v * ddv < 0.0){
					dd = -dd / rate1 / (d_u * d_u + d_v * d_v);
					du /= dd;
					dv /= dd;
				}
			}
			else{
				ddu = du - rate2 * d_u;
				ddv = dv - rate2 * d_v;
				if(d_u * ddu + d_v * ddv > 0.0){
					dd = dd / rate2 / (d_u * d_u + d_v * d_v);
					du /= dd;
					dv /= dd;
				}
			}
		}

		duu = dvu = d_u = du;
		dvv = duv = d_v = dv;

		if(du < 0.0){
			if(u + du < 0.0){
				duu = (0.0 - u) / 2.0;
				if(du < -delta_u)duv = duu / du * dv;
			}
		}
		if(du > 0.0){
			if(u + du > 1.0){
				duu = (1.0 - u) / 2.0;
				if(du > delta_v)duv = duu / du * dv;
			}
		}

		if(dv < 0.0){
			if(v + dv < 0.0){
				dvv = (0.0 - v) / 2.0;
				if(dv < -delta_v)dvu = dvv / dv * du;
			}
		}
		if(dv > 0.0){
			if(v + dv > 1.0){
				dvv = (1.0 - v) / 2.0;
				if(dv > delta_v)dvu = dvv / dv * du;
			}
		}

		iter++;
		if(iter == MAX_ITER)
			error_message(GENERAL_ERROR, "Projection to ellipsoid does not converge");

		u += (d_u = (fabs(duu) < fabs(dvu)) ? duu : dvu);
		v += (d_v = (fabs(dvv) < fabs(duv)) ? dvv : duv);

		if(u < 0.0)u = 0.0;
		if(v < 0.0)v = 0.0;
		if(u > 1.0)u = 1.0;
		if(v > 1.0)v = 1.0;
	}

/* move projection to original octant */
	invert_rotation_transformation(&trans);
	transform_point_from_global_to_local(&pnt_spec, &p, &(ellipsoid -> center), &trans);
	p.x *= sign_x;
	p.y *= sign_y;
	p.z *= sign_z;
	invert_rotation_transformation(&trans);
	transform_point_from_local_to_global(&p, pnt, &(ellipsoid -> center), &trans);

/* note: u and v are continuous (but not monotonic)
         u = 0 for y = 0, z = 0  v = <0-1> singular point
         v = 0 for z = 0
         u = 1 for x = 0
				 v = 1 for y = 0
	 to prevent confusion, octant is returned
         x+,y+,z+   1
				 x-,y+,z+   2
				 x-,y-,z+   3
				 x+,y-,z+   4
         x+,y+,z-   5
				 x-,y+,z-   6
				 x-,y-,z-   7
				 x+,y-,z-   8
   zeros are included into octants 1 and 7 */

	*u_par = u;
	*v_par = v;
	return(octant);
}



static void
get_ellipsoid_point_uv_octant(ellipsoid_rec *ellipsoid, int octant, double u, double v, point_rec *pnt)
{
	point_rec p;
	trans_matrix trans;

	get_ellipsoid_point_uv(ellipsoid, u, v, pnt);

/* move point to octant */
	if(octant != 1){
		copy_vec(trans.x, ellipsoid -> major);
		copy_vec(trans.y, ellipsoid -> minor);
		copy_vec(trans.z, ellipsoid -> middle);

		transform_point_from_global_to_local(pnt, &p, &(ellipsoid -> center), &trans);
		if(octant == 2 || octant == 3 || octant == 6 || octant == 7)p.x *= -1.0;
		if(octant == 4 || octant == 3 || octant == 8 || octant == 7)p.y *= -1.0;
		if(octant == 5 || octant == 6 || octant == 7 || octant == 8)p.z *= -1.0;

		invert_rotation_transformation(&trans);
		transform_point_from_local_to_global(&p, pnt, &(ellipsoid -> center), &trans);
	}
}


static void
get_ellipsoid_normal_uv_octant(ellipsoid_rec *ellipsoid, int octant, double u, double v, vector_rec *normal)
{
	vector_rec n;
	trans_matrix trans;

	get_ellipsoid_normal_uv(ellipsoid, u, v, normal);

/* move normal to octant */
	if(octant != 1){
		copy_vec(trans.x, ellipsoid -> major);
		copy_vec(trans.y, ellipsoid -> minor);
		copy_vec(trans.z, ellipsoid -> middle);

		transform_vector_from_global_to_local(normal, &n, &trans);
		if(octant == 2 || octant == 3 || octant == 6 || octant == 7)n.x *= -1.0;
		if(octant == 4 || octant == 3 || octant == 8 || octant == 7)n.y *= -1.0;
		if(octant == 5 || octant == 6 || octant == 7 || octant == 8)n.z *= -1.0;

		invert_rotation_transformation(&trans);
		transform_vector_from_local_to_global(&n, normal, &trans);
	}
}


static void
get_ellipsoid_point_uv(ellipsoid_rec *ellipsoid, double u, double v, point_rec *pnt)
{
	double weight = SQRT1_2;
	double B0u, B1u, B2u, B0v, B1v, B2v, Bx, By, Bz;
	double B00, B10, B20, B01, B11, B21, B02, B12, B22;
	double sum_B;

	B0u = B_0(u);
	B1u = B_1(u) * weight;
	B2u = B_2(u);

	B0v = B_0(v);
	B1v = B_1(v) * weight;
	B2v = B_2(v);

	B00 = B0u * B0v;
	B10 = B1u * B0v;
	B20 = B2u * B0v;

	B01 = B0u * B1v;
	B11 = B1u * B1v;
	B21 = B2u * B1v;

	B02 = B0u * B2v;
	B12 = B1u * B2v;
	B22 = B2u * B2v;

	sum_B = B00 + B10 + B20 + B01 + B11 + B21 + B02 + B12 + B22;

	Bx = B00 * ellipsoid -> pnt00.x + B10 * ellipsoid -> pnt10.x + B20 * ellipsoid -> pnt20.x
		 + B01 * ellipsoid -> pnt01.x + B11 * ellipsoid -> pnt11.x + B21 * ellipsoid -> pnt21.x
		 + B02 * ellipsoid -> pnt02.x + B12 * ellipsoid -> pnt12.x + B22 * ellipsoid -> pnt22.x;
	By = B00 * ellipsoid -> pnt00.y + B10 * ellipsoid -> pnt10.y + B20 * ellipsoid -> pnt20.y
		 + B01 * ellipsoid -> pnt01.y + B11 * ellipsoid -> pnt11.y + B21 * ellipsoid -> pnt21.y
		 + B02 * ellipsoid -> pnt02.y + B12 * ellipsoid -> pnt12.y + B22 * ellipsoid -> pnt22.y;
	Bz = B00 * ellipsoid -> pnt00.z + B10 * ellipsoid -> pnt10.z + B20 * ellipsoid -> pnt20.z
		 + B01 * ellipsoid -> pnt01.z + B11 * ellipsoid -> pnt11.z + B21 * ellipsoid -> pnt21.z
		 + B02 * ellipsoid -> pnt02.z + B12 * ellipsoid -> pnt12.z + B22 * ellipsoid -> pnt22.z;

	pnt -> x = Bx / sum_B;
	pnt -> y = By / sum_B;
	pnt -> z = Bz / sum_B;
}


static double
get_ellipsoid_u_tangent_uv(ellipsoid_rec *ellipsoid, double u, double v, vector_rec *grad)
{
	double weight = SQRT1_2;
	double B0u, B1u, B2u, B0v, B1v, B2v, dB0u, dB1u, dB2u;
	double B00, B10, B20, B01, B11, B21, B02, B12, B22;
	double dB00, dB10, dB20, dB01, dB11, dB21, dB02, dB12, dB22;
	double sum_B, sum_dB, sum_B2, Bx, By, Bz, dBx, dBy, dBz;

	B0u = B_0(u);
	B1u = B_1(u) * weight;
	B2u = B_2(u);

	B0v = B_0(v);
	B1v = B_1(v) * weight;
	B2v = B_2(v);

	dB0u = dB_0(u);
	dB1u = dB_1(u) * weight;
	dB2u = dB_2(u);

	B00 = B0u * B0v;
	B10 = B1u * B0v;
	B20 = B2u * B0v;

	B01 = B0u * B1v;
	B11 = B1u * B1v;
	B21 = B2u * B1v;

	B02 = B0u * B2v;
	B12 = B1u * B2v;
	B22 = B2u * B2v;

	dB00 = dB0u * B0v;
	dB10 = dB1u * B0v;
	dB20 = dB2u * B0v;

	dB01 = dB0u * B1v;
	dB11 = dB1u * B1v;
	dB21 = dB2u * B1v;

	dB02 = dB0u * B2v;
	dB12 = dB1u * B2v;
	dB22 = dB2u * B2v;

	sum_B = B00 + B10 + B20 + B01 + B11 + B21 + B02 + B12 + B22;
	sum_dB = dB00 + dB10 + dB20 + dB01 + dB11 + dB21 + dB02 + dB12 + dB22;

	sum_B2 = sum_B * sum_B;

	Bx = B00 * ellipsoid -> pnt00.x + B10 * ellipsoid -> pnt10.x + B20 * ellipsoid -> pnt20.x
		 + B01 * ellipsoid -> pnt01.x + B11 * ellipsoid -> pnt11.x + B21 * ellipsoid -> pnt21.x
		 + B02 * ellipsoid -> pnt02.x + B12 * ellipsoid -> pnt12.x + B22 * ellipsoid -> pnt22.x;
	By = B00 * ellipsoid -> pnt00.y + B10 * ellipsoid -> pnt10.y + B20 * ellipsoid -> pnt20.y
		 + B01 * ellipsoid -> pnt01.y + B11 * ellipsoid -> pnt11.y + B21 * ellipsoid -> pnt21.y
		 + B02 * ellipsoid -> pnt02.y + B12 * ellipsoid -> pnt12.y + B22 * ellipsoid -> pnt22.y;
	Bz = B00 * ellipsoid -> pnt00.z + B10 * ellipsoid -> pnt10.z + B20 * ellipsoid -> pnt20.z
		 + B01 * ellipsoid -> pnt01.z + B11 * ellipsoid -> pnt11.z + B21 * ellipsoid -> pnt21.z
		 + B02 * ellipsoid -> pnt02.z + B12 * ellipsoid -> pnt12.z + B22 * ellipsoid -> pnt22.z;

	dBx = dB00 * ellipsoid -> pnt00.x + dB10 * ellipsoid -> pnt10.x + dB20 * ellipsoid -> pnt20.x
		  + dB01 * ellipsoid -> pnt01.x + dB11 * ellipsoid -> pnt11.x + dB21 * ellipsoid -> pnt21.x
		  + dB02 * ellipsoid -> pnt02.x + dB12 * ellipsoid -> pnt12.x + dB22 * ellipsoid -> pnt22.x;
	dBy = dB00 * ellipsoid -> pnt00.y + dB10 * ellipsoid -> pnt10.y + dB20 * ellipsoid -> pnt20.y
		  + dB01 * ellipsoid -> pnt01.y + dB11 * ellipsoid -> pnt11.y + dB21 * ellipsoid -> pnt21.y
		  + dB02 * ellipsoid -> pnt02.y + dB12 * ellipsoid -> pnt12.y + dB22 * ellipsoid -> pnt22.y;
	dBz = dB00 * ellipsoid -> pnt00.z + dB10 * ellipsoid -> pnt10.z + dB20 * ellipsoid -> pnt20.z
		  + dB01 * ellipsoid -> pnt01.z + dB11 * ellipsoid -> pnt11.z + dB21 * ellipsoid -> pnt21.z
		  + dB02 * ellipsoid -> pnt02.z + dB12 * ellipsoid -> pnt12.z + dB22 * ellipsoid -> pnt22.z;

	grad -> x = (dBx * sum_B - Bx * sum_dB) / sum_B2;
	grad -> y = (dBy * sum_B - By * sum_dB) / sum_B2;
	grad -> z = (dBz * sum_B - Bz * sum_dB) / sum_B2;

	return(size_vec(*grad));
}


static double
get_ellipsoid_v_tangent_uv(ellipsoid_rec *ellipsoid, double u, double v, vector_rec *grad)
{
	double weight = SQRT1_2;
	double B0u, B1u, B2u, B0v, B1v, B2v, dB0v, dB1v, dB2v;
	double B00, B10, B20, B01, B11, B21, B02, B12, B22;
	double dB00, dB10, dB20, dB01, dB11, dB21, dB02, dB12, dB22;
	double sum_B, sum_dB, sum_B2, Bx, By, Bz, dBx, dBy, dBz;

	B0u = B_0(u);
	B1u = B_1(u) * weight;
	B2u = B_2(u);

	B0v = B_0(v);
	B1v = B_1(v) * weight;
	B2v = B_2(v);

	dB0v = dB_0(v);
	dB1v = dB_1(v) * weight;
	dB2v = dB_2(v);

	B00 = B0u * B0v;
	B10 = B1u * B0v;
	B20 = B2u * B0v;

	B01 = B0u * B1v;
	B11 = B1u * B1v;
	B21 = B2u * B1v;

	B02 = B0u * B2v;
	B12 = B1u * B2v;
	B22 = B2u * B2v;

	dB00 = B0u * dB0v;
	dB10 = B1u * dB0v;
	dB20 = B2u * dB0v;

	dB01 = B0u * dB1v;
	dB11 = B1u * dB1v;
	dB21 = B2u * dB1v;

	dB02 = B0u * dB2v;
	dB12 = B1u * dB2v;
	dB22 = B2u * dB2v;

	sum_B = B00 + B10 + B20 + B01 + B11 + B21 + B02 + B12 + B22;
	sum_dB = dB00 + dB10 + dB20 + dB01 + dB11 + dB21 + dB02 + dB12 + dB22;

	sum_B2 = sum_B * sum_B;

	Bx = B00 * ellipsoid -> pnt00.x + B10 * ellipsoid -> pnt10.x + B20 * ellipsoid -> pnt20.x
		 + B01 * ellipsoid -> pnt01.x + B11 * ellipsoid -> pnt11.x + B21 * ellipsoid -> pnt21.x
		 + B02 * ellipsoid -> pnt02.x + B12 * ellipsoid -> pnt12.x + B22 * ellipsoid -> pnt22.x;
	By = B00 * ellipsoid -> pnt00.y + B10 * ellipsoid -> pnt10.y + B20 * ellipsoid -> pnt20.y
		 + B01 * ellipsoid -> pnt01.y + B11 * ellipsoid -> pnt11.y + B21 * ellipsoid -> pnt21.y
		 + B02 * ellipsoid -> pnt02.y + B12 * ellipsoid -> pnt12.y + B22 * ellipsoid -> pnt22.y;
	Bz = B00 * ellipsoid -> pnt00.z + B10 * ellipsoid -> pnt10.z + B20 * ellipsoid -> pnt20.z
		 + B01 * ellipsoid -> pnt01.z + B11 * ellipsoid -> pnt11.z + B21 * ellipsoid -> pnt21.z
		 + B02 * ellipsoid -> pnt02.z + B12 * ellipsoid -> pnt12.z + B22 * ellipsoid -> pnt22.z;

	dBx = dB00 * ellipsoid -> pnt00.x + dB10 * ellipsoid -> pnt10.x + dB20 * ellipsoid -> pnt20.x
		  + dB01 * ellipsoid -> pnt01.x + dB11 * ellipsoid -> pnt11.x + dB21 * ellipsoid -> pnt21.x
		  + dB02 * ellipsoid -> pnt02.x + dB12 * ellipsoid -> pnt12.x + dB22 * ellipsoid -> pnt22.x;
	dBy = dB00 * ellipsoid -> pnt00.y + dB10 * ellipsoid -> pnt10.y + dB20 * ellipsoid -> pnt20.y
		  + dB01 * ellipsoid -> pnt01.y + dB11 * ellipsoid -> pnt11.y + dB21 * ellipsoid -> pnt21.y
		  + dB02 * ellipsoid -> pnt02.y + dB12 * ellipsoid -> pnt12.y + dB22 * ellipsoid -> pnt22.y;
	dBz = dB00 * ellipsoid -> pnt00.z + dB10 * ellipsoid -> pnt10.z + dB20 * ellipsoid -> pnt20.z
		  + dB01 * ellipsoid -> pnt01.z + dB11 * ellipsoid -> pnt11.z + dB21 * ellipsoid -> pnt21.z
		  + dB02 * ellipsoid -> pnt02.z + dB12 * ellipsoid -> pnt12.z + dB22 * ellipsoid -> pnt22.z;

	grad -> x = (dBx * sum_B - Bx * sum_dB) / sum_B2;
	grad -> y = (dBy * sum_B - By * sum_dB) / sum_B2;
	grad -> z = (dBz * sum_B - Bz * sum_dB) / sum_B2;

	return(size_vec(*grad));
}


static void
get_ellipsoid_normal_uv(ellipsoid_rec *ellipsoid, double u, double v, vector_rec *normal)
{
	vector_rec u_grad, v_grad;
	double norm;

	get_ellipsoid_u_tangent_uv(ellipsoid, u, v, &u_grad);
	get_ellipsoid_v_tangent_uv(ellipsoid, u, v, &v_grad);

	cross_product(*normal, u_grad, v_grad);
	norm = sqrt(size_vec(*normal));
	div_vec(*normal, norm);
}

#endif


static logic
check_point_inside_ellipse(point_rec *point, ellipse_rec *ellipse, double *value)
{
	point_rec pnt;
	trans_matrix trans;
	double x_rate, y_rate;

	copy_vec(trans.x, ellipse -> major);
	copy_vec(trans.y, ellipse -> minor);
	copy_vec(trans.z, ellipse -> normal);

	transform_point_from_global_to_local(point, &pnt, &(ellipse -> center), &trans);

	if(fabs(pnt.z) > EPSILON)error_message(GENERAL_ERROR, "Point is not in plane of ellipse");

	x_rate = pnt.x / ellipse -> max;
	y_rate = pnt.y / ellipse -> min;

	*value = x_rate * x_rate + y_rate * y_rate - 1.0;
	if(*value <= 0.0)return(YES);

	return(NO);
}


static int
get_ellipse_line_intersection(ellipse_rec *ellipse, point_rec *point, vector_rec *dir, point_rec *intersection, logic point_in_plane)
{
	point_rec pnt, pnt1, pnt2;
	vector_rec direction;
	trans_matrix trans;
	double a0, a1, a2, a0x, a0y, a1x, a1y, a2x, a2y;
	double D, max, min, t, t1, t2, norm;

	norm = size_vec(*dir);
	if(norm < EPSILON * EPSILON)error_message(GENERAL_ERROR, "Zero length direction");

	copy_vec(trans.x, ellipse -> major);
	copy_vec(trans.y, ellipse -> minor);
	copy_vec(trans.z, ellipse -> normal);

	transform_vector_from_global_to_local(dir, &direction, &trans);
	if(fabs(direction.z) > EPSILON)error_message(GENERAL_ERROR, "Direction is not in ellipse plane");

	transform_point_from_global_to_local(point, &pnt, &(ellipse -> center), &trans);
	if(point_in_plane == YES){
		if(fabs(pnt.z) > EPSILON)error_message(GENERAL_ERROR, "Point is not in ellipse plane");
	}

	max = ellipse -> max;
	min = ellipse -> min;

	a2x = direction.x / max;
	a2y = direction.y / min;
	a2 = a2x * a2x + a2y * a2y;

	a0x = pnt.x / max;
	a0y = pnt.y / min;
	a0 = a0x * a0x + a0y *a0y - 1.0;

	a1x = a2x * a0x;
	a1y = a2y * a0y;
	a1 = a1x + a1y;

	D = a1 * a1 - a0 * a2;
	
	if(D < 0.0)return(0);

	invert_rotation_transformation(&trans);

	if(D == 0.0){
		t = -a1 / a2;

		pnt.x += t * direction.x;
		pnt.y += t * direction.y;
		pnt.z = 0.0;

		transform_point_from_local_to_global(&pnt, &intersection[0], &(ellipse -> center), &trans);
		return(1);
	}

	D = sqrt(D);

	t1 = -(a1 - D) / a2;
	t2 = -(a1 + D) / a2;

	pnt1.x = pnt.x + t1 * direction.x;
	pnt1.y = pnt.y + t1 * direction.y;
	pnt1.z = 0.0;

	transform_point_from_local_to_global(&pnt1, &intersection[0], &(ellipse -> center), &trans);

	pnt2.x = pnt.x + t2 * direction.x;
	pnt2.y = pnt.y + t2 * direction.y;
	pnt2.z = 0.0;

	transform_point_from_local_to_global(&pnt2, &intersection[1], &(ellipse -> center), &trans);

	return(2);
}


static void
get_ellipse_line_center_intersection(ellipse_rec *ellipse, vector_rec *dir, point_rec *pnt1, point_rec *pnt2)
{
	vector_rec direction;
	trans_matrix trans;
	double a, b, t, norm;

	norm = size_vec(*dir);
	if(norm < EPSILON * EPSILON)error_message(GENERAL_ERROR, "Zero length direction");

	copy_vec(trans.x, ellipse -> major);
	copy_vec(trans.y, ellipse -> minor);
	copy_vec(trans.z, ellipse -> normal);

	transform_vector_from_global_to_local(dir, &direction, &trans);
	if(fabs(direction.z) > EPSILON)error_message(GENERAL_ERROR, "Line is not in plane of ellipse");

	a = direction.x / (ellipse -> max);
	b = direction.y / (ellipse -> min);

	t = sqrt(1.0 / (a * a + b * b));
	
	invert_rotation_transformation(&trans);

	mul_vec(direction, t);
	transform_point_from_local_to_global(&direction, pnt1, &(ellipse -> center), &trans);

	mul_vec(direction, -1.0);
	transform_point_from_local_to_global(&direction, pnt2, &(ellipse -> center), &trans);
}


#ifdef PARAMETRIC_ROUTINES

static int
get_ellipse_ellipse_intersection(ellipse_rec *ellipse1, ellipse_rec *ellipse2, point_rec *intersection)
{
	point_rec point, pnt;
	trans_matrix trans;
	double product, t, tt, dt, ddt, delta_t = EPSILON_T;
	double prev_t, next_t, prev_val, next_val, tn, tp, tpp, tnn, expand = 1.2;
	double max1, max2, max, dist, val, value, value_p, value_n, value_pp, value_nn;
	int q, prev_q, next_q, count, qq, qp, qn, qpp, qnn;
	logic found, last;

	product = dot_product(ellipse1 -> normal, ellipse2 -> normal);
	if(fabs(fabs(product) - 1.0) > EPSILON)error_message(GENERAL_ERROR, "Ellipses are not coplanar"); 

	copy_vec(trans.x, ellipse1 -> major);
	copy_vec(trans.y, ellipse1 -> minor);
	copy_vec(trans.z, ellipse1 -> normal);

	transform_point_from_global_to_local(&(ellipse2 -> center), &point, &(ellipse1 -> center), &trans);
	if(fabs(point.z) > EPSILON)error_message(GENERAL_ERROR, "Ellipses are not in the same plane"); 

	max1 = ellipse1 -> max;
	max2 = ellipse2 -> max;
	max = max1 + max2;

/* intersection of circumsribed circles */
	dist = dist_point(ellipse1 -> center, ellipse2 -> center);
	if(dist > max * max)return(0);

/* ellipse1 and ellipse2 are affine */
	if(fabs(point.x) < EPSILON && fabs(point.y) < EPSILON){
		product = dot_product(ellipse1 -> major, ellipse2 -> minor);
		if(fabs(product) < EPSILON){
			if(fabs(max1 / ellipse1 -> min - max2 / ellipse2 -> min) < EPSILON){
				if(fabs(max1 / max2 - 1.0) < EPSILON)return(-1);   /* coinciding ellipses */
				return(0);
			}
		}
	}

/* intersection of circle circumscribed to ellipse2 with ellipse1 */
	if(check_point_inside_ellipse(&(ellipse2 -> center), ellipse1, &val) == YES){
		if(fabs(point.x) > EPSILON && fabs(point.y) > EPSILON){
			project_point_to_ellipse_t_quadrant(&(ellipse2 -> center), ellipse1, &pnt, &t, NO);
			dist = dist_point(ellipse2 -> center, pnt);
			if(dist > max2 * max2)return(0);
		}
	}

/* intersection of circle circumscribed to ellipse1 with ellipse2 */
	if(check_point_inside_ellipse(&(ellipse1 -> center), ellipse2, &val) == YES){
		copy_vec(trans.x, ellipse2 -> major);
		copy_vec(trans.y, ellipse2 -> minor);
		copy_vec(trans.z, ellipse2 -> normal);

		transform_point_from_global_to_local(&(ellipse1 -> center), &point, &(ellipse2 -> center), &trans);

		if(fabs(point.x) > EPSILON && fabs(point.y) > EPSILON){
			project_point_to_ellipse_t_quadrant(&(ellipse1 -> center), ellipse2, &pnt, &t, NO);
			dist = dist_point(ellipse1 -> center, pnt);
			if(dist > max1 * max1)return(0);
		}
	}

	dt = 1.0 / 10;

	prev_q = 4;
	prev_t = dt / 2.0;

	get_ellipse_point_t_quadrant(ellipse1, prev_q, prev_t, &pnt);
	check_point_inside_ellipse(&pnt, ellipse2, &prev_val);
	while(fabs(prev_val) < EPSILON){
		prev_t -= delta_t;
		get_ellipse_point_t_quadrant(ellipse1, prev_q, prev_t, &pnt);
		check_point_inside_ellipse(&pnt, ellipse2, &prev_val);
	}

	next_q = q = 1;
	next_t = t = dt / 2.0;

	get_ellipse_point_t_quadrant(ellipse1, q, t, &pnt);
	check_point_inside_ellipse(&pnt, ellipse2, &val);
	while(fabs(val) < EPSILON){
		t += delta_t;
		get_ellipse_point_t_quadrant(ellipse1, q, t, &pnt);
		check_point_inside_ellipse(&pnt, ellipse2, &val);
	}

	count = 0;
	last = NO;
	while(YES){
		if(next_q == 1 || next_q == 3){
			next_t += dt;
			if(next_t > 1.0){
				++next_q;
				next_t = 2.0 - next_t;
			}
		}
		else{
			next_t -= dt;
			if(next_t < 0.0){
				if(++next_q == 5){
					next_q = 1;
					last = YES;
				}
				next_t = -next_t;
			}
		}

		get_ellipse_point_t_quadrant(ellipse1, next_q, next_t, &pnt);
		check_point_inside_ellipse(&pnt, ellipse2, &next_val);
		while(fabs(next_val) < EPSILON){
			
/* note: in(de)crementing by delta_t I can never cross quadrant boundary because I did not
	 start at the beginning of quadrant and because there may be at most four intersections */

			if(next_q == 1 || next_q == 3)
				next_t += delta_t;
			else
				next_t -= delta_t;
			get_ellipse_point_t_quadrant(ellipse1, next_q, next_t, &pnt);
			check_point_inside_ellipse(&pnt, ellipse2, &next_val);
		}

		if(next_val * val < 0.0){

/* intersection in the last interval */

			get_ellipse_ellipse_interval_intersection(ellipse1, ellipse2, q, t, val, dt, &intersection[count++]);
		}
		else{

/* note: check the equality only in the prev interval */
			if((prev_val <= val && next_val < val) || (prev_val >= val && next_val > val)){
				if(prev_val * val > 0.0){
					if(val + 1.0 < expand && val + 1.0 > 1.0 / expand){
				
/* extreme in the last two intervals */

						tt = t;
						qq = q;
						value = val;

						tpp = prev_t;
						qpp = prev_q;

						tnn = next_t;
						qnn = next_q;

						value_pp = prev_val;
						value_nn = next_val;
						
						ddt = dt / 2;
						
						while(YES){
							tn = tp = tt;
							qn = qp = qq;

							if(qq == 1 || qq == 3){
								tn += ddt;
								if(tn > 1.0){
									++qn;
									tn = 2.0 - tn;
								}
							}
							else{
								tn -= ddt;
								if(tn < 0.0){
									if(++qn == 5)qn = 1;
									tn = -tn;
								}
							}
							
							if(q == 2 || q == 4){
								tp += ddt;
								if(tp > 1.0){
									--qp;
									tp = 2.0 - tp;
								}
							}
							else{
								tp -= ddt;
								if(tp < 0.0){
									if(--qp == 0)qp = 4;
									tp = -tp;
								}
							}
							
							get_ellipse_point_t_quadrant(ellipse1, qp, tp, &pnt);
							check_point_inside_ellipse(&pnt, ellipse2, &value_p);
							while(fabs(value_p) < EPSILON){
								if(qp == 1 || qp == 3){
									tp += delta_t;
									if(tp > 1.0){
										++qp;
										tp = 2.0 - tp;
									}
								}
								else{
									tp -= delta_t;
									if(tp < 0.0){
										if(++qp == 5)qp = 1;
										tp = -tp;
									}
								}
								ddt += 2.0 * delta_t;
								get_ellipse_point_t_quadrant(ellipse1, qp, tp, &pnt);
								check_point_inside_ellipse(&pnt, ellipse2, &value_p);
							}
							
							get_ellipse_point_t_quadrant(ellipse1, qn, tn, &pnt);
							check_point_inside_ellipse(&pnt, ellipse2, &value_n);
							while(fabs(value_n) < EPSILON){
								if(qn == 1 || qn == 3){
									tn += delta_t;
									if(tn > 1.0){
										++qn;
										tn = 2.0 - tn;
									}
								}
								else{
									tn -= delta_t;
									if(tn < 0.0){
										if(++qn == 5)qn = 1;
										tn = -tn;
									}
								}
								ddt += 2.0 * delta_t;
								get_ellipse_point_t_quadrant(ellipse1, qn, tn, &pnt);
								check_point_inside_ellipse(&pnt, ellipse2, &value_n);
							}
							
							found = NO;
							if(value_pp * value_p < 0.0){
								found = YES;
								get_ellipse_ellipse_interval_intersection(ellipse1, ellipse2, qpp, tpp, value_pp, ddt, &intersection[count++]);
							}
							if(value * value_p < 0.0){
								found = YES;
								get_ellipse_ellipse_interval_intersection(ellipse1, ellipse2, qp, tp, value_p, ddt, &intersection[count++]);
							}
							if(value * value_n < 0.0){
								found = YES;
								get_ellipse_ellipse_interval_intersection(ellipse1, ellipse2, qq, tt, value, ddt, &intersection[count++]);
							}
							if(value_nn * value_n < 0.0){
								found = YES;
								get_ellipse_ellipse_interval_intersection(ellipse1, ellipse2, qn, tn, value_n, ddt, &intersection[count++]);
							}
							
							if(found == YES)break;
							
							ddt /= 2.0;
							if(ddt < delta_t)break;

							if((value <= value_p && value_pp <= value_p) || (value >= value_p && value_pp >= value_p)){
								tnn = tt;
								qnn = qq;
								value_nn = value;
								tt = tp;
								qq = qp;
								value = value_p;
								continue;
							}
							if((value <= value_n && value_nn <= value_n) || (value >= value_n && value_nn >= value_n)){
								tpp = tt;
								qpp = qq;
								value_pp = value;
								tt = tn;
								qq = qn;
								value = value_n;
								continue;
							}
							tpp = tp;
							qpp = qp;
							value_pp = value_p;
							tnn = tn;
							qnn = qn;
							value_nn = value_n;
						}
					}				
				}
			}
		}

		if(last == YES)break;

		prev_q = q;
		prev_t = t;
		prev_val = val;

		q = next_q;
		t = next_t;
		val = next_val;
	}

	return(count);
}


static void
get_ellipse_ellipse_interval_intersection(ellipse_rec *ellipse1, ellipse_rec *ellipse2, int q, double t, double val, double dt,
																					point_rec *point)
{
	point_rec pnt;
	int qq, qqq;
	double tt, ttt, value, delta_t = EPSILON_T;

	qq = qqq = q;
	tt = ttt = t;

	dt /= 2.0;
	while(YES){
		if(qq == 1 || qq == 3){
			tt += dt;
			if(tt > 1.0){
				++qq;
				tt = 2.0 - tt;
			}
		}
		else{
			tt -= dt;
			if(tt < 0.0){
				if(++qq == 5)qq = 1;
				tt = -tt;
			}
		}
		
		get_ellipse_point_t_quadrant(ellipse1, qq, tt, &pnt);
		check_point_inside_ellipse(&pnt, ellipse2, &value);
		
		if(fabs(value) < EPSILON || dt < delta_t){
			if(fabs(value) > EPSILON){
				if(qq == qqq){
					tt = ttt + (tt - ttt) * val / (val - value);
					if(tt >= 0.0 && tt <= 1.0){
						get_ellipse_point_t_quadrant(ellipse1, qq, tt, &pnt);
						check_point_inside_ellipse(&pnt, ellipse2, &value);
					}
				}
			}
						
			copy_vec(*point, pnt);
			return;
		}
					
		if(value * val > 0.0){
			qqq = qq;
			ttt = tt;
			val = value;
		}
		else{
			qq = qqq;
			tt = ttt;
			dt /= 2.0;
		}
	}
}

#endif

static void
project_point_to_ellipse(point_rec *point, ellipse_rec *ellipse, point_rec *pnt)
{
	point_rec p;
	trans_matrix trans;
	double max, min, max2, min2, max4, min4, xp2, yp2;
	double val, rate, k, dk = 1.0, fk, dfk;
	double a, b, c, d, e;
	double px, py, gx, gy, dx, dy;
	logic skip = NO;
	int iter = 0;
	
	copy_vec(trans.x, ellipse -> major);
	copy_vec(trans.y, ellipse -> minor);
	copy_vec(trans.z, ellipse -> normal);

	transform_point_from_global_to_local(point, &p, &(ellipse -> center), &trans);

	xp2 = p.x * p.x;
	yp2 = p.y * p.y;

	max = ellipse -> max;
	min = ellipse -> min;

#ifdef DEBUG
	if(max < min)error_message(GENERAL_ERROR, "Unexpected situation");		
#endif

	max2 = max * max;
	min2 = min * min;

	max4 = max2 * max2;
	min4 = min2 * min2;

	a = 16.0;
	b = 16.0 * (max2 + min2);
	c = 4.0 * (4.0 * max2 * min2 + max4 + min4 - xp2 * max2 - yp2 * min2);
	d = 4.0 * max2 * min2 * (max2 + min2 - xp2 - yp2);
	e = max4 * min4 - max2 * min2 * (xp2 * min2 + yp2 * max2);

/* get radial projection */

	val = xp2 / max2 + yp2 / min2;

	if(val < 1.0e-10){       /* point is in center */
		rate = 0.0;
		k = -min2 / 2.0;
		skip = YES;
	}
	else{
		rate = sqrt(1.0 / val);
		if(rate <= 1.0){       /* point is outside ellipse */
			px = p.x * rate;
			py = p.y * rate;

			gx = 2.0 * px / max2;
			gy = 2.0 * py / min2;

			dx = px - p.x;
			dy = py - p.y;

			k = sqrt((dx * dx + dy * dy) / (gx * gx + gy * gy));
		}
		else{                  /* point is inside ellipse */
			k = 0.0;

			if(p.y == 0.0){      /* point is on major axis */
				skip = YES;
				if(fabs(p.x) <= max * (1.0 - min2 / max2))    /* point is in outer half of minor osculating circle */
					k = -min2 / 2.0;
				else               /* point is between ellipse center and center of minor osculating circle */
					k = -max * (max - fabs(p.x)) / 2.0; 
			}
			else{
				if(p.x == 0.0){    /* point on minor axis */
					skip = YES;
					k = -min * (min - fabs(p.y)) / 2.0; 
				}
			}
		}
	}

	if(skip == NO){
		dk = 1.0;
		while(fabs(dk) > 1.0e-10){
			iter++;
			fk = a * k * k * k * k + b * k * k * k + c * k * k + d * k + e;
			dfk = 4.0 * a * k * k * k + 3.0 * b * k * k + 2.0 * c * k + d;
			dk = fk / dfk;
			k -= dk;

/* this check can cause failure of get_ellipsoid_line_square_distance */

			if(iter == MAX_ITER)error_message(GENERAL_ERROR, "Projection to ellipse does not converge");
		}
	}

	if(rate <= 1.0){
		if(rate == 0.0){
			p.x = 0.0;
			p.y = ellipse -> min;
		}
		else{
			p.x *= max2 / (max2 + 2.0 * k);
			p.y *= min2 / (min2 + 2.0 * k);
		}
	}
	else{
		p.x *= max2 / (max2 + 2.0 * k);
		if(p.y != 0.0)
			p.y *= min2 / (min2 + 2.0 * k);
		else if(p.x > max)
			p.y = 0.0;
		else
			p.y = sqrt((1.0 - p.x * p.x / max2) * min2);
	}

	invert_rotation_transformation(&trans);
	transform_point_from_local_to_global(&p, pnt, &(ellipse -> center), &trans);
}


#define NORMALIZE_LOCALLY

static void
get_ellipse_normal_point(ellipse_rec *ellipse, point_rec *point, vector_rec *normal)
{
	point_rec pnt;
	vector_rec nor;
	trans_matrix trans;
	double size;

	copy_vec(trans.x, ellipse -> major);
	copy_vec(trans.y, ellipse -> minor);
	copy_vec(trans.z, ellipse -> normal);

	transform_point_from_global_to_local(point, &pnt, &(ellipse -> center), &trans);

	nor.x = 2.0 * pnt.x / (ellipse -> max * ellipse -> max);
	nor.y = 2.0 * pnt.y / (ellipse -> min * ellipse -> min);
	nor.z = 0.0;

#ifdef NORMALIZE_LOCALLY
	size = 1.0 / sqrt(nor.x * nor.x + nor.y * nor.y);
	nor.x *= size;
	nor.y *= size;
#endif

	invert_rotation_transformation(&trans);

	transform_vector_from_local_to_global(&nor, normal, &trans);

#ifndef NORMALIZE_LOCALLY
	size = sqrt(size_vec(*normal));
	div_vec(*normal, size);
#endif
}


/* points point1 and point2 are expected to lie on the ellipse;
	 I do not compute control point as it may be infinite if weight is zero !!! */

static double
get_quadratic_elliptic_arc_weight(ellipse_rec *ellipse, point_rec *point1, point_rec *point2)
{
	point_rec pnt1, pnt2;
	double x1, y1, x2, y2, val1, val2, sz1, sz2, scale_y_e2c, weight, cosine, Radius, radius;
	trans_matrix trans_g2l; 

	Radius = ellipse -> max;
	radius = ellipse -> min;

	copy_vec(trans_g2l.x, ellipse -> major);
	copy_vec(trans_g2l.y, ellipse -> minor);
	copy_vec(trans_g2l.z, ellipse -> normal);
				
	transform_point_from_global_to_local(point1, &pnt1, &(ellipse -> center), &trans_g2l);
	transform_point_from_global_to_local(point2, &pnt2, &(ellipse -> center), &trans_g2l);

	if(fabs(pnt1.z) > 1.0e-10 || fabs(pnt1.z) > 1.0e-10)error_message(GENERAL_ERROR, "Unexpected situation");

	x1 = pnt1.x / Radius;
	y1 = pnt1.y / radius;

	x2 = pnt2.x / Radius;
	y2 = pnt2.y / radius;

	val1 = x1 * x1 + y1 * y1;
	val2 = x2 * x2 + y2 * y2;
	
	if(fabs(1.0 - val1) > 1.0e-10 || fabs(1.0 - val2) > 1.0e-10)error_message(GENERAL_ERROR, "Unexpected situation");

/* get radial projection of points pnt1 and pnt2 */

	val1 = sqrt(1.0 / val1);
	val2 = sqrt(1.0 / val2);

	scale_y_e2c = Radius / radius;

	pnt1.x *= val1;
	pnt2.x *= val2;

	pnt1.y *= val1 * scale_y_e2c;
	pnt2.y *= val2 * scale_y_e2c;

	pnt1.z = pnt2.z = 0.0;

	sz1 = size_vec(pnt1);
	sz2 = size_vec(pnt2);

	cosine = dot_product(pnt1, pnt2) / sqrt(sz1 * sz2);

	if(cosine >= 1.0)return(1.0);
	if(cosine <= -1.0)return(0.0);

	weight = sqrt((1.0 + cosine) / 2.0);

	return(weight);
}
	
	
#ifdef PARAMETRIC_ROUTINES

/* parameter spec is used to allow for handling the case when the point is on major axis 
	 between centers of minor osculating circles;
	 y-coordinate of the point is slightly perturbated */

static int
project_point_to_ellipse_t_quadrant(point_rec *point, ellipse_rec *ellipse, point_rec *pnt, double *t_par, logic spec)
{
	point_rec point_spec, pnt_spec, p;
	vector_rec grad, vec;
	trans_matrix trans;
	double xx, yy, dist, norm, rate1 = 0.25, rate2 = 1.5, delta_t = EPSILON_T;
	double t = 0.5, dt, dtt, ddt, d_t = 0.0, max_dt = 0.25, rate;
	int sign = 0, sign_x = 0, sign_y = 0, iter = 0, quadrant = 0;

	copy_vec(trans.x, ellipse -> major);
	copy_vec(trans.y, ellipse -> minor);
	copy_vec(trans.z, ellipse -> normal);

/* move point to the first quadrant */
	transform_point_from_global_to_local(point, &p, &(ellipse -> center), &trans);

	if(fabs(p.z) > EPSILON)error_message(GENERAL_ERROR, "Point is not in plane of ellipse");

	xx = p.x;
	yy = p.y;

	if(xx == 0.0 && yy == 0.0){
		if(spec == NO)error_message(GENERAL_ERROR, "Point is in center of ellipse");
		yy = ellipse -> min / 1.0e+10;
	}
	if(ellipse -> max != ellipse -> min){
		if(yy == 0.0){
/* point on major axis between center of ellipse and center of minor osculating circle cannot be uniquelly projected */
			rate = ellipse -> min / ellipse -> max;
			if(fabs(xx) < ellipse -> max * (1.0 - rate * rate)){
				if(spec == NO)error_message(GENERAL_ERROR, "Point lies on major axis of ellipse");
				yy = ellipse -> min / 1.0e+10;
			}
		}
	}

	p.x = fabs(xx);
	p.y = fabs(yy);

	if(xx != 0.0)sign_x = (xx > 0.0) ? 1 : -1;
	if(yy != 0.0)sign_y = (yy > 0.0) ? 1 : -1;

	if(sign_x > 0){
		if(sign_y > 0)
			quadrant = 1;
		else
			quadrant = 4;
	}
	else{
		if(sign_y > 0)
			quadrant = 2;
		else
			quadrant = 3;
	}

	if(sign_y == 0)sign = sign_x;
	if(sign_x == 0)sign = sign_y;

	if(sign != 0){
		if(sign > 0)
			quadrant = 1;
		else
			quadrant = 3;
	}

	invert_rotation_transformation(&trans);
	transform_point_from_local_to_global(&p, &point_spec, &(ellipse -> center), &trans);

/* make projection (in the first quadrant) */
	while(YES){
		get_ellipse_point_t(ellipse, t, &pnt_spec);
		norm = sqrt(get_ellipse_tangent_t(ellipse, t, &grad));
		sub_vec(vec, point_spec, pnt_spec);
		dist = dot_product(vec, grad) / norm;

		dt = dist / norm;
		if(fabs(dt) < delta_t)break;

		if((ddt = fabs(dt)) > max_dt)dt /= ddt / max_dt;

		if(iter != 0){
			if(dt * d_t < 0.0){
				if(fabs(dt) > rate1 * fabs(d_t))dt = -d_t * rate1;
			}
			else{
				if(fabs(dt) > rate2 * fabs(d_t))dt = d_t * rate2;
			}
		}
		
		dtt = d_t = dt;
				
		if(dt < 0.0){
			if(t + dt < 0.0)dtt = (0.0 - t) / 2.0;
		}
		if(dt > 0.0){
			if(t + dt > 1.0)dtt = (1.0 - t) / 2.0;
		}

		iter++;
		if(iter == MAX_ITER)
			error_message(GENERAL_ERROR, "Projection to ellipse does not converge");

		t += (d_t = dtt);
	}

/* move projection to original quadrant */
	invert_rotation_transformation(&trans);
	transform_point_from_global_to_local(&pnt_spec, &p, &(ellipse -> center), &trans);
	p.x *= sign_x;
	p.y *= sign_y;
	invert_rotation_transformation(&trans);
	transform_point_from_local_to_global(&p, pnt, &(ellipse -> center), &trans);

/* note: t is continuous (but not monotonic)
         t = 0 on major axis
				 t = 1 on minor axis
	 to prevent confusion, quadrant is returned
         x+,y+   1
				 x-,y+   2
				 x-,y-   3
				 x+,y-   4
   zeros are included into quadrants 1 and 3 */

	*t_par = t;
	return(quadrant);
}


static void
get_ellipse_point_t_quadrant(ellipse_rec *ellipse, int quadrant, double t, point_rec *pnt)
{
	point_rec p;
	trans_matrix trans;

	get_ellipse_point_t(ellipse, t, pnt);

/* move point to quadrant */
	if(quadrant != 1){
		copy_vec(trans.x, ellipse -> major);
		copy_vec(trans.y, ellipse -> minor);
		copy_vec(trans.z, ellipse -> normal);

		transform_point_from_global_to_local(pnt, &p, &(ellipse -> center), &trans);
		if(quadrant == 2 || quadrant == 3)p.x *= -1.0;
		if(quadrant == 4 || quadrant == 3)p.y *= -1.0;

		invert_rotation_transformation(&trans);
		transform_point_from_local_to_global(&p, pnt, &(ellipse -> center), &trans);
	}
}


static double
get_ellipse_tangent_t_quadrant(ellipse_rec *ellipse, int quadrant, double t, vector_rec *grad)
{
	vector_rec g;
	trans_matrix trans;
	double norm;

	norm = get_ellipse_tangent_t(ellipse, t, grad);

/* move grad to quadrant */
	if(quadrant != 1){
		copy_vec(trans.x, ellipse -> major);
		copy_vec(trans.y, ellipse -> minor);
		copy_vec(trans.z, ellipse -> normal);

		transform_vector_from_global_to_local(grad, &g, &trans);
		if(quadrant == 2 || quadrant == 3)g.x *= -1.0;
		if(quadrant == 4 || quadrant == 3)g.y *= -1.0;

		invert_rotation_transformation(&trans);
		transform_vector_from_local_to_global(&g, grad, &trans);
	}

	return(norm);
}


static void
get_ellipse_normal_t_quadrant(ellipse_rec *ellipse, int quadrant, double t, vector_rec *normal)
{
	vector_rec n;
	trans_matrix trans;

	get_ellipse_normal_t(ellipse, t, normal);

/* move normal to quadrant */
	if(quadrant != 1){
		copy_vec(trans.x, ellipse -> major);
		copy_vec(trans.y, ellipse -> minor);
		copy_vec(trans.z, ellipse -> normal);

		transform_vector_from_global_to_local(normal, &n, &trans);
		if(quadrant == 2 || quadrant == 3)n.x *= -1.0;
		if(quadrant == 4 || quadrant == 3)n.y *= -1.0;

		invert_rotation_transformation(&trans);
		transform_vector_from_local_to_global(&n, normal, &trans);
	}
}


static void
get_ellipse_point_t(ellipse_rec *ellipse, double t, point_rec *pnt)
{
	double weight = SQRT1_2;
	double B0, B1, B2, Bx, By, Bz;
	double sum_B;

	B0 = B_0(t);
	B1 = B_1(t) * weight;
	B2 = B_2(t);

	sum_B = B0 + B1 + B2;

	Bx = B0 * ellipse -> pnt0.x + B1 * ellipse -> pnt1.x + B2 * ellipse -> pnt2.x;
	By = B0 * ellipse -> pnt0.y + B1 * ellipse -> pnt1.y + B2 * ellipse -> pnt2.y;
	Bz = B0 * ellipse -> pnt0.z + B1 * ellipse -> pnt1.z + B2 * ellipse -> pnt2.z;
	
	pnt -> x = Bx / sum_B;
	pnt -> y = By / sum_B;
	pnt -> z = Bz / sum_B;
}


static double
get_ellipse_tangent_t(ellipse_rec *ellipse, double t, vector_rec *grad)
{
	double weight = SQRT1_2;
	double B0, B1, B2, dB0, dB1, dB2;
	double sum_B, sum_dB, sum_B2, Bx, By, Bz, dBx, dBy, dBz;

	B0 = B_0(t);
	B1 = B_1(t) * weight;
	B2 = B_2(t);

	dB0 = dB_0(t);
	dB1 = dB_1(t) * weight;
	dB2 = dB_2(t);

	sum_B = B0 + B1 + B2;
	sum_dB = dB0 + dB1 + dB2;

	sum_B2 = sum_B * sum_B;

	Bx = B0 * ellipse -> pnt0.x + B1 * ellipse -> pnt1.x + B2 * ellipse -> pnt2.x;
	By = B0 * ellipse -> pnt0.y + B1 * ellipse -> pnt1.y + B2 * ellipse -> pnt2.y;
	Bz = B0 * ellipse -> pnt0.z + B1 * ellipse -> pnt1.z + B2 * ellipse -> pnt2.z;

	dBx = dB0 * ellipse -> pnt0.x + dB1 * ellipse -> pnt1.x + dB2 * ellipse -> pnt2.x;
	dBy = dB0 * ellipse -> pnt0.y + dB1 * ellipse -> pnt1.y + dB2 * ellipse -> pnt2.y;
	dBz = dB0 * ellipse -> pnt0.z + dB1 * ellipse -> pnt1.z + dB2 * ellipse -> pnt2.z;

	grad -> x = (dBx * sum_B - Bx * sum_dB) / sum_B2;
	grad -> y = (dBy * sum_B - By * sum_dB) / sum_B2;
	grad -> z = (dBz * sum_B - Bz * sum_dB) / sum_B2;

	return(size_vec(*grad));
}


static void
get_ellipse_normal_t(ellipse_rec *ellipse, double t, vector_rec *normal)
{
	vector_rec grad;
	double norm;

	norm = get_ellipse_tangent_t(ellipse, t, &grad);
	norm = sqrt(norm);
	div_vec(grad, norm);

	cross_product(*normal, grad, ellipse -> normal);
}

#endif


/* delta_t */

static logic
check_point_on_ellipse_interval(int q, double t, int q1, double t1, int q2, double t2)
{
	double delta_t = EPSILON_T;

	if(q <= 0 || q > 4 || q1 <= 0 || q1 > 4 || q2 <= 0 || q2 > 4)error_message(GENERAL_ERROR, "Ivalid quadrant");

	if(q != q1 && q != q2){
		if(q1 == q2){
			if(q1 == 1 || q1 == 3){
				if(t1 < t2)return(NO);
			}
			else{
				if(t1 > t2)return(NO);
			}
		}
		else{
			if(q1 < q2){
				if(q < q1 || q > q2)return(NO);
			}
			else{
				if(q < q1){
					if(q > q2)return(NO);
				}
				if(q > q2){
					if(q < q1)return(NO);
				}
			}
		}

		return(YES);
	}

	if(q == q1){
		if(q == 1 || q == 3){
			if(t < t1 - delta_t)return(NO);
		}
		else{
			if(t > t1 + delta_t)return(NO);
		}
	}
	if(q == q2){
		if(q == 1 || q == 3){
			if(t > t2 + delta_t)return(NO);
		}
		else{
			if(t < t2 - delta_t)return(NO);
		}
	}

	return(YES);
}



static void
transform_vector_from_global_to_local(vector_rec *global, vector_rec *local, trans_matrix *trans_g2l)
{
	local -> x = dot_product(*global, trans_g2l -> x);
	local -> y = dot_product(*global, trans_g2l -> y);
	local -> z = dot_product(*global, trans_g2l -> z);
}


static void
transform_vector_from_local_to_global(vector_rec *local, vector_rec *global, trans_matrix *trans_l2g)
{
	global -> x = dot_product(*local, trans_l2g -> x);
	global -> y = dot_product(*local, trans_l2g -> y);
	global -> z = dot_product(*local, trans_l2g -> z);
}


static void
transform_point_from_global_to_local(point_rec *global, point_rec *local, point_rec *origin, trans_matrix *trans_g2l)
{
	vector_rec delta;

	delta.x = global -> x - origin -> x;
	delta.y = global -> y - origin -> y;
	delta.z = global -> z - origin -> z;

	local -> x = dot_product(delta, trans_g2l -> x);
	local -> y = dot_product(delta, trans_g2l -> y);
	local -> z = dot_product(delta, trans_g2l -> z);
}


static void
transform_point_from_local_to_global(point_rec *local, point_rec *global, point_rec *origin, trans_matrix *trans_l2g)
{
	global -> x = dot_product(*local, trans_l2g -> x) + origin -> x;
	global -> y = dot_product(*local, trans_l2g -> y) + origin -> y;
	global -> z = dot_product(*local, trans_l2g -> z) + origin -> z;
}


static void
invert_rotation_transformation(trans_matrix *trans)
{
	double tmp;

	swap(trans -> x.y, trans -> y.x, tmp);
	swap(trans -> x.z, trans -> z.x, tmp);
	swap(trans -> y.z, trans -> z.y, tmp);
}

static void
get_inverted_rotation_transformation(trans_matrix *trans_in, trans_matrix *trans_out)
{
	trans_out -> x.x = trans_in -> x.x;
	trans_out -> x.y = trans_in -> y.x;
	trans_out -> x.z = trans_in -> z.x;

	trans_out -> y.x = trans_in -> x.y;
	trans_out -> y.y = trans_in -> y.y;
	trans_out -> y.z = trans_in -> z.y;

	trans_out -> z.x = trans_in -> x.z;
	trans_out -> z.y = trans_in -> y.z;
	trans_out -> z.z = trans_in -> z.z;
}


static void
transform_local_from_ellipsoid_to_sphere(point_rec *ellipsoidal, point_rec *spherical, vector_rec *trans_e2s)
{
	spherical -> x = ellipsoidal -> x * trans_e2s -> x;
	spherical -> y = ellipsoidal -> y * trans_e2s -> y;
	spherical -> z = ellipsoidal -> z * trans_e2s -> z;
}


static void
transform_local_from_sphere_to_ellipsoid(point_rec *spherical, point_rec *ellipsoidal, vector_rec *trans_s2e)
{
	ellipsoidal -> x = spherical -> x * trans_s2e -> x;
	ellipsoidal -> y = spherical -> y * trans_s2e -> y;
	ellipsoidal -> z = spherical -> z * trans_s2e -> z;
}


static void
invert_scale_transformation(vector_rec *trans)
{
	trans -> x = 1.0 / trans -> x;
	trans -> y = 1.0 / trans -> y;
	trans -> z = 1.0 / trans -> z;
}

static void
get_inverted_scale_transformation(vector_rec *trans_in, vector_rec *trans_out)
{
	trans_out -> x = 1.0 / trans_in -> x;
	trans_out -> y = 1.0 / trans_in -> y;
	trans_out -> z = 1.0 / trans_in -> z;
}	



static char *
get_next_record_file(FILE *file, char *file_name, char *err_msg)
{
	char *buf = NULL;

	if((buf = fgets(line_buffer, BUFFER_SIZE, file)) == NULL){
		if(!feof(file))error_message(FILE_READ_ERROR, "File %s reading error", file_name);
		if(err_msg == NULL)return(NULL);
		error_message(INPUT_ERROR, "%s", err_msg);
	}
	return(buf);
}


static char *
get_next_relevant_record_file(FILE *file, char *file_name, char *err_msg)
{
	char *buf = NULL, *character = NULL;
	
	do{
		if((buf = get_next_record_file(file, file_name, err_msg)) == NULL)return(NULL);
		character = line_buffer;
		while(*character == SPACE)character++;
	}while(*character == NEWLINE || *character == CR || *character == REMARK);
	return(buf);
}


static void
write_record_file(FILE *file, char *file_name, char *format, ...)
{
	char buffer[1024];
	va_list args;

	va_start(args, format);
	vsprintf(buffer, format, args);
	va_end(args);

	if(fprintf(file, "%s", buffer) < 0)
		error_message(FILE_WRITE_ERROR, "File %s writing error", file_name);
}


static char *
get_next_record(char *err_msg)
{
	char *buf = NULL;

	if((buf = fgets(line_buffer, BUFFER_SIZE, geom_file)) == NULL){
		if(!feof(geom_file))error_message(FILE_READ_ERROR, "File %s reading error", geom_file_name);
		if(err_msg == NULL)return(NULL);
		error_message(INPUT_ERROR, "%s", err_msg);
	}
	return(buf);
}


static char *
get_next_relevant_record(char *err_msg)
{
	char *buf = NULL, *character = NULL;
	
	do{
		if((buf = get_next_record(err_msg)) == NULL)return(NULL);
		character = line_buffer;
		while(*character == SPACE)character++;
	}while(*character == NEWLINE || *character == CR || *character == REMARK);
	return(buf);
}


static void
write_record(char *format, ...)
{
	char buffer[1024];
	va_list args;

	va_start(args, format);
	vsprintf(buffer, format, args);
	va_end(args);

	if(fprintf(t3d_in_file, "%s", buffer) < 0)
		error_message(FILE_WRITE_ERROR, "File %s writing error", t3d_in_file_name);
}


static void
error_message(int exit_code, char *format, ...)
{
	char buffer[1024];
	va_list args;
	static logic first = YES;

	va_start(args, format);
	vsprintf(buffer, format, args);
	va_end(args);

	if(first == YES){
		fprintf(stderr, "\n");
		first = NO;
	}

	if(exit_code == NOTIFICATION)
		fprintf(stderr, "Notification: %s\n", buffer);
	else if(exit_code == WARNING)
		fprintf(stderr, "Warning: %s\n", buffer);
	else if(exit_code == ALARM)
		fprintf(stderr, "Alarm: %s\n", buffer);
	else{
		fprintf(stderr, "\nError: %s\n\n", buffer);

#ifdef ELIXIR
		if(draw == YES){
			if(ElixirInitialized() == YES){
				strcat(buffer, "\n\n");
				ESIEventLoop(YES, buffer);
			}
		}
#endif

		if(t3d_in_file != NULL)fclose(t3d_in_file);
		if(geom_file != NULL)fclose(geom_file);

		exit(exit_code);
	}
}


static void
sort_id_coord(long n, long *id, is_pnt_rec **data)
{
	long i, j, d = 1, tmp_id;
	is_pnt_rec *tmp_data = NULL;

	if(n <= 1)return;

	i = (long)(log((double)n + 0.5) / LN_2);
	while(i--)d *= 2;
	d -= 1;

	while(d != 0){
		for(i = 0; i < n - d; i++){
#ifdef DEBUG
			if(data[i] -> coord == data[i + d] -> coord)error_message(GENERAL_ERROR, "Unexpected situation");
#endif
			if(data[i] -> coord < data[i + d] -> coord)continue;
			tmp_data = data[i + d];
			data[i + d] = data[i];
			tmp_id = id[i + d];
			id[i + d] = id[i];
			if(i < d){
				data[i] = tmp_data;
				id[i] = tmp_id;
				continue;
			}
			for(j = i - d; j >= 0; j -= d){
#ifdef DEBUG
				if(tmp_data -> coord == data[j] -> coord)error_message(GENERAL_ERROR, "Unexpected situation");
#endif
				if(tmp_data -> coord > data[j] -> coord)break;
				data[j + d] = data[j];
				id[j + d] = id[j];
			}
			data[j + d] = tmp_data;
			id[j + d] = tmp_id;
		}
		d /= 2;
	}
}


#ifdef ELIXIR

static void
draw_number(point_rec *pnt, int layer, EPixel color, int number, char *prefix)
{
	WCRec point;
	GraphicObj *gr_obj;
	unsigned long mask = 0L;
	char string[16];

	if(draw == NO)return;

	sprintf(string, "%s%d", prefix, number);

  mask |= LAYER_MASK;
  mask |= COLOR_MASK;
	mask |= FONT_MASK;

	EASValsSetLayer(layer);
	EASValsSetColor(color);
	EASValsSetFontId(font);

	copy_vec(point, *pnt);

	if((gr_obj = CreateAnnText3D(&point, string)) == NULL){
		ERptErrMessage(ELIXIR_ERROR_CLASS, 2, "Failure of graphics creation", ERROR_GRADE);
		exit(ESISetExitCode(2));
	}

	EGWithMaskChangeAttributes(mask, gr_obj); 
	EMAddGraphicsToModel(ESIModel(), gr_obj);
}


static void
draw_point(point_rec *pnt, int layer, EPixel color)
{
	WCRec point[1];
	GraphicObj *gr_obj = NULL;
	unsigned long mask = 0L;
	
	if(draw == NO)return;

  mask |= LAYER_MASK;
  mask |= COLOR_MASK;
	mask |= MTYPE_MASK;
	mask |= MSIZE_MASK;

	EASValsSetLayer(layer);
	EASValsSetColor(color);
	EASValsSetMType(FILLED_CIRCLE_MARKER);
  EASValsSetMSize(6);	

	copy_vec(point[0], *pnt);

	if((gr_obj = CreateMarker3D(point)) == NULL){
		ERptErrMessage(ELIXIR_ERROR_CLASS, 2, "Failure of graphics creation", ERROR_GRADE);
		exit(ESISetExitCode(2));
	}

	EGWithMaskChangeAttributes(mask, gr_obj);
	EMAddGraphicsToModel(ESIModel(), gr_obj);
}



static void
draw_line(point_rec *pnt1, point_rec *pnt2, int layer, EPixel color)
{
	WCRec point[2];
	GraphicObj *gr_obj = NULL;
	unsigned long mask = 0L;
	
	if(draw == NO)return;
	
  mask |= LAYER_MASK;
  mask |= COLOR_MASK;
	mask |= STYLE_MASK;
  mask |= SHRINK_MASK;

	EASValsSetLayer(layer);
	EASValsSetColor(color);
	EASValsSetLineStyle(SOLID_STYLE);
	EASValsSetShrink(1.0);

	copy_vec(point[0], *pnt1);
	copy_vec(point[1], *pnt2);

	if((gr_obj = CreateLine3D(point)) == NULL){
		ERptErrMessage(ELIXIR_ERROR_CLASS, 2, "Failure of graphics creation", ERROR_GRADE);
		exit(ESISetExitCode(2));
	}
		
	EGWithMaskChangeAttributes(mask, gr_obj);
	EMAddGraphicsToModel(ESIModel(), gr_obj);
}



static void
draw_elliptic_arc(point_rec *pnt1, point_rec *pnt2, point_rec *pnt3, double w1, double w2, double w3, int layer, EPixel color)
{
	WCRec point[3];
	float weight[3];
	GraphicObj *gr_obj = NULL;
	unsigned long mask = 0L;
	
	if(draw == NO)return;
	
  mask |= LAYER_MASK;
  mask |= COLOR_MASK;
	mask |= STYLE_MASK;
	mask |= WIDTH_MASK;
	mask |= SHOW_POLY_MASK;
	mask |= TESSEL_INTERVALS_MASK;

	EASValsSetLayer(layer);
	EASValsSetColor(color);
	EASValsSetLineStyle(SOLID_STYLE);
	EASValsSetLineWidth(0);
	EASValsSetShowPoly(NO);
	EASValsSetTesselIntervals(32);

	copy_vec(point[0], *pnt1);
	copy_vec(point[1], *pnt2);
	copy_vec(point[2], *pnt3);

	weight[0] = w1;
	weight[1] = w2;
	weight[2] = w3;

	if((gr_obj = CreateRBezC3D(3, point, weight)) == NULL){
		ERptErrMessage(ELIXIR_ERROR_CLASS, 2, "Failure of graphics creation", ERROR_GRADE);
		exit(ESISetExitCode(2));
	}
		
	EGWithMaskChangeAttributes(mask, gr_obj);
	EMAddGraphicsToModel(ESIModel(), gr_obj);
}


static void
draw_ellipsoidal_surface(point_rec *pnt, double *w, int layer, EPixel color)
{
	WCRec point[9];
	float weight[9];
	int i;
	
	GraphicObj *gr_obj = NULL;
	unsigned long mask = 0L;
	
	if(draw == NO)return;
	
  mask |= LAYER_MASK;
  mask |= COLOR_MASK;
	mask |= STYLE_MASK;
	mask |= WIDTH_MASK;
	mask |= FILL_MASK;
	mask |= SHOW_POLY_MASK;
	mask |= TESSEL_U_MASK;
	mask |= TESSEL_V_MASK;

	EASValsSetLayer(layer);
	EASValsSetColor(color);
	EASValsSetLineStyle(SOLID_STYLE);
	EASValsSetLineWidth(0);
#ifdef FILLED
	EASValsSetFillStyle(FILL_SOLID);
	EASValsSetUTesselIntervals(32);
	EASValsSetVTesselIntervals(32);
#else
	EASValsSetFillStyle(FILL_HOLLOW);
	EASValsSetUTesselIntervals(16);
	EASValsSetVTesselIntervals(16);
#endif
	EASValsSetShowPoly(NO);

	for(i = 0; i < 9; i++){
		copy_vec(point[i], pnt[i]);
		weight[i] = w[i];
	}

	if((gr_obj = CreateRBezS3D(3, 3, point, weight)) == NULL){
		ERptErrMessage(ELIXIR_ERROR_CLASS, 2, "Failure of graphics creation", ERROR_GRADE);
		exit(ESISetExitCode(2));
	}
		
	EGWithMaskChangeAttributes(mask, gr_obj);
	EMAddGraphicsToModel(ESIModel(), gr_obj);
}


static void
draw_tile(tile_rec *tile, int layer, EPixel color)
{
	point_rec pnt[8];
	int sign_x[8] = {1, -1, -1, 1, 1, -1, -1, 1};
	int sign_y[8] = {1, 1, -1, -1, 1, 1, -1, -1};
	int sign_z[8] = {-1, -1, -1, -1, 1, 1, 1, 1};
	int i;
	
	if(draw == NO)return;

	for(i = 0; i < 8; i++){
		pnt[i].x = tile -> center.x + sign_x[i] * tile -> size.x * 0.5;
		pnt[i].y = tile -> center.y + sign_y[i] * tile -> size.y * 0.5;
		pnt[i].z = tile -> center.z + sign_z[i] * tile -> size.z * 0.5;
	}

	draw_line(&pnt[0], &pnt[1], layer, color);
	draw_line(&pnt[2], &pnt[3], layer, color);
	draw_line(&pnt[4], &pnt[5], layer, color);
	draw_line(&pnt[6], &pnt[7], layer, color);

	draw_line(&pnt[0], &pnt[3], layer, color);
	draw_line(&pnt[1], &pnt[2], layer, color);
	draw_line(&pnt[5], &pnt[6], layer, color);
	draw_line(&pnt[4], &pnt[7], layer, color);

	draw_line(&pnt[0], &pnt[4], layer, color);
	draw_line(&pnt[1], &pnt[5], layer, color);
	draw_line(&pnt[2], &pnt[6], layer, color);
	draw_line(&pnt[3], &pnt[7], layer, color);
}
	

static void
draw_inclusion(inclusion_rec *inclusion, int layer, EPixel color)
{
	if(draw == NO)return;

	draw_ellipsoid(&(inclusion -> ellipsoid), layer, color);

#ifdef DRAW_INCLUSION_CENTER
	draw_point(&(inclusion -> ellipsoid.center), layer, color);
	draw_number(&(inclusion -> ellipsoid.center), layer, color, inclusion -> id, " ");
#else
	draw_number(&(inclusion -> ellipsoid.center), layer, color, inclusion -> id, "");
#endif
}


static void
draw_ellipsoid(ellipsoid_rec *ellipsoid, int layer, EPixel color)
{
#ifdef SOLID
	point_rec point[9];
	double w[9];
	vector_rec vec_max, vec_min, vec_mid;
#else
	ellipse_rec ellipse;
#endif

	if(draw == NO)return;

#ifdef SOLID
	w[0] = 1.0;
	w[1] = SQRT1_2;
	w[2] = 1.0;
	w[3] = SQRT1_2;
	w[4] = 0.5;
	w[5] = SQRT1_2;
	w[6] = 1.0;
	w[7] = SQRT1_2;
	w[8] = 1.0;

	copy_mul_vec(vec_max, ellipsoid -> major, ellipsoid -> max);
	copy_mul_vec(vec_min, ellipsoid -> minor, ellipsoid -> min);
	copy_mul_vec(vec_mid, ellipsoid -> middle, ellipsoid -> mid);

	add_vec(point[0], ellipsoid -> center, vec_max);
	copy_vec(point[1], point[0]);
	copy_vec(point[2], point[0]);

	add_vec(point[6], ellipsoid -> center, vec_min);
	add_vec(point[8], ellipsoid -> center, vec_mid);
	add_vec(point[7], point[6], vec_mid);
		
	add_vec(point[3], point[6], vec_max);
	add_vec(point[4], point[7], vec_max);
	add_vec(point[5], point[8], vec_max);

	draw_ellipsoidal_surface(point, w, layer, color);

	sub_vec(point[6], ellipsoid -> center, vec_min);
	//add_vec(point[8], ellipsoid -> center, vec_mid);  /* does not change */
	add_vec(point[7], point[6], vec_mid);

	add_vec(point[3], point[6], vec_max);
	add_vec(point[4], point[7], vec_max);
	//add_vec(point[5], point[8], vec_max);  /* does not change */

	draw_ellipsoidal_surface(point, w, layer, color);

	//sub_vec(point[6], ellipsoid -> center, vec_min);  /* does not change */
	sub_vec(point[8], ellipsoid -> center, vec_mid);
	sub_vec(point[7], point[6], vec_mid);

	//add_vec(point[3], point[6], vec_max);  /* does not change */
	add_vec(point[4], point[7], vec_max);
	add_vec(point[5], point[8], vec_max);

	draw_ellipsoidal_surface(point, w, layer, color);

	add_vec(point[6], ellipsoid -> center, vec_min);
	//sub_vec(point[8], ellipsoid -> center, vec_mid);  /* does not change */
	sub_vec(point[7], point[6], vec_mid);

	add_vec(point[3], point[6], vec_max);
	add_vec(point[4], point[7], vec_max);
	//add_vec(point[5], point[8], vec_max);  /* does not change */

	draw_ellipsoidal_surface(point, w, layer, color);

	sub_vec(point[0], ellipsoid -> center, vec_max);
	copy_vec(point[1], point[0]);
	copy_vec(point[2], point[0]);

	sub_vec(point[3], point[6], vec_max);
	sub_vec(point[4], point[7], vec_max);
	sub_vec(point[5], point[8], vec_max);

	draw_ellipsoidal_surface(point, w, layer, color);

	sub_vec(point[6], ellipsoid -> center, vec_min);
	//sub_vec(point[8], ellipsoid -> center, vec_mid);  /* does not change */
	sub_vec(point[7], point[6], vec_mid);

	sub_vec(point[3], point[6], vec_max);
	sub_vec(point[4], point[7], vec_max);
	//sub_vec(point[5], point[8], vec_max);  /* does not change */

	draw_ellipsoidal_surface(point, w, layer, color);

	//sub_vec(point[6], ellipsoid -> center, vec_min);  /* does not change */
	add_vec(point[8], ellipsoid -> center, vec_mid);
	add_vec(point[7], point[6], vec_mid);

	//sub_vec(point[3], point[6], vec_max);  /* does not change */
	sub_vec(point[4], point[7], vec_max);
	sub_vec(point[5], point[8], vec_max);

	draw_ellipsoidal_surface(point, w, layer, color);

	add_vec(point[6], ellipsoid -> center, vec_min);
	//add_vec(point[8], ellipsoid -> center, vec_mid);  /* does not change */
	add_vec(point[7], point[6], vec_mid);

	sub_vec(point[3], point[6], vec_max);
	sub_vec(point[4], point[7], vec_max);
	//sub_vec(point[5], point[8], vec_max);  /* does not change */

	draw_ellipsoidal_surface(point, w, layer, color);
#else
	copy_vec(ellipse.center, ellipsoid -> center);

	copy_vec(ellipse.major, ellipsoid -> major);
	copy_vec(ellipse.minor, ellipsoid -> minor);
	copy_vec(ellipse.normal, ellipsoid -> middle);

	ellipse.max = ellipsoid -> max;
	ellipse.min = ellipsoid -> min;

	draw_ellipse(&ellipse, layer, color);

	copy_vec(ellipse.major, ellipsoid -> minor);
	copy_vec(ellipse.minor, ellipsoid -> middle);
	copy_vec(ellipse.normal, ellipsoid -> major);

	ellipse.max = ellipsoid -> min;
	ellipse.min = ellipsoid -> mid;

	draw_ellipse(&ellipse, layer, color);

	copy_vec(ellipse.major, ellipsoid -> middle);
	copy_vec(ellipse.minor, ellipsoid -> major);
	copy_vec(ellipse.normal, ellipsoid -> minor);

	ellipse.max = ellipsoid -> mid;
	ellipse.min = ellipsoid -> max;

	draw_ellipse(&ellipse, layer, color);
#endif

}


static void
draw_ellipse(ellipse_rec *ellipse, int layer, EPixel color)
{
	point_rec point[3];
	vector_rec vec_max, vec_min;

	if(draw == NO)return;

	copy_mul_vec(vec_max, ellipse -> major, ellipse -> max);
	copy_mul_vec(vec_min, ellipse -> minor, ellipse -> min);

	add_vec(point[0], ellipse -> center, vec_max);
	add_vec(point[2], ellipse -> center, vec_min);
	add_vec(point[1], point[0], vec_min);

	draw_elliptic_arc(&point[0], &point[1], &point[2], 1.0, SQRT1_2, 1.0, layer, color);

	sub_vec(point[2], ellipse -> center, vec_min);
	sub_vec(point[1], point[0], vec_min);
	
	draw_elliptic_arc(&point[0], &point[1], &point[2], 1.0, SQRT1_2, 1.0, layer, color);

	sub_vec(point[0], ellipse -> center, vec_max);
	sub_vec(point[1], point[0], vec_min);

	draw_elliptic_arc(&point[0], &point[1], &point[2], 1.0, SQRT1_2, 1.0, layer, color);

	add_vec(point[2], ellipse -> center, vec_min);
	add_vec(point[1], point[0], vec_min);
	
	draw_elliptic_arc(&point[0], &point[1], &point[2], 1.0, SQRT1_2, 1.0, layer, color);
}


static void
draw_elliptic_arc_segment(ellipse_rec *ellipse, int q, double ta, double tb, point_rec *pointa, point_rec *pointb, 
													trans_matrix *trans, int layer, EPixel color)
{
	point_rec pnt_loc, pnt_glob, point;
	double w0, w1, w2, w01a, w12a, wa, w01b, w12b, wb;
	int sign_x[4] = {1, -1, -1, 1}, sign_y[4] = {1, 1, -1, -1};

	if(draw == NO)return;

	w0 = 1.0;
	w1 = SQRT1_2;
	w2 = 1.0;

	tb = (tb - ta) / (1.0 - ta);

	w01a = (1.0 - ta) * w0 + ta * w1;
	w12a = (1.0 - ta) * w1 + ta * w2;
	wa = (1.0 - ta) * w01a + ta * w12a;

	w01b = (1.0 - tb) * wa + tb * w12a;
	w12b = (1.0 - tb) * w12a + tb * w2;
	wb = (1.0 - tb) * w01b + tb * w12b;
	
	pnt_loc.x = ellipse -> max * sign_x[q - 1] * (1.0 - ta) * w1 / ((1.0 - ta) * w1 + ta * w2);
	pnt_loc.y = ellipse -> min * sign_y[q - 1];
	pnt_loc.z = 0.0;

	transform_point_from_local_to_global(&pnt_loc, &pnt_glob, &(ellipse -> center), trans);

	point.x = ((1.0 - tb) * wa * pointa -> x + tb * w12a * pnt_glob.x) / ((1.0 - tb) * wa + tb * w12a);
	point.y = ((1.0 - tb) * wa * pointa -> y + tb * w12a * pnt_glob.y) / ((1.0 - tb) * wa + tb * w12a);
	point.z = ((1.0 - tb) * wa * pointa -> z + tb * w12a * pnt_glob.z) / ((1.0 - tb) * wa + tb * w12a);

	draw_elliptic_arc(pointa, &point, pointb, wa, w01b, wb, layer, color);
}



static void
draw_elliptic_arc_start_segment(ellipse_rec *ellipse, int q, double t, point_rec *point, trans_matrix *trans, int layer, EPixel color)
{
	point_rec pnt_loc, pnt_glob, point_loc, point_glob;
	double w0, w1, w2, w01, w12, w;
	int sign_x[4] = {1, -1, -1, 1}, sign_y[4] = {1, 1, -1, -1};

	if(draw == NO)return;

	w0 = 1.0;
	w1 = SQRT1_2;
	w2 = 1.0;

	w01 = (1.0 - t) * w0 + t * w1;
	w12 = (1.0 - t) * w1 + t * w2;
	w = (1.0 - t) * w01 + t * w12;

	point_loc.x = ellipse -> max * sign_x[q - 1];
	point_loc.y = 0.0;
	point_loc.z = 0.0;

	pnt_loc.x = ellipse -> max * sign_x[q - 1];
	pnt_loc.y = ellipse -> min * sign_y[q - 1] * t * w1 / ((1.0 - t) * w0 + t * w1);
	pnt_loc.z = 0.0;

	transform_point_from_local_to_global(&pnt_loc, &pnt_glob, &(ellipse -> center), trans);
	transform_point_from_local_to_global(&point_loc, &point_glob, &(ellipse -> center), trans);

	draw_elliptic_arc(&point_glob, &pnt_glob, point, w0, w01, w, layer, color);
}



static void
draw_elliptic_arc_end_segment(ellipse_rec *ellipse, int q, double t, point_rec *point, trans_matrix *trans, int layer, EPixel color)
{
	point_rec pnt_loc, pnt_glob, point_loc, point_glob;
	double w0, w1, w2, w01, w12, w;
	int sign_x[4] = {1, -1, -1, 1}, sign_y[4] = {1, 1, -1, -1};

	if(draw == NO)return;

	w0 = 1.0;
	w1 = SQRT1_2;
	w2 = 1.0;

	w01 = (1.0 - t) * w0 + t * w1;
	w12 = (1.0 - t) * w1 + t * w2;
	w = (1.0 - t) * w01 + t * w12;

	point_loc.x = 0.0;
	point_loc.y = ellipse -> min * sign_y[q - 1];
	point_loc.z = 0.0;
		
	pnt_loc.x = ellipse -> max * sign_x[q - 1] * (1.0 - t) * w1 / ((1.0 - t) * w1 + t * w2);
	pnt_loc.y = ellipse -> min * sign_y[q - 1];
	pnt_loc.z = 0.0;

	transform_point_from_local_to_global(&pnt_loc, &pnt_glob, &(ellipse -> center), trans);
	transform_point_from_local_to_global(&point_loc, &point_glob, &(ellipse -> center), trans);

	draw_elliptic_arc(point, &pnt_glob, &point_glob, w, w12, w2, layer, color);
}



static top_command table[] = {
	{CONST_CHAR_PTR "DISTANCE",   8, select_distance_points, CONST_CHAR_PTR "[# # #, # # #] \ncompute distance of two points"},
};



static void
augment_command_table(void)
{
  TypeInAugmentCommandTable(table, XtNumber(table));
}


void
ESICustomize(Widget parent_pane)
{
}


static void
select_distance_points(void)
{
	double dx, dy, dz, x1, y1, z1, x2, y2, z2, dist = 0.0;
	int i;
	char buffer[128];
	
	if(TypeInGetTokenType(2) == NO_TOKEN){
		EMInstallHandler(ESIModel(), distance_handler,
										 NULL, NULL,
										 NULL, NULL,
										 NULL, NULL,
										 NULL, NULL);
		EVSetSelectActionFunction(NULL, NULL, NULL);
		ERptPrompt(100, 100, "B1=Select first point|B3=Cancel");
	}
	else{
		for(i = 0; i < 6; i++){
			if(TypeInGetTokenType(i + 2) != NUMBER){
				TypeInBadSyntax(NULL);
				return;
			}
		}
		x1 = TypeInGetTokenAsDouble(2);
		y1 = TypeInGetTokenAsDouble(3);
		z1 = TypeInGetTokenAsDouble(4);
		x2 = TypeInGetTokenAsDouble(5);
		y2 = TypeInGetTokenAsDouble(6);
		z2 = TypeInGetTokenAsDouble(7);

		dx = x2 - x1;
		dy = y2 - y1;
		dz = z2 - z1;
		dist = dx * dx + dy * dy + dz * dz;
		sprintf(buffer, "Components [%.8e %.8e %.8e]\nDistance = %20.14e", dx, dy, dz, sqrt(dist));
		ERptStatusMessage(1, 1, buffer);
	}
}



static void
distance_handler(Widget w, EView *v_p, XEvent *event)
{
  static WCRec point1, point2, curr;
	static double total_dist = 0.0;
	static int pick = 0;
	static logic b2 = NO;
	double dist, dx, dy, dz;
	char buffer[128];
  
	switch(event->type){
  case ButtonPress:
    switch(event->xbutton.button){
    case Button1:
			pick++;
			if(pick == 1){
				EVInputPoint(v_p, event, &point1);
				ERptPrompt(100, 100, "B1=Select second point|B3=Cancel");
			}
			else{
				EVInputPoint(v_p, event, &point2);
				dx = point2.x - point1.x;
				dy = point2.y - point1.y;
				dz = point2.z - point1.z;
				dist = sqrt(dx * dx + dy * dy + dz * dz);
				total_dist += dist;
				sprintf(buffer, "Components [%.8e %.8e %.8e]\nDist = %20.14e    Total = %20.14e", 
								dx, dy, dz, dist, total_dist);
				ERptStatusMessage(1, 1, buffer);
				copy_vec(point1, point2);
				ERptPrompt(100, 100, "B1=Select next second point|B2=Repeat|B3=Cancel");
				pick = 1;
				b2 = YES;
			}
			break;
		case Button2:
			if(b2 == YES){
				ERptPrompt(100, 100, "B1=Select first point|B3=Cancel");
				if(pick == 0)total_dist = 0.0;
				sprintf(buffer, "\n%30s Total = %20.14e", "", total_dist);
				ERptStatusMessage(1, 1, buffer);
				pick = 0;
				b2 = NO;
			}
			break;
		case Button3:
			pick = 0;
			total_dist = 0.0;
			ERptStatusMessage(1, 1, "");
			EMUninstallHandler(EVGetModel(v_p));
			break;
		}
		break;
  case MotionNotify:
    EVInputPoint(v_p, event, &curr);
    ERptTrackPointer(&curr);
		break;
	case ClientMessage:
		if(pick == 0)ERptPrompt(100, 100, "B1=Select first point|B3=Cancel");
		if(pick == 1)ERptPrompt(100, 100, "B1=Select second point|B3=Cancel");
		break;
  }
}

#endif


static void
output_patch_boundary(long *edge_curves, long **edge_curve_array, int *edge_curve_sign, long side_patches, long *side_patch_array, long side_vertices, long *side_vertex_array)
{
	long count, cnt, kk, lines;
	int j, ind;

	count = 0;
	for(j = 0; j < 8; j++)count += edge_curves[j];

	lines = count / ITEMS_PER_LINE;
	cnt = count - lines * ITEMS_PER_LINE;
	if(cnt != 0)
		lines++;
	else
		cnt = ITEMS_PER_LINE;

	ind = 0;
	count = edge_curves[ind];
	kk = 0;
	while(lines--){
		while(cnt){
			if(count <= cnt){
				cnt -= count;
				while(count--)write_record(" %ld", edge_curve_array[ind][kk++] * edge_curve_sign[ind]);
				count = edge_curves[++ind];
				kk = 0;
			}
			else{
				count -= cnt;
				while(cnt--)write_record(" %ld", edge_curve_array[ind][kk++] * edge_curve_sign[ind]);
				break;
			}
		}
		write_record(" \\\n  ");
		cnt = ITEMS_PER_LINE;
	}

	if((count = side_patches) != 0){
		write_record(" subpatch");
		
		lines = count / ITEMS_PER_LINE;
		cnt = count - lines * ITEMS_PER_LINE;
		if(cnt != 0)
			lines++;
		else
			cnt = ITEMS_PER_LINE;
		
		kk = 0;
		while(lines--){
			if(count == cnt){
				while(count--)write_record(" %ld", side_patch_array[kk++]);
			}
			else{
				count -= cnt;
				while(cnt--)write_record(" %ld", side_patch_array[kk++]);
			}
			write_record(" \\\n  ");
			cnt = ITEMS_PER_LINE;
		}
	}

	if((count = side_vertices) != 0){
		write_record(" fixed vertex");
		
		lines = count / ITEMS_PER_LINE;
		cnt = count - lines * ITEMS_PER_LINE;
		if(cnt != 0)
			lines++;
		else
			cnt = ITEMS_PER_LINE;
		
		kk = 0;
		while(lines--){
			if(count == cnt){
				while(count--)write_record(" %ld", side_vertex_array[kk++]);
			}
			else{
				count -= cnt;
				while(cnt--)write_record(" %ld", side_vertex_array[kk++]);
			}
			write_record(" \\\n  ");
			cnt = ITEMS_PER_LINE;
		}
	}
}


static void
output_region_boundary(long *side_patches, long **side_patch_array, int *side_patch_sign, long tile_regions, long *tile_region_array)
{
	long count, cnt, kk, lines;
	int j, ind;

	count = 0;
	for(j = 0; j < 6; j++)count += side_patches[j];

	lines = count / ITEMS_PER_LINE;
	cnt = count - lines * ITEMS_PER_LINE;
	if(cnt != 0)
		lines++;
	else
		cnt = ITEMS_PER_LINE;

	ind = 0;
	count = side_patches[ind];
	kk = 0;
	while(lines--){
		while(cnt){
			if(count <= cnt){
				cnt -= count;
				while(count--)write_record(" %ld", side_patch_array[ind][kk++] * side_patch_sign[ind]);
				count = side_patches[++ind];
				kk = 0;
			}
			else{
				count -= cnt;
				while(cnt--)write_record(" %ld", side_patch_array[ind][kk++] * side_patch_sign[ind]);
				break;
			}
		}
		write_record(" \\\n  ");
		cnt = ITEMS_PER_LINE;
	}

	if((count = tile_regions) != 0){
		write_record(" subregion");
		
		lines = count / ITEMS_PER_LINE;
		cnt = count - lines * ITEMS_PER_LINE;
		if(cnt != 0)
			lines++;
		else
			cnt = ITEMS_PER_LINE;
			
		kk = 0;
		while(lines--){
			if(count == cnt){
				while(count--)write_record(" %ld", tile_region_array[kk++]);
			}
			else{
				count -= cnt;
				while(cnt--)write_record(" %ld", tile_region_array[kk++]);
			}
			write_record(" \\\n  ");
			cnt = ITEMS_PER_LINE;
		}
	}
}


void
print_geometry(void)
{
	fprintf(stderr, INDENT "Geometry description:\n");
	fprintf(stderr, INDENT "---------------------\n");
	fprintf(stderr, INDENT "problem identification\n");
	fprintf(stderr, INDENT "tiles inclusions codes\n");
	fprintf(stderr, INDENT "center[xyz] dimension[xyz] mat imsz bmsz (for reference tile)\n");
	fprintf(stderr, INDENT "id center[xyz] point[xyz] pnt[xyz] max min mid imsz bmsz mat ifmat (for each inclusion)\n");
	fprintf(stderr, INDENT "id codes[-+xyz] inclusions ctrl_points (for each particular tile)\n");
#ifdef SEPARATE_PER_REP
	fprintf(stderr, INDENT "id per_x per_y per_z (for each tile inclusion)\n");
#else
	fprintf(stderr, INDENT "id per_xyz (for each tile inclusion)\n");
#endif
	fprintf(stderr, INDENT "coords[xyz] msz (for each tile ctrl point)\n\n");
}


void
print_setting(void)
{
	fprintf(stderr, INDENT "Setting:\n");
	fprintf(stderr, INDENT "--------\n");
	fprintf(stderr, INDENT "DEF_REL_EPSILON = %e\n", DEF_REL_EPSILON);	
	fprintf(stderr, INDENT "MAX_REL_EPSILON = %e\n", MAX_REL_EPSILON);
	fprintf(stderr, INDENT "MAX_TILES = %d\n", MAX_TILES);
	fprintf(stderr, INDENT "OFFSET = %f\n", (double)OFFSET);
	fprintf(stderr, INDENT "MULTIPLE_EPS = %f\n", (double)MULTIPLE_EPS);
	fprintf(stderr, INDENT "BUFFER_SIZE = %d\n", BUFFER_SIZE);
#if SETUP == GRID_3D
	fprintf(stderr, INDENT "SETUP = GRID_3D\n");
#endif
#if SETUP == GRID_2D
	fprintf(stderr, INDENT "SETUP = GRID_2D\n");
#endif
#if SETUP == DIAGONAL
	fprintf(stderr, INDENT "SETUP = DIAGONAL\n");
#endif
#if EXCESS_SIZE == MAX_ALLOWED_SIZE
		fprintf(stderr, INDENT "EXCESS_SIZE = MAX_ALLOWED_SIZE\n");
#endif
#if EXCESS_SIZE == MAX_PRESENT_SIZE
		fprintf(stderr, INDENT "EXCESS_SIZE = MAX_PRESENT_SIZE\n");
#endif
#if EXCESS_SIZE == MAX_PRESENT_EXCESS
		fprintf(stderr, INDENT "EXCESS_SIZE = MAX_PRESENT_EXCESS\n");
#endif
#ifdef ARBITRARY_TILE_ID
	fprintf(stderr, INDENT "ARBITRARY_TILE_ID . . . . . . . defined\n");
#else
	fprintf(stderr, INDENT "ARBITRARY_TILE_ID . . . . . not defined\n");
#endif
#ifdef SEPARATE_PER_REP
	fprintf(stderr, INDENT "SEPARATE_PER_REP  . . . . . . . defined\n");
#else
	fprintf(stderr, INDENT "SEPARATE_PER_REP  . . . . . not defined\n");
#endif
#ifdef DETOUCH
	fprintf(stderr, INDENT "DETOUCH . . . . . . . . . . . . defined\n");
#ifdef OVERDETOUCH
	fprintf(stderr, INDENT "OVERDETOUCH . . . . . . . . . . defined\n");
#else
	fprintf(stderr, INDENT "OVERDETOUCH . . . . . . . . not defined\n");
#endif
#else
	fprintf(stderr, INDENT "DETOUCH . . . . . . . . . . not defined\n");
#endif
#ifdef ALLOW_OVERLAP
	fprintf(stderr, INDENT "ALLOW_OVERLAP . . . . . . . . . defined\n");
#else	
	fprintf(stderr, INDENT "ALLOW_OVERLAP . . . . . . . not defined\n");
#endif
#ifdef CHECK_OVERLAP
	fprintf(stderr, INDENT "CHECK_OVERLAP . . . . . . . . . defined\n");
#else	
	fprintf(stderr, INDENT "CHECK_OVERLAP . . . . . . . not defined\n");
#endif
#ifdef ALLOW_OUTSIDE
	fprintf(stderr, INDENT "ALLOW_OUTSIDE . . . . . . . . . defined\n");
#else	
	fprintf(stderr, INDENT "ALLOW_OUTSIDE . . . . . . . not defined\n");
#endif
#ifdef FIX_HOLE
	fprintf(stderr, INDENT "FIX_HOLE  . . . . . . . . . . . defined\n");
#else	
	fprintf(stderr, INDENT "FIX_HOLE  . . . . . . . . . not defined\n");
#endif
#ifdef USE_VIRTUAL_ELLIPSOID
	fprintf(stderr, INDENT "USE_VIRTUAL_ELLIPSOID . . . . . defined\n");
#else
	fprintf(stderr, INDENT "USE_VIRTUAL_ELLIPSOID . . . not defined\n");
#endif
#ifdef HALF_MAX_TILES
	fprintf(stderr, INDENT "HALF_MAX_TILES  . . . . . . . . defined\n");
#else
	fprintf(stderr, INDENT "HALF_MAX_TILES  . . . . . . not defined\n");
#endif
#ifdef USE_TILE
	fprintf(stderr, INDENT "USE_TILE  . . . . . . . . . . . defined\n");
#else
	fprintf(stderr, INDENT "USE_TILE  . . . . . . . . . not defined\n");
#endif
#ifdef USE_SIDE
	fprintf(stderr, INDENT "USE_SIDE  . . . . . . . . . . . defined\n");
#else
	fprintf(stderr, INDENT "USE_SIDE  . . . . . . . . . not defined\n");
#endif
#ifdef USE_CODE
	fprintf(stderr, INDENT "USE_CODE  . . . . . . . . . . . defined\n");
#else
	fprintf(stderr, INDENT "USE_CODE  . . . . . . . . . not defined\n");
#endif
#ifdef USE_SIMPLE
	fprintf(stderr, INDENT "USE_SIMPLE  . . . . . . . . . . defined\n");
#else
	fprintf(stderr, INDENT "USE_SIMPLE  . . . . . . . . not defined\n");
#endif
#ifdef DEBUG
	fprintf(stderr, INDENT "DEBUG . . . . . . . . . . . . . defined\n");
#ifdef DEBUG_PRINT
	fprintf(stderr, INDENT "DEBUG_PRINT . . . . . . . . . . defined\n");
#else
	fprintf(stderr, INDENT "DEBUG_PRINT . . . . . . . . not defined\n");
#endif
#ifdef ELIXIR
	fprintf(stderr, INDENT "ELIXIR  . . . . . . . . . . . . defined\n");
#else
	fprintf(stderr, INDENT "ELIXIR  . . . . . . . . . . not defined\n");
#endif
#ifdef IS_TESTING
	fprintf(stderr, INDENT "IS_TESTING  . . . . . . . . . . defined\n");
	fprintf(stderr, INDENT "MIRROR_X = %d\n", MIRROR_X);
	fprintf(stderr, INDENT "MIRROR_Y = %d\n", MIRROR_Y);
	fprintf(stderr, INDENT "MIRROR_Z = %d\n", MIRROR_Z);
	fprintf(stderr, INDENT "CYCLE_COORD = %d\n", CYCLE_COORD);
#ifdef PER_REP_ALL
	fprintf(stderr, INDENT "PER_REP_ALL . . . . . . . . . . defined\n");
#else	
	fprintf(stderr, INDENT "PER_REP_X = %d\n", PER_REP_X);
	fprintf(stderr, INDENT "PER_REP_Y = %d\n", PER_REP_Y);
	fprintf(stderr, INDENT "PER_REP_Z = %d\n", PER_REP_Z);
#endif
#else
	fprintf(stderr, INDENT "IS_TESTING  . . . . . . . . not defined\n");
#endif
#ifdef SPECIAL_TESTING
	fprintf(stderr, INDENT "SPECIAL_TESTING . . . . . . . . defined\n");
	fprintf(stderr, INDENT "SPLITTING_PLANE = %d\n", SPLITTING_PLANE);
	fprintf(stderr, INDENT "OPPOSITE_PLANE = %d\n", OPPOSITE_PLANE);
#else
	fprintf(stderr, INDENT "SPECIAL_TESTING . . . . . . not defined\n");
#endif
	fprintf(stderr, INDENT "MAX_RATE = %f\n", MAX_RATE);
	fprintf(stderr, INDENT "INC_RATE = %f\n", INC_RATE);
	fprintf(stderr, INDENT "cos(RIGHT_ANGLE) = %12.10f\n", RIGHT_ANGLE);
	fprintf(stderr, INDENT "cos(ZERO_ANGLE) = %12.10f\n", ZERO_ANGLE);
#else	
	fprintf(stderr, INDENT "DEBUG . . . . . . . . . . . not defined\n");
#endif
	fprintf(stderr, "\n");
}


#include "geom3d_x.c"
#include "geom3d_y.c"
#include "geom3d_z.c"

